/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

* $Id: application-cbr.cc,v 1.1.1.1 2006/12/13 13:09:35 dnguyen Exp $

*/

// Georgia Tech Network Simulator - Constant Bit Rate Application class
// George F. Riley.  Georgia Tech, Spring 2002

// Define a CBR Application class

// Uncomment below to enable debug level 0
//#define DEBUG_MASK 0x1
#include "debug.h"
#include "application-cbr.h"
#include "udp.h"
#include "node.h"
#include "simulator.h"
#include "mystat.h"

using namespace std;

// Constructors

CBRApplication::CBRApplication(IPAddr_t rIP,
                               PortId_t rPort,
                               PortId_t lPort,
                               Rate_t   rate,
                               Size_t   size,
                               const L4Protocol& l4)
                               
    : remoteIP(rIP), remotePort(rPort), localPort(lPort),
      cbrRate(rate), pktSize(size), l4Proto(l4.Copy()),
      connected(false), pendingEvent(nil)
{
  // We don't bind or do anything permanent here since we often
  // construct anonymous temporary objects.
}

CBRApplication::CBRApplication(const CBRApplication& c)
    : remoteIP(c.remoteIP), remotePort(c.remotePort), localPort(c.localPort),
      cbrRate(c.cbrRate), pktSize(c.pktSize),
      l4Proto(c.l4Proto->Copy()), connected(false), 
      pendingEvent(nil)
{
}

      
CBRApplication::~CBRApplication()
{
  delete l4Proto;
}


// Handler Methods
void CBRApplication::Handle(Event* e, Time_t t)
{ // This is only called on nonZero rate, when it's time to send again
  AppCBREvent* ev = (AppCBREvent*)e;
  switch (ev->event) {
    case AppCBREvent::SEND_PKT :
      DEBUG0((cout << "CBRApp " << this << " sending pkt time " 
              << Simulator::Now() << endl));
      l4Proto->Send(pktSize); // Send the packet
      Time_t nextTime = pktSize * 8 / cbrRate; // Time till next packet
      // Reuse the same event
      pendingEvent = ev;
      Simulator::instance->Schedule(ev, nextTime, this); // Schedule again
      return;
  }
  Application::Handle(e, t);  // May be application event
}

// NotifyHandler Methods
void CBRApplication::Notify(void*)
{ // Notify that packet has been transmitted
  // This will only be called when cbrRate == 0, indicating
  // we should send as fast as possible
  DEBUG0((cout << "AppCBR Notify Handler" << endl));
  l4Proto->Send(pktSize); // Send the next packet
}

// Application Methods
void CBRApplication::StartApp()    // Called at time specified by Start
{
  StopApp();                               // Insure no pending event
  if (!connected) l4Proto->Connect(remoteIP, remotePort);
  if (cbrRate == 0)
    { // Send as fast as possible.  Do this by first sendind a packet
      // immediately,  and requesting a notification every time a
      // packet transmission is complete.
      DEBUG0((cout << "CBRApp Adding notification, l4 " << l4Proto << endl));
      l4Proto->AddNotification(this);
      l4Proto->Send(pktSize); // Send the packet
    }
  else
    {
      Time_t nextTime = pktSize * 8 / cbrRate; // Time till next packet
      AppCBREvent* ev = new AppCBREvent(AppCBREvent::SEND_PKT);
      pendingEvent = ev;
      Simulator::instance->Schedule(ev, nextTime, this); // Schedule again
    }
}

void CBRApplication::StopApp()     // Called at time specified by Stop
{
  if (pendingEvent)
    { // Cancel the pending transmit event
      Simulator::instance->Cancel(pendingEvent);
      pendingEvent = NULL;
    }
}

void CBRApplication::AttachNode(Node* n) 
{
  l4Proto->Attach(n);
  n->Bind(l4Proto->Proto(), localPort, l4Proto);
  // We don't actually connect here, since the simulation might
  // not have already started.  Connect is done in StartApp.
}

Application* CBRApplication::Copy() const
{
  return new CBRApplication(*this);
}



