/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - On/Off Data Source Application class
// George F. Riley.  Georgia Tech, Spring 2002

// Define an on/off data source application class

// Uncomment below to enable debug level 0
//#define DEBUG_MASK 0x1
#include "debug.h"
#include "application-onoff.h"
#include "udp.h"
#include "tcp.h"
#include "node.h"
#include "simulator.h"

using namespace std;

// Defaults for rate/size
Rate_t OnOffApplication::defaultRate = 500000.0;
Size_t OnOffApplication::defaultSize = 512;

// Constructors

OnOffApplication::OnOffApplication(IPAddr_t rip, PortId_t rport,
                   const  Random& ontime, 
                   const  L4Protocol& p,
                   Rate_t r,
                   Size_t s)
    : l4Proto(p.Copy()),
    peerIP(rip), peerPort(rport),
    connected(false),
    onTime(ontime.Copy()), offTime(ontime.Copy()),
    cbrRate(r), pktSize(s), pendingEvent(nil), pendingOO(nil), 
    residualBits(0), lastStartTime(0),
    maxBytes(MAX_COUNT), totBytes(0)
{
}

OnOffApplication::OnOffApplication(IPAddr_t rip, PortId_t rport,
                   const  Random& ontime, 
                   const  Random& offtime,
                   const  L4Protocol& p,
                   Rate_t r,
                   Size_t s)
  : l4Proto(p.Copy()),
    peerIP(IPADDR_NONE), peerPort(NO_PORT),
    connected(false),
    onTime(ontime.Copy()), offTime(offtime.Copy()),
    cbrRate(r), pktSize(s), pendingEvent(nil), pendingOO(nil),
    residualBits(0), lastStartTime(0),
    maxBytes(MAX_COUNT), totBytes(0)
{
}

OnOffApplication::OnOffApplication(const OnOffApplication& c)
  : l4Proto(c.l4Proto->Copy()),
    peerIP(c.peerIP), peerPort(c.peerPort),
    connected(c.connected),
    onTime(c.onTime->Copy()), offTime(c.offTime->Copy()),
    cbrRate(c.cbrRate), pktSize(c.pktSize),
    pendingEvent(nil), pendingOO(nil),
    residualBits(c.residualBits),
    lastStartTime(c.lastStartTime),
    maxBytes(c.maxBytes), totBytes(c.totBytes)
{
}

// Handler Methods
void OnOffApplication::Handle(Event* e, Time_t t)
{
  AppOOEvent* ev = (AppOOEvent*)e;
  switch (ev->event) {
    case AppOOEvent::SEND_PKT :
      {
        l4Proto->Send(pktSize); // Send the packet
        DEBUG0((cout << "SendOO Pkt at time " << Simulator::Now() << endl));
        lastStartTime = Simulator::Now();
        residualBits = 0;
        totBytes += pktSize;
        ScheduleNextTx();
        return;
      }
    case AppOOEvent::START_GEN :
      {
        DEBUG0((cout << "StartGen at " << Simulator::Now() << endl));
        lastStartTime = Simulator::Now();
        ScheduleNextTx();
        Time_t onInterval = onTime->Value();
        pendingOO->event = AppOOEvent::STOP_GEN;
         // Schedule the stop event
        Simulator::instance->Schedule(pendingOO, onInterval, this);
        return;
      }
    case AppOOEvent::STOP_GEN :
      {
        DEBUG0((cout << "StopGen at " << Simulator::Now() << endl));
        if (totBytes < maxBytes)
          { // Only schedule if not execeeded maxBytes
            Time_t offInterval = offTime->Value();
            pendingOO->event = AppOOEvent::START_GEN;
            // Schedule the start event
            Simulator::instance->Schedule(pendingOO, offInterval, this);
          }
        if (pendingEvent)
          {
            // Calculate residual bits since last packet sent
            residualBits += (Size_t)(cbrRate*(Simulator::Now()-lastStartTime));
            Simulator::instance->Cancel(pendingEvent);
            delete pendingEvent;
            pendingEvent = nil;
          }
        return;
      }
  }
  Application::Handle(e, t);  // May be application event
}

// Application Methods
void OnOffApplication::StartApp()    // Called at time specified by Start
{
  // Bind local endpoint if not already
  if (l4Proto->Port() == NO_PORT)
    {
      l4Proto->Bind();  // Choose any available
      l4Proto->Connect(peerIP, peerPort);
    }

  StopApp();                         // Insure no pending event
  // Schedule the start generation event
  if (!connected && l4Proto->IsTCP())
    { // Need to establish connection if TCP
      l4Proto->Connect(peerIP, peerPort);
    }
  Time_t offInterval = offTime->Value();
  DEBUG0((cout << "Scheduling initial off time " << offInterval << endl));
  pendingOO = new AppOOEvent(AppOOEvent::START_GEN);
  // Schedule the start event
  Simulator::instance->Schedule(pendingOO, offInterval, this);
}

void OnOffApplication::StopApp()     // Called at time specified by Stop
{
  if (pendingEvent)
    { // Cancel the pending transmit event
      Simulator::instance->Cancel(pendingEvent);
      delete pendingEvent;
      pendingEvent = nil;
      // Calculate residual bits since last packet sent
      residualBits += (Size_t)(cbrRate * (Simulator::Now() - lastStartTime));
    }
  if (pendingOO)
    { // Cancel the pending on/off event
      Simulator::instance->Cancel(pendingOO);
      delete pendingOO;
      pendingOO = nil;
    }
}

void OnOffApplication::AttachNode(Node* n) 
{
  l4Proto->Attach(n);
  l4Proto->AttachApplication(this);
}

Application* OnOffApplication::Copy() const
{
  return new OnOffApplication(*this);
}

#ifdef OLD_DONT_NEED
// Private Methods
void OnOffApplication::ConstructorHelper( IPAddr_t rip, PortId_t rport)
{
  l4Proto->Attach(n1);              // Attach the protocol to local node
  peerProto = l4Proto->Copy();      // Get a second protocol object for n2
  if (l4Proto->Bind() == NO_PORT)   // Bind the local l4 protocol
    {
      cout << "ON-OFF Application Local Protoocl Bind Failed!" << endl;
    }
  l4Proto->AttachApplication(this); // Attached application

  peerProto->Attach(n2);              // Attach remote to it's node
  if (peerProto->Bind() == NO_PORT)   // Bind remote l4 proto to available port
    {
      cout << "ON-OFF Application Remote Protocol Bind Failed!" << endl;
    }
  peerIP = n2->GetIPAddr();
  peerPort = peerProto->Port();
  if (l4Proto->IsTCP())
    { // Tcp endpoints, connect later and set listening endpoint
      ((TCP*)peerProto)->Listen();
    }
  else
    { // Not tcp, connect immediately
      l4Proto->Connect(peerIP, peerPort);    // Set remote node/port
      connected = true;
    }
}
#endif

void OnOffApplication::ScheduleNextTx()
{
  if (totBytes < maxBytes)
    {
      DEBUG0((cout << "SNTX pktSize " << pktSize 
              << " residual " << residualBits << endl));
      Count_t bits = pktSize * 8 - residualBits;
      Time_t nextTime = bits / cbrRate; // Time till next packet
      DEBUG0((cout << "SNTX, bits " << bits 
              << " nextTime " << nextTime << endl));
      if (!pendingEvent)
        {
          pendingEvent = new AppOOEvent(AppOOEvent::SEND_PKT);
        }
      // Schedule the packet transmit
      Simulator::instance->Schedule(pendingEvent, nextTime, this);
    }
  else
    { // All done, cancel any pending events
      StopApp();
    }
}
