/*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
                                                                                
* Permission is hereby granted for any non-commercial use of this code
                                                                                
*/
                                                                                
// Boeing MANET OSPF6D port.  Copyright 2004, The Boeing Company

// Master ospf6d object that controls all threads and state machines

#ifndef __ospf6d_application_h__
#define __ospf6d_application_h__

#include <fstream> //for printing stats
#include <list>
#include <string>
#include <sys/types.h>
#include <sys/socket.h>
#include "assert.h"

#include "common-defs.h"
#include "application.h"
#include "l4demux.h"
#include "node.h"
#include "udp.h"
#include "interface.h"

#include "timer.h" //for timers
#include "thread.h" //for threads
#include "command.h" // for struct host
#include "linklist.h" // for struct List
#include "log.h" // for struct zlog

#include "ospf6d.h"
#include "ospf6_top.h"  //for OSPF6_STAT_LENGTH

#define MAX_STRING_LENGTH 100
// Below is unused under SIM_REPLACE_SCHEDULER
#define SCHED_GRANULARITY 0.001 // seconds

class OSPF6DInstance;
class L3Protocol;

typedef std::vector<OSPF6DInstance *>  vpOSPF6Instances;
typedef std::vector<struct OSPF6InterfaceConfig *> vpIntersConfig;

struct OSPF6InterfaceConfig
{
  char *name;
  int cost;
  int HelloInterval;
  int RxmtInterval;
  int AckInterval;
  int DeadInterval;
  int priority;
  int transmit_delay;
  int instance_id;
  int ifmtu;
  char *type;
  char *flooding;
  int area;
  bool diff_hellos;
  int jitter;
  int flood_delay;
  int PushbackInterval;
  bool smart_peering;
  bool unsynch_adj;
  bool topo_reduc;
  int BackupWaitInterval;
  int TwoHopRefresh;
  int HelloRepeatCount;
  char* AdjConnectivity;
  char* LSAFullness;
  bool NonPersistentMDR;
};

class OSPF6InterfacesConfig
{
  public:
    OSPF6InterfacesConfig();
    ~OSPF6InterfacesConfig();
    void SetInterfaceCost(char *name, int cost);
    void SetInterfaceArea(char *name, IPAddr_t ip);
    void SetInterfaceJitter(char *name, int jitter_msec);
    void SetInterfaceFloodDelay(char *name, int flood_delay_msec);
    void SetInterfaceMTU(char *name,int mtu_bytes);
    void CreateConfFileInterfaces(FILE *);
    void CreateConfFileAreaAttach(FILE *);
    void SetInterfaceType(char *name, char *type);
    void SetInterfaceFlooding(char *name, char *flooding);
    void SetInterfaceHelloInterval(char *name, int hello_int);
    void SetInterfaceDeadInterval(char *name, int dead);
    void SetInterfaceRetransmitInterval(char *name, int r);
    void SetInterfaceAckInterval(char *name, int ack_int);
    void SetInterfaceDiffHellos(char *name, bool diff);
    void SetInterfacePushBackInterval(char *name, int pb_int);
    void SetInterfaceSmartPeering(char *name, bool smart);
    void SetInterfaceUnsynchAdj(char *name, bool adj);
    void SetInterfaceTopoReduc(char *name, bool tr);
    void SetInterfaceBackupWaitInterval(char *name, int bw_int);
    void SetInterfaceTwoHopRefresh(char *name, int count);
    void SetInterfaceHelloRepeatCount(char *name, int count);
    void SetInterfaceAdjConnectivity(char *name, char* conn);
    void SetInterfaceLSAFullness(char *name, char* fullness);
    void SetInterfaceNonPersistentMDR(char *name, bool nonpers);
  
  private:
    struct OSPF6InterfaceConfig *FindInterface(char *name);
    struct OSPF6InterfaceConfig *AddInterface(char *name);
    vpIntersConfig vpInterfaces;
};


/*
 *  OSPF6DEvent
*/
class OSPF6DEvent : public TimerEvent {
  public:
    // types of the OSPF6 timer events
    typedef enum 
    {
      TYPE_MAIN, 
      TYPE_FETCH, 
      TYPE_EXECUTE,
      TYPE_TIMER,
			TYPE_START,
      TYPE_NEIGHBOR_LOG
    } OSPF6Timeout_t;

    OSPF6DEvent(Event_t e);
    virtual ~OSPF6DEvent();

    struct thread* t;
};

/*
 *  OSPF6DApplication
*/
//class OSPF6DApplication : public Application, public TimerHandler {
class OSPF6DApplication : public Application {
public:
  OSPF6DApplication();
  ~OSPF6DApplication();
  OSPF6DApplication(const OSPF6DApplication&); // Copy constructor
  virtual Application* Copy() const;// Make a copy of the application

  static OSPF6DApplication& instance() {
    return (*instance_);    // general access to OSPF6DApplication
  }

  OSPF6DInstance * AddNode(Node*);
  OSPF6DInstance * AddNode(Node*, char *);

	void CollectStats(const char *name, Time_t time);
	void CollectStats(const char *name, const char *args, Time_t time);
  void LogNeighbors(double time);
	void PrintStats(int num_nodes);
	void PrintOspf6PktLine(double, double, Time_t);
	void ExtractStats(OSPF6DInstance *ospf6inst);
	
  // Method to receive a packet from layer-3 (actually, from OSPFDemux)
	//  void OSPF6DApplication::Receive(Node*, Packet*, IPAddr_t, IPAddr_t, Interface*);
  void Receive(Node*, Packet*, IPAddr_t, IPAddr_t, Interface*);
  
  OSPF6DInstance * LookupInstance(struct ospf6 *);
	
private:
  vpOSPF6Instances vpInstances;
  static OSPF6DApplication* instance_;
  std::ofstream os;             // File for statistics
  bool stats_enabled;
  std::string statfile, argstring;
  Time_t start_stat_time;
  double statistics[OSPF6_STAT_LENGTH];
  double polling_interval; // seconds
};

/*
 *  OSPF6DInstance
 */

// Define this class so it could be used in STL containers
class OSPF6DInstance : public TimerHandler {
public: 
  OSPF6DInstance(Node *n, char *file); 
  OSPF6DInstance(const OSPF6DInstance& o);
  const OSPF6DInstance& operator= (const OSPF6DInstance& right);
  ~OSPF6DInstance();
  
  void Ospf6Main();
  void Start(double delay);
	void Finalize();
  int SendMessage(int s, const struct msghdr *msg, int flags);
  void ReceiveMessage(Packet *p, IPAddr_t src, IPAddr_t dst, Interface *i);

  void SetGlobals();
  Node * GetNode() {return node;}
  void SetApp(OSPF6DApplication* a) {app = a;} 

  //Interface configuration
  void SetHostName(char *name);
  void SetRouterId(IPAddr_t ip);
  void SetLogFileName(char *name);
  void SetMinLSInterval(int ls_int);
  void SetMinLSArrival(int ls_int);
  void SetInterfaceType(Interface *i, char *type);
  void SetInterfaceFlooding(Interface *i, char *flooding);
  void SetInterfaceCost(Interface *i, int cost);
  void SetInterfaceArea(Interface *i, IPAddr_t ip);
  void SetInterfaceJitter(Interface *i, int jitter_msec);
  void SetInterfaceFloodDelay(Interface *i, int flood_delay_msec);
  void SetInterfaceMTU(Interface *i, int mtu_bytes);
  void SetInterfaceHelloInterval(Interface *i, int hello_int);
  void SetInterfaceDeadInterval(Interface *i, int dead);
  void SetInterfaceRetransmitInterval(Interface *i, int r);
  void SetInterfaceAckInterval(Interface *i, int ack_int);
  void SetInterfaceDiffHellos(Interface *i, bool diff);
  void SetInterfacePushBackInterval(Interface *i, int pb_int);
  void SetInterfaceSmartPeering(Interface *i, bool smart);
  void SetInterfaceUnsynchAdj(Interface *i, bool adj);
  void SetInterfaceTopoReduc(Interface *i, bool tr);
  void SetInterfaceBackupWaitInterval(Interface *i, int bw_int);
  void SetInterfaceTwoHopRefresh(Interface *i, int count);
  void SetInterfaceHelloRepeatCount(Interface *i, int count);
  void SetInterfaceAdjConnectivity(Interface *i, char* conn);
  void SetInterfaceLSAFullness(Interface *i, char* fullness);
  void SetInterfaceNonPersistentMDR(Interface *i, bool nonpers);

  void CreateConfFile();
	
	//stat gathering
	void SetStartStatTime(Time_t time) {start_stat_time = time;}

  //L3 Protocol
  void SetL3Proto(L3Protocol *proto) {l3Proto = proto;}

  //Timer 
  virtual void Timeout(TimerEvent*); // Timer handle method
  void ScheduleTimer(Event_t, OSPF6DEvent*&, Time_t); //Timer generate
  void ScheduleTimerWithThread(Event_t, OSPF6DEvent*&, Time_t, struct thread *);

  // Access functions
  struct ospf6* get_inst_ospf6() {return inst_ospf6; }
  IPAddr_t get_router_id() {return routerid; }

  // Neighbor logging
  void SetNeighborPollingInterval(double interval);
  void UpdateNeighborStats();
  double GetNeighborRunningAverage() {return neighbor_running_average; }
  Time_t GetNeighborRunningAverageTime() {
    return neighbor_running_average_time; }
  double GetNeighborChanges() { return neighbor_changes; }
  Time_t GetNeighborChangesTime() { return neighbor_changes_time; }
private:
  //ospf6 instances of GLOBALS 
  OSPF6DApplication *app;
  Node*         node;      // Pointer to node that this instance runs on
  struct ospf6 *inst_ospf6;  //global at ospf6_top.c
  struct thread_master *inst_master; //global at  ospf6d/ospf6_main.c
  struct host * inst_phost; //global at lib/command.c
  pVector inst_cmdvec; //global at lib/command.c
  struct List *inst_iflist;  //global at lib/if.c
  struct route_table *inst_ifaddr_ipv4_table;  //global at lib/if.c 
  pVector inst_vtyvec;  //global at lib/vty.c
  pVector inst_Vvty_serv_thread; //global at lib/vty.c
  char *inst_vty_cwd; //global at lib/vty.c
  struct zlog *inst_zlog_default; //global at lib/log.c

  //stat start time
  Time_t start_stat_time;

  //L3 Protocol
  L3Protocol*   l3Proto;  // Pointer to IPv4 static class instance
  typedef enum { ProtoNumber =  89 } OSPFProto_t; // OSPF Protocol number

  //Timer variable
  Timer          timer;           // Timer for various OSPF6 timer events
  OSPF6DEvent * FetchTimeout; 
  OSPF6DEvent * StartTimeout; 

  
  struct thread thread;    // used for scheduling loop
  std::string logfile;
  std::string config_file;
  std::string hostname;
  IPAddr_t routerid;
  int minLSInterval;
  int minLSArrival;
  bool needcreateconf;

  //Interface Configuration
  OSPF6InterfacesConfig ifc;

  //Neighbor logging
  char* neighbor_list;
  char* last_neighbor_list;
  OSPF6DEvent * NeighborLogTimeout; 
  int num_nodes;
  double neighbor_polling_interval;
  Time_t neighbor_last_time;
  double neighbor_running_average; // units of neighbor-seconds
  Time_t neighbor_running_average_time; // units of seconds
  double neighbor_changes; // units of changes
  Time_t neighbor_changes_time; // units of seconds
};
#endif //__ospf6d_application_h__
