/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

* $Id: application-ping.cc,v 1.1.1.1 2006/12/13 13:09:35 dnguyen Exp $

*/

// Georgia Tech Network Simulator - Ping Application class
// George F. Riley.  Georgia Tech, Summer 2004

// Define a Ping Application class

// Uncomment below to enable debug level 0
//#define DEBUG_MASK 0x1
#include "debug.h"
#include "application-ping.h"
#include "node.h"
#include "icmp.h"
#include "simulator.h"

using namespace std;

// Static variables
Count_t PingApplication::defaultPingCount = 0;
Rate_t  PingApplication::defaultPingRate = 10.0;

// Constructors

PingApplication::PingApplication(IPAddr_t rIP)
    : remoteIP(rIP), localNode(nil), pendingEvent(nil),
      pingCount(defaultPingCount), pingRate(defaultPingRate)
{
  ICMP::Enable(true); // Insure ICMP messages are processed
}

PingApplication::PingApplication(const PingApplication& c)
    : remoteIP(c.remoteIP), localNode(c.localNode),
      pendingEvent(nil), 
      pingCount(c.pingCount), pingRate(c.pingRate)
{
}

      
PingApplication::~PingApplication()
{
}


// Handler Methods
void PingApplication::Handle(Event* e, Time_t t)
{ // Handle  simulator event..may be SEND_PING, or for base Application class
  AppPingEvent* ev = (AppPingEvent*)e;
  switch (ev->event) {
    case AppPingEvent::SEND_PING :
      DEBUG0((cout << "PingApp " << this << " sending ping time " 
              << Simulator::Now() << endl));
      // Send the packet
      ICMPHeader* icmp = new ICMPHeader(ICMP::ECHO);
      icmp->sequence = pingSeq++;
      ICMP::Echo(localNode, icmp, remoteIP);
      // Schedule the next packet if need more
      if (pingCount)
        {
          pingCount--;
          if (pingCount == 0)
            { // No more
              delete ev;
              pendingEvent = nil;
              return;
            }
        }
      
      Time_t nextTime = 1.0 / pingRate; // Time till next packet
      // Reuse the same event
      pendingEvent = ev;
      Simulator::instance->Schedule(ev, nextTime, this); // Schedule again
      return;
  }
  Application::Handle(e, t);  // May be application event
}

// Application Methods
// Data received
void PingApplication::Receive(Packet* p, L4Protocol*, Seq_t)
{
  IPV4Header* ipv4 = (IPV4Header*)p->PeekPDU(-1);
  ICMPHeader* icmp = (ICMPHeader*)p->PeekPDU();
  switch(icmp->type) {
    case ICMP::ECHO_REPLY :
      PingReplyRx(p, ipv4, icmp);
      break;
    default:
      break; // Ignore all others
  }
}

void PingApplication::StartApp()    // Called at time specified by Start
{
  if (!localNode)
    {
      cout << "HuH?  PingApplication::StartApp() with no localNode" << endl;
      return;
    }
  
  StopApp();                               // Insure no pending event
  AppPingEvent* ev = new AppPingEvent(AppPingEvent::SEND_PING);
  pendingEvent = ev;
  Simulator::instance->Schedule(ev, 0.0, this); // Send the packet now
  // Also need to request a callback from the ICMP handler
  ICMP::AddCallback(localNode, this);
}

void PingApplication::StopApp()     // Called at time specified by Stop
{
  if (pendingEvent)
    { // Cancel the pending transmit event
      Simulator::instance->Cancel(pendingEvent);
      pendingEvent = NULL;
    }
}

void PingApplication::AttachNode(Node* n) 
{
  localNode = n;
}

Application* PingApplication::Copy() const
{
  return new PingApplication(*this);
}

void PingApplication::PingReplyRx(Packet* p, IPV4Header* ip, ICMPHeader* icmp)
{ // Should be overridden by subclasses to allow additional action
  cout << "Ping reply at node " << (string)IPAddr(ip->dst)
       << " from " << (string)IPAddr(ip->src) 
       << " sequence " << icmp->sequence
       << endl;
  delete p;
}

  

