/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Syn-Flood DDOS Attack application for GTNetS
// George F. Riley, Georgia Tech, Summer 2003

#include <iostream>

#define DEBUG_MASK 0x03
#include "debug.h"
#include "application-synflood.h"
#include "tcp.h"
#include "packet.h"
#include "ipv4.h"

using namespace std;

Count_t     SynFlood::defaultSynPerSec = 10;
PortId_t    SynFlood::defaultDstPort = 80;

SynFlood::SynFlood() : synPerSec(defaultSynPerSec), dstPort(defaultDstPort),
                       timeoutEvent(nil), node(nil),
                       rngV(nil), rngS(nil), rngP(nil), active(false)
{
}

SynFlood::SynFlood(const SynFlood& r)
  : synPerSec(r.synPerSec), dstPort(r.dstPort),
    timeoutEvent(nil), node(nil),
    rngV(nil), rngS(nil), rngP(nil), active(false)
{
}

void SynFlood::Timeout(TimerEvent* ev)
{
  if (!node)
    {
      cout << "SynFlood::Timeout with no attached node" << endl;
      return;
    }
  if (victims.size() == 0)
    {
      cout << "SynFlood::Timeout with no victims" << endl;
      return;
    }
  
  DEBUG0((cout << "SynFlood::Timeout " << Simulator::Now() << endl));
  SendSyn(false); // Send the packet, no notification

  // Schedule the next syn
  timer.Schedule(ev, 1.0 / (double)synPerSec, this);
}

void SynFlood::Notify(void*)
{ // We are sending as fast possible, just send next unless stopped
  DEBUG(1,(cout << "Got notif, active " << active << endl));
  if (active) SendSyn(true);
}

void SynFlood::StartApp()
{ // Schedule the timer for the next syn-pkt
  if (active) return;
  cout << "Starting SF attack at time " << Simulator::Now() << endl;
  active = true;

  if (synPerSec == 0)
    { // 0 rate indicates "as fast as possible
      // In this case, use a notification when each packet is transmitted
      SendSyn(true);
    }
  else
    {
      if (!timeoutEvent)
        {
           timeoutEvent = new TimerEvent();
        }
      timer.Schedule(timeoutEvent, 1.0 / (double)synPerSec, this);
    }
}

void SynFlood::StopApp()
{ // Cancel the timer if active
  if (timeoutEvent) timer.Cancel(timeoutEvent);
  cout << "Stopping SF attack at time " << Simulator::Now() << endl;
  active = false;
}

void SynFlood::AttachNode(Node* n)
{
  node = n;
}


Application* SynFlood::Copy() const
{
  return new SynFlood(*this);
}

void SynFlood::Victim(IPAddr_t ip)
{ // Add victim ip
  victims.push_back(ip);
  delete rngV; // Any existing victim RNG is not valid
  rngV = nil;
}

void SynFlood::Spoof(IPAddr_t ip)
{ // Add spoofed IP address
  spoofs.push_back(ip);
  delete rngS; // Any existing victim RNG is not valid
  rngS = nil;
}

// Private Methods
void SynFlood::SendSyn(bool notif)
{
  // Create the packet and TCP header
  Packet* p = new Packet();
  // Set notification if requested
  if (notif)
    {
      p->notification = this;
    }

  TCPHeader* h = new TCPHeader();
  if (!rngP) rngP = new Uniform(10000, 50000); // Range for random ports
  h->sourcePort = rngP->IntValue();
  h->destPort = dstPort; // Victim's port
  h->sequenceNumber = 0;
  h->ackNumber = 0;
  h->flags = 0x02; // Syn
  h->window = 0xffff;
  h->fid = 0;
  h->dataLength = 0;
  // Add this header to the packet
  p->PushPDU(h);

  // Create the layer 3 request info
  if (!rngV) rngV = new Uniform(0, victims.size());
  IPAddr_t victimIP = victims[rngV->IntValue()]; // Get random victim
  IPAddr_t spoofIP;
  if (!spoofs.empty())
    { // Use a spoof'ed ip from the list
      if (!rngS) rngS = new Uniform(0, spoofs.size());
      spoofIP = spoofs[rngS->IntValue()]; // Get random spoof
    }
  else
    {
      if (!rngS) rngS = new Uniform(0xa0000000, 0xcfffffff); // IP Range
      spoofIP = rngS->IntValue(); // Get spoofed IP address
    }
  DEBUG(1,(cout << "SynFlood app " << this
           << " sending to victim " << (string)IPAddr(victimIP) << endl));
  IPV4ReqInfo r(victimIP, spoofIP, 64, 6); // Info for l3
  // Pass to layer 3 for further processing
  IPV4::Instance()->DataRequest(node, p, &r);
}

// Static Methods
void SynFlood::DefaultSynPerSec(Count_t sps)
{
  defaultSynPerSec = sps;
}

void SynFlood::DefaultDstPort(PortId_t dp)
{
  defaultDstPort = dp;
}
