/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Syn-Flood DDOS Attack application for GTNetS
// George F. Riley, Georgia Tech, Summer 2003

#ifndef __synflood_h__
#define __synflood_h__

#include <vector>

#include "application.h"
#include "timer.h"
#include "notifier.h"
#include "rng.h"

typedef std::vector<IPAddr_t> IPAddrVec_t;     // List of ip addresses

// SynFlood is not an "application" specifically, but we subclass
// it anyway to get start/stop semantics.

//Doc:ClassXRef
class SynFlood : public Application, public TimerHandler, public NotifyHandler {
  //Doc:Class Defines an application that a initiates a Syn--Flood denial
  //Doc:Class of service attack on one or more victim servers.
public:
  //Doc:Method
  SynFlood();
    //Doc:Desc Default constructor, no arguments.

  //Doc:Method
  SynFlood(const SynFlood&);
    //Doc:Desc Copy constructor.
    //Doc:Arg1 SynFlood application object to copy.

public:
  // TimerHandler 
  void Timeout(TimerEvent*);
  // Notification
  void Notify(void*);         // Called if packet tx notification specified
  // Inherited from Application
  virtual void StartApp();        // Called at time specified by Start
  virtual void StopApp();         // Called at time specified by Stop
  virtual void AttachNode(Node*); // Note which node attached to
  virtual Application* Copy() const; // Make a copy of application
 public:
  //Doc:Method
  void Victim(IPAddr_t);
    //Doc:Desc Add an \IPA\ to the list of potential victims.  When generating
    //Doc:Desc the Syn--Flood packets, the destination address is chosen
    //Doc:Desc uniformly from all specified victims.
    //Doc:Arg1 \IPA\ of potential victim.

  //Doc:Method
  void Spoof(IPAddr_t);
    //Doc:Desc Add an \IPA\ to the list of possible spoofed source addresses.
    //Doc:Desc When generating Syn--Flood packets, the spoofed source address
    //Doc:Desc is chosen uniformly from all specified spoofed addresses.
    //Doc:Arg1 \IPA\ to add to spoof list.

  //Doc:Method
  void VictimPort(PortId_t p) { dstPort = p;}
    //Doc:Desc Specify the desired port number for the Syn--Flood packets.
    //Doc:Arg1 Port number to use.

  //Doc:Method
  void SynPerSec(Count_t sps) { synPerSec = sps; }
    //Doc:Desc Set the "syns per second" rate.
    //Doc:Arg1 Syn per Second rate.
    //Doc:Arg1 Specifying zero indicates "as fast as possible"

private:
  void SendSyn(bool);



private:
  IPAddrVec_t victims;      // List of victims
  IPAddrVec_t spoofs;       // List of spoofed IP (choose random if empty)
  Count_t     synPerSec;    // Number of syn packets per second
  PortId_t    dstPort;      // Destination Port
  Timer       timer;        // SynPkt timer
  TimerEvent* timeoutEvent; // Pending timeout event
  Node*       node;         // Attached node
  Random*     rngV;         // Random number generator for selecting victims
  Random*     rngS;         // Random number generator for spoofs
  Random*     rngP;         // Random number generator for source port
  bool        active;       // True if application started
private:
  static Count_t     defaultSynPerSec;
  static PortId_t    defaultDstPort;
public:
  //Doc:Method
  static void        DefaultSynPerSec(Count_t);
    //Doc:Desc Specify the default number of Syn packets per second
    //Doc:Desc to be generated.
    //Doc:Desc This value is assigned when the SynFlood application objects
    //Doc:Desc are created, so any SynFlood applications already created when
    //Doc:Desc this is called are unaffected.
    //Doc:Arg1 Count of Syn packet per second to use.

  //Doc:Method
  static void        DefaultDstPort(PortId_t);
    //Doc:Desc Specify the destination port to use for newly created SynFlood
    //Doc:Desc applications.
    //Doc:Arg1 Port number to use.

};

#endif
