/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - TCP Bulk Data Receive Application class
// George F. Riley.  Georgia Tech, Spring 2002

// Define a TCP Bulk Data Receive class

// Uncomment below to enable debug 
//#define DEBUG_MASK 0xff
#include "debug.h"
#include "application-tcpreceive.h"
#include "tcp.h"
#include "simulator.h"
#include "globalstats.h"

using namespace std;

// TCP Type, peerIp, peerPort, RandomSize, Random Sleep
TCPReceive::TCPReceive(IPAddr_t ip, PortId_t p, 
                       const Random& rqRV, 
                       const Random& respRV, 
                       const TCP& t,
                       const Random& sleepTime,
                       Count_t loopCount)
  : TCPApplication(t),
    peerIP(ip), peerPort(p),
    rqSizeRV(rqRV.Copy()), respSizeRV(respRV.Copy()),
    sleepRV(sleepTime.Copy()), 
    lCount(loopCount), rCount(0), pendingEvent(0),
    respReq(0), respCount(0),
    connected(false)
{

}

TCPReceive::TCPReceive(const TCPReceive& c)
  : TCPApplication(c),
    peerIP(c.peerIP), peerPort(c.peerPort),
    rqSizeRV(c.rqSizeRV->Copy()), respSizeRV(c.respSizeRV->Copy()),
    sleepRV(c.sleepRV->Copy()),
    lCount(c.lCount), rCount(c.rCount),
    pendingEvent(0),
    respReq(c.respReq), respCount(c.respCount),
    connected(c.connected)
{ // Copy constructor
}


// Handler Methods
void TCPReceive::Handle(Event* e, Time_t t)
{
  AppTCPReceiveEvent* ev = (AppTCPReceiveEvent*)e;
  switch (ev->event) {
    case AppTCPReceiveEvent::SEND_REQUEST :
      DEBUG0((cout << "Send Data Event" << endl));
      SendData();                      // Send the next data request message
      delete e;
      return;
    default :
      break;
  }
  Application::Handle(e, t);  // May be application event
}

  // Upcalls from L4 protocol
void TCPReceive::Receive(Packet* p,L4Protocol*,Seq_t)
{ // Data received, see if we have all
  respCount += p->Size(); 
  if (respCount >= respReq)
    { // Got all, check loop count and re-schedule
      if (++rCount < lCount)
        { // Need wakeup event
          AppTCPReceiveEvent* ev = 
            new AppTCPReceiveEvent(AppTCPReceiveEvent::SEND_REQUEST);
          pendingEvent = ev;
          // Schedule again    
          Time_t st = sleepRV->Value(); // Sleep time
          DEBUG0((cout << "Sleeping for " << st << " secs" << endl));
          Simulator::instance->Schedule(ev, st, this);
        }
    }
  delete p; // Delete the packet
}

void TCPReceive::CloseRequest(L4Protocol*)
{ // Close request from peer
  DEBUG0((cout << "TCPReceive::CloseRequest() " << this << endl));
  l4Proto->Close(); // Close in response
  Delete(); // Delete the application object
}

void TCPReceive::Closed(L4Protocol*)
{ // Connection has closed
  // Should we delete the TCP object here?
  DEBUG0((cout << "TCPReceive::Closed() " << this << endl));
  Delete(); // Delete the application object
}

void TCPReceive::ConnectionComplete(L4Protocol*)
{ // Connection request succeeded
  connected = true;
  DEBUG0((cout << "TCPReceive::ConnectionComplete" << endl));
  SendData(); // Connection successful, send the data
}

void TCPReceive::ConnectionFailed(L4Protocol*)
{ // Connection request failed
  // Possibly print a debug msg
  cout << "TCPReceive::ConnectionFailed" << endl;
}

// Application Methods
void TCPReceive::StartApp()    // Called at time specified by Start
{
  StopApp();                       // Insure no pending event
  SendData();                      // Send the data block
}

void TCPReceive::StopApp()     // Called at time specified by Stop
{
  if (pendingEvent)
    { // Cancel the pending transmit event
      Simulator::instance->Cancel(pendingEvent);
      delete pendingEvent;
      pendingEvent = NULL;
    }
}

void TCPReceive::AttachNode(Node* n)
{ // Note attached node
  if (l4Proto) l4Proto->Attach(n);
}

Application* TCPReceive::Copy() const
{ // Make a copy of this application
  return new TCPReceive(*this);
}

// Private Methods

void TCPReceive::SendData()
{
  if (!connected)
    {
      l4Proto->Connect(peerIP, peerPort);
      return;
    }
  Count_t rq = rqSizeRV->IntValue();// Get request size
  respReq = respSizeRV->IntValue(); // Get response size
  if (!rq) rq++;                    // Insure non-zero
  if (!respReq) respReq++;          // Insure non-zero
  Data d(rq);
  d.msgSize = rq;                   // Total size of request
  d.responseSize = respReq;         // Note response size in data message  
  respCount = 0;                    // Reset received count
  l4Proto->Send(d);                 // Send the data
}

