/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - TCP Bulk Data Transfer Application class
// George F. Riley.  Georgia Tech, Spring 2002

// Define a TCP Bulk Data class

#ifndef __tcp_application_h__
#define __tcp_application_h__

#include "common-defs.h"
#include "application-tcp.h"
#include "tcp.h"
#include "rng.h"

// Define the start/stop events
class AppTCPEvent : public Event {
public:
  typedef enum { SEND_DATA = 100 } AppTCPEvent_t;
  AppTCPEvent() { }
  AppTCPEvent(Event_t ev) : Event(ev) { }
};

//Doc:ClassXRef
class TCPSend : public TCPApplication {
  //Doc:Class Class {\tt TCPSend} creates an application that sends a
  //Doc:Class random amount of data to a {\tt TCP} server.  The application
  //Doc:Class will optionally sleep for a random amount of time and
  //Doc:Class send some more data, up to a user specified limit on the
  //Doc:Class number of sending iterations.

public:
  // Constructor
  // TCP Type, peerIp, peerPort, RandomSize, Random Sleep
  //Doc:Method
  TCPSend(IPAddr_t, PortId_t,
          const Random&,
          const TCP& = TCP::Default(), 
          const Random& sleepTime = Constant(0),
          Count_t loopCount = 1);
    //Doc:Desc Constructor for the {\tt TCPSend} application.
    //Doc:Arg1 A reference to any {\tt TCP} variant to use for the {\tt TCP}
    //Doc:Arg1 endpoint.
    //Doc:Arg2 \IPA\ of the {\tt TCPServer} to connect to.
    //Doc:Arg3 Port number of the of the {\tt TCPServer} to connect to.
    //Doc:Arg4 Random variable specifying amount of data to send each
    //Doc:Arg4 iteration.
    //Doc:Arg4 Use a {\tt Constant} random variable if a fixed size is desired.
    //Doc:Arg5 Random variable specifying the sleep time between sending
    //Doc:Arg5 iterations.
    //Doc:Arg6 Limit on the number of times the data is sent.

  //Doc:Method
  TCPSend(const TCPSend& c);     // Copy constructor
    //Doc:Desc Copy Constructor.
    //Doc:Arg1 {\tt TCPSend} object to copy.

  void Handle(Event*, Time_t);
  // Upcalls from L4 protocol
  virtual void Receive(Packet*,L4Protocol*, Seq_t = 0);// Data received
  virtual void Sent(Count_t, L4Protocol*);     // Data has been sent
  virtual void CloseRequest(L4Protocol*);      // Close request from peer
  virtual void Closed(L4Protocol*);            // Connection has closed
  virtual void ConnectionComplete(L4Protocol*);// Connection request succeeded
  virtual void ConnectionFailed(L4Protocol*);  // Connection request failed
  // Called from base application class
  virtual void StartApp();           // Called at time specified by Start
  virtual void StopApp();            // Called at time specified by Stop
  virtual Application* Copy() const; // Make a copy of the application
  // Debug
  void DBDump(); 
public:
  IPAddr_t      peerIP;       // PeerIP
  PortId_t      peerPort;     // Peer port
  Random*       sizeRV;       // Random variable for size
  Random*       sleepRV;      // Random variable for sleep time
  Count_t       lCount;       // Loop Count
  Count_t       rCount;       // Number of times repeated
  AppTCPEvent*  pendingEvent; // Pending SendData event
  Count_t       sent;         // Bytes set this loop
  Count_t       sentAck;      // Bytes acknowledged this loop
  bool          connected;    // True if already connected
  // all below is debug
  bool          ended;        // True if all data sent
  IPAddr_t      srcip;
  IPAddr_t      dstip;
  PortId_t      srcPort;
  PortId_t      dstPort;
private:
  void SendData();            // Send the (next) data block
public:
  //Doc:Member
  static Count_t totalStarted;// Number of TCP applications started
    //Doc:Desc Count of total number of {\tt TCPSend} applications started.

  //Doc:Member
  static Count_t totalEnded;  // Number of TCP applications ended
    //Doc:Desc Count of the total number of {\tt TCPSend} applications  ended.
};

#endif

