/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - TCP Server Application class
// George F. Riley.  Georgia Tech, Spring 2002

// Define a TCP Server Application Class
// Just sends back the amount of data specified in datapdu, "responseSize"

// Uncomment below to enable debug
//#define DEBUG_MASK 0x01
#include "debug.h"
#include "application-tcpserver.h"
#include "tcp.h"
#include "simulator.h"
#include "globalstats.h"

using namespace std;

// Constructors
// TCP Type
TCPServer::TCPServer(const TCP& t)
  : TCPApplication(t),
    totSent(0), totAck(0), totRx(0),
    stats(nil), msgStart(0.0), timeKnown(false)
{
  TCP* tcp = (TCP*)GetL4();
  tcp->isServer = true;
  tcp->Listen(); // Listen
}

TCPServer::TCPServer(const TCPServer& c)
  : TCPApplication(c),
    totSent(c.totSent), totAck(c.totAck), totRx(c.totRx),
    stats(c.stats), msgStart(c.msgStart), timeKnown(c.timeKnown)
{ // Copy constructor
}


  // Upcalls from L4 protocol
void TCPServer::Receive(Packet* p, L4Protocol*,Seq_t)
{ // Data received
  Data* d = (Data*)p->PopPDU();  // Get the data pdu
  totRx += d->Size();            // Log bytes received
  DEBUG0((cout << "AppTCPServer " << this << " RxBytes " << d->Size()
          << " totRx " << totRx << " msgsize " << d->msgSize << endl));
  if (!timeKnown)
    { // Note start time
      msgStart = p->time;
      timeKnown = true;
    }
  if (totRx >= d->msgSize)
    {
      totRx -= d->msgSize;       // Subtract for next message
      totSent = d->responseSize; // Number of bytes to sent     
      totAck = 0;                // Clear ack count     
      if (closeOnEmpty)
        {
          l4Proto->CloseOnEmpty(true);
        }
      DEBUG0((cout << "AppTCPServer " << this << " responding with "
              << totSent << " bytes " << endl));
      if (stats)
        { // Record response time
          stats->Record(Simulator::Now() - msgStart);
        }
      timeKnown = false;
      if (totRx)
        { // If got partial of next
          msgStart = p->time;
          timeKnown = true;
        }
      if (totSent)
        {
          DEBUG0((cout << "AppTS " << this 
                  << " Rq " << d->msgSize
                  << " Rx " << d->responseSize
                  << " peer " << ((TCP*)l4Proto)->Peer() << endl));
          l4Proto->Send(totSent);// Send the response   
        }
    }
  delete p;                      // Delete the packet
}

void TCPServer::Sent(Count_t c, L4Protocol*)
{ // Data has been sent
  DEBUG0((cout << "TCPServer " << this 
          << " sent c " << c << " totAck " << totAck
          << " totSent " << totSent << endl));
  totAck += c; // Note received
  if (totAck >= totSent)
    { // All sent, see if we close on empty
      if (closeOnEmpty)
        {
          l4Proto->Close();
        }
    }
}

void TCPServer::CloseRequest(L4Protocol*)
{ // Close request from peer
  DEBUG0((cout << "TCPServer::CloseRequest() " << endl));
  l4Proto->Close(); // Close in response
  Delete();// Delete this application
}

void TCPServer::Closed(L4Protocol*)
{ // Connection has closed
  // Should we delete the TCP object here?
  DEBUG0((cout << "TCPServer::Closed() " << this 
          << " doc " << deleteOnComplete << endl));
  Delete();// Delete this application
}

void TCPServer::BindAndListen(Node* n, PortId_t port)
{ // Note attached node and listen on specified port
  l4Proto->Attach(n);
  l4Proto->Bind(port);
  ((TCP*)l4Proto)->Listen();
}

void TCPServer::BindAndListen(PortId_t port)
{ // Listen on specified port
  Node* n = l4Proto->localNode;
  if (!n)
    {
      cout << "HuH? BindAndListen with no attached node" << endl;
      exit(1);
    }
  BindAndListen(n, port);
}

Application* TCPServer::Copy() const
{ // Make a copy of this application
  return new TCPServer(*this);
}

