/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - TCP Server Application class
// George F. Riley.  Georgia Tech, Summer 2002

// Define a TCP Server Application Class
// Just sends back the amount of data specified in datapdu, "responseSize"

#ifndef __tcp_appserver_h__
#define __tcp_appserver_h__

#include "common-defs.h"
#include "application-tcp.h"
#include "statistics.h"
#include "tcp.h"
#include "rng.h"

//Doc:ClassXRef
class TCPServer : public TCPApplication {
  //Doc:Class Class {\tt TCPServer} creates a simple model of a
  //Doc:Class request/response
  //Doc:Class {\tt TCP} server.  The server binds to a specified port, and
  //Doc:Class listens for connection requests from {\tt TCP} peers.  The 
  //Doc:Class data received from the peers specifies how much data to send
  //Doc:Class or receive.

public:
  // Constructor
  //Doc:Method
  TCPServer(const TCP& = TCP::Default());   // TCP Type
    //Doc:Desc Constructor for {\tt TCPServer} objects.  Optionally
    //Doc:Desc specify a reference to any {\tt TCP} variant to use
    //Doc:Desc for the listening {\tt TCP} endpoint.
    //Doc:Arg1 A reference to any {\tt TCP} variant to use for
    //Doc:Arg1 the listening endpoint.

  //Doc:Method
  TCPServer(const TCPServer& c);               // Copy constructor
    //Doc:Desc Copy constructor.
    //Doc:Arg1 {\tt TCPServer} object to copy.

  // Upcalls from L4 protocol
  virtual void Receive(Packet*,L4Protocol*,Seq_t);   // Data received
  virtual void Sent(Count_t, L4Protocol*);     // Data has been sent
  virtual void CloseRequest(L4Protocol*);      // Close request from peer
  virtual void Closed(L4Protocol*);            // Connection has closed

  // Specific to TCPServer app
  //Doc:Method
  virtual void BindAndListen(Node*, PortId_t); // Bind to node/port
    //Doc:Desc Bind the application to the specified port and start 
    //Doc:Desc listening for connections.
    //Doc:Arg1 Node pointer to the node this application is associated with.
    //Doc:Arg2 Port number to bind to.

  //Doc:Method
  virtual void BindAndListen(PortId_t); // Bind to node/port
    //Doc:Desc Bind the application to the specified port and start 
    //Doc:Desc listening for connections.  Node is already known.
    //Doc:Arg1 Port number to bind to.

  virtual Application* Copy() const;           // Make a copy of application
  //Doc:Method
  void SetStatistics(Statistics* s) { stats = s;}
    //Doc:Desc Set a statistics object to use for statistics collection.
    //Doc:Desc This application will record the response time of each
    //Doc:Desc message sent to the statistics object.
    //Doc:Arg1 Pointer to a statistics object.

public:
  //Doc:Member
  Count_t       totSent;      // Total bytes sent
    //Doc:Desc Count of the total number of bytes sent by this application.

  //Doc:Member
  Count_t       totAck;       // Total bytes acknowledged
    //Doc:Desc Count of total bytes acknowledged by this application.

  //Doc:Member
  Count_t       totRx;        // Total bytes received
    //Doc:Desc Count of total bytes received by this application.
  Statistics*   stats;        // Stats object for logging response time

  //Doc:Member
  Time_t        msgStart;     // Time message request started
    //Doc:Desc Time the last message started transmission or receipt.

  //Doc:Member
  bool          timeKnown;    // True if above time is known
    //Doc:Desc True if the start time of the message is known.
};

#endif

