/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

* $Id: application-traceroute.cc,v 1.1.1.1 2006/12/13 13:09:35 dnguyen Exp $

*/

// Georgia Tech Network Simulator - Traceroute Application class
// George F. Riley.  Georgia Tech, Summer 2004

// Define a Traceroute Application class

// Uncomment below to enable debug level 0
//#define DEBUG_MASK 0x1
#include "debug.h"
#include "application-traceroute.h"
#include "node.h"
#include "icmp.h"
#include "udp.h"
#include "simulator.h"

using namespace std;

// Constructors

TracerouteApplication::TracerouteApplication(IPAddr_t rIP)
    : remoteIP(rIP), localNode(nil), pendingEvent(nil),
      hopCount(0), retryCount(0), probeCount(3), maxTTL(30),
      timeSent(0), respondIP(IPADDR_NONE), done(false),
      timeout(1.0)
{
  ICMP::Enable(true); // Insure ICMP messages are processed
}

TracerouteApplication::TracerouteApplication(const TracerouteApplication& c)
    : remoteIP(c.remoteIP), localNode(c.localNode),
      pendingEvent(nil), hopCount(c.hopCount), retryCount(c.retryCount),
      probeCount(c.probeCount), maxTTL(c.maxTTL), timeSent(c.timeSent),
      respondIP(c.respondIP), done(false), timeout(c.timeout)
{
}

      
TracerouteApplication::~TracerouteApplication()
{
}


// Handler Methods
void TracerouteApplication::Handle(Event* e, Time_t t)
{ // Handle  simulator event..may be SEND_TRACE, or for base Application class
  AppTracerouteEvent* ev = (AppTracerouteEvent*)e;
  switch (ev->event) {
    case AppTracerouteEvent::SEND_TRACE :
      DEBUG0((cout << "TracerouteApp " << this << " timeout at " 
              << Simulator::Now() << endl));
      TracerouteReplyRx(nil, nil, nil); // Notify user of timeout
      // This one has timed out, so use infinite time
      rttHist.push_back(INFINITE_TIME);
      SendProbePacket();
      return;
  }
  Application::Handle(e, t);  // May be application event
}

// Application Methods
// Data received
void TracerouteApplication::Receive(Packet* p, L4Protocol*, Seq_t)
{
  IPV4Header* ipv4 = (IPV4Header*)p->PeekPDU(-1);
  ICMPHeader* icmp = (ICMPHeader*)p->PeekPDU();
  switch(icmp->type) {
    case ICMP::DESTINATION_UNREACHABLE:
      // This means the target found the UDP packet, but no port
      done = true;
      // Fall through to continue
    case ICMP::TIME_EXCEEDED :
      respondIP = ipv4->src;
      // Log the RTT
      rttHist.push_back(Simulator::Now() - timeSent);
      TracerouteReplyRx(p, ipv4, icmp);
      // Cancel any pending timeout event, as we received the reply
      if (pendingEvent) Simulator::instance->Cancel(pendingEvent);
      SendProbePacket();
      return; // Return so we don't delete packet below
    default:
      break; // Ignore all others
  }
  delete p;
}

void TracerouteApplication::StartApp()    // Called at time specified by Start
{
  if (!localNode)
    {
      cout << "HuH?  TracerouteApplication::StartApp() with no localNode"
           << endl;
      return;
    }
  
  StopApp();         // Insure no pending event
  hopCount = 1;      // Start with TTL = 1
  retryCount = 0;    // No retries so far
  done = false;      // Not done
  cout << "Starting traceroute from " << (string)IPAddr(localNode->GetIPAddr())
       << " to " << (string)IPAddr(remoteIP)
       << endl;
  SendProbePacket(); // Send the first trace packet
  // Also need to request a callback from the ICMP handler
  ICMP::AddCallback(localNode, this);
}

void TracerouteApplication::StopApp()     // Called at time specified by Stop
{
  if (pendingEvent)
    { // Cancel the pending transmit event
      Simulator::instance->Cancel(pendingEvent);
      pendingEvent = NULL;
    }
}

void TracerouteApplication::AttachNode(Node* n) 
{
  localNode = n;
}

Application* TracerouteApplication::Copy() const
{
  return new TracerouteApplication(*this);
}

void TracerouteApplication::TracerouteReplyRx(
    Packet* p,
    IPV4Header* ip, ICMPHeader* icmp)
{ // Should be overridden by subclasses to allow additional action
  if (p) delete p;
}

void TracerouteApplication::SendProbePacket()
{
  if (retryCount++ == probeCount)
    { // Done with this TTL, print results and advance
      PrintResultsLine();
      rttHist.clear();
      respondIP = IPADDR_NONE;
      retryCount = 1;
      if (++hopCount == maxTTL || done)
        { // Max TTL value, or target responded
          if (pendingEvent) delete pendingEvent;
          pendingEvent = nil;
          cout << "Traceroute from " << (string)IPAddr(localNode->GetIPAddr())
               << " to " << (string)IPAddr(remoteIP)
               << " complete." << endl;
          return;
        }
    }
  // Create a UDP header, push on a packet, and send using IPV4
  // Using dest port 0x4355, just randomly hoping not used at destination
  UDPHeader* udp = new UDPHeader(0, 0x4355, 0, 0);
  Packet* p = new Packet();
  localNode->TracePDU(nil, udp, p, "-");
  p->PushPDU(udp);
  IPV4ReqInfo info(remoteIP, IPADDR_NONE, hopCount, udp->Proto());
  IPV4::Instance()->DataRequest(localNode, p, &info);
  timeSent = Simulator::Now();
  // Schedule a timeout event
  if (!pendingEvent) pendingEvent = 
      new AppTracerouteEvent(AppTracerouteEvent::SEND_TRACE);
  Simulator::instance->Schedule(pendingEvent, timeout, this); // Packet timeout
}

void TracerouteApplication::PrintResultsLine()
{
  cout << hopCount << " ";
  if (respondIP != IPADDR_NONE)
    cout << (string)IPAddr(respondIP) << " ";
  Count_t ms;
  for (TimeVec_t::size_type i = 0; i < rttHist.size(); ++i)
    {
      if (rttHist[i] == INFINITE_TIME)
        {
          cout << " *";
        }
      else
        {
          ms = (Count_t)(rttHist[i] * 1000.0);
          cout << " " << ms << " ms";
        }
    }
  cout << endl;
}

  
  
  
          
          
      

