/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// UDPStorm DDOS Attack application for GTNetS
// George F. Riley, Georgia Tech, Summer 2003

#ifndef __udpstorm_h__
#define __udpstorm_h__

#include <vector>

#include "application.h"
#include "timer.h"
#include "rng.h"

typedef std::vector<IPAddr_t> IPAddrVec_t;     // List of ip addresses

//Doc:ClassXRef
class UDPStorm : public Application, public TimerHandler {
  //Doc:Class Simulation model for a {\em UDP Storm} denial of service attack.
  //Doc:Class Each object of class {\tt UDPStorm} generates {\tt UDP} packets
  //Doc:Class at a specified rate, and addrsses them to an \IPA\ sampled
  //Doc:Class from a victim list specified by the user.  The destination
  //Doc:Class port is chosen randomly from a uniform distribution between
  //Doc:Class 10000 and 50000.  The source address is spoofed from a list
  //Doc:Class of \IPA{s} specified by the user.
public:
  //Doc:Method
  UDPStorm();
    //Doc:Desc Default constructor, no arguments.

  //Doc:Method
  UDPStorm(const UDPStorm&);
    //Doc:Desc Copy constructor.
    //Doc:Arg1 {\tt UDPStorm} object to copy.

public:
  // TimerHandler 
  void Timeout(TimerEvent*);
  // Inherited from Application
  virtual void StartApp();        // Called at time specified by Start
  virtual void StopApp();         // Called at time specified by Stop
  virtual void AttachNode(Node*); // Note which node attached to
  virtual Application* Copy() const; // Make a copy of application
 public:
  //Doc:Method
  void Victim(IPAddr_t);
    //Doc:Desc Add an \IPA\ to the list of potential victims.
    //Doc:Arg1 \IPA\ to add.

  //Doc:Method
  void Spoof(IPAddr_t);
    //Doc:Desc Add an \IPA\ to the list of spoofed source addresses.
    //Doc:Arg1 \IPA\ to add.

  void UseSpoofing(bool us) { useSpoofing = us; }

  void PacketsPerSecond(Count_t pkps) { pktsPerSec= pkps; }
  void PacketLength(Size_t pkl) { pktLength = pkl; }

public:
  IPAddrVec_t victims;      // List of victims
  IPAddrVec_t spoofs;       // List of spoofed IP (choose random if empty)
  Count_t     pktsPerSec;   // Number of UDP packets per second
  Size_t      pktLength;    // Length of storm packets
  bool        useSpoofing;  // True if using spoofing
  Timer       timer;        // Pkt timer
  TimerEvent* timeoutEvent; // Pending timeout event
  Node*       node;         // Attached node
  Random*     rngV;         // Random number generator for selecting victims
  Random*     rngS;         // Random number generator for spoofs
  Random*     rngP;         // Random number generator for port
private:
  static Count_t     defaultPktPerSec;
  static Count_t     defaultPktLength;
public:
  //Doc:Method
  static void        DefaultPktPerSec(Count_t);
    //Doc:Desc Specify the default packets per second to be generated by all
    //Doc:Desc {\tt UDPStorm} objects.
    //Doc:Arg1 Count of packets per second to generate.

  //Doc:Method
  static void        DefaultPktLength(Count_t);
    //Doc:Desc Specify the default length of the {\tt UDP} packets generated
    //Doc:Desc by the {\tt UDPStorm} objects.
    //Doc:Arg1 Size of {\tt UDP} packets.
};

#endif
