/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Web Server Application class
// George F. Riley.  Georgia Tech, Spring 2003

// Define a Web Server Application Class
// Responds to HTTP-style web GET commands.  

#include <stdio.h>
#include <map>

// Uncomment below to enable debug
//#define DEBUG_MASK 0x02
#include "debug.h"
#include "application-webserver.h"
#include "tcp.h"
#include "simulator.h"
#include "globalstats.h"

using namespace std;

// Static variables
PortId_t    WebServer::port = 80; // Port number to bind to
Count_t     WebServer::defaultMaxConnections = 128; // Default Max connections
Count_t     WebServer::defaultMaxIP          = 20;  // Default Max connections
Count_t     WebServer::defaultMaxURL         = 20;  // Default Max connections

// Constructors
// TCP Type
WebServer::WebServer(const TCP& t)
  : TCPApplication(t), 
    node(nil), gCache(false), logEnabled(false),
    maxConnections(defaultMaxConnections),
    maxIP(defaultMaxIP), maxURL(defaultMaxURL)
{
  CopyOnConnect(false); // Do not copy this application on connect requests
  // Debug only..poplulate the ipaddrs table
  gcIPAddrs.push_back(IPPort_t(IPAddr("192.168.100.100"), 6347));
}

WebServer::WebServer(const WebServer& c)
  : TCPApplication(c), 
    node(c.node), gCache(c.gCache), logEnabled(c.logEnabled),
    maxConnections(c.maxConnections), 
    maxIP(c.maxIP), maxURL(c.maxURL)
{ // Copy constructor
}

void WebServer::AttachNode(Node* n)
{
  node = n;
  // Inform l4protocol of node, bind to port, and listen for connections
  l4Proto->Attach(n);
  l4Proto->Bind(port);
  DEBUG0((cout << "Node " << (string)IPAddr(n->GetIPAddr())
          << " binding to port " << port 
          << " tcp " << l4Proto
          << endl));
  l4Proto->AttachApplication(this);
  ((TCP*)l4Proto)->Listen();
  ((TCP*)l4Proto)->CloseOnEmpty();
}

// Upcalls from L4 protocol
void WebServer::Receive(Packet* p, L4Protocol* l4, Seq_t)
{ // Data received
  // Find the HTTP info object
  DEBUG0((cout << "WebServer::Receive, " << this << " l4 " << l4 << endl));
  L4Map_t::iterator i = activeConnections.find(l4);
  if (i == activeConnections.end())
    {
      cout << "HuH? WebServer::Receive can't find active connection" << endl;
      return;
    }

  Data* d = (Data*)p->PeekPDU();  // Get the data pdu
  if (d->Contents())
    { // Actual HTTP Request
      DEBUG0((cout << "WebServer::Receive, dataSize " << d->Size()
              << " Contents " << d->Contents() << endl));
      ProcessHTTPRequest(i, p, l4);
      return;
    }

  // Just the dummy request type
  TCP* t = (TCP*)i->first;
  i->second.totRx += d->Size();          // Log bytes received
  DEBUG0((cout << "AppWebServer " << this << " RxBytes " << d->Size()
          << " totRx " << i->second.totRx << " msgsize " << d->msgSize <<
          endl));
  if (!i->second.timeKnown)
    { // Note start time
      i->second.msgStart = p->time;
      i->second.timeKnown = true;
    }
  if (i->second.totRx >= d->msgSize)
    {
      i->second.totRx -= d->msgSize;       // Subtract for next message
      i->second.totSent = d->responseSize; // Number of bytes to sent     
      i->second.totAck = 0;                // Clear ack count     
      DEBUG0((cout << "AppWebServer " << this << " responding with "
              << i->second.totSent << " bytes " << endl));
      if (i->second.stats)
        { // Record response time
          i->second.stats->Record(Simulator::Now() - i->second.msgStart);
        }
      i->second.timeKnown = false;
      if (i->second.totRx)
        { // If got partial of next
          i->second.msgStart = p->time;
          i->second.timeKnown = true;
        }
      if (i->second.totSent)
        {
          DEBUG0((cout << "AppTS " << this 
                  << " Rq " << d->msgSize
                  << " Rx " << d->responseSize
                  << " peer " << t->Peer() << endl));
          t->Send(i->second.totSent);// Send the response   
        }
    }
  delete p;                      // Delete the packet
}

void WebServer::Sent(Count_t c, L4Protocol* l4)
{ // Data has been sent
  L4Map_t::iterator i = activeConnections.find(l4);
  if (i == activeConnections.end())
    {
      cout << "HuH? WebServer::Sent " << this 
           << " can't find connection " << l4 << endl;
      cout << "acCount " << activeConnections.size() << endl;
      return;
    }
  HTTPInfo& hInfo = i->second; // Get the http info record
  hInfo.totAck += c; // Note received
  DEBUG0((cout << "WebServer " << this 
          << " sent c " << c << " totAck " << hInfo.totAck
          << " totSent " << hInfo.totSent << endl));
  if (hInfo.totAck >= hInfo.totSent)
    { // All sent, close
      DEBUG0((cout << "WebServer closing connection" << endl));
      l4->Close();
    }
}

void WebServer::CloseRequest(L4Protocol* l4)
{ // Close request from peer
  DEBUG0((cout << "WebServer::CloseRequest() " << this 
          << " l4 " << l4 
          << " time " << Simulator::Now()
          << endl));
  l4->Close(); // Close in response
  // Remove from the map of open connections
  L4Map_t::iterator i = activeConnections.find(l4);
  if (i == activeConnections.end())
    {
      cout << "HuH? WebServer::CloseRequest can't find connection" << endl;
      return;
    }
  activeConnections.erase(i); // Remove from the list
  LogTimeCount();
}

void WebServer::Closed(L4Protocol* l4)
{ // Connection has closed
  // Remove this connectionn from pending
  DEBUG0((cout << "WebServer::Closed() " << this 
          << " l4 " << l4 << endl));
  // Remove from the map of open connections
  L4Map_t::iterator i = activeConnections.find(l4);
  if (i == activeConnections.end())
    {
      cout << "HuH? WebServer::CloseRequest can't find connection" << endl;
      return;
    }
  activeConnections.erase(i); // Remove from the list
  LogTimeCount();
}

Application* WebServer::Copy() const
{ // Make a copy of this application
  return new WebServer(*this);
}

void WebServer::ConnectionFailed(L4Protocol* l4)
{ // The connection did not complete 3-way handshake
  // Find in the active list and remove
  L4Map_t::iterator i = activeConnections.find(l4);
  if (i == activeConnections.end()) return; // Nothing to do
  // Found it
  L4Protocol* t = i->first;
  // Unbind and delete
  DEBUG(1,(cout << "AppWS unbinding" << endl));
  t->localNode->Unbind(t->ProtocolNumber(), t->localPort, IPADDR_NONE, 
            t->peerPort, t->peerIP, t);
  activeConnections.erase(i);
  LogTimeCount();
  DEBUG(1,(cout << "Deleted failed connection, tcp " << t 
           << " count " << activeConnections.size()
           << endl));
  delete t;
}

bool WebServer::ConnectionFromPeer(L4Protocol* l4, IPAddr_t, PortId_t)
{ // Listener has connection request from peer
  // First see if we can accept more connections
  DEBUG(1, (cout << "WebServer::ConnectionFromPeer, ac.size " 
            << activeConnections.size() << " maxC " << maxConnections 
            << " tcp " << l4
            << " time " << Simulator::Now()
            << endl));
  if (activeConnections.size() == maxConnections) return false;
  // Next double check that specified is not already active (should not happen)
  L4Map_t::iterator i = activeConnections.find(l4);
  if (i != activeConnections.end())
    { // ?
      cout << "HuH?  WebServer::ConnectFromPeer already active" << endl;
      activeConnections.erase(i); // Erase it and continue
      LogTimeCount();
    }
  TCP* t = (TCP*)l4;
  t->DeleteOnTWait(); // Delete this endpoint when done
  activeConnections[l4] = HTTPInfo(); // Add a new entry
  LogTimeCount();
  return true;
}

void WebServer::EnableGCache(bool egc)
{
  gCache = egc;
}

void WebServer::EnableLog(bool el)
{ // Enable the logging of table size vs. time
  logEnabled = el;
}

void WebServer::MaxConnections(Count_t m)
{ // Set a limit on number of simulataneous connectoins
  maxConnections = m;
}

void WebServer::PrintTimeCount(std::ostream& os, char sep )
{  // Print the Time/Count history
  for (TimeCount_t::size_type i = 0; i < timeCountLog.size(); ++i)
    {
      os << timeCountLog[i].time << sep << timeCountLog[i].count << endl;
    }
}

void WebServer::MaxIP(Count_t m)
{ // Set a limit on number of IP's in the GCache
  maxIP = m;
}

void WebServer::MaxURL(Count_t m)
{ // Set a limit on number of URL's in the GCache
  maxURL = m;
}

// Private member functions
void  WebServer::ProcessHTTPRequest(
          L4Map_t::iterator i,
          Packet* p, L4Protocol* l4)
{
  Data* d = (Data*)p->PeekPDU();  // Get the data pdu
  string r(d->Contents());        // Get the request
  delete p;                       // Done with the packet
  DEBUG0((cout << "WebServer::ProcessHTTPRequest " << r << endl));
  if (gCache)
    { // Check for the gcache commands
      if (r.find("gcache.php?") == 0)
        { // Found a gcache command
          string s = r.substr(11, string::npos); // the command
          string::size_type j;
          if (s.find("hostfile=1") == 0)
            { // Found host file request
              cout << "Found hostfile request " << r << endl;
              SendHostFile(i, l4);
            }
          if (s.find("urlfile=1") == 0)
            { // Found url file request
              cout << "Found urlfile request " << r << endl;
              SendURLFile(i, l4);
            }
          j = s.find("ip=");
          if (j != string::npos)
            { // Found ip update request
              cout << "Found ip update request " << r << endl;
              j = s.find(':');
              string ipa = s.substr(3, j - 3);
              string port = s.substr(j+1, string::npos);
              gcIPAddrs.push_back(
                  IPPort_t(IPAddr(ipa.c_str()),
                           atol(port.c_str())));
              cout << "Added " << ipa << ":" << port << " to cache" << endl;
              if (gcIPAddrs.size() > maxIP)
                { // Too many, remove one
                  gcIPAddrs.pop_front();
                }
            }
          j = s.find("url=");
          if (j != string::npos)
            { // Found url update erquest
            }
        }
      return; // Done
    }
  // Check HTTP Requests
}

void  WebServer::SendHostFile(L4Map_t::iterator i, L4Protocol* l4)
{
  Data d;
  deque<IPPort_t>::iterator j;
  if (!gcIPAddrs.size())
    { // deubg...just add one
      gcIPAddrs.push_back(IPPort_t(IPAddr("192.168.100.100"), 6347));
    }
  for (j = gcIPAddrs.begin(); j != gcIPAddrs.end(); ++j)
    { // Send each ip address
      string ip(IPAddr(j->first));
      d.Add(ip.length(), ip.c_str());
      d.Add(1, ":");
      char work[128];
      sprintf(work, "%ld", j->second);
      d.Add(strlen(work), work);
      d.Add(1, "\n");
    }
  d.Add(1, "\n");
  d.Add(1, "\0");
  HTTPInfo& hInfo = i->second; // Get the http info record
  hInfo.totSent = d.Size();
  hInfo.totAck = 0;
  l4->Send(d); // Send the data
}

void  WebServer::SendURLFile(L4Map_t::iterator i, L4Protocol* l4)
{
}

void WebServer::LogTimeCount()
{
  if (logEnabled)
    {
      timeCountLog.push_back(
          TimeCount(Simulator::Now(), 
                    activeConnections.size()));
    }
}

// Static member functions
void  WebServer::DefaultMaxConnections(Count_t m) // Set maximum connections
{
  defaultMaxConnections = m;
}

void  WebServer::DefaultMaxIP(Count_t m) // Set maximum IP GCache
{
  defaultMaxIP = m;
}

void  WebServer::DefaultMaxURL(Count_t m) // Set maximum URL GCache
{
  defaultMaxURL = m;
}


