/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Web Server Application class
// George F. Riley.  Georgia Tech, Spring 2003

// Define a Web Server Application Class
// Responds to HTTP-style web GET commands.  

#ifndef __webserver_h__
#define __webserver_h__

#include <map>
#include <deque>
#include <vector>

#include "common-defs.h"
#include "application-tcp.h"
#include "statistics.h"
#include "tcp.h"
#include "rng.h"

// Define a class to track child TCP processes
//Doc:ClassXRef
class HTTPInfo {
public:
  HTTPInfo() : totSent(0), totAck(0), totRx(0), stats(nil),
               msgStart(0), timeKnown(false) { }
public:
  Count_t       totSent;      // Total bytes sent
  Count_t       totAck;       // Total bytes acknowledged
  Count_t       totRx;        // Total bytes received
  Statistics*   stats;        // Stats object for logging response time
  Time_t        msgStart;     // Time message request started
  bool          timeKnown;    // True if above time is known
};

// Keep track of open connections using protocol pointer -> info map
typedef std::map<L4Protocol*, HTTPInfo> L4Map_t;

// Class TimeCount is  used to log connection table size vs. time
class TimeCount {
public:
  TimeCount(Time_t t, Count_t s) : time(t), count(s) { };
  TimeCount(const TimeCount& c) : time(c.time), count(c.count) { }
public:
  Time_t   time;
  Count_t  count;
};
typedef std::vector<TimeCount>    TimeCount_t;    // Time/sequence logs

//Doc:ClassXRef
class WebServer : public TCPApplication {
  //Doc:Class Class {\tt WebServer} defines the behavior of a web server
  //Doc:Class application, including the Gnutella {\em GCache} scripts.
public:
  // Constructor
  //Doc:Method
  WebServer(const TCP& = TCP::Default());   // TCP Type
    //Doc:Desc Constructor for {\tt WebServer} object.  Single (optional)
    //Doc:Desc argument is a reference to a {\tt TCP} object to use
    //Doc:Desc for the listening {\tt TCP} endponit.
    //Doc:Arg1 Reference to a {\tt TCP} object to copy for listening endpoint.

  //Doc:Method
  WebServer(const WebServer& c);               // Copy constructor
    //Doc:Desc Copy constructor
    //Doc:Arg1 Object to copy

  // From application.h
  virtual void AttachNode(Node*);              // Note which node attached to
  // Upcalls from L4 protocol
  virtual void Receive(Packet*,L4Protocol*,Seq_t); // Data received
  virtual void Sent(Count_t, L4Protocol*);     // Data has been sent
  virtual void CloseRequest(L4Protocol*);      // Close request from peer
  virtual void Closed(L4Protocol*);            // Connection has closed
  virtual void ConnectionFailed(L4Protocol*);  // Connection request failed
  // Listener has cn rq from peer 
  virtual bool ConnectionFromPeer(L4Protocol*, IPAddr_t, PortId_t);
  virtual Application* Copy() const;           // Make a copy of application
  // Specific to WebServer objects
  //Doc:Method
  void    EnableGCache(bool egc = true);       // Enable Gnutalla Cache
    //Doc:Desc Specify enabling or disabling of the Gnutella GCache
    //Doc:Desc processing at this web server.
    //Doc:Arg1 True if GCache processing enabled.

  //Doc:Method
  void    EnableLog(bool el = true);           // Enable Time/Count log
    //Doc:Desc Specify enabling or disabling of the connection count versus
    //Doc:Desc time history logging.
    //Doc:Arg1 True if count vs. time logging desired.

  //Doc:Method
  void    MaxConnections(Count_t m);           // Limit number of connections
    //Doc:Desc Specify a limit on the maximum number of simulataneous
    //Doc:Desc connections that can be open for this server.
    //Doc:Arg1 Connection limit.

  //Doc:Method
  void    PrintTimeCount(std::ostream&, char sep = ' ' ); // Print the T/C
    //Doc:Desc Print the count versus time history log on an output stream.
    //Doc:Arg1 Output stream on which to log the count versus time history.
    //Doc:Arg2 Separator character to use.

  Node*   AttachedNode() { return node;}       // Return the node
  // GCache methods
  //Doc:Method
  void    MaxIP(Count_t m);                    // Limit number of IP's
    //Doc:Desc Specify the maximum number of \IPA{s} to log in the GCache.
    //Doc:Desc If more than this limit are posted, the earlier ones roll off.
    //Doc:Arg1 Limit on the number of posted \IPA{s}

  //Doc:Method
  void    MaxURL(Count_t m);                   // Limit number of URL's
    //Doc:Desc Specify the maximum number of URLs to log in the GCache.
    //Doc:Desc If more than this limit are posted, the earlier ones roll off.
    //Doc:Arg1 Limit on the number of posted URLs

  // Private methods
  void ProcessHTTPRequest(L4Map_t::iterator, Packet*, L4Protocol*);
  void SendHostFile(L4Map_t::iterator, L4Protocol*);
  void SendURLFile(L4Map_t::iterator, L4Protocol*);
  void LogTimeCount();                         // Log the time and count
  // Static Methods
  //Doc:Method
  static void  Port(PortId_t p) { port = p;};  // Allow non-standard port
    //Doc:Desc Specify a non--default port to use for newly created 
    //Doc:Desc {\tt WebServer} objects.
    //Doc:Arg1 Port number.

  //Doc:Method
  static void  DefaultMaxConnections(Count_t m); // Set maximum connections
    //Doc:Desc Specify a default limit on the maximum number of simulataneous
    //Doc:Desc connections for newly created  {\tt WebServer} objects.
    //Doc:Arg1 Connection limit.

  //Doc:Method
  static void  DefaultMaxIP(Count_t m);        // Set default IP limit
    //Doc:Desc Specify a default limit on the number of GCache posted \IPA{s}
    //Doc:Desc for newly created  {\tt WebServer} objects.
    //Doc:Arg1 GCache {IPA} limit.

  //Doc:Method
  static void  DefaultMaxURL(Count_t m);       // Set default URL Limit
    //Doc:Desc Specify a default limit on the number of GCache posted URLs
    //Doc:Desc for newly created  {\tt WebServer} objects.
    //Doc:Arg1 GCache URL limit.

private:
  Node*              node;                     // Attached node
  bool               gCache;                   // True if GCache enabled
  bool               logEnabled;               // Enable log of Time/Count
  Count_t            maxConnections;           // Max number concurrent conn.
  L4Map_t            activeConnections;        // Map of current connections
  TimeCount_t        timeCountLog;             // Vector of time/count entries
  // GCache variables
  std::deque<IPPort_t> gcIPAddrs;              // Deque of ip/ports
  Count_t            maxIP;                    // Maximum number of IP's
  Count_t            maxURL;                   // Maximum number of URL's
  // Static members
  static PortId_t    port;                     // Port number to bind to
  static Count_t     defaultMaxConnections;    // Default Max connections
  static Count_t     defaultMaxIP;             // Default Max IP Count
  static Count_t     defaultMaxURL;            // Default Max URL Count
};

#endif

