/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Layer 5 Application base class
// George F. Riley.  Georgia Tech, Spring 2002

// Define the layer 5 application interface

#ifndef __application_h__
#define __application_h__

#include "common-defs.h"
#include "handler.h"
#include "l4protocol.h"

//Doc:ClassXRef
class Application : public Handler, public Object  {
  //Doc:Class Class {\tt Application} is the base class for all
  //Doc:Class \GTNS\ applications.
  //Doc:Class It defines the interface between the application class and
  //Doc:Class the associated layer 4 protocols.  Applications can have one
  //Doc:Class or more layer 4 protocols assigned, to allow (for example)
  //Doc:Class a web browser model with multiple simulataneous connections.
public:
  //Doc:Method
  Application() : deleteOnComplete(false), copyOnConnect(true) { };
    //Doc:Desc The default constructor for the {\tt Application} class.

  //Doc:Method
  Application(const Application&);
    //Doc:Desc Copy constructor, used bt the {\tt Copy()} method for all
    //Doc:Desc applications.
    //Doc:Arg1 Application object to copy.

  virtual ~Application();
  // Starting, Stoping
  //Doc:Method
  void Start(Time_t);
    //Doc:Desc Schedule a {\tt Start} event for this application at the
    //Doc:Desc specified time.
    //Doc:Arg1 The simulation time to start the application.
    //Doc:Arg1 The {\t StartApp} method of the application will be called
    //Doc:Arg1 at the specified time.

  //Doc:Method
  void Stop(Time_t);
    //Doc:Desc Schedule a {\tt Stop} event for this application at the
    //Doc:Desc specified time.
    //Doc:Arg1 The simulation time to stop the application.
    //Doc:Arg1 The {\t StopApp} method of the application will be called
    //Doc:Arg1 at the specified time.

  // Handler functions
  void Handle(Event*, Time_t);

  // Upcalls from L4 protocol
  //Doc:Method
  virtual void Receive(Packet*,L4Protocol*, Seq_t = 0);   // Data received
    //Doc:Desc Called by an associated layer 4 protocol when data 
    //Doc:Desc is received.
    //Doc:Arg1 The packet received, with the layer 4 PDU removed.
    //Doc:Arg2 A pointer to the layer 4 protocol object that recieved the data.
    //Doc:Arg3 Optional sequence number for this data (TCP protocols only)

  //Doc:Method
  virtual void Sent(Count_t, L4Protocol*);     // Data has been sent
    //Doc:Desc Called by an associated layer 4 protocol when all some part 
    //Doc:Desc of the outstanding data has been sent.  For TCP protocols,
    //Doc:Desc this occurs when the acknowledgement is received from the peer.
    //Doc:Arg1 Count of number of bytes successfully sent.
    //Doc:Arg2 A pointer to the layer 4 protocol object that sent the data.

  //Doc:Method
  virtual void CloseRequest(L4Protocol*);      // Close request from peer
    //Doc:Desc Called by an associated layer 4 protocol when a connection
    //Doc:Desc close request has been receivd from a peer.  Applications should
    //Doc:Desc respond by calling the corresponding layer 4 ::Close() routine.
    //Doc:Arg1 A pointer to the layer 4 protocol object that closed.

  //Doc:Method
  virtual void Closed(L4Protocol*);            // Connection has closed
    //Doc:Desc Called by an associated layer 4 protocol when a connection
    //Doc:Desc has completely closed.
    //Doc:Arg1 A pointer to the layer 4 protocol object that closed.

  //Doc:Method
  virtual void ConnectionComplete(L4Protocol*);// Connection request succeeded
    //Doc:Desc Called by an associated layer 4 protocol when a previous
    //Doc:Desc connection request has successfully completed.
    //Doc:Arg1 A pointer to the layer 4 protocol object that connected.

  //Doc:Method
  virtual void ConnectionFailed(L4Protocol*);  // Connection request failed
    //Doc:Desc Called by an associated layer 4 protocol when a previous
    //Doc:Desc connection request has failed.
    //Doc:Arg1 A pointer to the layer 4 protocol object that failed.

  // Listener has cn rq from peer 
  //Doc:Method
  virtual bool ConnectionFromPeer(L4Protocol*, IPAddr_t, PortId_t);
    //Doc:Desc Called by an associated layer 4 protocol when a listening
    //Doc:Desc TCP protocol has received a connection request.
    //Doc:Arg1 A pointer to the listening TCP protocol object.
    //Doc:Arg2 \IPA\ of remote peer.
    //Doc:Arg3 Port number of report peer.
    //Doc:Return Return true if able to accept a new connection, false if not.

  // Manage the application
  //Doc:Method
  void DeleteOnComplete(bool doc = true) { deleteOnComplete = doc;}
    //Doc:Desc Specifies that this application object should automatically
    //Doc:Desc be deleted when the application has finished.
    //Doc:Arg1 True if delete on complete desired.

  //Doc:Method
  void CopyOnConnect(bool coc)           { copyOnConnect = coc;}
    //Doc:Desc Specifies that this application object should be copied
    //Doc:Desc when connection requests are processed.
    //Doc:Arg1 True if delete on copy on connect desired.

  //Doc:Method
  bool CopyOnConnect() { return copyOnConnect;}
    //Doc:Desc Return current status of copy on connect flag.
    //Doc:Return True if copy on connect, false if not.

  //Doc:Method
  virtual void StartApp() { };    // Called at time specified by Start
    //Doc:Desc Called at the specified applcation start time.

  //Doc:Method
  virtual void StopApp() { };     // Called at time specified by Stop
    //Doc:Desc Called at the specified applcation stop time.

  //Doc:Method
  virtual void AttachNode(Node*); // Note which node attached to
    //Doc:Desc Specify which node to which this application is attached.
    //Doc:Arg1 Node pointer to attached node.

  //Doc:Method
  virtual Application* Copy() const = 0;// Make a copy of the application
    //Doc:Desc Return a copy of this application. \purev
    //Doc:Return A pointer to a new copy of this application.

  // If the application has a single layer4 protocol, return it
  //Doc:Method
  virtual L4Protocol* GetL4() const { return nil;}
    //Doc:Desc Returns a pointer to the layer 4 protocol object,
    //Doc:Desc if this application has a single associated l4 protocol.
    //Doc:Return Pointer to the single layer 4 object, or {\tt nil} if 
    //Doc:Return none, or {\tt nil} if more than one.

public:
  //Doc:Member
  bool        deleteOnComplete;
    //Doc:Desc True if delete on complete requested.

  //Doc:Member
  bool        copyOnConnect;      // True if application copied
    //Doc:Desc True if copy on connect requested.
};

#endif


