/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Helper classes for command line argument processing
// George F. Riley, Georgia Tech, Winter 2002

#ifndef __args_h__
#define __args_h__

#include <string>
#include <vector>

#ifdef BOEING_ARGS
#include "node.h"
#include "rng.h"
#include "simulator.h"
#include "validation.h"
#include "wlan.h"
#endif //BOEING_ARGS

class Arg;
typedef std::vector<std::string> StringVec_t;
typedef std::vector<Arg>         ArgVec_t;

#ifdef BOEING_ARGS
static const int mobility_groups_max = 10;

struct ospf6_config_args {
  Count_t seed; 
  std::string tag;
  DCount_t start_time;
  DCount_t stop_time;
  bool trace;
  bool etrace;
  Time_t tracetopo;
  Time_t tracetopoonly;
  bool ospfv3log;

  Count_t pktrate;
  Count_t pktsize;

  Count_t nNodes;
  Count_t radio_range;
  double alpha;
  DCount_t grid_length;
  DCount_t velocity; 
  DCount_t pause_time;

  Count_t mobility_groups;
  Count_t mg_nNodes[mobility_groups_max];
  Count_t rr[mobility_groups_max];
  DCount_t v_max[mobility_groups_max];
  DCount_t v_min[mobility_groups_max];
  DCount_t p_max[mobility_groups_max];
  DCount_t p_min[mobility_groups_max];

  L2ProtocolType mac_protocol;

  Count_t w_int_type;
  Count_t w_flooding;
  Count_t HelloInterval;
  Count_t DeadInterval;
  Count_t RxmtInterval;
  Count_t MinLSInterval;
  Count_t MinLSArrival;
  Count_t AckInterval;
  bool diff_hellos;
  int cost; 
  int interface_jitter;
  int interface_flood_delay;

  Count_t PushbackInterval;
  bool smart_peering;
  bool unsynch_adj;

  bool topo_reduc;

  Count_t BackupWaitInterval;
  Count_t TwoHopRefresh;
  Count_t HelloRepeatCount;
  Count_t AdjConnectivity;
  Count_t LSAFullness;
  bool NonPersistentMDR;
};

bool set_ospf6_config_arguments(int argc, char** argv,
                             struct ospf6_config_args *ca);
#endif //BOEING_ARGS


//Doc:ClassXRef
class Arg {
  //Doc:Class Since it is often the case that a given simulation scenario
  //Doc:Class must be run multiple times with a variety of differnet
  //Doc:Class parameters, \GTNS\ provides a simple interface to handle
  //Doc:Class command line  arguments.  Class {\tt Arg} is used to 
  //Doc:Class specify an allowable set of equivalenced parameters, which
  //Doc:Class provide default values for parameters as well as explicitly
  //Doc:Class specified value.  Further, the {\tt Arg} interface allows
  //Doc:Class for non--equivalenced parameters, which can be queried
  //Doc:Class at runtime for presence or absence.
private:
  typedef enum {
    Int, Long, Double, String } Arg_t;
public:
  //Doc:Method
  Arg(const std::string& n, long&, long);
    //Doc:Desc Add an argument of type {\tt long}.
    //Doc:Arg1 Argument keyword.
    //Doc:Arg2 Reference to {\tt long} variable to receive equivalenced value.
    //Doc:Arg3 Default value if not specified.

  //Doc:Method
  Arg(const std::string& n, unsigned long&, unsigned long);
    //Doc:Desc Add an argument of type {\tt unsigned long}.
    //Doc:Arg1 Argument keyword.
    //Doc:Arg2 Reference to {\tt unsigned long} variable to
    //Doc:Arg2 receive equivalenced value.
    //Doc:Arg3 Default value if not specified.

  //Doc:Method
  Arg(const std::string& n, double&, double);
    //Doc:Desc Add an argument of type {\tt double}.
    //Doc:Arg1 Argument keyword.
    //Doc:Arg2 Reference to {\tt double} variable to
    //Doc:Arg2 receive equivalenced value.
    //Doc:Arg3 Default value if not specified.

  //Doc:Method
  Arg(const std::string& n, std::string&, const std::string&);
    //Doc:Desc Add an argument of type {\tt string}.
    //Doc:Arg1 Argument keyword.
    //Doc:Arg2 Reference to {\tt string} variable to
    //Doc:Arg2 receive equivalenced value.
    //Doc:Arg3 Default value if not specified.

public:
  //Doc:Method
  static bool ProcessArgs(int argc, char** argv); // Process the arguments
    //Doc:Desc Process the command line arguments.  Must be called before
    //Doc:Desc checking
    //Doc:Desc any argument values.
    //Doc:Arg1 {\tt argc} argument from {\tt main}.
    //Doc:Arg2 {\tt argv} argument from {\tt main}.
    //Doc:Return True if successfully processed.

  // Return number of non-equivalenced args      
  //Doc:Method
  static StringVec_t::size_type  NumberNonEquiv();
    //Doc:Desc Return the number of non--equivalenced arguments.
    //Doc:Return Count of non--equivalenced argument.

  // Return specified non-equivalended  
  //Doc:Method
  static std::string NonEquiv(StringVec_t::size_type);
    //Doc:Desc Return the specified non--equivalenced argument.
    //Doc:Arg1 Index from the non--equivalenced list to return.
    //Doc:Return The specified non--equivalenced argument.

  // True if specified string v was a non-equivalenced argument
  //Doc:Method
  static bool   Specified(const std::string&);
    //Doc:Desc Determine if a particular non--equivalenced argument was
    //Doc:Desc specified.
    //Doc:Return True if the argument was specified.

private:
  std::string    argName;       // Name of argument
  Arg_t          argType;       // Type of argument
  void*          pArg;          // Address (typeless) of argument
private:
  static ArgVec_t    args;      // The expected arguments list
  static StringVec_t nonEquiv;  // The non-equivalenced argument list
};


#endif
