/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Address resolution Protocol class
// Mohamed Abd Elhafez.  Georgia Tech, Spring 2004

// Implements the InternetAddress Resolution Protocol Layer 3

#ifndef __arp_h__
#define __arp_h__

#include <queue>


#include "common-defs.h"
#include "l3protocol.h"
#include "pdu.h"
#include "ipaddr.h"
#include "ipv4.h"
#include "node.h"
#ifdef HAVE_QT
#include "qnamespace.h"
#endif

#define ARP_PACKET 0x806             // the type field for the arp packet
#define MAX_RETRIES 5                // Maximum number of retries for resolving a certain entry
#define DEFAULT_RX_TIMEOUT 0.1       // default timeout for arp request in seconds

typedef std::pair<Packet*, int> PacketPair_t;
typedef std::queue<PacketPair_t> ARPQueue_t;


#ifdef HAVE_QT
class QColor;
#endif
class ARPHeader;

class ARPEntry {
public:
  typedef enum{FREE, PENDING, RESOLVED} State_t;

  ARPEntry(): state(FREE), mac(MACAddr::NONE), lifetime(nil), retries(0){}
  ARPEntry(State_t s, MACAddr m, Time_t t, Count_t r) : state(s), mac(m), lifetime(t), retries(r) 
    { 
    }
public:
  State_t        state;      // current state, free, rsolved or pending
  MACAddr        mac;        // mac address of the target
  Time_t         lifetime;   // lifetime of this entry
  Count_t        retries;    // number of arp attempts made for this entry
  ARPQueue_t     pQueue;     // Associated Queue
};


typedef std::map<IPAddr_t, ARPEntry> ARPCache_t;


// Define the event subclass for ARP events
//Doc:ClassXRef
class ARPEvent : public Event {
 public:
  typedef enum { RETRANSMIT } ARPEvent_t;
 public:
  ARPEvent() 
    : p(0), dst(0), iface(0)
    { }
  ARPEvent(Event_t ev) 
    : Event(ev), p(0),  dst(0), iface(0)
    { }
  ARPEvent(Event_t ev, Packet* pkt, IPAddr_t ip, Interface* interface) 
    : Event(ev), p(pkt), dst(ip), iface(interface)
    { }
  virtual ~ARPEvent() { }
public:
  Packet*     p;     // Associated packet
  IPAddr_t    dst;   // destination IP address
  Interface*  iface; // source interface
};



//Doc:ClassXRef
class ARP : public L3Protocol, public Handler{
//Doc:Class The class {\tt ARP} implements the Address Resolution Protocol
//Doc:Class It derives 	from the class {\tt L3Protocol}. 
	
public:
  typedef enum {NONE, REQUEST, REPLY} ARPPacket_t;

  //Doc:Method
  ARP();
  //Doc:Desc Default constructor

  // Handler
  virtual void Handle(Event*, Time_t);

  //Doc:Method
  void ARPRequest(Interface* , Packet*, int); // prepare an arp request for that target
  //Doc:Desc This method is used to generate an arp request for a certain IP adress
  //Doc:Arg1 A pointer to the interface which needs to resolve the IP address
  //Doc:Arg2 The packet that is queued for the destination
  //Doc:Arg3 The type of the packet


  //Doc:Method
  void DataIndication(Interface*, Packet*);    // From lower layer
  //Doc:Desc This method is used to pass on a packet from any layer 2 protocol
  //Doc:Desc up the stack.
  //Doc:Arg1 A pointer to the interface at which this packet is received
  //Doc:Arg2 A pointer to the packet that was received
  
   
  //Doc:Method
  virtual Proto_t Proto()   { return 0x806;}   // Get the protocol number
  //Doc:Desc This method returns the IANA assigned protocol number
  //Doc:Return the protocol number

  virtual void DataRequest(Node*, Packet*,   void*) {return;}
#ifdef SIM
  virtual void DataRequest(Node*, Packet*, void*, Interface*) {return;}
#endif //SIM
  virtual Interface* PeekInterface(Node*,   void*) {return nil;}
  virtual Count_t Version() {return 0;}

  //Doc:Method
  void SetARPTimeout(Time_t t){ timeout = t;}
  //Doc:Desc This method sets the timeout for the arp entries
  //Doc:Arg1 the new timeout in seconds

 public:
  ARPCache_t arpcache;    // The arp cache that stores all the entries
  Time_t     timeout;     // Timeout for the arp entries
  Time_t     rxtimeout;   // Timout for retransmitting arp request
private:
  static Trace::TraceStatus traceStatus; // For tracing
 
};

//Doc:ClassXRef
class ARPHeader : public PDU {
//Doc:Class	
public :
  //Doc:Method
  ARPHeader();
  //Doc:Desc This is the default constructor
 
  //Doc:Method 
  ARPHeader(const ARPHeader& r);
  //Doc:Desc This constructor takes the reference to an ARP header

  //Doc:Method
  virtual ~ARPHeader();
  //Doc:Desc Destructor
  
  //Doc:Method
  Size_t  Size() const;
  //Doc:Desc This method returns the size of the PDU
  //Doc:Return the size of the PDU
  
  //Doc:Method
  PDU*    Copy() { return new ARPHeader(*this);}
  //Doc:Desc This method creates a copy of this PDU
  //Doc:Return A pointer to the copy of this PDU
  
  //Doc:Method
  Layer_t Layer() { return 3;}         // Indicate layer 3 pdu
  //Doc:Desc This method returns the layer number it belongs in the ISO model
  //Doc:Return the layer number

  //Doc:Method
  void    Trace(Tfstream&, Bitmap_t);  // Trace the contents of this pdu
  //Doc:Desc This method traces the contents of the PDU
  //Doc:Arg1 File output stream descriptor
  //Doc:Arg2 Bitmap that specifies the level of detail of the trace

     
public:
  Word_t    hwmacaddrtype;  // The hardware address type, default is ethernet mac
  Word_t    protoaddrtype;  // The protocol address type, default is IP
  Byte_t    hwmacaddrsize;  // The size for the hardware address, default is 6 for ethernet mac
  Byte_t    protoaddrsize;  // The size for the protocol address, default is 4 for IPV4
  Word_t    op;             // Either Request or Reply
  MACAddr   srcmacaddr;     // The mac address of the source of the packet
  IPAddr    srcipaddr;      // The IP address of the source of the packet
  MACAddr   dstmacaddr;     // The mac address of the destination of the packet
  IPAddr    dstipaddr;      // The IP address of the destination of the packet
  Count_t   uid;            // Unique id (not part of ARP)
};


#endif

