/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// DropTail Queue class with IPA derivatives calculation
// This version calculates derivatives when the sources retransmit
// lost packets.
// George F. Riley, Georgia Tech, Spring 2004

#include "debug.h"
#include "droptail-ipa-retx.h"
#include "simulator.h"
#include "interface.h"
#include "link.h"

using namespace std;

Time_t DropTailIPAReTx::retxTime = 0.1;

// Constructors

DropTailIPAReTx::DropTailIPAReTx()
  : DropTail(), busyPeriod(false), lossyBusyPeriod(false),
    full(false), totalLosses(0), 
    c1(0), c2(0), dldq(0), retxWeight(0), retx(false), 
    startWeight(0), debugCount(0)
{
}

DropTailIPAReTx::DropTailIPAReTx(Count_t s) // Queue Length
  : DropTail(s), busyPeriod(false), lossyBusyPeriod(false),
    full(false), totalLosses(0), 
    c1(0), c2(0), dldq(0), retxWeight(0), retx(false),
    startWeight(0), debugCount(0)
{
}


DropTailIPAReTx::DropTailIPAReTx(const DropTailIPAReTx& c) // Copy constructor
  : DropTail(c), busyPeriod(false), lossyBusyPeriod(false),
    full(false), totalLosses(0), 
    c1(0), c2(0), dldq(0), retxWeight(0), retx(false),
    startWeight(0), debugCount(0)
{
  DEBUG0((cout << "DT Copy Const, limit " << Limit() << endl));
}


DropTailIPAReTx::~DropTailIPAReTx()
{
}

// Handler methods
void DropTailIPAReTx::Handle(Event* e, double t)     // Event handler
{
  DEBUG0((cout << "Hello from retx handler, time " << t << endl));
  ReTxEvent* retxev = (ReTxEvent*)e;
  if (Length() == 0)
    { // Retx on empty queue, do nothing except note weight of  retx pkt
      startWeight = retxev->weight;
      return;
    }
  
  retxWeight  = retxev->weight; // Set weight for re-tx event
  retx = true; // Note this is a retx
  // Enque it again
  Enque(retxev->p);
  // Re-tx event scheduled by Enque, so nothing else to do
  retxWeight = 0;
  retx = false;
}


// Public methods
bool DropTailIPAReTx::Enque(Packet* p)
{
  bool empty = Length() == 0;  // Initially empty
  bool r= DropTail::Enque(p);
  if (r)
    {
      full = false;
      if(!empty)c2 += retxWeight; // Count c2 if retx
      return true;
    }
  DEBUG0((cout << "Lost packet at time " << Simulator::Now()
          << ", Length " << Length() << endl));
  totalLosses++;      // Count losses
  if (!lossyBusyPeriod)
    { // Note a new lbp
      lossyBusyPeriod = true;
      DEBUG0((cout << "Found a new LBP starting at time " 
              << Simulator::Now() << endl));
      c1 = 1;
      if (startWeight) debugCount++;
    }
  c1 += c2; // Accumulate prior
  c1 += retxWeight;// And accumulate this one
  dldq += c1; // Calculate the derivative
  full = true;
  // Retransmit the dropped packet later
  ReTxEvent* e = new ReTxEvent();
  if (retx)
    { // Retransmitting an already retransmitted packet
      e->p = p; // Already have a copy of the packet
      DEBUG0((cout << "Retx of retx, weight " << e->weight << endl));
    }
  else
    { // A new retx packet
      e->p = p->Copy(); // Need a copy of the packet
      DEBUG0((cout << "New retx, weight " << e->weight << endl));
    }
  e->weight = c1;
  //Simulator::instance->Schedule(e, retxTime + Uniform(0,0.05).IntValue(), this);
  Simulator::instance->Schedule(e, retxTime, this);
  c1 = 0;
  c2 = 0;
  return false;           // Full queue, did not enque
}

Packet* DropTailIPAReTx::Deque()
{
  Packet* r = DropTail::Deque();
  if (!Length() || !r)
    { // End of busy period
      lossyBusyPeriod = false;
      c1 = 0;
      c2 = 0;
      DEBUG0((cout << "Queue empty at time " << Simulator::Now() << endl));
      startWeight = 0;
    }
  full = false;
  return r;
}

Queue* DropTailIPAReTx::Copy() const
{
  return new DropTailIPAReTx(*this);
}

// Methods specific to DropTailIPA Queues
NCount_t DropTailIPAReTx::DL1dB1()
{ // This is just the negative of the count of lossybusyperiods
  //return -(dldq + c1 + c2);
  return -(dldq);
}

Time_t DropTailIPAReTx::DQ1dB1()
{ // Includes partial time if still in the LBP
  return 0; // code later
}

Count_t DropTailIPAReTx::Losses()
{
  return totalLosses;
}


