/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - DropTail Queue class
// George F. Riley.  Georgia Tech, Spring 2002

#ifndef __droptail_h__
#define __droptail_h__

#include "queue.h"

class Interface;

#define DEFAULT_DROPTAIL_LIMIT 50000

//Doc:ClassXRef
class DropTail : public Queue {
//Doc:Class This method implements the simple droptail or FIFO queue.
public:

  //Doc:Method  
  DropTail() 
    : Queue(true), limit(Queue::DefaultLength()), pktLimit(0),
      size(0), pktCount(0), lastTime(0) { }
  //Doc:Desc the default constructor
  
  //Doc:Method  
  DropTail(Count_t l) 
    : Queue(true), limit(l), pktLimit(0),
      size(0), pktCount(0), lastTime(0) { }
  //Doc:Desc This constructor creates a queue of specified size
  //Doc:Arg1 the size of the queue in bytes
  
  //Doc:Method  
  DropTail(const DropTail& c) 
    : Queue(c), limit(c.limit), pktLimit(c.pktLimit),
      size(c.size), pktCount(c.pktCount), lastTime(c.lastTime) { }
  //Doc:Desc This constructor creates a queue identical to another queue
  //Doc:Arg1 Reference to an existing queue

  virtual ~DropTail() { }
  
  //Doc:Method  
  bool Enque(Packet*); // Return true if enqued, false if dropped
  //Doc:Desc This method is used to enque a packet in the queue
  //Doc:Arg1 the pointer to the packet to be added to the queue
  //Doc:Return a boolean to indicate if the enque operation succeeded or
  //Doc:Return if the packet was dropped.

  //Doc:Method  
  Packet* Deque();     // Return next element (NULL if none)
  //Doc:Desc This method deques a packet from the queue and returns it
  //Doc:Return Pointer to the packet

  //Doc:Method  
  Count_t Length();    // Length in bytes
  //Doc:Desc This method returns the size of the queue in terms of bytes
  //Doc:Return number of bytes

  //Doc:Method  
  Count_t LengthPkts();// Length in packets
  //Doc:Desc This method returns the size of the queue in terms of the
  //Doc:Desc number of packets
  //Doc:Return number of packets enqued

  //Doc:Method  
  void    SetInterface(Interface*);// Associated interface
  //Doc:Desc This method attaches the queue to an interface
  //Doc:Arg1 The interface to which it is to be attached
  

  //Doc:Method  
  Queue*  Copy() const;       // Make a copy
  //Doc:Desc This method makes a copy of the queue and returns a pointer
  //Doc:Desc to the copy
  //Doc:Return pointer to a copy of the queue
  
  //Doc:Method  
  virtual void SetLimit(Count_t l);
  //Doc:Desc This method sets a limit to the size of the queue. this limit
  //Doc:Desc is in terms of bytes
  //Doc:Arg1 size of the queue in bytes

  
  //Doc:Method  
  Count_t GetLimit() const { return limit;}
  //Doc:Desc This method returns the size of the queue in terms of the bytes
  //Doc:Return Max limit of the queue size in bytes
  
  //Doc:Method  
  virtual void SetLimitPkts(Count_t l) { pktLimit = l;} // Set packets limit
  //Doc:Desc This method sets a limit to the size of the queue. this limit
  //Doc:Desc is in terms of number of packets that can be enqued
  //Doc:Arg1 size of the queue in packets

  //Doc:Method  
  virtual Count_t GetLimitPkts() const { return pktLimit;} // Get pkt limit
  //Doc:Desc This method returns the size of the queue in terms of the packets
  //Doc:Return Max limit of the queue size in number of packets
  

  //Doc:Method  
  virtual Time_t  QueuingDelay();    // Calculate queuing delay at current lth
  //Doc:Desc This method calculates the current queuing delay
  //Doc:Return the queuing delay
  

  //Doc:Method  
  bool    Check(Size_t, Packet* = nil); // Test if buffer space available
  //Doc:Desc This method checks if buffer space (of a given size) is available
  //Doc:Arg1 The size to be checked
  //Doc:Arg2 The pointer to a packet to be queued.
  //Doc:Return boolean to indicate true or false

  //Doc:Method  
  Count_t Limit() { return limit;}
  //Doc:Desc This method returns the maximum limit of the queue size.
  //Doc:Return the limit (in bytes)
private:
  void    UpdateSize();       // Adjust size since last update
public:
  Count_t    limit;           // Limit of bytes in queue
  Count_t    pktLimit;        // Limit on number of pkts (if non-zero)
  Count_t    size;            // Current bytes in queue
  Count_t    pktCount;        // Current packets in queue
  Time_t     lastTime;        // Time size was calculated
public:
  PktList_t  pkts;            // List of packets in the queue
};

#endif
