/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Dual Queue Class Definition
// George F. Riley.  Georgia Tech, Spring 2003

// The "dual" queue implements a private queue for TCP traffic
// and another queue for all others.  The two queues are scheduled
// in a work-conserving round-robin method.

#ifndef __dualqueue_h__
#define __dualqueue_h__

#include "queue.h"

class Interface;

//Doc:ClassXRef
class DualQueue : public Queue {
public:
  DualQueue();
  DualQueue(Count_t l);
  DualQueue(const DualQueue& c);
  DualQueue(const Queue& t, const Queue& n); // Specify TCP and Non-TCP queues
  virtual ~DualQueue();
  // Inherited methods from Queue
  virtual bool Enque(Packet*);        // Return true if enqued, false if not
  virtual Packet* Deque();            // Return next element (NULL if none)
  // Most of the following should not be used for DualQueues.
  // Rather, the values should be retrived for the TCP/NonTCP queues directly
  virtual Count_t Length();           // How many bytes are enqued?
  virtual Count_t LengthPkts();       // How many packets are enqued?
  virtual Queue* Copy() const;        // Make a copy of this queue
  virtual void SetLimit(Count_t);     // Set new limit
  virtual Count_t GetLimit() const;   // Get limit (bytes)
  virtual void SetLimitPkts(Count_t); // Set packets limit
  virtual Count_t GetLimitPkts() const;  // Set packets limit
  virtual bool Check(Size_t, Packet* = nil); // Test if buffer space available
  virtual bool Detailed();            // True if detailed model w/ local queing
  virtual void Detailed(bool);        // Set detailed on/off
  virtual Time_t   QueuingDelay();    // Calculate queuing delay at current lth
  virtual DCount_t TotalWorkload();   // Return the byte-seconds (workload)
  virtual void     SetInterface(Interface*);  // Let the q know which interface
  virtual Count_t  DropCount();       // Number of dropped packets
  virtual Count_t  EnqueueCount();    // Number of enqueued packets
  virtual void     CountEnq(Packet* = nil); // Count an enque
  virtual void     ResetStats();      // Reset the statistics to zero
  // Methods specific to DualQueue
  void   TCPQueue(const Queue& q);
  void   NonTCPQueue(const Queue& q);
  Queue* TCPQueue() const { return pTCP;}
  Queue* NonTCPQueue() const { return pNonTCP;}
  bool   LastTCP() const { return lastTCP;}
  Count_t TCPEmpty() const { return tcpEmpty;}
  Count_t UDPEmpty() const { return udpEmpty;}
private:
  bool   IsTCP(Packet* p);            // True if packet has TCP header
private:
  Count_t limit;                      // Limit of bytes in queue
  Count_t pktLimit;                   // Limit on number of pkts (if non-zero)
  bool    lastTCP;                    // True if last deque was from TCP queue
  Count_t tcpEmpty;                   // Times TCP had no packet on it's turn
  Count_t udpEmpty;                   // Times UDP had no packet on it's turn
public:
  Queue*  pTCP;                       // Points to the TCP queue
  Queue*  pNonTCP;                    // Points to non-tcp queue
};

#endif
