/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Extended tracing (hacked in by Boeing) 

#include <iostream>
#include "netinet/in.h" //ntohl and ntohs
#include "arpa/inet.h" //inet_ntop

#include "estat.h"
#include "protograph.h"
#include "node.h"
#include "packet.h"
#include "interface.h"
#include "simulator.h"
#include "ipaddr.h"
#include "ipv4.h"
#include "l2proto802.11.h"
#include "l2proto802.3.h"
#include "datapdu.h"
#include "ospf6_message.h"
#include "ospf6_proto.h" //for ospf6_options_printbuf
#include "ospf6_lsa.h" //for ospf6_lsa_header
#include "zebra.h" //for CHECK_FLAG

using namespace std;

EStat*            EStat::instance;        // Single global stat object
bool              EStat::enabled = false; // Global enabled flag
bool              EStat::extended = false; // Global extended tracing flag
int               EStat::stat_event_level = 0;

EStat::EStat() 
  : stat(ProtocolGraph::N_LAYERS, DISABLED), //vector constructor
    node(0), start_stat_time(0)
{
	memset(statistics, 0, sizeof(double) * STAT_LENGTH);
}

EStat::~EStat()
{
  Close(); // Close the file if open
}

bool EStat::Open(const char* fn)
{
  if (os.is_open())
    ;
  else 
    os.open(fn, ios::app);
  return os;  // True if already open or opened successfully
}

void EStat::Close()
{
  if (os.is_open())
    {
      os.close();
    }
  enabled = false;
}

void EStat::LayerOn()
{ // Enable all
  for (ETSVec_t::size_type i = 0; i < stat.size(); ++i)
    {
      stat[i] = ENABLED;
    }
}

void EStat::LayerOn(Layer_t l)
{
  if (l > ProtocolGraph::N_LAYERS) return;  // Out of range
  stat[l-1] = ENABLED;
}

void EStat::LayerOff()
{ // Disable all
  for (ETSVec_t::size_type i = 0; i < stat.size(); ++i)
    {
      stat[i] = DISABLED;
    }
}

void EStat::LayerOff(Layer_t l)
{
  if (l > ProtocolGraph::N_LAYERS) return;  // Out of range
  stat[l-1] = DISABLED;
}

bool EStat::IsLayerOn(Layer_t l)
{ // Enquire stat status by layer
  if (l > ProtocolGraph::N_LAYERS) return false;  // Out of range
  if (!enabled) return false; 
  return stat[l-1] == ENABLED;
}

bool EStat::IsEnabled()
{
  return(os.is_open());
}

void EStat::NodeOff(int id)
{
	vNodeIds::iterator ids_iter;
	
	ids_iter = find(vnodeids.begin(), vnodeids.end(), id);
	if(ids_iter != vnodeids.end())
		return;
	vnodeids.push_back(id);
}

void EStat::NodeOn()
{
	vnodeids.clear();
}

void EStat::NodeOn(int id)
{
	vNodeIds::iterator ids_iter;
	
	ids_iter = find(vnodeids.begin(), vnodeids.end(), id);
	if(ids_iter == vnodeids.end())
		return;
	vnodeids.erase(ids_iter);
}

bool EStat::IsNodeOn(int id)
{
	vNodeIds::iterator ids_iter;
	
	ids_iter = find(vnodeids.begin(), vnodeids.end(), id);
	if(ids_iter == vnodeids.end())
		return true;
	return false;
}

void EStat::EStatPacket(Packet *p, Node* n, Layer_t L, char* str) 
{
  PDU* pdup;
	int offset = 0; // May not be needed
	string s = str;

	
	if (start_stat_time > Simulator::Now())
		return;
	if (!enabled) 
		return;
	if (!IsLayerOn(L))
		return;
	if (!IsNodeOn(n->Id()))
		return;
	switch (L) {
		case 3 : // Look at layer 3.  FindPDU(3) seems to be reliable call
			pdup = p->FindPDU(3);
      if (pdup) {
         offset += pdup->Size();
         EStatIPV4(pdup, s);
      }
#if 0
        IPV4Header* iph = (IPV4Header*) pdup;
				if ((iph) && (iph->protocol == 89)) {// Find and parse OSPF header
					//pdup = p->PeekPDU(1);
	        pdup = p->FindPDU(0,0); // should be OSPF PDU
  	      TraceOspf6(os, pdup);
				}
#endif
				break;
		default:
				break;
	}
}

void EStat::EStatIPV4(PDU* pdup, string str)
{
  IPV4Header* iph = (IPV4Header*) pdup;
  switch (iph->protocol) {
    case 89:
      break;
    case 17:
			if (str == "s" || str == "d") {
				statistics[UDP_SENT]++;
				statistics[UDP_BYTE_SENT] += iph->totalLength;
			}
			else if (str == "f") {
				statistics[UDP_FWD]++;
				statistics[UDP_BYTE_FWD] += iph->totalLength;
			}
			else if (str == "r") {
				statistics[UDP_RECV]++;
				statistics[UDP_BYTE_RECV] += iph->totalLength;
			}
      break;
    default:
      break;
  }
}

void EStat::EStatPrint()
{
	if (!Open(statfile.c_str()))
		return;

	double del_ratio = ((statistics[UDP_SENT] == 0) ? 0 :
			statistics[UDP_RECV]/statistics[UDP_SENT]);
	
	os.setf(ios_base::fixed, ios_base::floatfield);
	os.precision(0);
	os << "User Traffic Analysis" << endl;
	os << "--------------------------------------------------" << endl;
	os << "UDP sends: " << statistics[UDP_SENT] << " pkts, " 
		<< statistics[UDP_BYTE_SENT] << " bytes" << endl;
	os << "UDP receives: " << statistics[UDP_RECV] << " pkts, " 
		<< statistics[UDP_BYTE_RECV] << " bytes" << endl;
	os << "UDP forwards: " << statistics[UDP_FWD] << " pkts, " 
		<< statistics[UDP_BYTE_FWD] << " bytes" << endl;

	os.precision(3);
	os << endl << "Delivery Ratio:  " << del_ratio << endl;

}

void EStat::CollectEStats(const char *name, Time_t time)
{
	start_stat_time = time;
	statfile = name;	
	enabled = true;
	LayerOn();
}

void EStat::UpdateAveragePower(double power)
{
  statistics[POWER] += 10 * log10(power);  //convert to dBm
  statistics[POWER_COUNT]++;
}

void EStat::UpdateContentionRatio(double ratio)
{
  statistics[CONTENTION_RATIO]+= ratio;
  statistics[CONTENTION_RATIO_COUNT]++;
}

// Static methods
EStat* EStat::Instance()
{
  // If no instance, create one
  if (!instance) instance = new EStat();
  return instance;
}

