/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Ethernet LAN class
// George F. Riley.  Georgia Tech, Spring 2002

// Define the Ethernet LAN Class

#include <iostream>

#include "debug.h"
#include "ethernet.h"
#include "node.h"
#include "queue.h"

using namespace std;


#ifdef ETHERNET_BOEING

Ethernet::Ethernet(Count_t n,               // Node count
                   IPAddr_t i, Mask_t m,    // IP addr start, mask
                   SystemId_t id,           // Responsible system
                   Rate_t r, Time_t dly, Detail_t d)
  : LinkReal(r, dly), gw(nil), gwIf(nil), receiveOwnBroadcast(false)
{
  ConstructorHelper(n, nil, i, m, id, nil, r, dly, d);
}

Ethernet::Ethernet(Count_t n,             // Node count
                   IPAddr_t i, Mask_t m,  // IP addr start, mask
                   const L4Protocol& l4,  // L4 protocol to bind to each
                   SystemId_t id,         // Responsible system
                   Rate_t r, Time_t dly, Detail_t d)
  : LinkReal(r, dly), gw(nil), gwIf(nil),
    receiveOwnBroadcast(false)
{
  ConstructorHelper(n, nil, i, m, id, l4.Copy(), r, dly, d);
}

Ethernet::Ethernet(Count_t n, Node* g,    // Node count, gateway node
                   IPAddr_t i, Mask_t m,  // IP addr start, mask
                   SystemId_t id,         // Responsible system
                   Rate_t r, Time_t dly, Detail_t d)
  : LinkReal(r, dly), gw(g), gwIf(nil),
    receiveOwnBroadcast(false)
{
  ConstructorHelper(n, g, i, m, id, nil, r, dly, d);
}

Ethernet::Ethernet(Count_t n, Node* g,    // Node count, gateway node
                   IPAddr_t i, Mask_t m,  // IP addr start, mask
                   const L4Protocol& l4,  // L4 protocol to bind to each
                   SystemId_t id,         // Responsible system
                   Rate_t r, Time_t dly, Detail_t d)
  : LinkReal(r, dly), gw(g), gwIf(nil),
    receiveOwnBroadcast(false)
{
  ConstructorHelper(n, g, i, m, id, l4.Copy(), r, dly, d);
}

Ethernet::~Ethernet()
{ // Destructor, nothing to do
}

// Link methods
Link* Ethernet::Copy() const
{
  return new Ethernet(*this);
}

Interface* Ethernet::GetPeer(Packet* p)
{
  if (p)
    { // Packet is specified, peek and find the neighbor
      L2802_3Header* l2pdu = (L2802_3Header*)p->PeekPDU();
      return GetIfByMac(l2pdu->dst);  
    }
  // No packet?
  return nil;
}

Interface* Ethernet::GetPeer(Count_t i)
{ // Get the i'th peer's interface
  return GetIf(i);
}

void  Ethernet::SetPeer(Interface* ifc) 
{
  Node* n = ifc->GetNode();
  if (NodeIsPeer(n))
  {
    cout << "Associated node is already peer " << endl;
    return;
  }
  nodes.push_back(n);
  ifc->SetIPAddr(nextIP);
  nextIP++;
  ifc->SetLink(this);          // Uses this link

  Queue* q = ifc->GetQueue();
  if (q)
		q->Detailed(true);  // Can't use non-detailed q

  if (l4proto)
	{ // Layer 4 specified
		PortId_t port = l4proto->localPort;  // Port to use
		L4Protocol* p = l4proto->Copy();     // Get a copy of this protocol
		p->Attach(n);                // Set attached node
		p->Bind(port);                  // Bind to specified port
	}

  // Sending update node events to all nodes connected to this link 
  if(detail == FULL)
	{
		for (Count_t j = 0; j < nodes.size(); ++j)
		{ 	  
			Interface* ifc = GetIf(j);
			EthernetEvent* upd = new EthernetEvent(EthernetEvent::UPDATE_NODES);
			Simulator::instance->Schedule(upd, 0.0, ifc);
		}
	}    
}

Count_t  Ethernet::Size()
{ // Return number of peers
  return nodes.size(); // subtract one since not self peer
}

Count_t  Ethernet::PeerCount()
{ // Return number of peers

	if (nodes.size() < 1)
		return 0;
  return nodes.size() - 1; // subtract one since not self peer
}

bool Ethernet::NodeIsPeer(Node* n)
{ // Return true of node is a peer
  if (n == gw) 
    return true; // Definitely true if gateway

  NodeId_t id = n->Id();
  for(NodeVec_t::size_type  i = 0; i < nodes.size(); ++i)
	{
		if( id == nodes[i]->Id())
			return true;
	}
  return false; 
}

Count_t  Ethernet::NodePeerIndex(Node* n)
{ // Return the peer index.  Assumed to be in range
  NodeId_t id = n->Id();

  if (n == gw) return 0;  
 
  for(NodeVec_t::size_type  i = 0; i < nodes.size(); ++i)
  {
    if(n->Id() == nodes[i]->Id())
      return i;
  }
  return MAX_COUNT;
}

IPAddr_t Ethernet::NodePeerIP(Node* n)
{ // Get peer IP Address for the specified node
  if (n == gw) 
		return GetIP(0);

  for (Count_t i = 0; i < nodes.size(); ++i)
    {
      Node* peerNode = GetNode(i);
      if (peerNode == n) 
				return GetIP(i); // Found it, return IP Addresss
    }
  return IPADDR_NONE;        // Not found, return none
}

MACAddr  Ethernet::IPToMac(IPAddr_t ip)
{ // Convert peer IP to peer MAC
  for(NodeVec_t::size_type  j = 0; j < nodes.size(); ++j)
    {
      if(ip == nodes[j]->GetIfByLink(this)->GetIPAddr())
	{
	  return nodes[j]->GetIfByLink(this)->GetMACAddr();
	}
    } 
  return MACAddr::NONE;
}

Node* Ethernet::GetNode(Count_t i)
{ // Get the i'th node
  if (gw && i == 0) 
    return gw;         // Gateway is zero'th

	if (i >= nodes.size())
		return nil;

	return nodes.at(i);
}

IPAddr_t   Ethernet::GetIP(Count_t i)
{ // Get the i'th ip address

	if (i >= nodes.size())
		return IPADDR_NONE;

	Node *n = nodes.at(i);

	return n->GetIfByLink(this)->GetIPAddr();

}

Interface* Ethernet::GetIfByMac(MACAddr m)
{ // Get the interface corresponding to the specified MAC address
  for(NodeVec_t::size_type j = 0; j < nodes.size(); ++j)
	{
		if(m == nodes[j]->GetIfByLink(this)->GetMACAddr())
			return nodes[j]->GetIfByLink(this);
	} 
  return nil;                         // Out of range
}

void  Ethernet::AddNode(Node* n)
{
  if(NodeIsPeer(n))
	{
		cout << "N" << n->Id() << " is already in list " << endl;
		return;
	}

  // Add interface
  Interface* iface;
  if(detail == FULL)
    iface = n->AddInterface(L2Proto802_3(), InterfaceEthernet(), nextIP, mask);
  else
    iface = n->AddInterface(L2Proto802_3(), nextIP, mask);

	SetPeer(iface);
}


void Ethernet::ConstructorHelper(
           Count_t n,               // Node count
           Node*   g,               // Gateway node (nil if none)
           IPAddr_t i, Mask_t mask_, // IP addr start, mask
           SystemId_t id,
           L4Protocol* l4,          // L4 protocol to bind to each
           Rate_t r, Time_t dly, Detail_t d)
{
  DEBUG0((cout << "Creating Ethernet, firstIP " << (string)IPAddr(i)
       << " nnodes " << n << " detail "<< d << endl));

  detail = d;                       // Detail level
  bw = r;                           // Bandwidth
  delay = dly;                      // Delay
  l4proto = l4;
  nextIP = i;
  mask = mask_;

  if (g)
  { // Gateway specified, add the interface
		  if(detail == FULL)
		  {	    
			   gwIf = g->AddInterface(L2Proto802_3(), InterfaceEthernet(), nextIP, mask_);      g->SetLocation(0, 0);
    }
    else
    {
	     gwIf = g->AddInterface(L2Proto802_3(), nextIP, mask_);
  		}
		  gwIf->SetLink(this);           // Uses this link
  		Queue* q = gwIf->GetQueue();
		  if (q) 
  		{
   			q->Detailed(true);  // Can't use non-detailed q
  		}
		  nextIP++;                          // Advance the ip address
  		nodes.push_back(g);
	 }
	
  for (Count_t j = 0; j < n; ++j)
  { // Create the nodes
    Node* node = new Node(id);
    AddNode(node);
  }

  // Sending update node events to all nodes connected to this link
  if(detail == FULL)
  {
    UseSeqEvents(false);
    for (Count_t j = 0; j < nodes.size(); ++j)
    { 	  
      Interface* iface = GetIf(j);
      EthernetEvent* upd = new EthernetEvent(EthernetEvent::UPDATE_NODES);
      Simulator::instance->Schedule(upd, 0.0, iface);
    }
  }    
} 

bool Ethernet::Transmit(Packet* p, Interface* i, Node* n)
{
  L2802_3Header* l2pdu = (L2802_3Header*)p->PeekPDU(); // Get a ptr to l2 hdr
  Interface* targ_if = GetIfByMac(l2pdu->src);
  if (l2pdu->dst.IsBroadcast())
    { // Broadcast, pass interface of source, not dst
      return TransmitHelper(p, i, n, targ_if, true);
    }
  else
    {
      Interface* targ_if = GetPeer(p);
      if (!i)
        {
          cout << "HuH?  Ethernet::Transmit null peer i/f" << endl;
          L2802_3Header* l2pdu = (L2802_3Header*)p->PeekPDU();
          cout << "Mac dst " << l2pdu->dst << endl;
          IPV4Header* l3pdu = (IPV4Header*)p->PeekPDU(1);
          cout << "Src IP "  << (string)IPAddr(l3pdu->src)
               << " dst IP " << (string)IPAddr(l3pdu->dst) << endl;
          return false;
        }
      return TransmitHelper(p, i, n, targ_if);
    }
}

bool Ethernet::Transmit(Packet* p, Interface* i, Node* n, Rate_t)
{
  return Transmit(p, i, n);
}


IPAddr_t Ethernet::PeerIP(Count_t npeer)
{ // Get peer IP Address (if known)
  Node* n = GetNode(npeer);  // Get the node    
  if (n) 
		return n->GetIPAddr(); // get the node's ip address
  return IPADDR_NONE;        // Problem, return none
}


IPAddr_t Ethernet::DefaultPeer(Interface* fromIf)
{ // Get IP Addr of default peer
  if (gw)
    {
      IPMaskVec_t ipmasks;
      gw->IPAddrs(ipmasks); // Get the list of ipaddresses for gateway
      // Find the one matching my subnet
      IPAddr_t ip = fromIf->GetIPAddr();
      DEBUG0((cout << "Ethernet::DefaultPeer, fromIF " 
              << (string)IPAddr(ip) << endl));
      
      for (IPMaskVec_t::size_type i = 0; i < ipmasks.size(); ++i)
        {
          DEBUG0((cout << "Ethernet::DefaultPeer, checking "
                  << (string)IPAddr(ipmasks[i].ip)
                  << " vs " << (string)IPAddr(ip) << endl));
          IPAddr_t pip = ipmasks[i].ip;
          Mask_t   m   = ipmasks[i].mask;
          if ((ip & m) == (pip & m))
            {
              DEBUG0((cout << "Ethernet::DefaultPeer, returning " 
                      << (string)IPAddr(pip) << endl));
              return pip; // found it
            }
        }
    }
  DEBUG0((cout << "Ethernet::DefaultPeer returning IPADDR_NONE" << endl));
  return IPADDR_NONE;  // No default peer if no gateway
}

Count_t  Ethernet::NeighborCount(Node* n)
{ // Number of routing neighbors
  if (gw)
  {
    if (gw != n)
    return 1;// If a gw system and not the gw, return just one
  }
  return PeerCount(); // Otherwise return number of peers
}

void Ethernet::Neighbors(NodeWeightVec_t& nwv, Interface* i, bool forceAll)
{ // Add to neighbor list
  // If a gateway system, all leafs just claim the gateway as
  // a single neighbor.  The gateway itself claims no neighbors.
  if (gw && !forceAll)
  {
    Node* n = i->GetNode();
    if (n == gw) 
      return;  // This is the gateway node, no neighbors
    nwv.push_back(NodeIfWeight(gw, i, Weight())); // Otherwise gw is neighbor
    return;
  }

  // Not a single gateway, a bit more complex
  // Add each node in the list.  Note that the list does include
  // all nodes, which means certain route calculations must take
  // care to ignore "self" adjacenies
  //for (NodeId_t ni = first; ni < last; ++ni)
  for(NodeVec_t::size_type ni = 0; ni < nodes.size(); ++ni)
    nwv.push_back(NodeIfWeight(Node::GetNode(ni), i, Weight()));
}

void Ethernet::AllNeighbors(NodeWeightVec_t& nwv)
{ // Make a list of all neighbors.  Note the the "Interface" pointer
  // in the NodeIfWeight vector is the TARGET interface, not the
  // source as in Neighbors above.  This function is used by L2
  // broadcast to determine where to schedule receive events.
  // 
  for(NodeVec_t::size_type ni = 0; ni < nodes.size(); ++ni)
  {
    Interface* iface = GetIf(ni);
    if (!iface) 
      cout << "HuH?  Ethernet::AllNeighbors nill iface for " << ni << endl;
    nwv.push_back(NodeIfWeight(Node::GetNode(ni), GetIf(ni), Weight()));
  }
}

Interface* Ethernet::GetIf(Count_t i)
{ // Get the interface for the i'th node
	if (i >= nodes.size())
		return nil;
	Node *n = nodes.at(i);
  return n->GetIfByLink(this);
}

Node* Ethernet::Gateway()
{
  return gw; // Return the gateway node
}

void  Ethernet::SetLeafQLimit(Count_t l)
{
  for (Count_t t = 0; t < nodes.size(); ++t)
  {
    Node* n = GetNode(t);
    if (n)
    {
      Queue* q = n->GetQueue(); // Get the queue for single interface
      if (q)
        q->SetLimit(l); // Found single interface node, set specified limit
    }
  }
}

#else //ETHERNET_BOEING
//Begin GTNets Ethenet

Ethernet::Ethernet(Count_t n,               // Node count
                   IPAddr_t i, Mask_t m,    // IP addr start, mask
                   SystemId_t id,           // Responsible system
                   Rate_t r, Time_t dly, Detail_t d)
  : LinkReal(r, dly), gw(nil), gwIf(nil),
    receiveOwnBroadcast(false)
{
  ConstructorHelper(n, nil, i, m, id, nil, r, dly, d);
}

Ethernet::Ethernet(Count_t n,             // Node count
                   IPAddr_t i, Mask_t m,  // IP addr start, mask
                   const L4Protocol& l4,  // L4 protocol to bind to each
                   SystemId_t id,         // Responsible system
                   Rate_t r, Time_t dly, Detail_t d)
  : LinkReal(r, dly), gw(nil), gwIf(nil),
    receiveOwnBroadcast(false)
{
  ConstructorHelper(n, nil, i, m, id, l4.Copy(), r, dly, d);
}

Ethernet::Ethernet(Count_t n, Node* g,    // Node count, gateway node
                   IPAddr_t i, Mask_t m,  // IP addr start, mask
                   SystemId_t id,         // Responsible system
                   Rate_t r, Time_t dly, Detail_t d)
  : LinkReal(r, dly), gw(g), gwIf(nil),
    receiveOwnBroadcast(false)
{
  ConstructorHelper(n, g, i, m, id, nil, r, dly, d);
}

Ethernet::Ethernet(Count_t n, Node* g,    // Node count, gateway node
                   IPAddr_t i, Mask_t m,  // IP addr start, mask
                   const L4Protocol& l4,  // L4 protocol to bind to each
                   SystemId_t id,         // Responsible system
                   Rate_t r, Time_t dly, Detail_t d)
  : LinkReal(r, dly), gw(g), gwIf(nil),
    receiveOwnBroadcast(false)
{
  ConstructorHelper(n, g, i, m, id, l4.Copy(), r, dly, d);
}

Ethernet::~Ethernet()
{ // Destructor, nothing to do
}

// Link methods
Link* Ethernet::Copy() const
{
  return new Ethernet(*this);
}

Interface* Ethernet::GetPeer(Packet* p)
{
  if (p)
    { // Packet is specified, peek and find the neighbor
      L2802_3Header* l2pdu = (L2802_3Header*)p->PeekPDU();
      return GetIfByMac(l2pdu->dst);  
    }
  // No packet?
  return nil;
}

Interface* Ethernet::GetPeer(Count_t i)
{ // Get the i'th peer's interface
  return GetIf(i);
}

//////////// modified by Mohamed //////////////

void  Ethernet::SetPeer(Interface* ifc) 
{
  Node* n = ifc->GetNode();
  if (NodeIsPeer(n))
    {
      cout << "Associated node is already peer " << endl;
      return;
    }
  nodes.push_back(n);
  size++;
  ifc->SetIPAddr(nextIP);
  nextIP++;
  ifc->SetLink(this);          // Uses this link
  Queue* q = ifc->GetQueue();
  if (q)
    {
      q->Detailed(true);  // Can't use non-detailed q
    }
  if (l4proto)
    { // Layer 4 specified
      PortId_t port = l4proto->localPort;  // Port to use
      L4Protocol* p = l4proto->Copy();     // Get a copy of this protocol
      p->Attach(n);                // Set attached node
      p->Bind(port);                  // Bind to specified port
    }

  
  // Sending update node events to all nodes connected to this link 
  if(detail == FULL)
    {
      for (Count_t j = 0; j < size; ++j)
	{ 	  
	  Interface* ifc = GetIf(j);
	  EthernetEvent* upd = new EthernetEvent(EthernetEvent::UPDATE_NODES);
	  Simulator::instance->Schedule(upd, 0.0, ifc);
	}
    }    
}


Count_t  Ethernet::Size()
{ // Return number of peers
  return size; // subtract one since not self peer
}

// Count_t  Ethernet::PeerCount()
// { // Return number of peers
//   Count_t r = last - first - 1; // Subtract one since not self-peer
//   if (gw) r++;                  // Include gateway
//   return r;
// }

Count_t  Ethernet::PeerCount()
{ // Return number of peers
  return size - 1; // subtract one since not self peer
}

// bool     Ethernet::NodeIsPeer(Node* n)
// { // Return true of node is a peer
//   if (n == gw) return true; // Definitely true if gateway
//   NodeId_t id = n->Id();
//   return (id >= first && id < last); // True if in range
// }

bool     Ethernet::NodeIsPeer(Node* n)
{ // Return true of node is a peer
  if (n == gw) return true; // Definitely true if gateway

  NodeId_t id = n->Id();
  if(id >= first && id < last) return true; // True if in range

  for(NodeVec_t::size_type  i = 0; i < nodes.size(); ++i)
	{
		if( id == nodes[i]->Id())
			return true;
	}
  return false; 
}


// Count_t  Ethernet::NodePeerIndex(Node* n)
// { // Return the peer index.  Assumed to be in range
//   NodeId_t id = n->Id();
//   if (gw)
//     { // Gateway is peer 0, all others up by one
//       if (n == gw) return 0;
//       return id - first + 1;
//     }
//   return id - first;
// }

Count_t  Ethernet::NodePeerIndex(Node* n)
{ // Return the peer index.  Assumed to be in range
  NodeId_t id = n->Id();
  Count_t extra = gw ? 1 : 0;

  if (n == gw) return 0;  
  if (id >= first && id < last)
    {          
      return id - first + extra;
    }
 
  for(NodeVec_t::size_type  i = 0; i < nodes.size(); ++i)
    {
      if( n->Id() == nodes[i]->Id())
	{
	  return last + extra + i - first;
	}
    } 
  return MAX_COUNT ;
}

// IPAddr_t Ethernet::NodePeerIP(Node* n)
// { // Get peer IP Address for the specified node
//   if (n == gw) return firstIP;
//   for (Count_t i = 0; i < PeerCount(); ++i)
//     {
//       Node* peerNode = GetNode(i);
//       if (peerNode == n) return GetIP(i); // Found it, return IP Addresss
//     }
//   return IPADDR_NONE;        // Not found, return none
// }

IPAddr_t Ethernet::NodePeerIP(Node* n)
{ // Get peer IP Address for the specified node
  if (n == gw) return firstIP;
  for (Count_t i = 0; i < size; ++i)
    {
      Node* peerNode = GetNode(i);
      if (peerNode == n) return GetIP(i); // Found it, return IP Addresss
    }
  return IPADDR_NONE;        // Not found, return none
}


// MACAddr  Ethernet::IPToMac(IPAddr_t ip)
// { // Convert peer IP to peer MAC
//   Count_t extra = (gw) ? 1 : 0;
//   DEBUG0((cout << "Ethernet::IPToMac, ip " << (string)IPAddr(ip)
//           << " firstIP " << (string)IPAddr(firstIP)
//           << " lastIP " << (string)IPAddr(firstIP + (last - first + extra))
//           << endl));
//   if (ip < firstIP)
//     {
//       return MACAddr::NONE; // Out of range
//     }
//   if (ip >= (firstIP + (last - first + extra)))
//     {
//       return MACAddr::NONE;
//     }
//   Count_t delta = ip - firstIP;
//   return firstMac + delta;
// }

MACAddr  Ethernet::IPToMac(IPAddr_t ip)
{ // Convert peer IP to peer MAC
  Count_t extra = (gw) ? 1 : 0;
  DEBUG0((cout << "Ethernet::IPToMac, ip " << (string)IPAddr(ip)
          << " firstIP " << (string)IPAddr(firstIP)
          << " lastIP " << (string)IPAddr(firstIP + (last - first + extra))
          << endl));

  if (ip >= firstIP && ip < (firstIP + (last - first + extra)))
    {
      Count_t delta = ip - firstIP;
      return firstMac + delta;
    }

  for(NodeVec_t::size_type  j = 0; j < nodes.size(); ++j)
    {
      if(ip == nodes[j]->GetIfByLink(this)->GetIPAddr())
	{
	  return nodes[j]->GetIfByLink(this)->GetMACAddr();
	}
    } 
  return MACAddr::NONE;
}


// Ethernet Methods
// Node* Ethernet::GetNode(Count_t i)
// { // Get the i'th node
//   Count_t extra = (gw) ? 1 : 0;
//   if (gw && i == 0) return gw;         // Gateway is zero'th
//   if ((first + i) >= (last+extra)) return nil; // Out of range
//   return Node::GetNode(first + i - extra);     // Return the specified node
// }

Node* Ethernet::GetNode(Count_t i)
{ // Get the i'th node
  Count_t extra = (gw) ? 1 : 0;
  if (gw && i == 0) return gw;         // Gateway is zero'th
  if ((first + i) < (last + extra))  // Original range 
    {
      return Node::GetNode(first + i - extra);     // Return the specified node
    }
  //cout << "Node not in original range " << endl;
  //cout << "index " << i << " last " << last << " size " << size << endl;
  if( (i - last - extra + first) < nodes.size() )
    {
      //cout << "Return node index " << i - last - extra + first << endl; 
      return nodes[i - last - extra + first];
    }
      
  return nil;
}

// IPAddr_t   Ethernet::GetIP(Count_t i)
// { // Get the i'th ip address
//   Count_t extra = (gw) ? 1 : 0;
//   if ((first + i) >= (last+extra)) return IPADDR_NONE; // Out of range
//   return (IPAddr_t)(firstIP + i);
// }

IPAddr_t   Ethernet::GetIP(Count_t i)
{ // Get the i'th ip address
  Count_t extra = (gw) ? 1 : 0;
  if ((first + i) < (last+extra)) // Original range
    return (IPAddr_t)(firstIP + i);

  if( (i - last - extra + first) < nodes.size() )
    return nodes[i - last - extra + first]->GetIfByLink(this)->GetIPAddr();

  return IPADDR_NONE;
}

// Interface* Ethernet::GetIfByMac(MACAddr m)
// { // Get the interface corresponding to the specified MAC address
//   Node* n = GetNode(m - firstMac);    // First get the node
//   if (n) return n->GetIfByLink(this); // Find the interface for this MAC
//   return nil;                         // Out of range
// }

Interface* Ethernet::GetIfByMac(MACAddr m)
{ // Get the interface corresponding to the specified MAC address
  if(gw)
    {
      if (m == firstMac)
	return gwIf;
    }
  if(((m - firstMac) >= 0) && (m - firstMac) <= (last - first)) 
    {
      Node* n = GetNode(m - firstMac);    // First get the node
      if (n) return n->GetIfByLink(this); // Find the interface for this MAC
    }
  
  for(NodeVec_t::size_type j = 0; j < nodes.size(); ++j)
    {
      if(m == nodes[j]->GetIfByLink(this)->GetMACAddr())
	{
	  return nodes[j]->GetIfByLink(this);
	}
    } 

  return nil;                         // Out of range
}

void  Ethernet::AddNode(Node* n)
{
  if(NodeIsPeer(n))
    {
      cout << "N" << n->Id() << " is already in list " << endl;
      return;
    }
  nodes.push_back(n);
  size++;

  // Add interface
  Interface* iface;

  if(detail == FULL)
  {	   
    iface = n->AddInterface(L2Proto802_3(), InterfaceEthernet(), nextIP, mask);
  }
  else
  {
    iface = n->AddInterface(L2Proto802_3(), nextIP, mask);
  }
  nextIP++;
  iface->SetLink(this);          // Uses this link
  Queue* q = iface->GetQueue();
  if (q)
  {
    q->Detailed(true);  // Can't use non-detailed q
  }
  if (l4proto)
  { // Layer 4 specified
    PortId_t port = l4proto->localPort;  // Port to use
    L4Protocol* p = l4proto->Copy();     // Get a copy of this protocol
    p->Attach(n);                // Set attached node
    p->Bind(port);                  // Bind to specified port
  }

  
  // Sending update node events to all nodes connected to this link 
  if(detail == FULL)
  {
    for (Count_t j = 0; j < size; ++j)
		{ 	  
			Interface* ifc = GetIf(j);
			EthernetEvent* upd = new EthernetEvent(EthernetEvent::UPDATE_NODES);
			Simulator::instance->Schedule(upd, 0.0, ifc);
		}
	}    
}


// Private methods

// void Ethernet::ConstructorHelper(
//            Count_t n,               // Node count
//            Node*   g,               // Gateway node (nil if none)
//            IPAddr_t i, Mask_t mask, // IP addr start, mask
//            SystemId_t id,
//            L4Protocol* l4,          // L4 protocol to bind to each
//            Rate_t r, Time_t dly, Detail_t d)
// {
//   DEBUG0((cout << "Creating Ethernet, firstIP " << (string)IPAddr(i)
//           << " nnodes " << n << endl));
//   first = Node::nextId;             // Set the address of first node
//   firstMac = MACAddr::next;         // First MAC address
//   firstIP = i;                      // First IP Address
//   detail = d;                       // Detail level
//   bw = r;                           // Bandwidth
//   delay = dly;                      // Delay
//   if (g)
//     { // Gateway specified, add the interface
//       gwIf = g->AddInterface(L2Proto802_3(), i, mask);
//       gwIf->SetLink(this);           // Uses this link
//       Queue* q = gwIf->GetQueue();
//       if (q) 
//         {
//           q->Detailed(true);  // Can't use non-detailed q
//         }
//       i++;                          // Advance the ip address
//     }
//   for (Count_t j = 0; j < n; ++j)
//     { // Create the nodes
//       Node* node = new Node(id);
//       Interface* iface = node->AddInterface(L2Proto802_3(), i, mask);
//       iface->SetLink(this);          // Uses this link
//       Queue* q = iface->GetQueue();
//       if (q)
//         {
//           q->Detailed(true);  // Can't use non-detailed q
//         }
//       if (l4)
//         { // Layer 4 specified
//           PortId_t port = l4->localPort;  // Port to use
//           L4Protocol* p = l4->Copy();     // Get a copy of this protocol
//           p->Attach(node);                // Set attached node
//           p->Bind(port);                  // Bind to specified port
//         }
//       ++i;                          // Advance IP Address
//     }
//   last = Node::nextId;              // Last id + 1
// } 

void Ethernet::ConstructorHelper(
           Count_t n,               // Node count
           Node*   g,               // Gateway node (nil if none)
           IPAddr_t i, Mask_t mask, // IP addr start, mask
           SystemId_t id,
           L4Protocol* l4,          // L4 protocol to bind to each
           Rate_t r, Time_t dly, Detail_t d)
{
  DEBUG0((cout << "Creating Ethernet, firstIP " << (string)IPAddr(i)
       << " nnodes " << n << " detail "<< d << endl));
  first = Node::nextId;             // Set the address of first node
  firstMac = MACAddr::next;         // First MAC address
  firstIP = i;                      // First IP Address
  detail = d;                       // Detail level
  bw = r;                           // Bandwidth
  delay = dly;                      // Delay

  if (g)
  { // Gateway specified, add the interface
		if(detail == FULL)
		{	    
			gwIf = g->AddInterface(L2Proto802_3(), InterfaceEthernet(), i, mask );  
			g->SetLocation(0, 0);
		}
    else
		{
	   gwIf = g->AddInterface(L2Proto802_3(), i, mask);
		}
		gwIf->SetLink(this);           // Uses this link
		Queue* q = gwIf->GetQueue();
		if (q) 
		{
			q->Detailed(true);  // Can't use non-detailed q
		}
		i++;                          // Advance the ip address
	}
  for (Count_t j = 0; j < n; ++j)
  { // Create the nodes
    Node* node = new Node(id);
    Interface* iface;

    if(detail == FULL)
		{	   
			iface = node->AddInterface(L2Proto802_3(), InterfaceEthernet(), i, mask );
			node->SetLocation(0, j + 1 );
		}
		else
		{
			iface = node->AddInterface(L2Proto802_3(), i, mask);
		}
		iface->SetLink(this);          // Uses this link
		Queue* q = iface->GetQueue();
		if (q)
		{
			q->Detailed(true);  // Can't use non-detailed q
		}
		if (l4)
		{ // Layer 4 specified
			PortId_t port = l4->localPort;  // Port to use
			L4Protocol* p = l4->Copy();     // Get a copy of this protocol
			p->Attach(node);                // Set attached node
			p->Bind(port);                  // Bind to specified port
		}
		++i;                          // Advance IP Address
	}
  last = Node::nextId;              // Last id + 1
  nextIP = i;
  // Sending update node events to all nodes connected to this link
  if(detail == FULL)
  {
    UseSeqEvents(false);
    for (Count_t j = 0; j < size; ++j)
		{ 	  
			Interface* iface = GetIf(j);
			EthernetEvent* upd = new EthernetEvent(EthernetEvent::UPDATE_NODES);
			Simulator::instance->Schedule(upd, 0.0, iface);
		}
	}    
} 


//////////// End of Mohamed's Modifications ///////////////


bool Ethernet::Transmit(Packet* p, Interface* i, Node* n)
{
  L2802_3Header* l2pdu = (L2802_3Header*)p->PeekPDU(); // Get a ptr to l2 hdr
  Interface* targ_if = GetIfByMac(l2pdu->src);
  if (l2pdu->dst.IsBroadcast())
    { // Broadcast, pass interface of source, not dst
      return TransmitHelper(p, i, n, targ_if, true);
    }
  else
    {
      Interface* targ_if = GetPeer(p);
      if (!i)
        {
          cout << "HuH?  Ethernet::Transmit null peer i/f" << endl;
          L2802_3Header* l2pdu = (L2802_3Header*)p->PeekPDU();
          cout << "Mac dst " << l2pdu->dst << endl;
          IPV4Header* l3pdu = (IPV4Header*)p->PeekPDU(1);
          cout << "Src IP "  << (string)IPAddr(l3pdu->src)
               << " dst IP " << (string)IPAddr(l3pdu->dst) << endl;
          return false;
        }
      return TransmitHelper(p, i, n, targ_if);
    }
}

bool Ethernet::Transmit(Packet* p, Interface* i, Node* n, Rate_t)
{
  return Transmit(p, i, n);
}


IPAddr_t Ethernet::PeerIP(Count_t npeer)
{ // Get peer IP Address (if known)
  Node* n = GetNode(npeer);  // Get the node    
  if (n) return n->GetIPAddr(); // get the node's ip address
  return IPADDR_NONE;        // Problem, return none
}


IPAddr_t Ethernet::DefaultPeer(Interface* fromIf)
{ // Get IP Addr of default peer
  if (gw)
    {
      IPMaskVec_t ipmasks;
      gw->IPAddrs(ipmasks); // Get the list of ipaddresses for gateway
      // Find the one matching my subnet
      IPAddr_t ip = fromIf->GetIPAddr();
      DEBUG0((cout << "Ethernet::DefaultPeer, fromIF " 
              << (string)IPAddr(ip) << endl));
      
      for (IPMaskVec_t::size_type i = 0; i < ipmasks.size(); ++i)
        {
          DEBUG0((cout << "Ethernet::DefaultPeer, checking "
                  << (string)IPAddr(ipmasks[i].ip)
                  << " vs " << (string)IPAddr(ip) << endl));
          IPAddr_t pip = ipmasks[i].ip;
          Mask_t   m   = ipmasks[i].mask;
          if ((ip & m) == (pip & m))
            {
              DEBUG0((cout << "Ethernet::DefaultPeer, returning " 
                      << (string)IPAddr(pip) << endl));
              return pip; // found it
            }
        }
    }
  DEBUG0((cout << "Ethernet::DefaultPeer returning IPADDR_NONE" << endl));
  return IPADDR_NONE;  // No default peer if no gateway
}

Count_t  Ethernet::NeighborCount(Node* n)
{ // Number of routing neighbors
  if (gw)
    {
      if (gw != n)
        {
          return 1;// If a gw system and not the gw, return just one
        }
    }
  return PeerCount(); // Otherwise return number of peers
}

void     Ethernet::Neighbors(NodeWeightVec_t& nwv, Interface* i, bool forceAll)
{ // Add to neighbor list
  // If a gateway system, all leafs just claim the gateway as
  // a single neighbor.  The gateway itself claims no neighbors.
  if (gw && !forceAll)
    {
      Node* n = i->GetNode();
      if (n == gw) return;  // This is the gateway node, no neighbors
      nwv.push_back(NodeIfWeight(gw, i, Weight())); // Otherwise gw is neighbor
      return;
    }
  // Not a single gateway, a bit more complex
  // Add each node in the list.  Note that the list does include
  // all nodes, which means certain route calculations must take
  // care to ignore "self" adjacenies
  for (NodeId_t ni = first; ni < last; ++ni)
    {
       nwv.push_back(NodeIfWeight(Node::GetNode(ni), i, Weight()));
    }
}

void     Ethernet::AllNeighbors(NodeWeightVec_t& nwv)
{ // Make a list of all neighbors.  Note the the "Interface" pointer
  // in the NodeIfWeight vector is the TARGET interface, not the
  // source as in Neighbors above.  This function is used by L2
  // broadcast to determine where to schedule receive events.
  // 
  for (NodeId_t ni = first; ni < last; ++ni)
    {
      Interface* iface = GetIf(ni);
      if (!iface) cout << "HuH?  Ethernet::AllNeighbors nill iface for " 
                       << ni << endl;
      nwv.push_back(NodeIfWeight(Node::GetNode(ni), GetIf(ni), Weight()));
    }
}




Interface* Ethernet::GetIf(Count_t i)
{ // Get the interface for the i'th node
  Node* n = GetNode(i);
  if (!n) return nil;
  return n->GetIfByLink(this);
}


Node* Ethernet::Gateway()
{
  return gw; // Return the gateway node
}

void  Ethernet::SetLeafQLimit(Count_t l)
{
  Count_t r = last - first;     // Number nodes
  if (gw) r++;                  // Include gateway
  for (Count_t t = 0; t < r; ++t)
    {
      Node* n = GetNode(t);
      if (n)
        {
          Queue* q = n->GetQueue(); // Get the queue for single interface
          if (q)
            { // Found single interface node, set specified limit
              q->SetLimit(l);
            }
        }
    }
}
#endif //ETHERNET_BOEING
