/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Extended tracing (hacked in by Boeing) 

#include <iostream>
#include "netinet/in.h" //ntohl and ntohs
#include "arpa/inet.h" //inet_ntop

#include "etrace.h"
#include "protograph.h"
#include "node.h"
#include "packet.h"
#include "interface.h"
#include "simulator.h"
#include "tfstream.h"
#include "ipaddr.h"
#include "ipv4.h"
#include "l2proto802.11.h"
#include "l2proto802.3.h"
#include "datapdu.h"
#include "ospf6_message.h"
#include "ospf6_proto.h" //for ospf6_options_printbuf
#include "ospf6_lsa.h" //for ospf6_lsa_header
#include "zebra.h" //for CHECK_FLAG
#ifdef USAP_ADDON
#include "l2protousap.h"
#endif // USAP_ADDON

#include "application-ospf6d.h"  // For position tracing
#include "ospf6_area.h"
#include "ospf6_interface.h"
#ifdef OSPF6_MANET_MDR_FLOOD
#include "ospf6_mdr.h"
#endif //OSPF6_MANET_MDR_FLOOD
#include "ospf6_neighbor.h"

using namespace std;

ETrace*            ETrace::instance;        // Single global trace object
bool              ETrace::enabled = false; // Global enabled flag
bool              ETrace::extended = false; // Global extended tracing flag
int               ETrace::trace_event_level = 0;
ETrace::Callback_t ETrace::callback = nil;  // Tracing callback

ETrace::ETrace() 
  : trace(ProtocolGraph::N_LAYERS, DISABLED), node(0), trace_on_time(0),
    trace_off_time(0), tracetopo(0)
{
}

ETrace::~ETrace() 
{
  Close(); // Close the file if open
}

bool ETrace::Open(const char* fn)
{
  os.open(fn, ios::out);
  if (os) {
		enabled = true;
		LayerOn();
	}
  return os;  // True if opened successfully
}

void ETrace::Close()
{
  if (os.is_open())
    {
      os.AppendEOL();
      os.close();
    }
  enabled = false;
}

void ETrace::LayerOn()
{ // Enable all
  for (ETSVec_t::size_type i = 0; i < trace.size(); ++i)
    {
      trace[i] = ENABLED;
    }
}

void ETrace::LayerOn(Layer_t l)
{
  if (l > ProtocolGraph::N_LAYERS) return;  // Out of range
  trace[l-1] = ENABLED;
}

void ETrace::LayerOff()
{ // Disable all
  for (ETSVec_t::size_type i = 0; i < trace.size(); ++i)
    {
      trace[i] = DISABLED;
    }
}

void ETrace::LayerOff(Layer_t l)
{
  if (l > ProtocolGraph::N_LAYERS) return;  // Out of range
  trace[l-1] = DISABLED;
}

bool ETrace::IsLayerOn(Layer_t l)
{ // Enquire trace status by layer
  if (l > ProtocolGraph::N_LAYERS) return false;  // Out of range
  if (!os.is_open()) return false;  // Not enabled if trace file not open
  return trace[l-1] == ENABLED;
}

bool ETrace::IsEnabled()
{
  return(os.is_open());
}

void ETrace::NodeOff(int id)
{
	vNodeIds::iterator ids_iter;
	
	ids_iter = find(vnodeids.begin(), vnodeids.end(), id);
	if(ids_iter != vnodeids.end())
		return;
	vnodeids.push_back(id);
}

void ETrace::NodeOn()
{
	vnodeids.clear();
}

void ETrace::NodeOn(int id)
{
	vNodeIds::iterator ids_iter;
	
	ids_iter = find(vnodeids.begin(), vnodeids.end(), id);
	if(ids_iter == vnodeids.end())
		return;
	vnodeids.erase(ids_iter);
}

bool ETrace::IsNodeOn(int id)
{
	vNodeIds::iterator ids_iter;
	
	ids_iter = find(vnodeids.begin(), vnodeids.end(), id);
	if(ids_iter == vnodeids.end())
		return true;
	return false;
}

void ETrace::SetTraceEventLevel(int i)
{
	trace_event_level = i;
}

void ETrace::TraceEvent(Node *n, char *str, int level)
{
	if (!enabled)
		return;
  if (!IsNodeOn(n->Id()))
    return;
	if (level > trace_event_level)
		return;
  if (trace_off_time && (Simulator::Now() > trace_off_time || 
      Simulator::Now() < trace_on_time))
    return;
  if (tracetopo)
    return; // XXX-- should not be mutually incompatible, but doing this
            // as a quick and dirty way to enable this, so the tracefiles
            // don't get too large

  // This looks a bit crufty, but is what is needed to prepend str to the
  // traceline
	os.pendingEOL = true;
	SetNode(n);   
	os << "e ";
  os.pendingEOL = false;
	NewNode(n);   
	
	os << " " << str << "\n";
}

void ETrace::TracePacket(Packet *p, Node* n, Interface* i, Layer_t l, char* str) 
{
  PDU* pdup;
	int offset = 0; // May not be needed

	
	if (!enabled) 
		return;
	if (!IsLayerOn(l))
		return;
	if (!IsNodeOn(n->Id()))
		return;
  if (trace_off_time && (Simulator::Now() > trace_off_time || 
      Simulator::Now() < trace_on_time))
    return;

  // This looks a bit crufty, but is what is needed to prepend str to the
  // traceline
	os.pendingEOL = true;
	SetNode(n);   
	os << str << " ";
  os.pendingEOL = false;
	NewNode(n);   

	os << " -int " << i->GetIfIndex();
  os << " -pkt_id " << p->uid << " ";

	switch (l) { // based on layer to trace
		case 1 : 
      // XXX this is a hack to turn off the "S", "D", and "R" tracing
			{// Look at layer 3.  FindPDU(3) seems to be reliable call
				pdup = p->FindPDU(3);
    		if (pdup) {
  	   	   offset += pdup->Size();
	     	   TraceIPV4(os, pdup); 
      	} 
    	  IPV4Header* iph = (IPV4Header*) pdup;
	  if ((iph) && (iph->protocol == 89)) {// Find and parse OSPF header
	      //pdup = p->PeekPDU(1); 
	      pdup = p->FindPDU(0,0); // should be OSPF PDU
	      TraceOspf6(os, pdup);
	  }
				break;
			}
			
		case 2 : 
			{// Look at layer 2
				pdup = p->FindPDU(2);	
				if (pdup) {
					offset = pdup->Size();
					// Look at the type
					switch(pdup->Proto()) {
						case 3: 
							Trace8023(os, pdup); 
							break;
      	   case 11:
    	     	  Trace80211(os, pdup); 
  	        	break;
#ifdef USAP_ADDON
					 case 18:
					    TraceUsap(os,pdup);
						  break;
#endif // USAP_ADDON
	         default:
        	 	 cerr << "Unknown l2 protocol number " << pdup->Proto() << endl;
      	   	 exit(1);
					}
				} 
			} // no break here, we want both layer 2 and 3

		case 3 : 
			{// Look at layer 3.  FindPDU(3) seems to be reliable call
				pdup = p->FindPDU(3);
    		if (pdup) {
  	   	   offset += pdup->Size();
	     	   TraceIPV4(os, pdup); 
      	} 
    	  IPV4Header* iph = (IPV4Header*) pdup;
	  if ((iph) && (iph->protocol == 89)) {// Find and parse OSPF header
	      //pdup = p->PeekPDU(1); 
	      pdup = p->FindPDU(0,0); // should be OSPF PDU
	      TraceOspf6(os, pdup);
	  }
				break;
			}
	}
	AppendEOL(); 
}

#ifdef USAP_ADDON
void ETrace::TraceUsap(Tfstream& os, PDU* pdup)
{
  L2HeaderUsap* l2h = (L2HeaderUsap*) pdup;
  os << "-l2proto Usap ";
  switch (l2h->type) {
    case USAP_UCST_DATA:
      os << "-type UCST_DATA ";
      break;
    case USAP_BCST_DATA:
      os << "-type BCST_DATA ";
      break;
    case USAP_UCST_DATA_ACK:
      os << "-type ACK ";
      break;
    case USAP_CTL:
      os << "-type CTL ";
      break;

  }
  os << "-ra " << IPAddr::ToDotted(l2h->ra);
}
#endif // USAP_ADDON

void ETrace::Trace80211(Tfstream& os, PDU* pdup)
{
  L2Header802_11* l2h = (L2Header802_11*) pdup;
  os << "-l2proto 802.11 ";
  switch (l2h->type) {
    case RTS:
      os << "-type RTS ";
      break;
    case CTS:
      os << "-type CTS ";
      break;
    case DATA:
      os << "-type DATA ";
      break;
    case ACK:
      os << "-type ACK ";
      break;

  }
  os << "-ra " << IPAddr::ToDotted(l2h->ra);
}

void ETrace::Trace8023(Tfstream& os, PDU* pdup)
{
  L2802_3Header* l2h = (L2802_3Header*) pdup;
   
  os << "-l2proto 802.3 -dst " << l2h->dst;
  os << " -src " << l2h->src;
  os << " -len " << l2h->length;
}

void ETrace::TraceIPV4(Tfstream& os, PDU* pdup)
{
  IPV4Header* iph = (IPV4Header*) pdup;
  //os << " -len " << l2h->length;
  os << " -l3proto IPv4 -src " << IPAddr::ToDotted(iph->src);
  os << " -dst " << IPAddr::ToDotted(iph->dst);
  os << " -len " << iph->totalLength;
  os << " -ttl " << iph->ttl;
  switch (iph->protocol) {
    case 89:
      os << " -proto OSPFv3 ";
      break;
    case 17:
      os << " -proto UDP ";
      break;
    default:
      os << " -proto " << iph->protocol;
      break;
  }
}

void ETrace::TraceOspf6(Tfstream& os, PDU *pdup)
{
	Data *d = (Data*) pdup;
	struct ospf6_header *oh = (struct ospf6_header *) d->data;

	switch (oh->type)
  {
    case OSPF6_MESSAGE_TYPE_HELLO:
			os << "-type HELLO ";
      break;
		case OSPF6_MESSAGE_TYPE_DBDESC:
			os << "-type DBDESC ";
			break;
		case OSPF6_MESSAGE_TYPE_LSREQ:
			os << "-type LSR ";
			break;
		case OSPF6_MESSAGE_TYPE_LSUPDATE:
			os << "-type LSU ";
			break;
		case OSPF6_MESSAGE_TYPE_LSACK:
			os << "-type LSAck ";
			break;
		default:
			break;
	}

	if (!extended)
		return;
	
	os << "-len " << ntohs(oh->length) << " ";
	os << "-rid " << IPAddr::ToDotted(ntohl(oh->router_id)) << " ";
	os << "-aid " << IPAddr::ToDotted(ntohl(oh->area_id)) << " ";

	switch (oh->type)
  {
    case OSPF6_MESSAGE_TYPE_HELLO:
			TraceOspf6Hello(os, oh);
      break;
		case OSPF6_MESSAGE_TYPE_DBDESC:
			TraceOspf6Dbdesc(os, oh);
			break;
		case OSPF6_MESSAGE_TYPE_LSREQ:
			TraceOspf6Lsr(os, oh);
			break;
		case OSPF6_MESSAGE_TYPE_LSUPDATE:
			TraceOspf6Lsu(os, oh);
			break;
		case OSPF6_MESSAGE_TYPE_LSACK:
			TraceOspf6Lsa(os, oh);
			break;
		default:
			break;
	}
}

void ETrace::TraceOspf6Hello(Tfstream& os, struct ospf6_header *oh)
{
  struct ospf6_hello *hello;
  char options[36];
  char drouter[16], bdrouter[16], neighbor[16];
  char *p;
  struct ospf6_LLS_header * lls_ptr;
  struct ospf6_TLV_header *tlv_header;
  int length_lls = 0, tlv_len=0;

  assert (oh->type == OSPF6_MESSAGE_TYPE_HELLO);

  hello = (struct ospf6_hello *) (oh +1);

  inet_ntop (AF_INET, &hello->drouter, drouter, sizeof (drouter));
  inet_ntop (AF_INET, &hello->bdrouter, bdrouter, sizeof (bdrouter));
#ifdef QUAGGA_0_96_5
  ospf6_options_printbuf ((char *) hello->options, options, sizeof (options));
#else
  ospf6_options_printbuf ( hello->options, options, sizeof (options));
#endif //QUAGGA_0_96_5
  
  os << "-I/FId " << ntohl(hello->interface_id) << " "; 
  //os << "-Prio " << hello->priority << " ";
  os << "-Opt " << options << " ";
  os << "-HInt " << ntohs(hello->hello_interval) << " ";
  os << "-DInt " << ntohs (hello->dead_interval) << " ";
  os << "-DR " << drouter << " ";
  os << "-BDR "<< bdrouter << " ";
  os << "-Neigh: ";

  for (p = (char *) (hello + 1);
       p + sizeof (u_int32_t) <= OSPF6_MESSAGE_END (oh);
       p += sizeof (u_int32_t))
    {
      inet_ntop (AF_INET, (void *) p, neighbor, sizeof (neighbor));
			os << neighbor << " ";
    }
#ifdef OSPF6_MANET
	/* process TLVs */
	/* set LLS pointer */
	if (!OSPF6_OPT_ISSET(hello->options, OSPF6_OPT_L, 1))
		return;
	lls_ptr = (struct ospf6_LLS_header *) p;

  length_lls =  ntohs(lls_ptr->len) * 4 - sizeof(struct ospf6_LLS_header);
  tlv_header = (struct ospf6_TLV_header *) (lls_ptr + 1);

  TraceOspf6LLS(os, lls_ptr);

  while(length_lls)
  {
    TraceOspf6Tlv(os, tlv_header);
    tlv_len = sizeof(tlv_header) + ntohs(tlv_header->len);
    length_lls -= tlv_len;
    tlv_header = (struct ospf6_TLV_header *) ((char *) tlv_header + tlv_len);
  }		
#endif //OSPF6_MANET
}

void ETrace::TraceOspf6Dbdesc(Tfstream& os, struct ospf6_header *oh)
{
	struct ospf6_dbdesc *dbdesc;
  char options[36];
  char *p;
  struct ospf6_LLS_header * lls_ptr;
  struct ospf6_TLV_header *tlv_header;
  int length_lls = 0, tlv_len=0;

  assert (oh->type == OSPF6_MESSAGE_TYPE_DBDESC);

  dbdesc = (struct ospf6_dbdesc *) (oh + 1);

#ifdef QUAGGA_0_96_5
  ospf6_options_printbuf ((char *)dbdesc->options, options, sizeof (options));
#else
  ospf6_options_printbuf ((u_char *)dbdesc->options, options, sizeof (options));
#endif //QUAGGA_0_96_5

  os << "-Opt " << options << " ";
  os << "-IfMTU " << ntohs (dbdesc->ifmtu) << " ";
  os << "-Bits ";
  os << (CHECK_FLAG (dbdesc->bits, OSPF6_DBDESC_IBIT) ? "I" : "*") << " ";
  os << (CHECK_FLAG (dbdesc->bits, OSPF6_DBDESC_MBIT) ? "M" : "*") << " ";
  os << (CHECK_FLAG (dbdesc->bits, OSPF6_DBDESC_MSBIT) ? "m" : "s") << " ";
  os << "-Seq " << ntohl (dbdesc->seqnum) << " ";

  for (p = (char *) ((caddr_t) dbdesc + sizeof (struct ospf6_dbdesc));
       p + sizeof (struct ospf6_lsa_header) <= OSPF6_MESSAGE_END (oh);
       p += sizeof (struct ospf6_lsa_header))
    TraceOspf6LsaHeader(os, (ospf6_lsa_header *) p);

  if (p != OSPF6_MESSAGE_END (oh))
     os << "Trailing garbage exists "; 

  if (!OSPF6_OPT_ISSET(dbdesc->options, OSPF6_OPT_L, 1))
    return;
  lls_ptr = (struct ospf6_LLS_header *) p;

  length_lls =  ntohs(lls_ptr->len) * 4 - sizeof(struct ospf6_LLS_header);
  tlv_header = (struct ospf6_TLV_header *) (lls_ptr + 1);
  TraceOspf6LLS(os, lls_ptr);

  while(length_lls)
  {
    TraceOspf6Tlv(os, tlv_header);
    tlv_len = sizeof(tlv_header) + ntohs(tlv_header->len);
    length_lls -= tlv_len;
    tlv_header = (struct ospf6_TLV_header *) ((char *) tlv_header + tlv_len);
  }		
}

void ETrace::TraceOspf6Lsr(Tfstream& os, struct ospf6_header *oh)
{
  char id[16], adv_router[16];
  char *p;

  assert (oh->type == OSPF6_MESSAGE_TYPE_LSREQ);

  for (p = (char *) ((caddr_t) oh + sizeof (struct ospf6_header));
       p + sizeof (struct ospf6_lsreq_entry) <= OSPF6_MESSAGE_END (oh);
       p += sizeof (struct ospf6_lsreq_entry))
    {
      struct ospf6_lsreq_entry *e = (struct ospf6_lsreq_entry *) p;
      inet_ntop (AF_INET, &e->adv_router, adv_router, sizeof (adv_router));
      inet_ntop (AF_INET, &e->id, id, sizeof (id));
#ifdef QUAGGA_0_96_5
      os << "-Lsa " << OSPF6_LSTYPE_NAME (e->type) << " ";
#else
      os << "-Lsa " << ospf6_lstype_name(e->type) << " ";
#endif //QUAGGA_0_96_5

			os << "-id " <<  id << " ";
			os << "-advrt " << adv_router << " ";
    }

  if (p != OSPF6_MESSAGE_END (oh))
    os << "Trailing garbage exists ";
}

void ETrace::TraceOspf6Lsu(Tfstream& os, struct ospf6_header *oh)
{
  struct ospf6_lsupdate *lsupdate;
  u_long num;
  char *p;

  assert (oh->type == OSPF6_MESSAGE_TYPE_LSUPDATE);

  lsupdate = (struct ospf6_lsupdate *) (oh + 1);

  num = ntohl (lsupdate->lsa_number);
  os << "-Num " <<  num << " ";

  for (p = (char *) ((caddr_t) lsupdate + sizeof (struct ospf6_lsupdate));
       p < OSPF6_MESSAGE_END (oh) &&
       p + OSPF6_LSA_SIZE (p) <= OSPF6_MESSAGE_END (oh);
       p += OSPF6_LSA_SIZE (p))
    {
			TraceOspf6LsaHeader(os, (ospf6_lsa_header *) p);
      if (OSPF6_LSA_SIZE (p) < sizeof (struct ospf6_lsa_header))
        {
          os << "Malformed LSA length, quit printing ";
          break;
        }
    }

  if (p != OSPF6_MESSAGE_END (oh))
    {
      char buf[32];

      int num = 0;
      memset (buf, 0, sizeof (buf));

      os << "Trailing garbage exists ";
      while (p < OSPF6_MESSAGE_END (oh))
        {
          snprintf (buf, sizeof (buf), "%s %2x", buf, *p++);
          num++;
          if (num == 8)
            {
              os << " " << buf << " ";
              memset (buf, 0, sizeof (buf));
              num = 0;
            }
        }
      if (num)
        os << " " << buf << " ";
    }
}

void ETrace::TraceOspf6Lsa(Tfstream& os, struct ospf6_header *oh)
{
  char *p;

  assert (oh->type == OSPF6_MESSAGE_TYPE_LSACK);

  for (p = (char *) ((caddr_t) oh + sizeof (struct ospf6_header));
       p + sizeof (struct ospf6_lsa_header) <= OSPF6_MESSAGE_END (oh);
       p += sizeof (struct ospf6_lsa_header))
		TraceOspf6LsaHeader(os, (ospf6_lsa_header *) p);

  if (p != OSPF6_MESSAGE_END (oh))
    os << "Trailing garbage exists ";
}

void ETrace::TraceOspf6LsaHeader(Tfstream& os, struct ospf6_lsa_header *header)
{
  char id[16], adv_router[16];
  inet_ntop (AF_INET, &header->id, id, sizeof (id));
  inet_ntop (AF_INET, &header->adv_router, adv_router,
             sizeof (adv_router));
#ifdef QUAGGA_0_96_5
  os << "-Lsa " << OSPF6_LSTYPE_NAME (header->type) << " ";
#else
		os << "-Lsa " << ospf6_lstype_name(header->type) << " ";
#endif //QUAGGA_0_96_5
	os << "-id " << id << " ";
	os << "-advrt " << adv_router << " ";
#ifdef SIM_ETRACE_STAT
	os << "-age " << ntohs (age_mask(header)) << " ";
#else
	os << "-age " << ntohs (header->age) << " ";
#endif //SIM_ETRACE_STAT
	os << "-seq " << ntohl (header->seqnum) << " ";
	os << "-len " << ntohs (header->length) << " ";
}

#ifdef OSPF6_MANET
void ETrace::TraceOspf6LLS(Tfstream& os, struct ospf6_LLS_header * lls_header)
{
  os << "-LLSlen " << (ntohs(lls_header->len) * 4) << " ";
  os << "-LLSchk " << ntohs(lls_header->cksum) << " ";
}

void ETrace::TraceOspf6Tlv(Tfstream& os, struct ospf6_TLV_header * tlv_header)
{
  int i;

  switch (ntohs(tlv_header->type))
  {
    case OSPF6_TLV_TYPE_OPTIONS:
      {
				os << "-TLVOPT -len " << ntohs(tlv_header->len) << " ";
        struct ospf6_options_TLV *tlv_val_opt;
        tlv_val_opt = (struct ospf6_options_TLV *) (tlv_header + 1);
				os << "-Opt " << tlv_val_opt->options[0] << "|";
				os << tlv_val_opt->options[1] << "|";
				os << tlv_val_opt->options[2] << "|";
				os << tlv_val_opt->options[3] << " ";
        break;
      }
#ifdef OSPF6_MANET_DIFF_HELLO
    case OSPF6_TLV_TYPE_SCS:
      {
				os << "-TLVSCS -len " << ntohs(tlv_header->len) << " ";
        struct ospf6_scs_TLV *tlv_val_scs;
        tlv_val_scs = (struct ospf6_scs_TLV *) (tlv_header + 1);
				os << "-Seq " << ntohs(tlv_val_scs->number) << " ";
				os << "-Bits ";
				os << (OSPF6_TLV_SCS_OPT_ISSET(tlv_val_scs->bits, OSPF6_TLV_SCS_OPT_R, 0)?"R":"*") << "|";
				os << (OSPF6_TLV_SCS_OPT_ISSET(tlv_val_scs->bits, OSPF6_TLV_SCS_OPT_FS, 0)?"FS":"*") << "|";
				os << (OSPF6_TLV_SCS_OPT_ISSET(tlv_val_scs->bits, OSPF6_TLV_SCS_OPT_N, 0)?"N":"*") << " ";

        break;
      }
    case OSPF6_TLV_TYPE_NEIGHDROP:
      {
        os << "-TLVNDROP -len " << ntohs(tlv_header->len) << " ";
        u_int32_t *tlv_val_drop = NULL;
        tlv_val_drop = (u_int32_t*) (tlv_header + 1);
        os << "-Neigh: ";
        for (i = 0; i < ntohs(tlv_header->len)/4; i++)
          os << IPAddr::ToDotted(ntohl(tlv_val_drop[i])) << " ";
        break;
      }
    case OSPF6_TLV_TYPE_REQUEST:
    {
      os << "-TLVREQ -len " << ntohs(tlv_header->len) << " ";
      u_int32_t *tlv_val_req = NULL;
      tlv_val_req = (u_int32_t*) (tlv_header + 1);
      os << "-Neigh: ";
      for (i = 0; i < ntohs(tlv_header->len)/4; i++)
        os << IPAddr::ToDotted(ntohl(tlv_val_req[i])) << " ";
      break;
    }
#endif
#ifdef OSPF6_MANET_MPR_FLOOD
#ifdef OSPF6_MANET_MPR_HELLO_OSPF_MPR
  case OSPF6_TLV_TYPE_FLOODING_MPR:
    {
      os << "-TLVREL -len " << ntohs(tlv_header->len) << " ";
      struct ospf6_flooding_mpr_TLV *tlv_val_flooding_mpr;
      u_int32_t *flooding_mpr_ptr;
      int nbFloodingMPR;

      tlv_val_flooding_mpr = (struct ospf6_flooding_mpr_TLV*)(tlv_header + 1);
      flooding_mpr_ptr = (u_int32_t *)(tlv_val_flooding_mpr + 1);
      nbFloodingMPR = (ntohs(tlv_header->len) - sizeof(struct ospf6_flooding_mpr_TLV)) / sizeof(u_int32_t);

      os << "-Willingness " << (int)tlv_val_flooding_mpr->willingness << " ";
      os << "-nbSymNeigh " << (int)tlv_val_flooding_mpr->nbSymNeigh << " ";
      os << "-MPR: ";
      for (i=0; i < nbFloodingMPR; i++)
	os << IPAddr::ToDotted(ntohl(flooding_mpr_ptr[i])) << " ";
      break;
    }
#else //OSPF6_MANET_MPR_HELLO_OSPF_MPR
    case OSPF6_TLV_TYPE_RELAY:
      {
				os << "-TLVREL -len " << ntohs(tlv_header->len) << " ";
        struct ospf6_relay_TLV *tlv_val_relay;
        u_int32_t *relay_id_ptr;

        tlv_val_relay = (struct ospf6_relay_TLV*) (tlv_header + 1);
        relay_id_ptr = (u_int32_t *) (tlv_val_relay + 1);

	os << "-add " << (int) tlv_val_relay->added << " -Rel: ";
	for (i = 0; i < ntohs(tlv_header->len)/4-1; i++)
	  os << IPAddr::ToDotted(ntohl(relay_id_ptr[i])) << " ";
        break;
      }
    case OSPF6_TLV_TYPE_WILLINGNESS:
      {
				os << "-TLVWIL -len " << ntohs(tlv_header->len) << " ";
        struct ospf6_will_TLV *tlv_val_will;
        tlv_val_will  = (struct ospf6_will_TLV*) (tlv_header + 1);
				os << "-will " << (int) tlv_val_will->will << " ";
        break;
      }
#endif //OSPF6_MANET_MPR_HELLO_OSPF_MPR
#endif //OSPF6_MANET_MPR_FLOOD
#ifdef OSPF6_MANET_MDR_FLOOD
    case OSPF6_TLV_TYPE_HNL:
      {
        os << "-TLVHNL -len " << ntohs(tlv_header->len) << " ";
        u_int32_t *tlv_val_hnl = NULL;
        tlv_val_hnl = (u_int32_t*) (tlv_header + 1);
        os << "-Neigh: ";
        for (i = 0; i < ntohs(tlv_header->len)/4; i++)
          os << IPAddr::ToDotted(ntohl(tlv_val_hnl[i])) << " ";
        break;
      }
    case OSPF6_TLV_TYPE_RNL:
      {
        os << "-TLVRNL -len " << ntohs(tlv_header->len) << " ";
        u_int32_t *tlv_val_rnl = NULL;
        tlv_val_rnl = (u_int32_t*) (tlv_header + 1);
        os << "-Neigh: ";
        for (i = 0; i < ntohs(tlv_header->len)/4; i++)
          os << IPAddr::ToDotted(ntohl(tlv_val_rnl[i])) << " ";
        break;
      }
    case OSPF6_TLV_TYPE_LNL:
      {
        os << "-TLVLNL -len " << ntohs(tlv_header->len) << " ";
        u_int32_t *tlv_val_lnl = NULL;
        tlv_val_lnl = (u_int32_t*) (tlv_header + 1);
        os << "-Neigh: ";
        for (i = 0; i < ntohs(tlv_header->len)/4; i++)
          os << IPAddr::ToDotted(ntohl(tlv_val_lnl[i])) << " ";
        break;
      }
    case OSPF6_TLV_TYPE_DD:
      {
        os << "-TLVMDR -len " << ntohs(tlv_header->len) << " ";
        struct ospf6_mdr_TLV *mdr_tlv;
        mdr_tlv = (struct ospf6_mdr_TLV *) (tlv_header + 1);
        os << "-ID1 ";
        os << IPAddr::ToDotted(ntohl(mdr_tlv->id1)) << " ";
        os << "-ID2 ";
        os << IPAddr::ToDotted(ntohl(mdr_tlv->id2)) << " ";
        break;
      }
#ifdef OSPF6_MANET_DIFF_HELLO
    case OSPF6_TLV_TYPE_HS:
      {
        os << "-TLVHS -len " << ntohs(tlv_header->len) << " ";
        struct ospf6_seq_TLV *seq_tlv;
        seq_tlv = (struct ospf6_seq_TLV *) (tlv_header + 1);
        os << "-seq ";
        os << ntohs(seq_tlv->number);
        break;
      }
#endif //OSPF6_MANET_DIFF_HELLO
#endif //OSPF6_MANET_MDR_FLOOD
    default:
      {
				os << "-TLV " << ntohs(tlv_header->type) << " ";
				os << "-len " << ntohs(tlv_header->len) << " ";
        break;
      }
  }

}
#endif //OSPF6_MANET

void ETrace::AppendEOL()
{
  if(os)os.AppendEOL(); // Add the end of line if needed
}

void ETrace::NewNode(Node* n)
{ // A packet has arrived at a new node, see if log needed
  os.NewNode(n);
}

void ETrace::SetNode(Node* n)
{ // Set node pointer for subsequent traces
  os.SetNode(n);
}

// Static methods
ETrace* ETrace::Instance()
{
  // If no instance, create one
  if (!instance) instance = new ETrace();
  return instance;
}

// Topology methods
void ETrace::TraceTopo(Time_t interval)
{
  tracetopo = interval; 
  if (tracetopo)
    StartTraceTopo();
}


void ETrace::StartTraceTopo() {
  if (tracetopo) {
    if (!timeoutevent) timeoutevent = new TimerEvent();
    timer.Schedule(timeoutevent, tracetopo, this);
  }
}

void ETrace::Timeout(TimerEvent *ev) {
  TraceTopology();
  timer.Schedule(timeoutevent, tracetopo, this);
}

void ETrace::TraceTopology() {
//  char buf[ETRACE_MAXSTR];
	if (!enabled)
		return;
  if (trace_off_time && (Simulator::Now() > trace_off_time || 
      Simulator::Now() < trace_on_time))
    return;
  // Loop through all nodes
  //
  // For nodes, print out 
  // "n <timestamp> <node> <xcoord> <ycoord> <dr> <bdr>"
  // where <dr> and <bdr> are flags (0 or 1)
  // 
  // For adjacencies, print out if the node is the lower numbered node of
  // the pair (to avoid duplicate entries) 
  // "a <timestamp> <node> <xcoord> <ycoord> <xcoord> <ycoord>"
  // 

  // This looks a bit crufty, but is what is needed to prepend str to the
  // traceline
  const NodeVec_t& nodes = node->GetNodes();
  for (NodeVec_t::const_iterator i = nodes.begin(); i != nodes.end(); ++i)
  {
      struct ospf6_interface *oi = NULL;
      struct ospf6_interface *poi = NULL;
      struct ospf6* pospfinst = NULL;
      struct ospf6_area *poa = NULL;
      struct ospf6_interface *oi_nonloopback = NULL;
      struct ospf6_interface *poi_nonloopback = NULL;
      struct listnode *j, *k, *l;
      Node* pn = NULL;

      Node* n = *i;
	    os.pendingEOL = true;
	    SetNode(n);   
	    os << "n ";
      os.pendingEOL = false;
	    NewNode(n);   

      os << " -xcoord " << n->LocationX();
      os << " -ycoord " << n->LocationY();
      struct ospf6* ospfinst = n->GetOspf6dInst()->get_inst_ospf6();
      if (ospfinst) {
        struct ospf6_area *oa = NULL;
        for (j = listhead (ospfinst->area_list); j; nextnode (j)) {
          oa = (struct ospf6_area *) getdata (j);
          for (k = listhead (oa->if_list); k; nextnode (k)) {
            oi = (struct ospf6_interface *) getdata (k);
            if (oi->type != OSPF6_IFTYPE_LOOPBACK) {
              oi_nonloopback = oi;
#ifdef OSPF6_MANET_MDR_FLOOD
              os << " -dr " << (oi->mdr_level == OSPF6_MDR);
              os << " -bdr " << (oi->mdr_level == OSPF6_BMDR);
#endif //OSPF6_MANET_MDR_FLOOD
            }
          }
        }
      }
	    AppendEOL(); 
      // Loop through the neighbors, and add a trace for each adjacency
      if (ospfinst && oi_nonloopback) {
        oi = oi_nonloopback;
        for (j = listhead(oi->neighbor_list); j; nextnode(j)) {
          struct ospf6_neighbor *on = (struct ospf6_neighbor *) getdata (j);
          if (on->state == OSPF6_NEIGHBOR_FULL) {
            // get peer interface, node pointers
             pn = Node::RouterIdToNode((int)((ntohl(on->router_id)) & 0xff));
             if (!pn) {
               cerr << "No pn" << endl;
               exit(1);
             }
             // compare node ids.  Only need one listing per adjacency.
             // list it only if this node's id is smaller
             if (n->Id() > pn->Id())
               continue;
             pospfinst = pn->GetOspf6dInst()->get_inst_ospf6();
             if (pospfinst) {
               for (k = listhead (pospfinst->area_list); k; nextnode (k)) {
                 poa = (struct ospf6_area *) getdata (k);
                 for (l = listhead (poa->if_list); l; nextnode (l)) {
                   poi = (struct ospf6_interface *) getdata (l);
                   if (poi->type != OSPF6_IFTYPE_LOOPBACK) {
                      poi_nonloopback = poi;
                   }
                 }
               }
             }
             poi = poi_nonloopback;
             if (!poi) {
               cerr << "No poi" << endl;
               exit(1);
             }
	           os.pendingEOL = true;
	           SetNode(n);   
#ifdef OSPF6_MANET_MDR_FLOOD
             if ((poi->mdr_level == OSPF6_MDR || poi->mdr_level == OSPF6_BMDR) 
                && (oi->mdr_level == OSPF6_MDR || oi->mdr_level == OSPF6_BMDR))
                // backbone link
	              os << "b ";
             else
                os << "a ";
#endif //OSPF6_MANET_MDR_FLOOD
             os.pendingEOL = false;
	           NewNode(n);   
             os << " -xcoord " << n->LocationX();
             os << " -ycoord " << n->LocationY();
             os << " -xcoord " << pn->LocationX();
             os << " -ycoord " << pn->LocationY();
             //os << " -n1 " << n->Id() << " -mdr " << oi->mdr_level;
             //os << " -n2 " << pn->Id() << " -mdr " << poi->mdr_level;
	           AppendEOL(); 
          }
        }
      }
  }
}
