/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Circular Queue for Sequential Events
// George F. Riley.  Georgia Tech, Fall 2002

// Defines an efficient circular queue for events that are inserted
// and removed in timestamp order.

#ifndef __eventcq_h__
#define __eventcq_h__

#include <vector>

#include "common-defs.h"
#include "simulator.h"

typedef std::vector<EventPair*> EventCQVec_t;

class EventCQ {
public:
  typedef enum {DefaultInitialSize = 4} EvCQSize_t;
public:
  EventCQ();

  bool empty()
  {
    return in == out;
  }

  void push_back(EventPair* e)         // Add a new entry (grow size if needed)
  {
    timeLastIn = e->time;
    EventCQVec_t::size_type newin = NextIn();
#ifdef DONT_NEED
    // Don't need this since we grow the queue if needed after then enqueue
    if (newin == out)
      {
        Grow();               // Need more room
        newin = NextIn();
      }
#endif
    events[in] = e;           // Store at the in location
    in = newin;               // Advance in
    if (full()) Grow();       // Insure room for next
  }

  void pop_front()                     // Remove the front entry
    { // Remove front element
      out = NextOut(); 
    }

  EventPair*   front()                 // Return the front entry
  { // Get a reference to the front object
    return events[out];
  }

  EventPair*   next()                  // Next entry to be pushed
  {
    return events[in];
  }

  Count_t      size()                  // Number of entries in the list
  {
    if (in >= out) return in - out;
    return limit - (out - in);
  }

  bool         full()                  // True if no more room
  {
    return (NextIn() == out);
  }
  
  EventPair*   operator[](Count_t i)   // Index as array
  {
    EventCQVec_t::size_type k = out + i;
    if(k >= limit) k-=limit;
    return events[k];
  }

private:
  EventCQVec_t::size_type NextIn()     // Get advanced in pointer
  { // Get the advanced in pointer, modulo limit
    EventCQVec_t::size_type r = in + 1;
    if (r != limit) return r;
    return 0;
  }

  EventCQVec_t::size_type NextOut()    // Get advanced out pointer
  { // Get the advanced out pointer, modulo limit
    EventCQVec_t::size_type r = out + 1;
    if (r != limit) return r;
    return 0;
  }

  void       Grow();                   // Increase size by power of 2
private:
  public: // debug..change later
  EventCQVec_t            events;      // Pending events in the circular queue
  EventCQVec_t::size_type limit;       // Current size (always a power of 2)
  EventCQVec_t::size_type in;          // In pointer
  EventCQVec_t::size_type out;         // Out pointer
  Time_t timeLastIn;
};

#endif


