/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - FTP Client Class
// George F. Riley.  Georgia Tech, Spring 2002

// Define the FTP Client Class

#ifndef __ftp_client_h__
#define __ftp_client_h__

#include "common-defs.h"
#include "rng.h"
#include "application.h"
#include "timer.h"
#include "tcp.h"
#include <vector>

//Doc:ClassXRef
class FTPAction {
public:
  typedef enum { NONE, OPEN, GET, PUT, CLOSE, SLEEP, REPEAT } Actions_t;
  FTPAction();
  FTPAction(Actions_t a);
  FTPAction(const FTPAction&);            // Copy constructor
  FTPAction& operator=(const FTPAction&); // Assignment operator
  ~FTPAction();
  void Reset();
public:
  Actions_t action;    // This ftp action
  IPAddr_t  peerIP;    // IP Address of server
  Count_t   reqSz;     // Size of request (if not random)
  Count_t   fileSz;    // Size of reply (if not random)
  Random*   reqRv;     // Random variable for request size
  Random*   fileRv;    // Random variable for fileonse size
  Time_t    sleepTime; // Sleep time
  Random*   sleepRv;   // Sleep time random variable
  Count_t   nRepeat;   // Number of repeats desired (0 if infinite)
  Count_t   cRepeat;   // Number of repeats so far
  Count_t   stepRepeat;// Step number to repeat
};

typedef std::vector<FTPAction> ActionVec_t;

//Doc:ClassXRef
class FTPClient : public Application, public TimerHandler  {
public:
  FTPClient(const TCP& = TCP::Default()); // Specify which TCP to use
  ~FTPClient();
  FTPClient* Copy() const { return new FTPClient(*this); }
  
  // TimerHandler 
  void Timeout(TimerEvent*);

  // Upcalls from L4 protocol
  void Receive(Packet*,L4Protocol*);    // Data received
  void Sent(Count_t, L4Protocol*);      // Data has been sent
  void Closed(L4Protocol*);             // Connection has closed
  void ConnectionComplete(L4Protocol*); // Connection request succeeded
  void ConnectionFailed(L4Protocol*);   // Connection request failed

  // All actions return step number, that can be used as a repeat target
  Count_t Open(IPAddr_t);          // Open session with peer
  Count_t Get(Count_t, Count_t);   // Get, size of req, size of file
  // Get, size of req/file (random)    
  Count_t Get(const Random&, const Random&);
  Count_t Put(Count_t, Count_t);   // Get, size of req, size of file
  // Get, size of req/rile (random)    
  Count_t Put(const Random&, const Random&);
  Count_t Close();                 // Close connection
  Count_t Sleep(Time_t);           // Sleep for specified time
  Count_t Sleep(Random*);          // Sleep for random time
  Count_t Repeat(Count_t,Count_t); // Repeat steps, starting at specified

public:
  virtual void StartApp();       // Called at time specified by Start
  virtual void StopApp();        // Called at time specified by Stop
private:
  typedef enum { CLOSED, CLOSING, OPENING, OPEN, 
                 SENDING_GET, GETTING, SENDING_PUT, PUTTING } State_t;
  typedef enum { SUCCESS, FAILED, FINISHED } ActionStatus_t;
  typedef enum { FTP_PORT = 21 } FTPPort_t;
  State_t     state;
  Count_t     currentStep;
  Count_t     getputSize;        // Size of get or put
  Count_t     getputCurrent;     // Number bytes tx/rx so far
  bool        started;           // True if application is active
  TCP*        l4Proto;           // TCP to use
  ActionVec_t actions;           // Pending actions
  FTPAction*  currentAction;     // Current action
  Timer       timer;             // Sleep timer
  TimerEvent* timeoutEvent;      // Pending timeout event
private:
  ActionStatus_t NextAction();   // Process next action
};

#endif


