/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Histogram Class
// George F. Riley.  Georgia Tech, Spring 2002

// Defines a Histogram and CDF statistics collection

#include "debug.h"
#include "histogram.h"

using namespace std;

// Constructors
Histogram::Histogram()
    : min(0), max(0), outOfRange(0)
{
}

Histogram::Histogram(double m)
    : min(0), max(m), outOfRange(0)
{
}

Histogram::Histogram(double m, Count_t n, double mn)
    : bins(n), min(mn), max(m), outOfRange(0)
{
  Bins(m, n, mn);
}

Histogram::~Histogram()
{
}

void Histogram::Bins(double m, Count_t n, double mn)
{
  bins.clear();                         // Insure empty
  bins.insert(bins.begin(), n, 0);      // Insert "n" empty elements
  outOfRange = 0;
  max = m;
  min = mn;
}

bool Histogram::Record(double v)
{ // Add an entry to histogram with value v
  if ((max-min) == 0)
    {
      outOfRange++;
      return false;
    }
  if (v < min || v > max)
    {
      outOfRange++;
      return false;
    }
  
  int binnum = (int)((bins.size()/(max-min))*(v-min));
  DEBUG0((cout << "Histo::Count, v " << v << " binnum " << binnum << endl));
  if ((vector<Count_t>::size_type)binnum >= bins.size() || binnum < 0)
    {
      outOfRange++;
      return false;
    }
  bins[binnum]++; // Add to this bin
  return true;
}

void Histogram::Reset()
{ // Reset to initial state
  for (vector<Count_t>::size_type i = 0; i < bins.size(); ++i)
    {
      bins[i] = 0;
    }
  outOfRange = 0;
}

void Histogram::Log(ostream& os, const char* hdr, const char sep)
{
  double perBin = (max-min) / bins.size();
  double f = min;
  if (hdr) os << hdr << endl;
  for (vector<Count_t>::size_type i = 0; i < bins.size(); ++i)
    {
      os << f << sep << f + perBin << sep << bins[i] << endl;
      f += perBin;
    }
}

Statistics* Histogram::Copy() const
{
  return new Histogram(*this);
}

void Histogram::CDF(ostream& os, const char* hdr, const char sep)
{
  double perBin = (max-min) / bins.size();
  double f = min;
  // Find total number of observations
  Count_t tot = outOfRange;
  for (vector<Count_t>::size_type i = 0; i < bins.size(); ++i)
    {
      tot += bins[i];
    }
  Count_t cum = 0;
  if (hdr) os << hdr << endl;
  os << "0" << sep << "0.0" << endl; // always 0 prob. < 0
  for (vector<Count_t>::size_type i = 0; i < bins.size(); ++i)
    {
      f += perBin;
      cum += bins[i]; // Cumulative total
      double pr = (double)cum/(double)tot;
      os << f << sep << pr << endl;
    }
  os << "Out of Range " << outOfRange << endl;
}

