/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - ICMP Class
// George F. Riley.  Georgia Tech, Summer 2004

#ifndef __icmp_h__
#define __icmp_h__

#include <vector>
#include <map>

#include "pdu.h"
#include "l4demux.h"
#include "l4protocol.h"
#include "datapdu.h"

class Node;
class IPV4Header;
class ICMPHeader;
class Application;
class Packet;
class Data;
class L3Protocol;

// Define a data structure to manage ICMP callbacks
typedef std::vector<Application*> ApplVec_t;
typedef std::map<Node*, ApplVec_t> ICMPCallback_t;

//Doc:ClassXRef
class ICMP : public L4Demux {
  //Doc:Class Class {\tt ICMP} defines a model of the 
  //Doc:Class Internet Control Message Protocol.
  //Doc:Class {\tt ICMP} inherits from {\tt L4Demux} even though
  //Doc:Class it's conceptually a Layer 3 protocol; however it does use the
  //Doc:Class services of IPV4.
public:
  typedef enum { ProtoNumber =  1 } ICMPProto_t; // ICMP Layer4 Protocol number
  // Types of ICMP messages
  typedef enum {
      ECHO_REPLY = 0,
      DESTINATION_UNREACHABLE = 3,
      SOURCE_QUENCH = 4,
      REDIRECT = 5,
      ECHO = 8,
      TIME_EXCEEDED = 11,
      PARAMETER_PROBLEM = 12,
      TIMESTAMP = 13,
      TIMESTAMP_REPLY = 14,
      INFORMATION_REQUEST = 15,
      INFORMATION_REPLY = 16 } ICMPMsg_t;

  // Various codes
  typedef enum {
      NET_UNREACHABLE,
      HOST_UNREACHABLE,
      PROTOCOL_UNREACHABLE,
      PORT_UNREACHABLE,
      CANT_FRAGMENT,
      SOURCE_ROUTE_FAILED } DU_t; // Destination Unreachable codes
  typedef enum {
      TTL_EXCEEDED,
      REASSEMBLY_EXCEEDED } TE_t; // Time exceeded codes
  
  // Tracing options
  typedef enum { TYPE, CODE, IDENTIFIER, SEQUENCE,
                 ORIGIN_TIME, RECEIVE_TIME, TRANSMIT_TIME } ICMPTrace_t;

  // Constructors
  //Doc:Method
  ICMP();
    //Doc:Desc Default constructor, no arguments.

  // Destructor
  virtual ~ICMP(); 
public:
  // L4Demux methods
  // Upcall From lower layer
#ifdef SIM
  void DataIndication(Node*, Packet*, IPAddr_t, L3Protocol&, Interface*, IPAddr_t = 0); 
#else
  void DataIndication(Node*, Packet*, IPAddr_t, L3Protocol&, Interface*); 
#endif //SIM
  // Static Methods
  static void Enable(bool e = true); // Enable/Disable ICMP for this simulation
  static void EnableQuench(bool e = true);  // Enable/Disable Quench  messages
  static ICMP* Instance() {return instance;}// Get a pointer to ICMP instance

  // Methods for each ICMP message generation, used mostly by IPV4
  static void DestinationUnreachable(Node*, Packet*, IPV4Header*, DU_t);
  static void TimeExceeded(Node*, Packet*, IPV4Header*, TE_t);
  // ParameterProblem not implemented
  static void SourceQuench(Node*, Packet*, IPV4Header*);
  // Redirect not implemented
  static void EchoReply(Node*, Packet*, IPV4Header*, ICMPHeader*);
  static void TimestampReply(Node*, Packet*, IPV4Header*, ICMPHeader*);
  // InformationRequest not implemented
  
  // Static methods used by applications to generate ICMP requests
  static void Echo(Node*, ICMPHeader*, IPAddr_t); // Generate an echo message
  static void Timestamp(Node*, ICMPHeader*, IPAddr_t);//Generate timestamp req

  // Request a callback on ICMP message receipt
  static void AddCallback(Node*, Application*);

private:
  // Static members
  static bool enabled;   // Default false;
  static bool quench;    // Default false
  static ICMP* instance; // Points to one and only instance of ICMP
  static ICMPCallback_t callbacks;  // List of Node/Application callbacks
};


//Doc:ClassXRef
class ICMPHeader : public L4PDU {
public :
  ICMPHeader();
  ICMPHeader(ICMP::ICMPMsg_t t);
  ICMPHeader(const ICMPHeader&);
  Size_t  Size() const;
  PDU*    Copy();
  Proto_t Proto() const { return 1;}   // ICMP is protocol 1
  // Serialization
  Size_t  SSize();                     // Size needed for serialization
  char*   Serialize(char*, Size_t&);   // Serialize to a buffer
  char*   Construct(char*, Size_t&);   // Construct from buffer
  void    Trace(Tfstream&, Bitmap_t, // Trace the contents of this pdu
                Packet* = nil, const char* = nil);
public:
  ICMP::ICMPMsg_t type;
  unsigned long   code;
  Count_t         identifier;
  Count_t         sequence;
  Time_t          originateTimestamp;
  Time_t          receiveTimestamp;
  Time_t          transmitTimestamp;
  IPV4Header*     ipv4Hdr;
};


#endif
