/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Interface class
// George F. Riley.  Georgia Tech, Spring 2002

// Define class to model wireless network link interfaces

#ifndef __interface_wireless_h__
#define __interface_wireless_h__

#include "interface-real.h"
#include <string>

using namespace std;

//Doc:ClassXRef
class InterfaceWireless : public InterfaceReal {
  //Doc:Class Class InterfaceWireless extends the functionality of the 
  //Doc:Class basic wired interface to include information  required
  //Doc:Class by the layer  2 wireless protocols.

public:
  typedef enum { IDLE = 0, TX, RX, CX, RX_ME, RX_ZZ } LinkState_t;
  typedef enum {
	NONE = 0, PHY_TXSTART, PHY_TXEND, PHY_RXSTART, PHY_RXEND, PHY_RXEND_ERR
  } PhyInd_t;

  typedef enum {ADHOC, BSS, HOSTAP} Opmode_t;
public:
  //Doc:Method
  InterfaceWireless(const L2Proto& l2 = L2Proto802_3(),
		    IPAddr_t i        = IPADDR_NONE,
		    Mask_t   m        = MASK_ALL,
		    MACAddr  mac      = MACAddr::NONE,
		    bool bootstrap    = false
		    );
  //Doc:Desc Constructor for InterfaceWireless.
  //Doc:Arg1 Layer 2 protocol for this interface.
  //Doc:Arg2 \IPA\ for this interface.
  //Doc:Arg3 \IPA\ mask for this interface.
  //Doc:Arg4 MAC address
  
  // Notifier
  virtual void Notify(void*);

  //Doc:Method
  LinkState_t GetLinkState() const { return linkState;}
    //Doc:Desc   Get the link state for the wireless link.
    //Doc:Return Link state.

  //Doc:Method
  void        SetLinkState(LinkState_t ls) { linkState = ls;}
    //Doc:Desc Set the link state for the wireless link
    //Doc:Arg1 Link state to set.

  //Doc:Method
  Time_t      GetLinkFreeTime() const {return linkFreeTime;}
    //Doc:Desc Get the link free time for the associated link
    //Doc:Return Time this link will be free

  //Doc:Method
  void        SetLinkFreeTime(Time_t t) {linkFreeTime = t;}
    //Doc:Desc Set a new link free time for the associated link.
    //Doc:Arg1   New time for link free

  LinkEvent*  GetRxPacketEvent() const {return evRx;}
    //Doc:Desc Get the rx event for earliest pending packet receipt event
    //Doc:Return Rx Packet event

  //Doc:Method
  void        SetRxPacketEvent(LinkEvent* p) {evRx = p;}
    //Doc:Desc Set a new rx event for earliest pending packet receipt event
    //Doc:Arg1   New packet rx event

  double      GetRxPower() const {return pr;}
    //Doc:Desc Get the rx power for earliest pending packet receipt
    //Doc:Return Rx Power

  //Doc:Method
  void        SetRxPower(double p) {pr = p;}
    //Doc:Desc Set new rx power for earliest pending packet receipt
    //Doc:Arg1   New packet rx power

  // Override the PacketRx{Start|End} function from InterfaceReal
  virtual void        PacketRxStart(LinkEvent*);
  virtual void        PacketRxEnd(LinkEvent*, Size_t);
  
  //Doc:Method
  virtual bool     IsWireless() const { return true;}
    //Doc:Desc Test if this interface is wireless 
    //Doc:Return Always true for InterfaceWireless nodes

  void        setTxPower(double p) {txPower = p;}
  double      getTxPower(void) {return txPower;}
  void        setRxPower(double p) {rxPower = p;}
  double      getRxPower(void) {return rxPower;}

  string      getSSID(void) {return ssid;}
  void        setSSID(string ssid_) {ssid = ssid_;}

  void        setRTSTxPower(double p) {rtsTxPow = p;}
  void        setCTSTxPower(double p) {ctsTxPow = p;}
  void        setACKTxPower(double p) {ackTxPow = p;}
  double      getRTSTxPower(void) {return rtsTxPow;}
  double      getCTSTxPower(void) {return ctsTxPow;}
  double      getACKTxPower(void) {return ackTxPow;}

  //Doc:Method
  static void SetDefaultRadioRange(Meters_t);
    //Doc:Desc Specify the default radio range for all wireless interfaces
    //Doc:Arg1 Default range desired


  Opmode_t    getOpMode(void) { return opmode;}
  void        setOpMode(Opmode_t opmode_);

public:
  Time_t       last_time;     // To keep track of when we
                              // updated energy calculations last  
  bool         bootstrap;     // Do we need bootstrapping functionality
  Opmode_t     opmode;        // Are we an AP or STA or Ad-Hoc
  string       ssid;
  
private:
  Time_t       linkFreeTime;  // Time when link will get free
  LinkEvent*   evRx;          // Receive event for the 1st scheduled pkt
  double       pr;            // rx power of the 1st scheduled pkt
  LinkState_t  linkState;     // Wireless link state
  double       txPower;       // the tx power of this iface
  double       rxPower;       // the rx power level

  double       rtsTxPow;
  double       ctsTxPow;
  double       ackTxPow;

  Meters_t     radioRange;    // Transmit range (meters)

 public:
  static Meters_t defaultRadioRange;
};

#endif

