/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Interface class
// George F. Riley.  Georgia Tech, Spring 2002

// Define base class for all network link interfaces
// Implements the Layer 2 protocols

#include <iostream>

#include "debug.h"
#include "hex.h"
#include "interface.h"
#include "simulator.h"
#include "link.h"
#include "link-real.h"
#include "qtwindow.h"

using namespace std;

bool               Interface::recordMacAddresses = false;
vector<Interface*> Interface::macIfVec;
bool Interface::usearp = false;

Interface::Interface()
    : ip(IPADDR_NONE), mask(Mask(32)), macAddr(0,0),
      pNode(nil), evList(nil), cbenable(false), down(false), usewormcontainment(false)
{
  DEBUG0((cout << "Interface default constructor" << endl));
}

Interface::Interface(const L2Proto& l2,
                     IPAddr_t i,
                     Mask_t   m,
                     MACAddr  mac,
		     bool     bootstrap)
    : ip(i), mask(m), macAddr(mac),
      pNode(nil), evList(nil), cbenable(false), down(false), usewormcontainment(false)
#ifdef SIM
	,spfgroup(false), drgroup(false)
#endif //SIM
{
  DEBUG0((cout << "Interface constructor " << this << " evList "
          << evList << endl));
}

Interface::Interface(const Interface& i)
    : ip(i.GetIPAddr()), mask(i.GetIPMask()), macAddr(i.GetMACAddr()),
      pNode(i.GetNode()), evList(nil), cbenable(false), down(false), usewormcontainment(i.usewormcontainment)
{
  DEBUG0((cout << "Interface copy constructor " << endl));
}

Interface::~Interface()
{
}

// Handler methods
void Interface::Handle(Event*, Time_t)
{
}

// Notifier
void Interface::Notify(void*)
{
}

// Local event list management
// We go ahead and include the pending rx event list in the
// base class, which allows animation to work properly wiht
// the distrbuted GTNetS.  Ghost nodes at the end of an rlink
// will maintain this list, so packets can animate when trasversing
// an rlink.

void Interface::ScheduleRx(LinkEvent* e, Time_t t, bool schedule)
{ // Schedule packet receipt event in the local fifo queue
  if (!evList)
    {
      evList = new EventCQ();
      DEBUG0((cout << "Allocated new eventcq " << evList 
              << " size " << evList->size()
              << endl));
    }
  
  bool needSchedule = evList->empty();
  e->fifoEvent = true; // Note this event is in fifo queue
  Time_t when = Simulator::Now() + t;
  DEBUG(2,(cout << "Interface::SRX, node " << GetNode()->Id()
           << " pushing time " << when << endl));
  EventPair* evp = evList->next();
  if (!evp)
    {
      DEBUG(3,(cout << "eventcq " << evList
               << " got nill, allocating new " << endl));
      evp = new EventPair(when, e);
    }
  else
    {
      DEBUG(3,(cout << "eventcq " << evList
               << " reusing evp " << evp << " old time "
               << evp->time << " new time " << when << endl));
      evp->time = when;
      evp->event = e;
    }
  evList->push_back(evp);
  DEBUG(3,(cout << "eventcq " << evList << " after pb, in " << evList->in
           << " out " << evList->out
           << " limit " << evList->limit
           << " size " << evList->size() 
           << " full " << evList->full()
           << endl));
  DBG((Stats::sequentialEvList++));
  if (needSchedule && schedule)
    { // List was empty, need to schedule in main simulator event list
      DEBUG(2,(cout << "Interface::SRX, node " << GetNode()->Id()
               << " when " << when << endl));
      Simulator::instance->Schedule(e, t, this);
    }
}

// Failure management
void Interface::Down()
{
  down = true;
  Simulator::instance->TopologyChanged();
#ifdef HAVE_QT
  // May need to redisplay as well
  QTWindow* qtw = Simulator::instance->GetQTWindow();
  Link* link = GetLink();
  if (qtw && link)
    {
      Count_t nc = link->PeerCount();
      if (nc == 1)
        { // Point to point link, get destination interface
          Interface* ifaceDst = link->GetPeer((Count_t)0);
          qtw->RedrawP2P(this, ifaceDst);
        }
    }
#endif
}

void Interface::Up()
{
  down = false;
  Simulator::instance->TopologyChanged();
#ifdef HAVE_QT
  // May need to redisplay as well
  QTWindow* qtw = Simulator::instance->GetQTWindow();
  Link* link = GetLink();
  if (qtw && link)
    {
      Count_t nc = link->PeerCount();
      if (nc == 1)
        { // Point to point link, get destination interface
          Interface* ifaceDst = link->GetPeer((Count_t)0);
          qtw->RedrawP2P(this, ifaceDst);
        }
    }
#endif
}

// Static methods
// Add the specified interface to list of mac addresses
void Interface::Add(Interface* i, MACAddr::MACAddr_t m)
{
  if (recordMacAddresses)
    {
      while(macIfVec.size() < m) macIfVec.push_back(nil);
      macIfVec.push_back(i);
    }
}

Interface* Interface::GetByMac(MACAddr::MACAddr_t m)
{ // Get the logged interface by mac address
  if (m >= macIfVec.size()) return nil;
  return macIfVec[m];
}

// 
// void Interface::CBListAdd(SwitchCallBack sc)
// {
//   //cbList.push_back(sc);
// }

// void    Interface::CBListDelete(SwitchCallBack)
// {
//   // ?
// }
