/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Interface class
// George F. Riley.  Georgia Tech, Spring 2002

// Define class to model network link interfaces
// Implements the Layer 2 protocols

#ifndef __interface_h__
#define __interface_h__

#include "common-defs.h"
#include "handler.h"
#include "notifier.h"
#include "eventcq.h"
#include "mask.h"
#include "node.h"
#include "macaddr.h"

class VirusThrottle;
class PortMatching;
class L2Proto;
class Node;
class Queue;
class LinkEvent;
class Link;

//Doc:ClassXRef
class Interface : public Handler, public NotifyHandler {
  //Doc:Class Class interface is a base class for all interfaces.
  //Doc:Class The various methods for interfaces are documented
  //Doc:Class in the appropriate subclasses.

#ifdef MOVED_TO_WIRELESS
public:
  typedef enum { IDLE = 0, TX, RX, CX, RX_ME, RX_ZZ } LinkState_t;
  typedef enum {
	NONE = 0, PHY_TXSTART, PHY_TXEND, PHY_RXSTART, PHY_RXEND, PHY_RXEND_ERR
  } PhyInd_t;
#endif

public:
  Interface();
  Interface(const L2Proto& l2 = L2Proto802_3(),
            IPAddr_t i   = IPADDR_NONE,
            Mask_t   m   = MASK_ALL,
            MACAddr  mac = MACAddr::NONE,
	    bool  bootstrap = false);
  Interface(const Interface&);
  virtual ~Interface();

  virtual bool Send(Packet*, IPAddr_t, int) = 0; // Enque and transmit
  virtual bool Send(Packet*, const MACAddr&, int) = 0; // Enque and transmit
  virtual bool Send(Packet*, Word_t) = 0; // Enque and transmit
  virtual bool Xmit(Packet*, IPAddr_t, int){ return false;} // Enque and transmit

  // Handler
  virtual void Handle(Event*, Time_t);
  // Notifier
  virtual void Notify(void*);
  

 //Doc:Method
  virtual void AddARPEntry(IPAddr_t, MACAddr) {}
  //Doc:Desc Adds a new arp entry to the interface arp cache
  //Doc:Arg1 The Ip address of the neighbour
  //Doc:Arg2 The corresponding Mac address


  //Doc:Method
  void         SetIPAddr(IPAddr_t i) { ip = i;}   // Set interface IPAddr
    //Doc:Desc Specify an \IPA\ associated with this interface.
    //Doc:Arg1 The \IPA\ to assign this interface.

  //Doc:Method
  IPAddr       GetIPAddr() const { return ip;}    // Get the associated ip addr
    //Doc:Desc Return the \IPA\ associated with this interface.
    //Doc:Return The \IPA\ associated with this interface.
    //Doc:Return Returns {\tt IPADDR_NONE}
    //Doc:Return if no \IPA\ is assigned.
	
#ifdef SIM
  //Doc:Method
  void         SetName(char *n) {strcpy(name, n);}   // Set interface Name
    //Doc:Desc Specify an Name associated with this interface.
    //Doc:Arg1 The Name to assign this interface.

  //Doc:Method
  char *       GetName() { return name;}    // Get the associated Name
    //Doc:Desc Return the Name associated with this interface.
    //Doc:Return The Name associated with this interface.
	
  //Doc:Method
  void         SetIfIndex(unsigned int i) {ifindex = i;} // Set inferface index
    //Doc:Desc Specify an index associated with this interface.
    //Doc:Arg1 The index to assign this interface.

  //Doc:Method
  unsigned int GetIfIndex() const { return ifindex;} // Get the interface index
    //Doc:Desc Return the index associated with this interface.
    //Doc:Return The index associated with this interface.
	
  //Doc:Method
  void         SetSpfGroup(bool b) {spfgroup = b;} // Set SpfGroup
    //Doc:Desc Specify if SpfGroup associated with this interface.
    //Doc:Arg1 If the SpfGroup is assigned to this interface.

  //Doc:Method
  unsigned int IsSpfGroup() const { return spfgroup;}//Is the interface SpfGroup
    //Doc:Desc Return true if the spfgroup is associated with this interface.
    //Doc:Return true or false
	
  //Doc:Method
  void         SetDrGroup(bool b) {drgroup = b;} // Set DrGroup
    //Doc:Desc Specify if DrGroup associated with this interface.
    //Doc:Arg1 If the DrGroup is assigned to this interface.

  //Doc:Method
  unsigned int IsDrGroup() const { return drgroup;} // Is the interface DrGroup
    //Doc:Desc Return true if the Drgroup is associated with this interface.
    //Doc:Return true or false
#endif //SIM

  //Doc:Method
  Mask_t       GetIPMask() const { return mask;} // Get the ip addr mask
    //Doc:Desc Return the \IPA\ mask associated with this interface.
    //Doc:Return The \IPA\ mask associated with this interface.

  //Doc:Method
  void         SetMACAddr(const MACAddr& m);      // Set MAC addresses
    //Doc:Desc Set the {\em MAC} address associated with this interface.
    //Doc:Arg1 The {\em MAC} address for this interface.
  //Doc:Method
  MACAddr      GetMACAddr() const { return macAddr;} // Return local MAC addr
    //Doc:Desc Get the {\em MAC} address associated with this interface.
    //Doc:Return The {\em MAC} address for this interface.

  virtual Count_t  PeerCount() const = 0;         // Number of peers
  virtual IPAddr   PeerIP(int) const = 0;         // Get peer IP address
  virtual IPAddr   NodePeerIP(Node*) const = 0;   // Get peer IP address
  virtual bool     NodeIsPeer(Node*) const = 0;   // Find out if node is a peer
  virtual Count_t  NodePeerIndex(Node*) const = 0;// Get peer index for node

  //Doc:Method
  Node*            GetNode() const { return pNode;}// Get the associated node
    //Doc:Desc Get the  node pointer to which this interface is associated.
    //Doc:Return A pointer to the node this interface is associated with.

  //Doc:Method
  void             SetNode(Node* n) { pNode = n; }
    //Doc:Desc Set the  node pointer to which this interface is associated.
    //Doc:Arg1 Pointer to the node this interface is associated with.

  virtual Link*    GetLink() const = 0;       // Return the link pointer
  virtual void     SetLink(Link*) = 0;        // Set the link pointer
  virtual Queue*   GetQueue() const = 0;      // Return the queue pointer
  virtual void     SetQueue(const Queue& q) = 0;  // Set a new queue object
  virtual bool     QueueDetailed() const = 0;     // True if queue is detailed
  virtual L2Proto* GetL2Proto() const = 0;        // Layer 2 proto object
  virtual void     SetL2Proto(L2Proto*) = 0;      // Set the layer 2 protocol
  virtual void     AddNotify(NotifyHandler*,void*) = 0;// Add a notify client
  virtual void     CancelNotify(NotifyHandler*) = 0;   // Cancel pending notif.
  virtual MACAddr  IPToMac(IPAddr_t) const = 0;        // Get mac addr for ip
  virtual Count_t  NeighborCount(Node*) const = 0; // Number of rtg neighbors
  // Get a list of directly connected neighbors on this interface.
  virtual void     Neighbors(NodeWeightVec_t&, bool = false) = 0;
  virtual IPAddr_t DefaultPeer() = 0;         // Get default peer from link
  virtual bool     IsLocalIP(IPAddr_t) const = 0; // True if IP is on this i/f
  virtual Mult_t   TrafficIntensity() const = 0;  // TotBytesSent/LineSpeed
  // Add to list of remote IP
  virtual void     AddRemoteIP(IPAddr_t, Mask_t, Count_t = 0) = 0;
  virtual void     HandleLLCSNAP(Packet*) = 0;    // Process the LLCSnap header
  /////////// modified by Mohamed //////////
  virtual bool     IsWireless() const = 0;        // True if wireless interface
  virtual bool     IsEthernet() const { return false; }       // True if ethernet interface
  virtual bool     IsReal() const = 0;            // True if not Ghost i/f
  virtual void     RegisterRxEvent(LinkEvent*){return;} ; // Register the last rx event issued by this interface 
  virtual void     AddRxEvent(LinkEvent*){return;} ; // add to the list of rx events , used in case of broadcast


  //Doc:Method
  virtual void     ScheduleRx(LinkEvent*, Time_t, bool = true);// Sched pkt rx in fifo q
    //Doc:Desc Schedule a packet receipt event at a future  time.
    //Doc:Arg1 The link event to add.
    //Doc:Arg2 Simulation time when the receipt will occur.
    //Doc:Arg3 list.  Only false for trace file playbacks.

  //Doc:Method
  EventCQ*            GetEventCQ() { return evList; }
    //Doc:Desc Get a pointer to the pending RX event circular queue
    //Doc:Desc for this interface.
    //Doc:Return Pointer to EventCQ for this interface.

  
  //Doc:Method
  // void      CBListAdd();
  //Doc:Desc Add a callback function to this node.  Callbacks allow
  //Doc:Desc a function to examine every packet recieved by this node.
  //Doc:Arg1 Callback function to add.
  
  //Doc:Method
  //void      CBListDelete();
  //Doc:Desc Remove a callback function from this node.
  //Doc:Arg1 Callback function to remove.
  
  virtual void CBEnable(bool b){cbenable = b;}
 
  virtual bool CBEnable() const {return cbenable;}

  static void UseARP(bool b) {Interface::usearp = b;}


  virtual void UseWormContainment(bool b) {usewormcontainment = b;}
  virtual bool UseWormContainment(){ return usewormcontainment;}
  virtual void SetWormContainment(WormContainment*) = 0;
  virtual WormContainment* GetWormContainment() = 0;

  //Doc:Method
  const CallBackList_t& GetCBList();
  //Doc:Desc Get a reference to the callback list.
  //Doc:Return Reference to the callback list


  bool IsDown() { return down; }
  void Down();
  void Up();

protected:
  IPAddr                  ip;                      // Associated IP Address
#ifdef SIM
  char name[5];
  unsigned int ifindex;
	bool spfgroup;  //HACK for mulitcast
	bool drgroup;
#endif //SIM
  Mask_t                  mask;                    // Network mask
  MACAddr                 macAddr;                 // Associated MAC address
  Node*                   pNode;                   // Associated Node
  EventCQ*                evList;                  // Pending fifo events
  bool                    cbenable;                // enable call back list
  bool                    down;                    // True if interface is DOWN
  static bool             usearp;                  // use ARP protocol
  bool                    usewormcontainment;      // use wormcontainment module

public:
  // True if a global list of MAC address to Interface pointers is needed
  // This is used by the playback from trace file feature
  bool                    bootstrap;  
  static bool recordMacAddresses;
  static std::vector<Interface*> macIfVec;
  static void Add(Interface*, MACAddr::MACAddr_t);
  static Interface* GetByMac(MACAddr::MACAddr_t);
};

#endif
