/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Layer 2 Protocol base class
// George F. Riley.  Georgia Tech, Spring 2002

#ifndef __l2proto_h__
#define __l2proto_h__

#include "common-defs.h"
#include "protocol.h"
#include "notifier.h"

class Interface;

typedef enum {
	ieee802_11 = 0,
	usap       = 1,
  simple     = 2
} L2ProtocolType;

//Doc:ClassXRef
class L2Proto : public Protocol, public NotifyHandler {
  //Doc:Class Class {\tt L2Proto} is the bse class for all layer two 
  //Doc:Class protocols. these implement specifically the Medium Access 
  //Doc:Class Control and associated functions. This class closely 
  //Doc:Class interacts with the {\tt Link} class to provide the low 
  //Doc:Class level details of a network stack
public:

  //Doc:Method
  L2Proto();
    //Doc:Desc Constructor.

  //Doc:Method
  virtual Layer_t Layer () { return 2; }
    //Doc:Desc This member function returns the layer number of the protocol
    //Doc:Desc stack corresponding to the ISO model.
    //Doc:Return the layer number (in this case 2).

  //Doc:Method
  virtual void BuildDataPDU(MACAddr, MACAddr, Packet*) = 0;
    //Doc:Desc Build and append a layer2 pdu to the specified packet.
    //Doc:Arg1 Source MAC address
    //Doc:Arg2 Destination MAC address
    //Doc:Arg3 Packet pointer to append this pdu to.


  //Doc:Method
  virtual void DataRequest( Packet*) = 0;
  //Doc:Desc This member function is called by the higher layer requesting 
  //Doc:Desc a transfer of  data to a peer.
  //Doc:Arg1 The packet to be sent, with l2 pdu already appended/

  //Doc:Method
  virtual void DataIndication( Packet*) = 0;    // from lower layer
  //Doc:Desc This method is called when the direction of protocol stack
  //Doc:Desc traversal is upwards, i.e, a packet has just been received
  //Doc:Desc at the underlying link
  //Doc:Arg1 The packet that has been received

  //Doc:Method
  virtual void SetInterface(Interface* i) { iface = i; }
    //Doc:Desc Specify which interface this protocol is attached to.
    //Doc:Arg1 Interface poiner for interface.

  //Doc:Method
  virtual bool Busy() = 0;                  // True if proto/link is busy
    //Doc:Desc Determine if the protocol is busy processing a packet.
    //Doc:Return True if busy.

  //Doc:Method
  virtual L2Proto* Copy() const = 0;                      // Make a copy
  //Doc:Desc This method is called to make a copy of the layer 2 protocol
  //Doc:Desc object

  //Doc:Method
  virtual bool IsWireless() const = 0;
    //Doc:Desc Query if this protocol needs a wireless interface.
    //Doc:Return True if wireless protocol.

  virtual void Notify(void*) {}

  virtual void Bootstrap(void) = 0;

   // Over-rides from the tracing methods from protocol.h
   // We do those a bit differently for l2 protocols, since
   // there is not a single "instance" that can globally set the
   // trace on/off status.  Instead we use a static variable
   // common to all instances.
  void    SetTrace(Trace::TraceStatus ts)// Set trace level this proto
    { globalTraceStatus = ts; }
  Trace::TraceStatus GetTrace()          // Get trace level this proto
    { return globalTraceStatus;}
  void DetailOn(Bitmap_t b)
    { globalTraceDetails |= (1L << b);}
  void DetailOff(Bitmap_t b)
    { globalTraceDetails &= (ALLBITS ^ (1L << b));}
  Bitmap_t Details()
    { return globalTraceDetails;}


  // Static methods for managing global trace
  static void GlobalSetTrace(Trace::TraceStatus ts)
    { globalTraceStatus = ts; }
  static Trace::TraceStatus GlobalGetTrace()
    { return globalTraceStatus;}
  static void GlobalDetailOn(Bitmap_t b)
    { globalTraceDetails |= (1L << b);}
  static void GlobalDetailOff(Bitmap_t b)
    { globalTraceDetails &= (ALLBITS ^ (1L << b));}
  static Bitmap_t GlobalDetails()
    { return globalTraceDetails;}
  
 public:
  Interface* iface;  // Interface for this protocol
 private:
  static Trace::TraceStatus globalTraceStatus;
  static Bitmap_t           globalTraceDetails;  // Bitmap of trace info
};

#endif

