/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Layer 2 802.3 class

// Implementation for 802.11

#include <iostream>

#include "debug.h"
#include "ieee802.11.h"
#include "l2proto802.11.h"
#include "interface.h"
#include "ipv4.h"
#include "trace.h"
#include "node.h"
#include "hex.h"
#include "macaddr.h"
#include "queue.h"
#include "link.h"
#ifdef ETRACE
#include "etrace.h"
#endif 

// Uncomment below to enable debug level 0
//#define DEBUG_MASK 0x01

#define MYIP	((string)iface->GetIPAddr())

using namespace std;

// Default setting
Word_t L2Proto802_11::defaultSeqCacheSize = SEQ_CACHE_SIZE;

// Only one copy of protocol object per simulator

Random* L2Proto802_11::backoffRng = new Uniform();

L2Proto802_11::L2Proto802_11() : L2Proto()
{ // Constructor
  navTimer = 0.0;
  backoffTimer = 0.0;
  cw = PMIB.CWMin;
  sifs = PMIB.SIFSTime;
  difs = sifs + 2.0*PMIB.SlotTime;
  eifs = sifs + difs + (8 * ETHER_ACK_LEN / PMIB.PLCPDataRate);
  basicRate = 11000000.0;
  dataRate =  11000000.0;
  //dataRate = 11000000.0;
  //dataRate = iface->pLink->Bandwidth();    // iface not available here
  shortRetryCount = 0;
  longRetryCount = 0;
  pendingPkt = nil;
  navTimeout = nil;
  backoffTimeout = nil;
  ctsTimeout = nil;
  ackTimeout = nil;
  dataTimeout = nil;
  mac_state = IDLE;
  xmitFailHandler = nil;
  mevFirstAttempt = false;
  mevBackoff = nil;
  backingOff = false;
  backoffTimerState = NONE;
  firstAttempt = true;  // very 1st pkt in a session
  nextCacheEntry = 0;   // sequence cache
  seqNo = 0;            // data frame sequence number
  ibfTime = 100;        // 100 * 1024 usecs; the std is ambiguous about it
}

L2Proto802_11::~L2Proto802_11()
{
  CancelTimer(backoffTimeout, true);
  CancelTimer(navTimeout, true);
  CancelTimer(ctsTimeout, true);
  CancelTimer(ackTimeout, true);
  CancelTimer(dataTimeout, true);
}

void L2Proto802_11::BuildDataPDU(MACAddr src, MACAddr dst, Packet* p)
{
  Word_t dur = usec(sifs + txtime(ETHER_ACK_LEN, basicRate)
		      + DSSS_MaxPropagationDelay);
  DATA_frame* l2pdu = new DATA_frame(dur, dst, src, ++seqNo, 0);
  p->PushPDU(l2pdu);
}

void L2Proto802_11::DataRequest(Packet* p)
{
  DEBUG0(cout << MYIP << ": DataRequest uid " << p->uid
	   << " at " << Simulator::Now() << endl);

  shortRetryCount = 0;
  longRetryCount = 0;
  cw = PMIB.CWMin;    // revert the cw back
  backingOff = false;
  backoffTimerState = NONE;

  if (pendingPkt) {
	/* This cannot happen because if pendingPkt is not null,
	   L2 protocol is busy and DataRequest() is never invoked
	*/
    cout << "HuH??? Data request with pending packet\n";
  }
  // Hold here till the RTS/CTS exchange is over
  pendingPkt = p;

  if (!MediumBusy()) {
	// The medium is idle
	DEBUG0(cout << "  medium is idle\n");
	if (firstAttempt)
	  InitiateFirstAttempt();
	else
	  InitiateBackoff();
  }
  else {
	// The medium is busy
	DEBUG0(cout << "  medium is busy\n");
	if (VcsBusy())
	  ScheduleTimer(MACTimerEvent::NAV_TIMEOUT, navTimeout,
					navTimer - Simulator::Now());
  }
  firstAttempt = false;
}

void L2Proto802_11::DataIndication(Packet* p)
{
#ifdef ETRACE
  ETrace::Instance()->TracePacket(p, iface->GetNode(), iface,2, "r"); // 2 for layer 2
#endif
  L2Header802_11* l2pdu = (L2Header802_11*)p->PopPDU();
  MACAddr dst;

  iface->GetNode()->TracePDU(this, l2pdu);
  switch(l2pdu->type)
    {
    case RTS: {
      if (mac_state != IDLE) {
        DEBUG0(cout << MYIP << ": RTS received in wrong state\n");
		delete p;
		break;
      }

      RTS_frame* rtsFrame = (RTS_frame*)l2pdu;
      if (rtsFrame->ra == iface->GetMACAddr()) {
		// This is for me
		if (VcsBusy()) {
		  // Don't respond if the NAV indicates busy
		  delete p;
		  break;
		}
		mac_state =  RTS_RCVD;
		DEBUG0(cout << MYIP << ": got RTS uid " << p->uid
			   << " at " << Simulator::Now() << endl);
		Word_t cts_time = usec(sifs + DSSS_MaxPropagationDelay
				       + txtime(ETHER_CTS_LEN, basicRate));
		SendCTS(rtsFrame->ta, rtsFrame->duration - cts_time);
      }
      else {
		// Else set my NAV timer
		Time_t new_nav = Simulator::Now() + rtsFrame->duration / 1e6;
		if (new_nav > navTimer) {
		  navTimer = new_nav;
		  CancelTimer(navTimeout);
		  ScheduleTimer(MACTimerEvent::NAV_TIMEOUT, navTimeout,
						navTimer - Simulator::Now());
		}
      }
	  delete p;
      break;
	}
    case CTS: {
      CTS_frame* ctsFrame = (CTS_frame*)l2pdu;
      if (ctsFrame->ra == iface->GetMACAddr()) {
		// This is for me
		if (mac_state != CTS_WAIT) {
	      DEBUG0(cout << MYIP << ": CTS received in wrong state\n");
	      delete p;
	      break;
	    }
		DEBUG0(cout << MYIP << ": got CTS uid " << p->uid
			   << " at " << Simulator::Now() << endl);
		mac_state = CTS_RCVD;
		shortRetryCount = 0;
		CancelTimer(ctsTimeout);
		SendDATA();
	  }
      else {
		// Update the NAV in case we haven't heard from original sender
		Time_t new_nav = Simulator::Now() + ctsFrame->duration / 1e6;
		if (new_nav > navTimer) {
		  navTimer = new_nav;
		  CancelTimer(navTimeout);
		  ScheduleTimer(MACTimerEvent::NAV_TIMEOUT,  navTimeout,
						navTimer - Simulator::Now());
		}
	  }
	  delete p;
      break;
	}
    case DATA: {
      DATA_frame* dataFrame = (DATA_frame*)l2pdu;
      dst = dataFrame->addr1;
      if (dst.IsBroadcast()) {
		// Broadcast
		if (mac_state != IDLE) {
	      DEBUG0(cout << MYIP << ": received BC-DATA in wrong state\n");
		  delete p;
		  break;
		}
		DEBUG0(cout << MYIP << ": received BC-DATA uid " << p->uid
			   << " at " << Simulator::Now() << endl);
		iface->HandleLLCSNAP(p);
	  }
      else if (dst == iface->GetMACAddr()) {
		// Unicast, it's for me
		if (mac_state != DATA_WAIT) {
	      DEBUG0(cout << MYIP << ": received DATA in wrong state\n");
		  delete p;
		  break;
		}
		DEBUG0(cout << MYIP << ": received DATA uid " << p->uid
			   << " at " << Simulator::Now() << endl);
		/* The spec says ACK transmission shall commence after a SIFS period
		   without regard to the busy/idle state of the medium
		   Refer to chapter 9.2.8, IEEE Std 802.11, 1999 Edition
		*/
		CancelTimer(dataTimeout);
		mac_state = DATA_RCVD;
		SendACK(dataFrame->addr2); // src-dest pair

		/* Need to check if this data frame is retry and received before
		   If so, drop silently
		*/
		if (!dataFrame->fc.fc_retry) { // first try
		  // first received
		  AddSeqCache(dataFrame->addr2, dataFrame->seqno);
		  iface->HandleLLCSNAP(p);
		}
		else { // retry
		  if (FindSeqCache(dataFrame->addr2, dataFrame->seqno)) {
			// already received
			delete p;
		  }
		  else {
			// first received
			AddSeqCache(dataFrame->addr2, dataFrame->seqno);
			iface->HandleLLCSNAP(p);
		  }
		}
	  }
	  else {
		// Unicast, not for me
		delete p;
	  }
      break;
	}
    case ACK: {
      ACK_frame* ackFrame = (ACK_frame*)l2pdu;
      if (ackFrame->ra == iface->GetMACAddr()) {
		if (mac_state != ACK_WAIT) {
		  DEBUG0(cout << MYIP << ": received ACK in wrong state\n");
		  delete p;
		  break;
		}
		DEBUG0(cout << MYIP << ": received ACK uid " << p->uid
			   << " at " << Simulator::Now() << endl);
		DEBUG0(cout << "  Data send complete\n");
		mac_state = IDLE;
		DEBUG0(cout << "  mac_state set to IDLE\n");
		longRetryCount = 0;
		CancelTimer(ackTimeout);
		if (pendingPkt) delete pendingPkt;
		pendingPkt = nil;

		// L2 protocol became idle
		// Process next packet if existing
		DataRequestFromIFQ();
	  }
	  delete p;
	  break;
	  // we are not intended for this ack so ignore
	}
    case BEACON:
      /* Will take this debug off at final commit */
      cout << Simulator::Now() << " Received a beacon packet at "
	   << iface->GetNode() << endl;
      break;
    default:
      cout << "HuH? unknown 802.11 packet type\n";
      delete p;
    }
}

bool L2Proto802_11::Busy()
{
  // Determine if protocol is busy
  return (!!pendingPkt || mac_state != IDLE);
}

bool L2Proto802_11::VcsBusy()
{
  return navTimer > Simulator::Now();
}

bool L2Proto802_11::PcsBusy()
{
  return ((InterfaceWireless*)iface)->GetLinkState() != InterfaceWireless::IDLE;
}

bool L2Proto802_11::MediumBusy()
{
  // Determine if the medium is busy by
  // either virtual or physical carrier sense
  return (VcsBusy() || PcsBusy());
}

// L2802_11Header methods

L2Proto* L2Proto802_11::Copy() const
{
  return new L2Proto802_11(*this);
}

// Sending the packets

bool L2Proto802_11::SendRTS(MACAddr ta, MACAddr ra, Word_t dur)
{
  RTS_frame* rf = new RTS_frame(dur, ra, ta, 0);
  Packet* p = new Packet();
  p->PushPDU(rf);

  ScheduleSendAfterIFS(p, difs);
#ifdef BUGFIX 
	delete(p);
#endif //BUGFIX
  return true;
}

bool L2Proto802_11::SendCTS(MACAddr ra, Word_t dur)
{
  CTS_frame* cf = new CTS_frame(dur, ra, 0);
  Packet* p = new Packet();
  p->PushPDU(cf);

  ScheduleSendAfterIFS(p, sifs);
#ifdef BUGFIX 
	delete(p);
#endif //BUGFIX
  return true;
}

bool L2Proto802_11::SendDATA()
{
  if (!pendingPkt) {
    cout << "HuH? null data pointer\n";
	mac_state = IDLE;
    return false;
  }

  DATA_frame* df = (DATA_frame*)pendingPkt->PeekPDU();
  MACAddr dst = df->addr1;

  /* Unicast means we seized the medium and we need to keep it
	 Thus we use the smallest gap, SIFS
  */
  Time_t ifs = dst.IsBroadcast() ? difs : sifs;
  ScheduleSendAfterIFS(pendingPkt, ifs);
  return true;
}

bool L2Proto802_11::SendACK(MACAddr ra)
{
  ACK_frame* af = new ACK_frame(0, ra, 0);
  Packet *p = new Packet();
  p->PushPDU(af);

  ScheduleSendAfterIFS(p, sifs);
#ifdef BUGFIX 
	delete(p);
#endif //BUGFIX
  return true;
}

void L2Proto802_11::BackoffProcedure()
{
  DEBUG0(cout << MYIP << ": backing off at " << Simulator::Now() << endl);
  incr_cw();
  mac_state = IDLE;
  backingOff = true;
  backoffTimerState = NONE;
  CancelTimer(backoffTimeout);
  ScheduleTimer(MACTimerEvent::BACKOFF_TIMEOUT, backoffTimeout,
				backoffRng->Value()*cw*PMIB.SlotTime);
}

bool L2Proto802_11::ScheduleSendAfterIFS(Packet* p, Time_t t)
{
  MACEvent* mev = new MACEvent(MACEvent::PACKET_TX, p);
  Simulator::instance->Schedule(mev, t, this);
  return true;
}

void L2Proto802_11::Handle(Event* ev, Time_t t )
{
  /* Need to send the packet down to link layer and for that we need
	 interface handle
	 Now we are at the IFS slot boundary and need to check if the medium
	 is idle in such case we are allowed to transmit
  */

  Time_t txtime1, txtime2, to_value;
  MACEvent* mev = (MACEvent*)ev;
  switch (mev->event)
    {
			case MACEvent::PACKET_TX: {
				mevFirstAttempt = false;
				L2Header802_11* l2pdu = (L2Header802_11*)mev->p->PeekPDU();
			
				switch (l2pdu->type)
				{
					case RTS:
						if (MediumBusy()) {
							mac_state = IDLE;
							break;
						}
#ifdef ETRACE
						ETrace::Instance()->TracePacket(mev->p, iface->GetNode(), iface,2, "s");
#endif 
						iface->GetLink()->Transmit(
                      mev->p, iface, iface->GetNode(), basicRate);
						txtime1 = (ETHER_RTS_LEN * 8) / basicRate;
						txtime2 = (ETHER_CTS_LEN * 8) / basicRate;
						to_value = txtime1 + sifs + txtime2
							+ DSSS_MaxPropagationDelay +  DSSS_MaxPropagationDelay;
						ScheduleTimer(MACTimerEvent::CTS_TIMEOUT, ctsTimeout, to_value);
						mac_state = CTS_WAIT;
						break;
					
					case CTS:
						if (MediumBusy()) {
							DEBUG0(cout << MYIP << ": linkBusy CTS "
									<< ((InterfaceWireless*)iface)->GetLinkState()
									<< endl);
							mac_state = IDLE;
							break;
						}
#ifdef ETRACE
						ETrace::Instance()->TracePacket(mev->p, iface->GetNode(), 
								iface,2, "s");
#endif 
						iface->GetLink()->Transmit(
                      mev->p, iface, iface->GetNode(), basicRate);
						iface->GetNode()->TracePDU(this, l2pdu);
						to_value = l2pdu->duration*1e-6 - sifs
							+ DSSS_MaxPropagationDelay + DSSS_MaxPropagationDelay;
						ScheduleTimer(MACTimerEvent::DATA_TIMEOUT, dataTimeout, to_value);
						mac_state = DATA_WAIT;
						break;

					case DATA: {
						 if (!pendingPkt) {
							 // No data pkt while l2 is about to send it
							 cout << "HuH? no pendingPkt to send\n";
							 mac_state = IDLE;
							 break;
						 }
						 DATA_frame* df = (DATA_frame*)l2pdu;
						 if (df->addr1.IsBroadcast()) {
							 if (MediumBusy()) {
								 mac_state = IDLE;
								 break;
							 }
#ifdef ETRACE
							 ETrace::Instance()->TracePacket(pendingPkt, iface->GetNode(), 
									 iface,2, "s");
#endif
							 iface->GetLink()->Transmit(
									 pendingPkt, iface, iface->GetNode(), dataRate);
							 delete pendingPkt;
							 pendingPkt = nil;
							 break;
						 }
						 if (MediumBusy()) {
							 DEBUG0(cout << MYIP << ": linkBusy DATA "
									 << ((InterfaceWireless*)iface)->GetLinkState()
									 << endl);
							 //longRetryCount++;
							 mac_state = IDLE;
							 break;
						 }
#ifdef ETRACE
						 ETrace::Instance()->TracePacket(mev->p, iface->GetNode(), 
								 iface,2, "s");
#endif 
						 iface->GetLink()->Transmit(mev->p, iface, iface->GetNode(), 
								 dataRate);
						 iface->GetNode()->TracePDU(this, l2pdu); 
						 txtime1 = ((Rate_t)pendingPkt->Size() * 8) / dataRate;
						 txtime2 = (ETHER_ACK_LEN * 8) / basicRate;
						 to_value = txtime1 + sifs + txtime2
							 + DSSS_MaxPropagationDelay + DSSS_MaxPropagationDelay;
						 ScheduleTimer(MACTimerEvent::ACK_TIMEOUT, ackTimeout, to_value);
						 mac_state = ACK_WAIT;
						 break;
					}
					case ACK:
						if (MediumBusy()) {
							DEBUG0(cout << MYIP << ": linkBusy ACK "
                               << ((InterfaceWireless*)iface)->GetLinkState()
                               << endl);
						}
						else {
#ifdef ETRACE
							ETrace::Instance()->TracePacket(mev->p, iface->GetNode(), 
									iface,2, "s");
#endif 
							iface->GetLink()->Transmit(mev->p, iface, iface->GetNode(), 
									basicRate);
						}
						mac_state = IDLE;
						if (!Busy()) {
							// L2 protocol became idle
							DataRequestFromIFQ();
						}
						break;

		case BEACON:
		  if (MediumBusy()) {
			DEBUG0(cout << MYIP << ": linkBusy BEACON "
                               << ((InterfaceWireless*)iface)->GetLinkState()
                               << endl);
		  }
		  else {
		    /* Schedule another before sending this */
		    SendBeacon(ibfTime* 0.001); /* absolute time here */
#ifdef ETRACE
      ETrace::Instance()->TracePacket(mev->p, iface->GetNode(), iface,2, "s");
#endif 
		    iface->GetLink()->Transmit(mev->p, iface,
					       iface->GetNode(),
					       basicRate);
		  }
		default:
		  DEBUG0(cout << "HuH? unknown packet type while sending\n");
		} //switch (l2pdu->type)

	  delete mev->p;
	  DEBUG0(cout.precision(8));
	  DEBUG0(cout << MYIP << ": transmit"
			 << " at " << Simulator::Now() << endl);
          break;
	} //case MACEvent::PACKET_TX

	case MACEvent::BACKOFF:
	  delete mev;
	  mevBackoff = nil;
	  if (MediumBusy()) {
		mac_state = IDLE;
		break;
	  }
	  if (!backingOff) {  // need to start backoff procedure
		BackoffProcedure();
	  }
	  else {  // already in backoff, need to resume backoff timer
		if (backoffTimerState == SUSPENDED)
		  ResumeTimer(backoffTimeout);
	  }
	  break;

    default:
      break;
    } //switch
}

void L2Proto802_11::Timeout(TimerEvent *ev)
{
  MACTimerEvent* mtev = (MACTimerEvent*)ev;
  switch (mtev->event)
    {
    case MACTimerEvent::BACKOFF_TIMEOUT:
	  /* Backoff timer expried, which means we can access the medium
		 Send immediately if there is a pending MSDU
	  */
	  DEBUG0(cout << MYIP << ": BACKOFF_TIMEOUT at " << Simulator::Now()
			 << endl);
	  if (MediumBusy()) {
		/* This is a double check, remove later
		   This timeout cannot occur if the medium became busy
		   because Idle2Busy() suspends this timer event
		*/
		cout << "HuH? medium busy at the BACKOFF_TIMEOUT\n";
		cout << "  PCS: "
                     << ((InterfaceWireless*)iface)->GetLinkState() << endl
                     << "  VCS: NAV " << navTimer
                     << " now " << Simulator::Now()
                     << endl;
		cout << "  mac_state " << mac_state << endl;
	  }
	  backingOff = false;
	  backoffTimerState = NONE;
      if (pendingPkt) {
		InitiateDataSequence();
	  }
      break;

    case MACTimerEvent::NAV_TIMEOUT:
      /* The NAV timer expired, which means virtual CS indicates idle
		 Thus need to ckeck if the medium became idle
	  */
	  if (navTimer != Simulator::Now()) {
		cout << "HuH? navTimer != now at the NAV_TIMEOUT"
			 << "  NAV " << navTimer << " now " << Simulator::Now() << endl;
	  }
	  DEBUG0(cout << MYIP << ": NAV_TIMEOUT at " << Simulator::Now() << endl);
	  if (!MediumBusy()) {
		Busy2Idle();
	  }
      break;

    case MACTimerEvent::CTS_TIMEOUT: {
	  DEBUG0(cout << MYIP << ": CTS_TIMEOUT at " << Simulator::Now() << endl);
	  shortRetryCount++;
	  mac_state = IDLE;
	  backingOff = false;
	  backoffTimerState = NONE;
	  CancelTimer(backoffTimeout);
	  if (shortRetryCount > MMIB.ShortRetryLimit) {  // drop
		DropPacket();
		break;
	  }
	  // Retry
	  DATA_frame* df = (DATA_frame*)pendingPkt->PeekPDU();
	  df->fc.fc_retry = 1;
	  if (!MediumBusy()) InitiateBackoff();
      break;
	}
    case MACTimerEvent::ACK_TIMEOUT: {
	  DEBUG0(cout << MYIP << ": ACK_TIMEOUT at " << Simulator::Now() << endl);
	  longRetryCount++;
	  mac_state = IDLE;
	  backingOff = false;
	  backoffTimerState = NONE;
	  CancelTimer(backoffTimeout);
	  if (longRetryCount > MMIB.LongRetryLimit) {  // drop
		DropPacket();
		break;
	  }
	  // Retry
	  DATA_frame* df = (DATA_frame*)pendingPkt->PeekPDU();
	  df->fc.fc_retry = 1;
	  if (!MediumBusy()) InitiateBackoff();
      break;
	}
    case MACTimerEvent::DATA_TIMEOUT:
	  DEBUG0(cout << MYIP << ": DATA_TIMEOUT at " << Simulator::Now() << endl);
      DEBUG0(cout << "  mac_state set to IDLE\n");
      mac_state = IDLE;
	  //if (pendingPkt && !MediumBusy()) InitiateBackoff();
	  //if (!Busy()) DataRequestFromIFQ();
      break;

    default:
      break;
    }
}

void L2Proto802_11::ScheduleTimer(Event_t e, MACTimerEvent*& ev, Time_t t)
{
  if (t <= 0) {
	cout << "HuH? schedule negative time " << t << endl;
	return;
  }

  Time_t now = Simulator::Now();
  if (ev)
    {
      if (ev->Time() == t + now)
		return;
      CancelTimer(ev);
    }
  if (!ev)
    {
      ev = new MACTimerEvent(e);
    }
  timer.Schedule(ev, t, this);
}

void L2Proto802_11::CancelTimer(MACTimerEvent*& ev, bool deltimer)
{
  if (ev)
    {
      timer.Cancel(ev);
      if (deltimer)
		{
		  delete ev;
		  ev = nil;
		}
    }
}

void L2Proto802_11::SuspendTimer(MACTimerEvent* ev)
{
  // This function assumes backoff timer
  if (!ev) {
	cout << "HuH? no timer event to suspend\n";
	return;
  }
  timer.Pause(ev);
  backoffTimerState = SUSPENDED;
}

void L2Proto802_11::ResumeTimer(MACTimerEvent* ev)
{
  // This function assumes backoff timer
  if (!ev) {
	cout << "HuH? no timer event to resume\n";
	return;
  }
  timer.Resume(ev);
  backoffTimerState = RESUMED;
}

void L2Proto802_11::InitiateDataSequence()
{
  DATA_frame* df = (DATA_frame*)pendingPkt->PeekPDU();
  MACAddr src = df->addr2;
  MACAddr dst = df->addr1;

  if (dst.IsBroadcast() 
#ifdef SIM
#endif //SIM
			) {
#ifdef ETRACE
      ETrace::Instance()->TracePacket(pendingPkt, iface->GetNode(), iface,2, "s");
#endif 
	iface->GetLink()->Transmit(pendingPkt, iface, iface->GetNode(), dataRate);
	delete pendingPkt;
	pendingPkt = nil;
  }
  else {
	Word_t duration = usec(sifs + txtime(ETHER_CTS_LEN, basicRate)
						   + DSSS_MaxPropagationDelay
						   + sifs
						   + txtime(pendingPkt->Size(), dataRate)
						   + DSSS_MaxPropagationDelay
						   + sifs
						   + txtime(ETHER_ACK_LEN, basicRate)
						   + DSSS_MaxPropagationDelay);
	RTS_frame* rf = new RTS_frame(duration, dst, src, 0);
	Packet* p = new Packet();
	p->PushPDU(rf);
	DEBUG0(cout << MYIP << ": sent RTS uid " << p->uid
		   << " for uid " << pendingPkt->uid
		   << " at " << Simulator::Now() << endl);
#ifdef ETRACE
      ETrace::Instance()->TracePacket(p, iface->GetNode(), iface,2, "s");
#endif 
	iface->GetLink()->Transmit(p, iface, iface->GetNode(), basicRate);
	Time_t txtime1 = ETHER_RTS_LEN * 8 / basicRate;
	Time_t txtime2 = ETHER_CTS_LEN * 8 / basicRate;
	Time_t to_value = txtime1 + sifs + txtime2
	  + DSSS_MaxPropagationDelay +  DSSS_MaxPropagationDelay;
	ScheduleTimer(MACTimerEvent::CTS_TIMEOUT, ctsTimeout, to_value);
	mac_state = CTS_WAIT;
	delete p;
  }
}

void L2Proto802_11::Idle2Busy()
{
  // This is called whenever the medium becomes busy
  DEBUG0(cout << MYIP << ": Idle2Busy() at " << Simulator::Now() << endl);

  if (!MediumBusy()) {  // this cannot happen
	cout << "HuH? medium is idle in Idle2Busy()\n";
  }

  if (backingOff) {  // in backoff procedure, need to suspend backoff timer
	if (backoffTimerState != SUSPENDED) {
	  SuspendTimer(backoffTimeout);
	  DEBUG0(cout << "  backoff timer suspened"
			 << " remaining_time " << backoffTimeout->remainingTime << endl);
	}
  }
}

void L2Proto802_11::Busy2Idle()
{
  // This is called whenever the medium becomes idle
  DEBUG0(cout << MYIP << ": Busy2Idle() at " << Simulator::Now() << endl);

  if (MediumBusy()) {  // this cannot happen
	cout << "HuH? medium is busy in Busy2Idle()\n";
  }

  if (!pendingPkt) return;

  if (mac_state != IDLE) return;
  /* mevFirstAttempt and mevBackoff are mutually excusive
	 Both cannot be on simultaneously
	 If mevFirstAttempt is on, don't initiate backoff procedure
  */
  if (mevFirstAttempt) return;
  InitiateBackoff();
}

void L2Proto802_11::DataRequestFromIFQ()
{
  // Get the next packet from ifque
  if (iface->GetQueue()->LengthPkts() > 0)
    {
      Packet* p = iface->GetQueue()->Deque();
      DataRequest(p);
    }
}

void L2Proto802_11::InitiateFirstAttempt()
{
  if (mevBackoff) {
	/* This should NOT happen!!!
	   because mevFirstAttempt is mutually exclusive with mevBackoff
	*/
	cout << "HuH? backoff initiated with 1st attempt\n";
  }

  if (mevFirstAttempt) // MACEvent::PACKET_TX for 1st attempt already scheduled
	return;
  mevFirstAttempt = true;  // this is on for RTS or BC-DATA
  DATA_frame* df = (DATA_frame*)pendingPkt->PeekPDU();
  MACAddr dst = df->addr1;

  if (dst.IsBroadcast()) {
	SendDATA();
  }
  else {
	MACAddr src = df->addr2;
	Word_t duration = usec(sifs + txtime(ETHER_CTS_LEN, basicRate)
						   + DSSS_MaxPropagationDelay
						   + sifs
						   + txtime(pendingPkt->Size(), dataRate)
						   + DSSS_MaxPropagationDelay
						   + sifs
						   + txtime(ETHER_ACK_LEN, basicRate)
						   + DSSS_MaxPropagationDelay);
	SendRTS(src, dst, duration);
  }
}

void L2Proto802_11::InitiateBackoff()
{
  if (mevFirstAttempt) {
	/* This should NOT happen!!!
	   because mevBackoff is mutually exclusive with mevFirstAttempt
	*/
	cout << "HuH? 1st attempt initiated with backoff initiation\n";
  }

  if (mevBackoff) return;    // MACEvent::BACKOFF already scheduled
  mevBackoff = new MACEvent(MACEvent::BACKOFF);
  Simulator::instance->Schedule(mevBackoff, difs, this);
}

void L2Proto802_11::DropPacket()
{
  shortRetryCount = 0;
  longRetryCount = 0;

  if (xmitFailHandler) {
	// L2 notifies L3 of transmit failure
#ifdef ETRACE
  ETrace::Instance()->TracePacket(pendingPkt, iface->GetNode(), iface,2, "d");
#endif
	pendingPkt->PopPDU();    // Remove L2 header
	pendingPkt->PopPDU();    // Remove LLCSNAP header
	xmitFailHandler->Notify(pendingPkt);
  }
  else
	delete pendingPkt;
  pendingPkt = nil;

  // L2 protocol became idle
  // Process next packet if existing
  DataRequestFromIFQ();
}

void L2Proto802_11::AddSeqCache(MACAddr ta, Word_t sn)
{
  if (seqCache.size() >= defaultSeqCacheSize) {
	// Cache is full, replace the oldest entry first
	seqCache[nextCacheEntry++] = SeqCacheEntry(ta, sn);
  }
  else {
	seqCache.push_back(SeqCacheEntry(ta, sn));
	nextCacheEntry++;
  }
  nextCacheEntry %= defaultSeqCacheSize;
}

bool L2Proto802_11::FindSeqCache(MACAddr ta, Word_t sn)
{
  for (SeqCache_t::iterator i = seqCache.begin(); i != seqCache.end(); ++i)
	if (i->ta == ta && i->sn == sn) return true;
  return false;
}

void L2Proto802_11::incr_cw()
{
  //cout << "CON_WIN " << cw << endl;
  cw = (cw << 1) + 1;
  if (cw > PMIB.CWMax) cw = PMIB.CWMax;
}

void L2Proto802_11::Notify(void* v)
{
  /* Calling this function means MAC received
	 a PHY-RXSTART, PHY-RXEND, PHY-TXSTART, or PHY-TXEND.indication,
	 which means physical CS indicates busy/idle
	 Thus need to check if the medium became busy/idle
  */
  InterfaceWireless::PhyInd_t* phyInd = (InterfaceWireless::PhyInd_t*)v;
  switch (*phyInd)
	{
	case InterfaceWireless::PHY_RXSTART:
	case InterfaceWireless::PHY_TXSTART:
	  /* This is a PHY-RXSTART or PHY-TXSTART.indication
		 PCS: was idle --> is busy
		 VCS: was idle ?
	  */
	  if (!VcsBusy()) Idle2Busy();
	  break;
	case InterfaceWireless::PHY_RXEND_ERR:
	case InterfaceWireless::PHY_RXEND:
	case InterfaceWireless::PHY_TXEND:
	  /* This is a PHY-RXEND or PHY-TXEND.indication
		 PCS: was busy --> is idle
		 VCS:          --> is idle ?
	  */
	  if (!VcsBusy()) Busy2Idle();
	  break;
	default:
	  break;
	} //switch
}

void L2Proto802_11::Bootstrap(void)
{
  /* Verify if we are a access point OR a station */
  InterfaceWireless* ifw = (InterfaceWireless*) iface;
  Uniform* rng = new Uniform();
  Time_t time = rng->Value();
  if (ifw->getOpMode() == InterfaceWireless::HOSTAP) {
    SendBeacon(time);
   }
}

bool L2Proto802_11::SendBeacon(Time_t time)
{
  
  Beacon_frame* bf = new Beacon_frame();
  
  bf->timestamp = Simulator::Now();
  bf->beacon_int = ibfTime;
  bf->cap_info = 0x00;

  InterfaceWireless* ifw = (InterfaceWireless*) iface;
  if(ifw->getOpMode() == InterfaceWireless::HOSTAP)  
    bf->cap_info &= 0x8000;
  else                        /* Beacons can also be sent in Ad-Hoc mode */
    bf->cap_info &= 0x4000;
  
  bf->ssidel.el_id = 0;
  bf->ssidel.len = 0x20;
  bf->ssidel.ssid = ((InterfaceWireless*)(iface))->getSSID();
  
  bf->ssrel.el_id = 1;
  bf->ssrel.len = 8;
  bf->ssrel.supp[0] = 0x82;
  bf->ssrel.supp[1] = 0x04;
  bf->ssrel.supp[2] = 0x0B;
  bf->ssrel.supp[3] = 0x16;

  bf->dsps.el_id = 3;
  bf->dsps.len = 1;
  bf->dsps.currChan = 6;

  bf->tim.el_id = 5;
  bf->tim.len = 0; /* No such info until i'm clear about functionality */
  
  Packet* p = new Packet();
  
  p->PushPDU(bf);
  
  ScheduleSendAfterIFS(p, time);

  return true;
}

// RTS_frame
RTS_frame::RTS_frame (Word_t dur, MACAddr rcv, MACAddr xmt, Long_t fcs_)
  : L2Header802_11(dur, fcs_, RTS)
{
  fc.fc_proto_ver  = MAC_ProtocolVersion;
  fc.fc_type       = MAC_Type_Control;
  fc.fc_subtype    = MAC_Subtype_RTS;
  fc.fc_to_ds      = 0;
  fc.fc_from_ds    = 0;
  fc.fc_more_frag  = 0;
  fc.fc_retry      = 0;
  fc.fc_pwr_mgt    = 0;
  fc.fc_more_data  = 0;
  fc.fc_wep        = 0;
  fc.fc_order      = 0;

  ra   = rcv;
  ta   = xmt;
}

Size_t RTS_frame::Size() const
{
  return ETHER_RTS_LEN;
}

// CTS_frame
CTS_frame::CTS_frame(Word_t dur, MACAddr rcv, Long_t fcs_)
  : L2Header802_11(dur, fcs_, CTS)
{
  fc.fc_proto_ver  = MAC_ProtocolVersion;
  fc.fc_type       = MAC_Type_Control;
  fc.fc_subtype    = MAC_Subtype_CTS;
  fc.fc_to_ds      = 0;
  fc.fc_from_ds    = 0;
  fc.fc_more_frag  = 0;
  fc.fc_retry      = 0;
  fc.fc_pwr_mgt    = 0;
  fc.fc_more_data  = 0;
  fc.fc_wep        = 0;
  fc.fc_order      = 0;

  ra = rcv;
}

Size_t CTS_frame::Size() const
{
  return ETHER_CTS_LEN;
}

// ACK_frame
ACK_frame::ACK_frame(Word_t dur, MACAddr rcv, Long_t fcs_)
  : L2Header802_11(dur, fcs_, ACK)
{
  fc.fc_proto_ver  = MAC_ProtocolVersion;
  fc.fc_type       = MAC_Type_Control;
  fc.fc_subtype    = MAC_Subtype_ACK;
  fc.fc_to_ds      = 0;
  fc.fc_from_ds    = 0;
  fc.fc_more_frag  = 0;
  fc.fc_retry      = 0;
  fc.fc_pwr_mgt    = 0;
  fc.fc_more_data  = 0;
  fc.fc_wep        = 0;
  fc.fc_order      = 0;

  ra = rcv;
}

Size_t ACK_frame::Size() const
{
  return ETHER_ACK_LEN;
}

// DATA_frame
DATA_frame::DATA_frame(Word_t dur, MACAddr rcv, MACAddr xmt, Word_t sqn, Long_t fcs_)
  : L2Header802_11(dur, fcs_, DATA)
{
  fc.fc_proto_ver  = MAC_ProtocolVersion;
  fc.fc_type       = MAC_Type_Data;
  fc.fc_subtype    = MAC_Subtype_DATA;
  fc.fc_to_ds      = 0;
  fc.fc_from_ds    = 0;
  fc.fc_more_frag  = 0;
  fc.fc_retry      = 0;
  fc.fc_pwr_mgt    = 0;
  fc.fc_more_data  = 0;
  fc.fc_wep        = 0;
  fc.fc_order      = 0;

  /* could juggle with to_ds and from_ds fields here */

  ra = addr1 = rcv;
  addr2 = xmt;
  addr3 = MACAddr::NONE;
  addr4 = MACAddr::NONE;
  seqno = sqn;
}

Size_t DATA_frame::Size() const
{
  return ETHER_DATA_LEN;
}

Size_t Beacon_frame::Size() const
{
  return 100;
}

Size_t Disassoc_frame::Size() const
{
  return 100;
}

Size_t Assocreq_frame::Size() const
{
  return 100;
}

Size_t Reassocreq_frame::Size() const
{
  return 100;
}

Size_t Reassocresp_frame::Size() const
{
  return 100;
}

Size_t Probereq_frame::Size() const
{
  return 100;
}

Size_t Proberesp_frame::Size() const
{
  return 100;
}

Size_t Auth_frame::Size() const
{
  return 100;
}

Size_t Deauth_frame::Size() const
{
  return 100;
}

