/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Layer 2 802.11 class
// George F. Riley.  Georgia Tech, Spring 2002

#ifndef __l2proto_802_11_h
#define __l2proto_802_11_h
#include <sys/types.h>

#include "math.h"
#include "l2proto.h"
#include "tfstream.h"
#include "macaddr.h"
#include "timer.h"
#include "rng.h"
#include "interface-wireless.h"

#define MAC_ProtocolVersion     0x00

#define MAC_Type_Management     0x00
#define MAC_Type_Control        0x01
#define MAC_Type_Data           0x02
#define MAC_Type_Reserved       0x03

#define MAC_Subtype_RTS         0x0B
#define MAC_Subtype_CTS         0x0C
#define MAC_Subtype_ACK         0x0D
#define MAC_Subtype_DATA        0x00

using namespace std;

class NotifyHandler;

//Doc:ClassXRef
class MACTimerEvent : public TimerEvent {
  //Doc:Class The class {\tt MACTimerEvent} defines the MAC timer events for
  //Doc:Class the 802.11 class which are essential for managing the various
  //Doc:Class timeouts
public:
  // The types of 802.11 timer events
  typedef enum { BACKOFF_TIMEOUT,
		 NAV_TIMEOUT,
		 CTS_TIMEOUT,
                 ACK_TIMEOUT,
		 DATA_TIMEOUT } MACTimeout_t;
public:
  //Doc:Method
  MACTimerEvent(Event_t ev) : TimerEvent(ev) { };
  //Doc:Desc This is the default constructor for this class
  //Doc:Arg1 The event type object

  virtual ~MACTimerEvent() {}
};

//Doc:ClassXRef
class MACEvent : public Event {
  //Doc:Class The class {\tt MACEvent} defines all other events that are
  //Doc:Class associated with the packet itself.
public:
  typedef enum { PACKET_TX, BACKOFF, SEND_PROBE } MACEvent_t;
public:
  //Doc:Method
  MACEvent() : p(nil) {}
  //Doc:Desc This method is the default constructor

  //Doc:Method
  MACEvent(Event_t ev, Packet* pkt=nil)
    : Event(ev) { p = pkt ? pkt->Copy() : nil; }
  //Doc:Desc This constructor creates a MACEvent object out of an {\tt Event_t}
  //Doc:Desc type and an associated packet object
  //Doc:Arg1 Event_t ev
  //Doc:Arg2 Pointer to the associated packet object

public:
  Packet* p;
};

#define SEQ_CACHE_SIZE	10

class SeqCacheEntry {
public:
  SeqCacheEntry() : ta(MACAddr::NONE), sn(0) {}
  SeqCacheEntry(MACAddr mac, Word_t no) : ta(mac), sn(no) {}

public:
  MACAddr ta;  // tx address, empty slot if NONE
  Word_t  sn;  // sequence number
};

//Doc:ClassXRef
class L2Proto802_11 : public L2Proto, public TimerHandler, public Handler {
  //Doc:Class The {\tt L2Proto802_11 } class derives from the {\tt L2Proto}
  //Doc:Class class and defines the behavior of the 802.11 MAC protocol

public:
  //Doc:Method
  L2Proto802_11(); // Constructor
  //Doc:Desc This method is the default constructor

  //Doc:Method
  virtual ~L2Proto802_11();
  //Doc:Desc This method is the default destructor

  //Doc:Method
  virtual void BuildDataPDU(MACAddr, MACAddr, Packet*);
  //Doc:Desc This method builds a Data PDU using the packet that the upper
  //Doc:Desc layers hand over before pushing it in the interface queue.
  //Doc:Arg1 Source MAC address
  //Doc:Arg2 Destination MAC Address

  //Doc:Method
  virtual void DataRequest(Packet*);
  //Doc:Desc This member function is called by the higher layer requesting
  //Doc:Desc a transfer of  data to a peer.
  //Doc:Arg1 The packet to be sent

  //Doc:Method
  virtual void DataIndication(Packet*);    // from lower layer
  //Doc:Desc This method is called when the direction of protocol stack
  //Doc:Desc traversal is upwards, i.e, a packet has just been received
  //Doc:Desc at the underlying link
  //Doc:Arg1 The packet that has been received

  //Doc:Method
  virtual bool Busy();                     // True if proto/link is busy
  //Doc:Desc Check if the protocol is busy doing processing
  //Doc:Return true if busy; false otherwise

  bool VcsBusy();
  bool PcsBusy();
  bool MediumBusy();

  //Doc:Method
  L2Proto* Copy() const;
  //Doc:Desc Check if the protocol is busy doing processing
  //Doc:Return true if busy; false otherwise

  //Doc:Method
   bool IsWireless() const { return true;}
    //Doc:Desc Query if this protocol needs a wireless interface.
    //Doc:Return Always returns true

  //Doc:Method
  bool SendRTS(MACAddr , MACAddr , Word_t dur);
  //Doc:Desc This method builds and sends the RTS frame with a give src and
  //Doc:Desc dest MAC address with a given duration
  //Doc:Arg1 source MAC Address
  //Doc:Arg2 destination MAC Address
  //Doc:Arg3 duration in microsecs
  //Doc:Return true if send is successful

  //Doc:Method
  bool SendCTS(MACAddr, Word_t dur );
  //Doc:Desc This method builds and sends the RTS frame with a give src and
  //Doc:Desc dest MAC address with a given duration
  //Doc:Arg1 destination MAC Address
  //Doc:Arg2 duration in microsecs
  //Doc:Return true if send is successful

  //Doc:Method
  bool SendDATA();
  //Doc:Desc This method  sends the DATA Frame
  //Doc:Return true if send is successful

  //Doc:Method
  bool SendACK(MACAddr );
  //Doc:Desc This method sends the ACK message
  //Doc:Arg1 The destination MAC address
  //Doc:Return true if send is successful

  //Doc:Method
  bool SendBeacon(Time_t time);
  //Doc:Desc This method sends periodic beacon messages in the interface is
  //Doc:Desc part of an access point
  //Doc:Return Return true on a successful send
  
  // To wait for an IFS (DIFS/EIFS) before sending the packet
  //Doc:Method
  bool ScheduleSendAfterIFS(Packet* p, Time_t t);
  //Doc:Desc This method schedules an actual sending of the frame after a given
  //Doc:Desc time duration. (generally after either SIFS/DIFS).
  //Doc:Arg1 The Packet to be sent
  //Doc:Arg2 The time to wait before sending
  //Doc:Return true if send is successful

  // The Timerhandler method << copying from tcp.h
  //Doc:Method
  void Timeout(TimerEvent*);
  //Doc:Desc This method is the timeout handler for the timer event passed as
  //Doc:Desc the arguement
  //Doc:Arg1 the timer event

  //Doc:Method
  void ScheduleTimer(Event_t, MACTimerEvent*&, Time_t);
  //Doc:Desc This method schedules a MAC timer with a timeout value
  //Doc:Arg1 The event to be used
  //Doc:Arg2 The MAC Timer event reference
  //Doc:Arg3 The time duration

  //Doc:Method
  void CancelTimer(MACTimerEvent*& ev, bool deltimer = false);
  //Doc:Desc This method cacels the scheduled MAC timer
  //Doc:Arg1 Reference to the MACTimer object
  //Doc:Arg2 bool to indicate whether we need to delete the timer

  //Doc:Method
  void BackoffProcedure();
  //Doc:Desc This method is the generic backoff procedure

  // to handle IFS delays
  //Doc:Method
  virtual void Handle(Event* ev, Time_t t);
  //Doc:Desc This method handles the actual MACEvents for sending packets
  //Doc:Arg1 Pointer to the event object
  //Doc:Arg2 The time at which this event is being handled

  void Idle2Busy();

  void Busy2Idle();

  void InitiateDataSequence();

  void InitiateFirstAttempt();

  void InitiateBackoff();

  void SuspendTimer(MACTimerEvent* ev);

  void ResumeTimer(MACTimerEvent* ev);

  void DataRequestFromIFQ();

  void DropPacket();

  void AddSeqCache(MACAddr, Word_t);

  bool FindSeqCache(MACAddr, Word_t);

  void Notify(void*);

  void Bootstrap(void);
  
private:
  enum {
    IDLE, RTS_RCVD, CTS_WAIT, CTS_RCVD, DATA_WAIT, DATA_RCVD, ACK_WAIT
  } mac_state;
  enum { NONE, SUSPENDED, RESUMED } backoffTimerState;
  typedef vector<SeqCacheEntry> SeqCache_t;

  SeqCache_t  seqCache;
  Word_t      nextCacheEntry;

public:
  Packet*         pendingPkt; // as obtained from the upper layer + l2pdu
                              // buffered to be sent after rts/cts/xchange
  Word_t          cw;
  bool            firstAttempt;
  bool            backingOff;
  MACTimerEvent*  backoffTimeout;
  MACTimerEvent*  navTimeout;
  MACTimerEvent*  ctsTimeout;
  MACTimerEvent*  ackTimeout;
  MACTimerEvent*  dataTimeout;
  bool            mevFirstAttempt;
  MACEvent*       mevBackoff;
  Time_t          backoffTime;
  Time_t          navTimer;
  Time_t          backoffTimer;
  Time_t          difs;
  Time_t          eifs;
  Time_t          sifs;
  Timer           timer;

  Rate_t          basicRate;
  Rate_t          dataRate;
  Word_t          shortRetryCount;
  Word_t          longRetryCount;
  Word_t          seqNo;  // data frame sequence number

  Word_t          bseqNo; // beacon sequence number
  Word_t          ibfTime; // inter beacon frame spacing in 1000 usecs
  NotifyHandler*  xmitFailHandler;    // xmit failure handler

  static Word_t   defaultSeqCacheSize;
  
  static Random*  backoffRng;
private:
  long   usec(double t) { return long(floor((t *= 1e6)+0.5)); }
  double txtime(long a, double b) { return a*8/b; }
  void   incr_cw();

};

struct Frame_control
{
public:
  u_char fc_subtype         : 4;
  u_char fc_type            : 2;
  u_char fc_proto_ver       : 2;

  u_char fc_order           : 1;
  u_char fc_wep             : 1;
  u_char fc_more_data       : 1;
  u_char fc_pwr_mgt         : 1;
  u_char fc_retry           : 1;
  u_char fc_more_frag       : 1;
  u_char fc_from_ds         : 1;
  u_char fc_to_ds           : 1;
};


// Add other types as the need arises.
// Its easier this way than to guess them from frame control

typedef enum {
	NONE,
	RTS,
	CTS,
	DATA,
	ACK,
	ASSOCREQ,
	ASSOCRESP,
	REASSOCREQ,
	REASSOCRESP,
	PROBEREQ,
	PROBERESP,
	BEACON,
	DISASSOC,
	AUTH,
	DEAUTH
} Frametype_t;

// framecontrol, duration and fcs are common to all frames

//Doc:ClassXRef
class L2Header802_11 : public PDU  // base class for all 802.11 frames
//Doc:Class The class {\tt L2Header802_11 } defines the base class containing
//Doc:Class all the commonality in the 802.11 frames
{
public:
  //Doc:Method
  L2Header802_11 () : duration(0), fcs(0), type(NONE) {}
  //Doc:Desc Default constructor

  L2Header802_11 (Word_t dur_, Long_t fcs_)
	: duration(dur_), fcs(fcs_), type(NONE) {}
  L2Header802_11 (Word_t dur_, Long_t fcs_, Frametype_t typ)
	: duration(dur_), fcs(fcs_), type(typ) {}

  virtual Size_t Size() const = 0;
  //  PDU* Copy() {return new L2Header802_11(*this);}
  void    setDuration(Word_t dur) { duration = dur; }
#ifdef ETRACE
  //Doc:Method
  virtual Proto_t Proto() { return 11;} // For 802.11        
  //Doc:Desc Most of the protocols are identified by the numeric values as
  //Doc:Desc listed by IANA. thie method returns that protocol number
  //Doc:Return the protocol number

  //Doc:Method
  virtual Layer_t Layer() { return 2;}         // Indicate layer 2 pdu
  //Doc:Desc This method returns the layer number it belongs in the ISO model
  //Doc:Return the layer number
  
#endif

public:
  Frame_control fc;
  Word_t        duration;
  Long_t        fcs;

  Frametype_t   type; /*its not needed if type/subtype fields are correct*/
  MACAddr       ra;
};


//Doc:ClassXRef
class RTS_frame : public L2Header802_11
//Doc:Class The class {\tt RTS_frame} defines the RTS frame
{
public:
  //Doc:Method
  RTS_frame (): L2Header802_11() { type = RTS; }
  //Doc:Desc This is the default constructor

  //Doc:Method
  RTS_frame (Word_t dur, MACAddr rcv, MACAddr xmt, Long_t fcs_);
  //Doc:Desc This method constructs the RTS frame PDU with a give set of
  //Doc:Desc arguments
  //Doc:Arg1 duration field in micro seconds
  //Doc:Arg2 Dest MAC Address
  //Doc:Arg3 Source MAC Address
  //Doc:Arg4 the checksum

  //Doc:Method
  Size_t  Size() const;
  //Doc:Desc The Size of the PDU
  //Doc:Return size of the pdu in bytes

  //Doc:Method
  PDU*    Copy() { return new RTS_frame(*this); }
  //Doc:Desc This method returns a pointer to the copy of the PDU
  //Doc:Return Pointer to the copy

public:
  MACAddr ta;
};

//Doc:ClassXRef
class CTS_frame : public L2Header802_11
//Doc:Class The class {\tt CTS_frame} defines the CTS frame
{
public:
  //Doc:Method
  CTS_frame() : L2Header802_11() { type = CTS; }
  //Doc:Desc  Default Constructor

  //Doc:Method
  CTS_frame(Word_t dur, MACAddr rcv, Long_t fcs_);
  //Doc:Desc This method constructs the CTS frame PDU with a give set of
  //Doc:Desc arguments
  //Doc:Arg1 duration field in micro seconds
  //Doc:Arg2 Dest MAC Address
  //Doc:Arg3 the checksum

  //Doc:Method
  Size_t  Size() const;
  //Doc:Desc The Size of the PDU
  //Doc:Return size of the pdu in bytes

  //Doc:Method
  PDU*    Copy() { return new CTS_frame(*this); }
  //Doc:Desc This method returns a pointer to the copy of the PDU
  //Doc:Return Pointer to the copy

};

//Doc:ClassXRef
class ACK_frame : public L2Header802_11
//Doc:Class The class {\tt ACK_frame} defines the ACK frame
{
public:
  //Doc:Method
  ACK_frame() : L2Header802_11() { type = ACK; }
  //Doc:Desc Default Constructor

  //Doc:Method
  ACK_frame(Word_t dur, MACAddr rcv, Long_t fcs_);
  //Doc:Desc This method constructs the ACK frame PDU with a give set of
  //Doc:Desc arguments
  //Doc:Arg1 duration field in micro seconds
  //Doc:Arg2 Dest MAC Address
  //Doc:Arg3 the checksum

  //Doc:Method
  Size_t  Size() const;
  //Doc:Desc The Size of the PDU
  //Doc:Return size of the pdu in bytes

  //Doc:Method
  PDU*    Copy() { return new ACK_frame(*this); }
  //Doc:Desc This method returns a pointer to the copy of the PDU
  //Doc:Return Pointer to the copy
};


//Doc:ClassXRef
class DATA_frame : public L2Header802_11
//Doc:Class The class {\tt DATA_frame} defines the DATA frame
{
public:
  //Doc:Method
  DATA_frame() : L2Header802_11() { type = DATA; }
  //Doc:Desc Default constructor

  //Doc:Method
  DATA_frame(Word_t dur, MACAddr rcv, MACAddr xmt, Word_t sqn, Long_t fcs_);
  //Doc:Desc This constructs a data frame from a give set of arguments
  //Doc:Arg1 Duration in us
  //Doc:Arg2 Destination MAC Address
  //Doc:Arg3 Source MAC Address
  //Doc:Arg4 the sequence number in case of fragmented frames
  //Doc:Arg5 The checksum length

  //Doc:Method
  Size_t  Size() const;
  //Doc:Desc The Size of the PDU
  //Doc:Return size of the pdu in bytes

  //Doc:Method
  PDU*    Copy() { return new DATA_frame(*this); }
  //Doc:Desc This method returns a pointer to the copy of the PDU
  //Doc:Return Pointer to the copy

public:
  MACAddr  addr1;  /* is dst for to_ds = 0 ; GENERALLY DEST
					  ra if to_ds = from_ds = 1;
					  else bssid */
  MACAddr  addr2;  /* is src for from_ds = 0; GENERALLY SRC
					  ta for to_ds=from_ds=1;
					  else bssid */
  MACAddr  addr3;  /* is da  for to_ds = 1 ;
					  bssid if to_ds = from_ds = 0 ;else sa*/
  MACAddr  addr4;  /* is sa  for to_ds = from_ds = 1 ; else NULL*/
  Word_t   seqno;
};



/*
  The Management frames' MAC header can be abstracted out but there is
  little to be gained from that.
*/

/*
  Information elements
 */

class Info_element {
 public:
	u_char el_id;
	u_char len;
};

class supp_rates_element : public Info_element {
 public:
  u_char supp[8];
};

class ssid_element: public Info_element {
 public:
  string ssid; /* not to exceed 32 chars */
};

class fh_paramset_element: public Info_element {
 public:
	Word_t dwell_time;
	u_char hop_set;
	u_char hop_pattern;
	u_char hop_index;
};

class ds_paramset_element: public Info_element {
 public:
	u_char currChan;
};

class cf_paramset_element: public Info_element {
 public:
	u_char cfp_count;
	u_char cfp_period;
	Word_t cfp_maxdur;
	Word_t cfp_remdur;
};

class tim_element : public Info_element {
 public:
	u_char dtim_count;
	u_char dtim_period;
	u_char bitmap_control;
	u_char pv_bitmap[256];
};

class ibss_paramset_element : public Info_element {
 public:
	Word_t aim_window;
};

class chall_text_element : public Info_element {
 public:
	u_char text[253];
};


/* The actual frames */

class Beacon_frame: public L2Header802_11 {
 public:
  Beacon_frame() : L2Header802_11() {
    type = BEACON;
  }

  //Doc:Method
  Size_t  Size() const;
  //Doc:Desc The Size of the PDU
  //Doc:Return size of the pdu in bytes

  //Doc:Method
  PDU*    Copy() { return new Beacon_frame(*this); }
  //Doc:Desc This method returns a pointer to the copy of the PDU
  //Doc:Return Pointer to the copy
  
 public:
	double timestamp;
	Word_t beacon_int;
	Word_t cap_info;
	ssid_element ssidel; /* size is 32 and is taken care on using Size() */
	supp_rates_element ssrel;
	fh_paramset_element fhps;
	ds_paramset_element dsps;
	ibss_paramset_element ibssps;
	tim_element  tim;
	MACAddr bssid;
	MACAddr sa;
	Word_t seqno;
};

class Disassoc_frame: public L2Header802_11 {
 public:
  Disassoc_frame() : L2Header802_11() {
    type = DISASSOC;
  }

  
  //Doc:Method
  Size_t  Size() const;
  //Doc:Desc The Size of the PDU
  //Doc:Return size of the pdu in bytes

  //Doc:Method
  PDU*    Copy() { return new Disassoc_frame(*this); }
  //Doc:Desc This method returns a pointer to the copy of the PDU
  //Doc:Return Pointer to the copy
  
 public:
	MACAddr bssid;
	MACAddr sa;
	Word_t seqno;
	Word_t reason_code;
};

class Assocreq_frame: public L2Header802_11 {
 public:
  Assocreq_frame() : L2Header802_11() {
    type = ASSOCREQ;
  }

  //Doc:Method
  Size_t  Size() const;
  //Doc:Desc The Size of the PDU
  //Doc:Return size of the pdu in bytes

  //Doc:Method
  PDU*    Copy() { return new Assocreq_frame(*this); }
  //Doc:Desc This method returns a pointer to the copy of the PDU
  //Doc:Return Pointer to the copy

 public:
	
	MACAddr bssid;
	MACAddr sa;
	Word_t  seqno;
	Word_t  cap_info;
	Word_t  listen_int;
	ssid_element ssid;
	supp_rates_element ssrel;
};

class Assocresp_frame: public L2Header802_11 {
 public:
  Assocresp_frame() : L2Header802_11() {
    type = ASSOCRESP;
  }

  //Doc:Method
  Size_t  Size() const;
  //Doc:Desc The Size of the PDU
  //Doc:Return size of the pdu in bytes

  //Doc:Method
  PDU*    Copy() { return new Assocresp_frame(*this); }
  //Doc:Desc This method returns a pointer to the copy of the PDU
  //Doc:Return Pointer to the copy

 public:
	MACAddr bssid;
	MACAddr sa;
	Word_t  seqno;
	Word_t  cap_info;
	Word_t  status_code;
	Word_t  aid;
	supp_rates_element ssrel;
};

class Reassocreq_frame: public L2Header802_11 {
 public:
  Reassocreq_frame() : L2Header802_11() {
    type = REASSOCREQ;
  }

  //Doc:Method
  Size_t  Size() const;
  //Doc:Desc The Size of the PDU
  //Doc:Return size of the pdu in bytes

  //Doc:Method
  PDU*    Copy() { return new Reassocreq_frame(*this); }
  //Doc:Desc This method returns a pointer to the copy of the PDU
  //Doc:Return Pointer to the copy
  
 public:
	MACAddr bssid;
	MACAddr sa;
	Word_t seqno;
	Word_t  cap_info;
	Word_t  listen_int;
	MACAddr current_ap;
	ssid_element ssid;
	supp_rates_element ssrel;
};

class Reassocresp_frame: public L2Header802_11 {
 public:
  Reassocresp_frame() : L2Header802_11() {
    type = REASSOCRESP;
  }

  //Doc:Method
  Size_t  Size() const;
  //Doc:Desc The Size of the PDU
  //Doc:Return size of the pdu in bytes

  //Doc:Method
  PDU*    Copy() { return new Reassocresp_frame(*this); }
  //Doc:Desc This method returns a pointer to the copy of the PDU
  //Doc:Return Pointer to the copy

 public:
	MACAddr bssid;
	MACAddr sa;
	Word_t seqno;
	Word_t  cap_info;
	Word_t  status_code;
	Word_t  aid;
	supp_rates_element ssrel;
};

class Probereq_frame: public L2Header802_11 {
 public:
  Probereq_frame() : L2Header802_11() {
    type = PROBEREQ;
  }

  //Doc:Method
  Size_t  Size() const;
  //Doc:Desc The Size of the PDU
  //Doc:Return size of the pdu in bytes

  //Doc:Method
  PDU*    Copy() { return new Probereq_frame(*this); }
  //Doc:Desc This method returns a pointer to the copy of the PDU
  //Doc:Return Pointer to the copy

  
 public:
	MACAddr bssid;
	MACAddr sa;
	Word_t seqno;
	ssid_element ssid;
	supp_rates_element ssrel;
};

class Proberesp_frame: public L2Header802_11 {
 public:
  Proberesp_frame() : L2Header802_11() {
    type = PROBERESP;
  }
  
  //Doc:Method
  Size_t  Size() const;
  //Doc:Desc The Size of the PDU
  //Doc:Return size of the pdu in bytes

  //Doc:Method
  PDU*    Copy() { return new Proberesp_frame(*this); }
  //Doc:Desc This method returns a pointer to the copy of the PDU
  //Doc:Return Pointer to the copy
  
 public:
	MACAddr bssid;
	MACAddr sa;
	Word_t  seqno;
	Word_t  timestamp;
	Word_t  beacon_int;
	Word_t  cap_info;
	ssid_element ssid;
	supp_rates_element ssrel;
	fh_paramset_element fhps;
	ds_paramset_element dsps;
	cf_paramset_element cfps;
	ibss_paramset_element ibssps;
};


class Auth_frame: public L2Header802_11 {
 public:
  Auth_frame() : L2Header802_11() {
    type = AUTH;
  }

    //Doc:Method
  Size_t  Size() const;
  //Doc:Desc The Size of the PDU
  //Doc:Return size of the pdu in bytes

  //Doc:Method
  PDU*    Copy() { return new Auth_frame(*this); }
  //Doc:Desc This method returns a pointer to the copy of the PDU
  //Doc:Return Pointer to the copy
  
 public:
	MACAddr bssid;
	MACAddr sa;
	Word_t seqno;
	Word_t algo_nr;
	Word_t algo_trans_nr;
	Word_t status;
	chall_text_element cte;
};

class Deauth_frame: public L2Header802_11 {
 public:
  Deauth_frame() : L2Header802_11 () {
    type = DEAUTH;
  }

  //Doc:Method
  Size_t  Size() const;
  //Doc:Desc The Size of the PDU
  //Doc:Return size of the pdu in bytes

  //Doc:Method
  PDU*    Copy() { return new Deauth_frame(*this); }
  //Doc:Desc This method returns a pointer to the copy of the PDU
  //Doc:Return Pointer to the copy
  
 public:
	MACAddr bssid;
	MACAddr sa;
	Word_t  seqno;
	Word_t  reason;
};

#endif
