/*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
                                                                                
* Permission is hereby granted for any non-commercial use of this code
*/

// USAP protocol implementation based on US Patent: 6504829,5719868
#ifdef USAP_ADDON

#include <iostream>

#include "debug.h"
#include "l2protousap.h"
#include "interface.h"
#include "ipv4.h"
#include "trace.h"
#include "node.h"
#include "hex.h"
#include "macaddr.h"
#include "queue.h"
#include "link.h"
#include "assert.h"

#ifdef ETRACE
#include "etrace.h"
#endif // ETRACE

USAP_DATA_frame::USAP_DATA_frame()
{  
	 type       = USAP_BCST_DATA;
	 hdr_length = sizeof(sa)+sizeof(ra)+sizeof(type)+sizeof(seqno);
}

USAP_DATA_frame::USAP_DATA_frame(MACAddr rcv, MACAddr xmt, Word_t sqn )
{
	 type  = USAP_BCST_DATA;
   sa    = xmt;
	 ra    = rcv;
	 seqno = sqn;
	 hdr_length = sizeof(sa)+sizeof(ra)+sizeof(type)+sizeof(seqno);
}



L2ProtoUsap::L2ProtoUsap() : L2Proto()
{
}

Size_t USAP_DATA_frame::Size() const
{
  return hdr_length;
}

L2ProtoUsap::~L2ProtoUsap() 
{

}

void L2ProtoUsap::DataRequest(Packet* p)
{
	delete p;
}


void L2ProtoUsap::BuildDataPDU(MACAddr src, MACAddr dst, Packet* p)
{   
  USAP_DATA_frame* l2pdu = new USAP_DATA_frame(dst, src, ++seqNo);
  p->PushPDU(l2pdu);
} 

void L2ProtoUsap::DataIndication(Packet* p)
{
#ifdef ETRACE 
	ETrace::Instance()->TracePacket(p, iface->GetNode(), iface,2, "r"); // 2 for layer 2
#endif
	L2HeaderUsap* l2pdu = (L2HeaderUsap*)p->PopPDU();
	MACAddr dst;

	iface->GetNode()->TracePDU(this, l2pdu);

	switch(l2pdu->type) {
		case USAP_BCST_DATA: {
			USAP_DATA_frame* dataFrame = (USAP_DATA_frame*)l2pdu;
			//iface->HandleLLCSNAP(p);
			
			dst = dataFrame->ra;
	    if (dst.IsBroadcast()) {
				iface->HandleLLCSNAP(p);
			}
			else if (dst == iface->GetMACAddr()) {
				iface->HandleLLCSNAP(p);
			}
			else { // overheard unicast packets
        delete (p);
			}
			break;
		}
		default:
			delete (p); // unknow type
	}

}


bool L2ProtoUsap::Busy()
{
  // Determine if protocol is busy
  return true;
}

void L2ProtoUsap::Notify(void* v)
{

  /* Calling this function means MAC received
	 a PHY-RXSTART, PHY-RXEND, PHY-TXSTART, or PHY-TXEND.indication,
  */
  InterfaceWireless::PhyInd_t* phyInd = (InterfaceWireless::PhyInd_t*)v;
  switch (*phyInd)
	{
	case InterfaceWireless::PHY_RXSTART:
	case InterfaceWireless::PHY_TXSTART:
	case InterfaceWireless::PHY_RXEND_ERR:
	case InterfaceWireless::PHY_RXEND:
	case InterfaceWireless::PHY_TXEND:
	default:
	  break;
	} //switch
}

void L2ProtoUsap::Bootstrap(void)
{
}

void L2ProtoUsap::Handle(Event* ev, Time_t t )
{
}

void L2ProtoUsap::Timeout(TimerEvent *ev)
{

    int previousStatus = currentStatus;

    assert(ev->event        == USAPTimerEvent::MSG_MAC_TimerExpired);
    assert(Simulator::Now() == timerExpirationTime);

    currentStatus   = nextStatus;
	  currentSlotType = nextSlotType;
    currentSlotId   = nextSlotId;
	  currentFrameId  = nextFrameId;

    assert((currentStatus != previousStatus) ||
           (currentStatus == USAP_STATUS_TX &&
            previousStatus == USAP_STATUS_TX) || (currentSlotId == 0) );

	if (currentSlotId == 0) { // new frame
		UpdateFrameDescriptor();
		currentStatus   = frameDescriptor[currentSlotId].status;
		currentSlotType = frameDescriptor[currentSlotId].slotType;
#ifdef X_DEBUG_USAP 
		PrintFrameDescriptor();
#endif // DEBUG_USAP
	}

  // TDMA state machine is here

   switch (currentStatus) {

        case USAP_STATUS_RX: { // all RXs are handled by DataIndication
            break;
        }

        case USAP_STATUS_IDLE: {
            break;
        }

        case USAP_STATUS_TX: {
            Time_t transmissionDelay;

#ifdef X_DEBUG_USAP
            printf("USAP DEBUG: -n %d in TRANSMIT slot %d\n",
                    (iface->GetNode())->Id(), currentSlotId);
#endif // DEBUG_USAP

			      if (frameDescriptor[currentSlotId].slotType == USAP_CTL_SLOT) {
				     // Place holder for USAP control data: synchronization, slot scheduling info
				       break;
			      }

            //
            // Dequeue packet from the interface: upper layer is forced to queue 
						// as Busy() is always true
            //
						if (iface->GetQueue()->LengthPkts() > 0)
						{
							Packet* p = iface->GetQueue()->Deque();
							transmissionDelay = p->Size()*8/dataRate;
							if (transmissionDelay > frameDescriptor[currentSlotId].slotDuration) {
								cout << "node ["<<(iface->GetNode())->Id()<<"] : transmission delay = ";
								cout << transmissionDelay <<" packet size= "<<p->Size();
								cout << "slot Duration = "<<frameDescriptor[currentSlotId].slotDuration<<endl;
								exit(0);
							}
#ifdef ETRACE
						  ETrace::Instance()->TracePacket(p, iface->GetNode(), iface,2, "s");
#endif //ETRACE

							iface->GetLink()->Transmit(p, iface, iface->GetNode(), dataRate);
#ifdef DEBUG_USAP 
							{
								USAP_DATA_frame* df = (USAP_DATA_frame*) p->PeekPDU();
								cout << "***TRANSMITTING*** node ["<<(iface->GetNode())->Id()<<"] at ";
								cout << Simulator::Now() <<" during SLOT["<<currentSlotId;
								cout << "] src="<< df->sa <<" dst="<<df->ra<<endl;
							}
#endif // DEBUG_USAP
						}

            break;
        }

        default: {
            abort();
        }
    }

	// Update timer for next TX
	if(ev)
		delete ev;

	UpdateTimer();
}


L2Proto* L2ProtoUsap::Copy() const
{
  return new L2ProtoUsap(*this);
}


void L2ProtoUsap::PrintFrameDescriptor(void)
{
	  int i;

    printf("-n %d -t %f:\n", (iface->GetNode())->Id(), (Time_t) Simulator::Now());

    for (i = 0; i < numSlotsPerFrame; i++) {
		printf("(");

		switch (frameDescriptor[i].slotType){
			case USAP_CTL_SLOT: {
				printf("C,");
				break;
			}
			case USAP_BROADCAST_SLOT: {
				printf("B,");
				break;
			}
			case USAP_DATA_SLOT: {
				printf("D,");
				break;
			}
			default:
				assert(false); abort();
		} // switch (frameDescriptor[i].slotType)

		switch (frameDescriptor[i].status){
			case USAP_STATUS_TX: {
				printf("T,");
				break;
			}
			case USAP_STATUS_RX: {
				printf("R,");
				break;
			}
			case USAP_STATUS_IDLE: {
				printf("I,");
				break;
			}
			default:
				assert(false); abort();
		} // switch (frameDescriptor[i].status)

    	printf("%f) ",frameDescriptor[i].slotDuration);

		if ((i-4) % 5 == 0 ) printf("\n");
    }

    printf("\n");

}

void L2ProtoUsap::UpdateFrameDescriptor(void) {
	  int i;

    for (i = 0; i < numSlotsPerFrame; i++) {

		if (i < numCtlSlotsPerFrame) { // control slots part 

			frameDescriptor[i].slotType     = USAP_CTL_SLOT;
			frameDescriptor[i].slotDuration = ctlSlotDuration;

			if ((iface->GetNode())->Id() == 
					(currentFrameId % numFramesForCtlSlotsPerCycle) 
					* numCtlSlotsPerFrame
					+(i) )  
				  frameDescriptor[i].status = USAP_STATUS_TX;
			else
			  	frameDescriptor[i].status = USAP_STATUS_RX;
		}
		else if (i >= numCtlSlotsPerFrame &&
				 i < (numCtlSlotsPerFrame + numBcSlotsPerFrame) ) { // broadcast slots 

			frameDescriptor[i].slotType     = USAP_BROADCAST_SLOT;
			frameDescriptor[i].slotDuration = bcSlotDuration;

			if ( (iface->GetNode())->Id() == 
				(currentFrameId % numFramesForBcSlotsPerCycle)
				* numBcSlotsPerFrame
				+(i-numCtlSlotsPerFrame))
				
				frameDescriptor[i].status = USAP_STATUS_TX;
			else
				frameDescriptor[i].status = USAP_STATUS_RX;
     }
		 else if ( i>= (numCtlSlotsPerFrame + numBcSlotsPerFrame) ) { // Data slots

			frameDescriptor[i].slotType     = USAP_DATA_SLOT;
			frameDescriptor[i].slotDuration = dataSlotDuration;

			if ( (iface->GetNode())->Id() ==
				 (currentFrameId % numFramesForDataSlotsPerCycle)
				 * numDataSlotsPerFrame
				 + (i-numCtlSlotsPerFrame - numBcSlotsPerFrame))
				frameDescriptor[i].status = USAP_STATUS_TX;
			else
				frameDescriptor[i].status = USAP_STATUS_RX;
      }
    }
}

void L2ProtoUsap::InitializeFrameDescriptor(void)
{
    frameDescriptor = new UsapSlotDescriptor[numSlotsPerFrame];

    UpdateFrameDescriptor ();
#ifdef X_DEBUG_USAP
    PrintFrameDescriptor();
#endif // DEBUG_USAP
}


void L2ProtoUsap::InitializeTimer(void) {
    int i;

    int initialStatus = (int)frameDescriptor[0].status;

#ifdef X_DEBUG_USAP
     printf("First timer is being initialized\n");
#endif // DEBUG_USAP

    if (initialStatus == USAP_STATUS_TX) {
        i = 0;
    }
    else {
        for (i = 0; i < numSlotsPerFrame; i++) {
            if ((int)frameDescriptor[i].status != initialStatus) {
                break;
            }
        }
    }

    if (i == numSlotsPerFrame) {
        // Schedule no timer
#ifdef DEBUG_USAP
            printf("No initial timer sent.  "
                   "Initial Status matches all slots in frame\n");
#endif // DEBUG_USAP       
    }
    else {
      Time_t delay = 0;
		  int j;
			for (j = 0; j < i; j++) 
				delay += (frameDescriptor[j].slotDuration + guardTime);

      currentStatus   = initialStatus;
      currentSlotId   = 0;
		  currentSlotType = frameDescriptor[0].slotType;

      nextStatus      = frameDescriptor[i].status;
		  nextSlotType    = frameDescriptor[i].slotType;
      nextSlotId      = i;

      timerExpirationTime = delay + Simulator::Now();

			ScheduleTimer(USAPTimerEvent::MSG_MAC_TimerExpired,UsapMacLayer,delay);

#ifdef X_DEBUG_USAP
       printf("Initial timer sent for node [%d] with a delay of %f at %f\n\n",(iface->GetNode())->Id(),
                delay,(Time_t) Simulator::Now());
#endif // DEBUG_USAP
    }
}


void L2ProtoUsap::UpdateTimer(void) {
    int i;
    int slotId = currentSlotId;
    Time_t delay = 0;

    assert(currentStatus == frameDescriptor[currentSlotId].status);

    for (i = 0; i < numSlotsPerFrame; i++) {
        if (slotId == numSlotsPerFrame - 1) {
            nextSlotId = 0;
			      nextFrameId = currentFrameId + 1;
        }
        else {
            nextSlotId = slotId + 1;
        }
        nextStatus   = frameDescriptor[nextSlotId].status;
        nextSlotType = frameDescriptor[nextSlotId].slotType;

        delay += frameDescriptor[slotId].slotDuration + guardTime;

        if (nextSlotId == 0) {
            delay += interFrameTime;
        }

		// break if status change, transmission or new frame starts
		// USAP changes the frame discription every frame

        if ((nextStatus != currentStatus) ||
            (nextStatus == USAP_STATUS_TX && currentStatus == USAP_STATUS_TX) ||
            (nextSlotId == 0))
			 
        {
 #ifdef X_DEBUG_USAP
                printf("USAP DEBUG: -n %d, Next Timer Being Sent, CurrentStatus : %d NextStatus : %d "
                       "CurrentSlotId: %d, NextSlotId : %d, CurrentFrameId : %d NextFrameId : %d\n",
                       (iface->GetNode())->Id(), currentStatus, nextStatus,
					             currentSlotId, nextSlotId, currentFrameId, nextFrameId);

 #endif // DEBUG_USAP

            break;
        }

        slotId++;

    	assert(slotId != numSlotsPerFrame);
    }

    assert(i != numSlotsPerFrame);

    timerExpirationTime = delay + Simulator::Now();

    ScheduleTimer(USAPTimerEvent::MSG_MAC_TimerExpired,UsapMacLayer,delay);
}


void L2ProtoUsap::Init(void)
                 
{
	 const NodeVec_t& nodes = Node::GetNodes();

   numCtlSlotsPerFrame               = USAP_DEFAULT_NUM_CTL_SLOTS_PER_FRAME;
   numBcSlotsPerFrame                = USAP_DEFAULT_NUM_BC_SLOTS_PER_FRAME;

   if (nodes.size() <= USAP_DEFAULT_NUM_DATA_SLOTS_PER_FRAME)
		 numDataSlotsPerFrame = nodes.size();
	 else 
		 numDataSlotsPerFrame = USAP_DEFAULT_NUM_DATA_SLOTS_PER_FRAME;

	 numSlotsPerFrame                  = numCtlSlotsPerFrame +
							                         numBcSlotsPerFrame  +
							                         numDataSlotsPerFrame ;

	 numFramesForCtlSlotsPerCycle      = max (1, (int) ceil((float) nodes.size()/numCtlSlotsPerFrame));
   numFramesForBcSlotsPerCycle       = max (1, (int) ceil((float) nodes.size()/numBcSlotsPerFrame));
   numFramesForDataSlotsPerCycle     = max (1, (int) ceil((float) nodes.size()/numDataSlotsPerFrame));

   ctlSlotDuration                   = USAP_DEFAULT_CTL_SLOT_DURATION;
   bcSlotDuration                    = USAP_DEFAULT_BC_SLOT_DURATION;
   dataSlotDuration                  = USAP_DEFAULT_DATA_SLOT_DURATION;
   guardTime                         = USAP_DEFAULT_GUARD_TIME;
   interFrameTime                    = USAP_DEFAULT_INTERFRAME_TIME;

   currentFrameId                    = 0;
	 nextFrameId                       = 0;

	 dataRate                          = 11000000.0; // 2 Mbps same as l2proto802.11.cc
	 seqNo                             = 0;         // data frame sequence number

   InitializeFrameDescriptor();
   InitializeTimer();
}


void L2ProtoUsap::ScheduleTimer(Event_t e, USAPTimerEvent*& ev, Time_t t)
{
	assert ( t >= 0 );

  ev = new USAPTimerEvent(e);

  timer.Schedule(ev, t, this);
}
#endif // USAP_ADDON
