/*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
                                                                                
* Permission is hereby granted for any non-commercial use of this code
*/

// USAP protocol based on US Patent: 6504829,5719868

// a global flag used in initiliazation testing
#define usap_is_enabled false

#ifdef USAP_ADDON

#undef usap_is_enabled
#define usap_is_enabled true

#ifndef __l2proto_usap_h
#define __l2proto_usap_h
#include <sys/types.h>

#include "math.h"
#include "l2proto.h"
#include "tfstream.h"
#include "macaddr.h"
#include "timer.h"
#include "rng.h"
#include "interface-wireless.h"

#define USAP_DEFAULT_NUM_CHANNELS 1

#define USAP_DEFAULT_MAX_NUM_NODES 100

#define USAP_DEFAULT_NUM_CTL_SLOTS_PER_FRAME   13  //number of boostrap/control slots
#define USAP_DEFAULT_NUM_BC_SLOTS_PER_FRAME    2    //number of broadcast slots per frame
#define USAP_DEFAULT_NUM_DATA_SLOTS_PER_FRAME  40 //number of Reservation slots and rotating broadcast slots

#define USAP_DEFAULT_NUM_FRAMES_FOR_CTL_PER_CYCLE  8 // repeat cycle for ctl slots: 13x4=52
#define USAP_DEFAULT_NUM_FRAMES_FOR_BC_PER_CYCLE   10 // repeat cycle for ctl slots: 25x2=100
#define USAP_DEFAULT_NUM_FRAMES_FOR_DATA_PER_CYCLE 1 // repeat cycle for data slots: 100x1=100

#define MICRO_SECOND                     (0.000001)         
#define USAP_DEFAULT_CTL_SLOT_DURATION   (50*MICRO_SECOND) //0.1 msec
#define USAP_DEFAULT_BC_SLOT_DURATION    (1250*MICRO_SECOND) //1.2 msec
#define USAP_DEFAULT_DATA_SLOT_DURATION  (1250*MICRO_SECOND) //1.2 msec

#define USAP_DEFAULT_GUARD_TIME 		     (0*MICRO_SECOND)
#define USAP_DEFAULT_INTERFRAME_TIME     (0*MICRO_SECOND)

//#define DEBUG_USAP 0


using namespace std;

typedef enum {
  USAP_STATUS_TX,
  USAP_STATUS_RX,
  USAP_STATUS_IDLE
} UsapStatus;

typedef enum {
	USAP_CTL_SLOT,
  USAP_BROADCAST_SLOT,
	USAP_DATA_SLOT
} UsapSlotType;

typedef struct usap_slot_slot_descriptor_str {
	UsapStatus status;
	UsapSlotType slotType;
	double slotDuration;
} UsapSlotDescriptor;

// Type of MAC frame
typedef enum {
	    USAP_CTL              = 0x01, // 00 0001
	    USAP_UCST_DATA        = 0x20, // 10 0000
	    USAP_UCST_DATA_ACK    = 0x21, // 10 0001
		  USAP_BCST_DATA        = 0x22, // 10 0010

	    USAP_NONE             = 0x3F  // 11 1111 Reserved
} UsapMacFrameType;


class L2HeaderUsap : public PDU  // base class for all usap frames
{
public:
  L2HeaderUsap () : type(USAP_NONE) {}

  virtual Size_t Size() const = 0;
#ifdef ETRACE
  virtual Proto_t Proto() { return 18;} // 18 for usap
  virtual Layer_t Layer() { return 2;}  // Indicate layer 2 pdu
#endif

public:

  UsapMacFrameType   type; 
  MACAddr            sa;    // sender address  
  MACAddr            ra;    // receiver address
};


class USAP_DATA_frame : public L2HeaderUsap
{
public:
  USAP_DATA_frame();  
  USAP_DATA_frame(MACAddr rcv, MACAddr xmt, Word_t sqn );

  Size_t  Size() const;

  PDU*    Copy() { return new USAP_DATA_frame(*this); }

public:
  Word_t   seqno;
	Size_t   hdr_length;
};

class USAPTimerEvent : public TimerEvent {

public:
  // The types of usap timer events
  typedef enum 
	{
		MSG_MAC_TimerExpired
	} MACTimeout_t;

  USAPTimerEvent(Event_t ev) : TimerEvent(ev) { };
  virtual ~USAPTimerEvent() {}
};

class L2ProtoUsap : public L2Proto, public TimerHandler, public Handler {
	public:  
		L2ProtoUsap();
		virtual ~L2ProtoUsap();

//  required functions from base classes
    virtual void BuildDataPDU(MACAddr, MACAddr, Packet*);
    virtual void DataRequest(Packet*);    // from upper layer
		virtual void DataIndication(Packet*); // from lower layer
    virtual bool Busy();                  // True if proto/link is busy
		L2Proto* Copy() const;
		bool IsWireless() const { return true;}
		void Bootstrap(void);
		void Notify(void*);
		void Timeout(TimerEvent*);
		virtual void Handle(Event* ev, Time_t t);
		void Init(void);
		void InitializeTimer(void);
		void InitializeFrameDescriptor(void);
		void UpdateTimer(void);
		void UpdateFrameDescriptor(void);
	  void PrintFrameDescriptor(void);
		void ScheduleTimer(Event_t e, USAPTimerEvent*& ev, Time_t t);

	public:
		Timer           timer;           // Timer for TDMA scheduling
		USAPTimerEvent* UsapMacLayer; 

	private:

    int currentStatus;
	  int currentSlotType;
    int currentSlotId;
	  int currentFrameId;

    int nextStatus;
    int nextSlotType;
    int nextSlotId;
	  int nextFrameId;

    int currentReceiveSlotId;

    Time_t timerExpirationTime;

    int numCtlSlotsPerFrame;
    int numBcSlotsPerFrame;
    int numDataSlotsPerFrame;

    int numSlotsPerFrame;

	  int numFramesForCtlSlotsPerCycle;
	  int numFramesForBcSlotsPerCycle;
	  int numFramesForDataSlotsPerCycle;

    int lowestDataRateType;
    int highestDataRateType;
    int highestDataRateTypeForBC;

    UsapSlotDescriptor*     frameDescriptor;

    Time_t ctlSlotDuration;
    Time_t bcSlotDuration;
    Time_t dataSlotDuration;

    Time_t guardTime;
    Time_t interFrameTime;

	  Rate_t dataRate;

		Word_t seqNo;  // data frame sequence number

};

#endif // __l2proto_usap_h

#endif // USAP_ADDON
