/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Link base class
// George F. Riley.  Georgia Tech, Spring 2002

#include <math.h>
#include <stdlib.h>

//#define DEBUG_MASK 0x10
#include "debug.h"
#include "link.h"
#include "simulator.h"
#include "event.h"
#include "ratetimeparse.h"
#include "linkp2p.h"
#include "droptail.h"
#include "hex.h"



using namespace std;

// Constructors
LinkReal::LinkReal()
    : Link(), notify(nil)
{
  ConstructorHelper(Link::DefaultRate(), Link::DefaultDelay());
}

LinkReal::LinkReal(Rate_t b, Time_t d)
    : Link(b, d), notify(nil)
{
  ConstructorHelper( b, d);
}

LinkReal::LinkReal(const LinkReal& c)
    : Link(c), notify(c.notify)
{ // Copy constructor
  busy = false;
  jitter = c.jitter->Copy();
  bw = c.bw;
  delay = c.delay;
  weight = c.weight;
  bitErrorRate = c.bitErrorRate;
  bytesSent = 0;
  pktsSent = 0;
  utilizationStart= 0.0;
  debugVerbose = c.debugVerbose;
  DBG((Stats::linksCreated++));
}

// Destructor
LinkReal::~LinkReal()
{
  if (jitter) delete jitter;
}

bool LinkReal::TransmitWithDelay(Packet* p, Time_t extraDelay)
{
  // Schedule receipt of new packet
  // at time extraDelay + txTime + delay + jitter;
  Time_t txTime = ((Rate_t)p->Size() * 8) / Bandwidth(); // Time to transmit
  Time_t txDelay = Delay();
  Time_t jit = jitter->Value();
  DEBUG0((cout << "Link " << this
          << " transmitting with extraDelay " << extraDelay
          << " txtime " << txTime
          << " delay " << txDelay
          << " bandwidth " << Bandwidth() << endl));
  Time_t rxTime = extraDelay + txTime + txDelay + jit;
  // Create a new packet receipt event
  LinkEvent* evRx = new LinkEvent(LinkEvent::PACKET_RX, p);
  if(bitErrorRate != 0.0)
    { // random loss requested
      Mult_t lossProb = 1.0 - pow(1.0 - bitErrorRate, (double)(p->Size() * 8));
      if (!berRng) berRng = new Uniform();
      Random_t rv = berRng->Value();
      evRx->hasBitError = (rv < lossProb);
    }
  Interface* i = GetPeer(p);
  DEBUG0((cout << "Scheduling RX interface " << i
          << " now " << Simulator::Now()
          << " rxTime " << rxTime
          << " time " << Simulator::Now() + rxTime << endl));
  i->ScheduleRx(evRx,rxTime);
  Stats::pktsTransmitted++;
  return true;
}

void LinkReal::Handle(Event* e, Time_t)
{
  LinkEvent* le = (LinkEvent*)e;
  switch(le->event) {
  case LinkEvent::PACKET_TX :
    TransmitComplete(le->numberBytes);
    break;
  default:
    break;
  }
  delete le;
}

void LinkReal::ResetUtilization()
{ // Start a new utilization measurement interval
  bytesSent = 0;
  pktsSent = 0;
  utilizationStart = Simulator::Now();
}

Mult_t LinkReal::Utilization()
{
  Time_t now = Simulator::Now();
  if (now == utilizationStart) return 0.0; // Avoid divide by zero
  return (bytesSent * 8) / ((now - utilizationStart) * bw);
}

void LinkReal::AddNotify(NotifyHandler* n, void* v)
{ // Add a notify client
  //notifications.push_back(Notification(n,v));
  if (notifications.empty())
    notifications.push_back(Notification(n,v));
  else {
    NList_t::iterator i;
    for (i = notifications.begin(); i != notifications.end(); ++i)
      if (*((Time_t*)v) < *((Time_t*)i->userData)) break;         
    notifications.insert(i, Notification(n,v));
  }
}

void LinkReal::Jitter(const Random& j)
{ // Set the desired transmit time jitter random variable
  if (jitter) delete jitter;
  jitter = j.Copy();
}

Time_t LinkReal::NextFreeTime()
{
  Time_t now = Simulator::Now();
  if (now > nextFreeTime)
    { // Time is in the past, adjust to now
      nextFreeTime = now;
      return now;
    }
  return nextFreeTime;
}

// Protected methods

void   LinkReal::TransmitComplete(Count_t b) // Previous transmit completed
{
  DEBUG0((cout << "adding " << b << " to bytesSent " << bytesSent << endl));
  bytesSent += b;     // Count total bytes sent
  pktsSent++;         // Count total packets sent
  busy = false;       // No longer busy
  // The SendNotification() below may result in another packet being
  // transmitted, which may overwrite the notification request.  Therefore
  // we need to copy it locally before calling SendNotificatiion()
  NotifyHandler* copyNotify = notify;


  SendNotification(); // Notify the interface(s) link not busy

  if (copyNotify)
    {
      copyNotify->Notify(nil); // Notify sender that tx is complete
    }


  DEBUG0((cout << "LinkReal::TransmitComplete on link " << endl));
  if (debugVerbose) cout << "LinkReal::TXC time " << Simulator::Now() << endl;
}


void LinkReal::SendNotification()
{ // Notify the entry in the notificatin list
  if (notifications.size() == 0) return;   // No pending
  Notification n = notifications.front();  // Get next notifier
  notifications.pop_front();               // And remove it
  n.handler->Notify(n.userData);           // Call the notification handler
}

#ifdef  TEST_BACKPLANE
Packet* TestImportExport(Packet* p);
#endif

bool LinkReal::TransmitHelper(Packet* p, Interface* self_if, Node* n, 
                              Interface* i, bool bCast)
{
  bool lost = false;
  if(bitErrorRate != 0.0)
    { // random loss requested
      Mult_t lossProb = 1.0 - pow(1.0 - bitErrorRate, (double)(p->Size() * 8));
      if (!berRng) berRng = new Uniform();
      Random_t rv = berRng->Value();
      lost = (rv < lossProb);
    }
  // Note the object to notify when tx complete
  notify = p->notification;
  p->notification = nil;
  
  busy = true; // Set link busy
  Time_t txTime = ((Rate_t)p->Size() * 8) / Bandwidth(); // Time to transmit
  // Create and schedule the packet transmit complete event
  LinkEvent* evTx = new LinkEvent(LinkEvent::PACKET_TX);
  evTx->numberBytes = p->Size();
  Simulator::instance->Schedule(evTx, txTime, this);
  DEBUG(4,(cout << "Link txTime " << txTime
           << " bandwidth " << Bandwidth()
           << " delay " << Delay() 
           << " pktSize " << p->Size() 
           << " pktUid " << p->uid << endl));
  if (debugVerbose) cout << "Link scheduling link free at " 
                         << Simulator::Now() << " for future "
                         << txTime << endl;
  if (debugVerbose) cout << "pktSize " << p->Size()
                         << " bw " << Bandwidth()
                         << endl;
  nextFreeTime = Simulator::Now() + txTime; // Note time link will be free
  // Packet arrives at destination at the sum of three times
  // 1) Time to transmit the bits (size/bw)
  // 2) Propagation delay
  // 3) Jitter (random variation specified by caller)
  Time_t rxTime = txTime + Delay() + jitter->Value();
  //Simulator::instance->Schedule(evRx, rxTime, i);
  // New way below...more efficient than old way
  // Pass this directly to the interface for inclusion in fifo event queue
  // This works since p2p and ethernet links are FIFO.  Packets are received
  // at the far end of the link in the order sent.

  Time_t* ptime = new Time_t(nextFreeTime);
  //Interface* self_if = n->GetIfByLink(this); // INEFFICIENT!
  AddNotify(self_if, ptime); // GFR Comment out later

  if (!bCast)
    { // Not broadcast, just send single packet
      // Create and schedule the packet receive event
#ifdef  TEST_BACKPLANE
      // For testing the backplane, export and import the message
      Packet* p1 = TestImportExport(p);
      p = p1;
#endif
      LinkEvent* evRx = new LinkEvent(LinkEvent::PACKET_RX, p);
      evRx->hasBitError = lost; // Note possible bit error
      if (useSeqEvents)
        { // Bypass central event list, using a fifo queue at receiver
          i->ScheduleRx(evRx,rxTime);
        }
      else
        { 
          Simulator::instance->Schedule(evRx, rxTime, i); 
        }
    }
  else
    { // Send to all neighbors on this link
      NodeWeightVec_t nwv;
      AllNeighbors(nwv);
      //Node* thisNode = i->GetNode(); // For ignoring self
      for (NodeWeightVec_t::size_type j = 0; j < nwv.size(); ++j)
        {
          Node* node = nwv[j].node;
          Interface* iface = nwv[j].iface;
          if (!iface)
            {
              cout << "HuH?  Link broadcast has nil interface" << endl;
              continue;
            }
          if (node != n || RxBroadcast())
            { // Don't sent to self unless we receive our own broadcasts
              LinkEvent* evRx = new LinkEvent(LinkEvent::PACKET_RX, p->Copy());
              evRx->hasBitError = lost; // Note possible bit error
              // debug follows
              if (useSeqEvents)
                {
                  iface->ScheduleRx(evRx, rxTime);
                }
              else
                {
                  Simulator::instance->Schedule(evRx, rxTime, iface); 
                }
            }
        }
      delete p; // Since we sent only copies of the packet, we must delete
    }
  Stats::pktsTransmitted++;
  return true;
}

// Private methods
void LinkReal::ConstructorHelper(Rate_t r, Time_t d)
{
  busy = false;
  if (defaultJitter)
    jitter = defaultJitter->Copy();
  else
    jitter = new Constant(0);
  bw = r;
  delay = d;
  weight = 1.0;
  bitErrorRate = 0.0;
  bytesSent = 0;
  pktsSent = 0;
  utilizationStart= 0.0;
  nextFreeTime = 0.0;
  debugVerbose = false;
  DBG((Stats::linksCreated++));
}

#ifdef  TEST_BACKPLANE
#include "backplane.h"
#include "tcp.h"
#include "ipv4.h"
#include "llcsnap.h"
#include "l2proto802.3.h"

#include "packet.h"
char*  bpBuf = nil;
int    bpBufSize = 0;

Packet* TestImportExport(Packet* p)
{
  DEBUG0((cout << "Hello from TestImportExport" << endl));
  if (!GetDSimHandle())
    { // Not initialized
      DSHandle_t dsimHandle = InitializeDSim(0);
      // Register the TCP and IP data items
      TCP::RegisterBackplane(dsimHandle);
      IPV4::RegisterBackplane(dsimHandle);
      cout << "Done Registering Items " << endl;
    }

  DEBUG0((cout << "Inquiring Message Size " << endl));
  int mSize = InquireMessageSize(); // Number of bytes needed
  DEBUG0((cout << "Message size is " << mSize << endl));
  if (bpBufSize < mSize)
    { // Existing buffer not big enough
      delete bpBuf;
      cout << "Allocating new bpbuf size " << mSize << endl;
      bpBuf = new char[mSize];
      bpBufSize = mSize;
    }
  DEBUG0((cout << "bpBuf is " << Hex8((unsigned long)bpBuf)
          << " size " << bpBufSize << endl));
  int outSize = mSize;
  DEBUG0((cout << "Exporting message" << endl));
#ifdef VERBOSE
  cout << "Exporting message, packet pdu size " << p->PDUs.size()
       << " pkt size " << p->Size()
       << " top " << p->top
       << endl;

  for (PDUVec_t::size_type k = 0; k < p->PDUs.size(); ++k)
    {
      PDU* pdu = p->PDUs[k];
      cout << "PDU " << k << " layer " << pdu->Layer()
           << " size " << pdu->Size() << endl;
    }
#endif
  ExportMessageWrapper((char*)p, nil, 0, bpBuf, &outSize);

  // Now import it back
  int outlth1 = 0; // no baggage
  Packet* p1 = new Packet();
  ImportMessageWrapper(bpBuf, outSize, (char*)p1,  nil, &outlth1);
  // The import creates the TCP and IP headers..now recreate the packet
  TCPHeader* tcp = TCP::bpHeader;
  IPV4Header* ip = IPV4::bpHeader;
  Size_t totalLength = ip->totalLength;
  Size_t dataLength = totalLength - tcp->Size() - ip->Size();
  // Add the data
  if (dataLength) p1->PushPDU(new Data(dataLength)); // Application data
  p1->PushPDU(tcp);                  // TCP Header
  p1->PushPDU(ip);                   // IP Header
  p1->PushPDU(new LLCSNAPHeader());  // LLCSNap header
  // L2 Header (802.3)
  p1->PushPDU(new L2802_3Header(MACAddr(), MACAddr(), p1->Size()));
#ifdef VERBOSE
  cout << "Imported message, packet pdu size " << p1->PDUs.size()
       << " pkt size " << p1->Size()
       << " top " << p1->top
       << endl;
  for (PDUVec_t::size_type k = 0; k < p1->PDUs.size(); ++k)
    {
      PDU* pdu = p1->PDUs[k];
      cout << "PDU " << k << " layer " << pdu->Layer()
           << " size " << pdu->Size() << endl;
    }
#endif
  delete p;
  // We need to nil out the backplane  headers for the next import
  TCP::bpHeader = nil;
  IPV4::bpHeader = nil;
  return p1;
}
#endif

