/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Real Link base class
// George F. Riley.  Georgia Tech, Spring 2002

// Virtual Base class for all Real Link types.

#ifndef __linkreal_h__
#define __linkreal_h__

#include <deque>

#include "common-defs.h"
#include "link.h"
#include "handler.h"
#include "l2proto802.3.h"
#include "event.h"
#include "notifier.h"
#include "rng.h"
#include "node.h"

// Define the event subclass for link events
//Doc:ClassXRef
class LinkEvent : public Event {
public:
typedef enum { 
  PACKET_RX,
  PACKET_TX,  // Packet transmit completed, link free for another
  NOTIFY      // Notify clients of available buffer space
} LinkEvent_t;
public:
  LinkEvent() 
    : p(0), numberBytes(0), hasBitError(false), fifoEvent(false)
    { }
  LinkEvent(Event_t ev) 
    : Event(ev), p(0), numberBytes(0), hasBitError(false), fifoEvent(false)
    { }
  LinkEvent(Event_t ev, Packet* pkt) 
    : Event(ev), p(pkt), numberBytes(0), hasBitError(false), fifoEvent(false)
    { }
  virtual ~LinkEvent() { }
public:
  Packet* p; // Associated packet
  Count_t numberBytes; // Number of bytes (for PACKET_TX event)
  bool    hasBitError; // True if receiver should detect bit error
  bool    fifoEvent;   // True if event is in local interface's fifo queue
};

// Link class
// This is a base class; actual links (peer-to-peer, Ethernet, 802.11)
// derive from this.
class Queue;

class LinkReal : public Link {
public:
  // Constructors
  LinkReal();
  LinkReal(Rate_t b, Time_t d);
  LinkReal(const LinkReal&); // Copy constructor
  virtual ~LinkReal();
  // Make a copy of this link
  virtual Link*  Copy() const = 0;  // All subclasses must provide a copy
  
  // Managing the link
  virtual void  SetPeer(Interface* i) = 0;
  virtual Interface* GetPeer(Packet*) = 0;// Get peer interface for packet
  virtual Interface* GetPeer(Count_t) = 0;// Get peer interface for peer index
  virtual  bool Transmit(Packet*, Interface*, Node*) = 0;// Transmit a packet
  virtual  bool Transmit(Packet*, Interface*, Node*, Rate_t) = 0;
  virtual  bool TransmitWithDelay(Packet*, Time_t);
  void     Handle(Event*, Time_t);
  void     Bandwidth(Rate_t b) { bw = b;}
  Rate_t   Bandwidth() const   { return bw; }
  void     Delay(Time_t d)     { delay = d;}
  Time_t   Delay() const       { return delay;}
  bool     Busy() const        { return busy;}
  virtual  Count_t  PeerCount() = 0;         // Return number of peers
  virtual  IPAddr_t PeerIP(Count_t) = 0;     // Get peer IP Address (if known)
  virtual  IPAddr_t NodePeerIP(Node*) = 0;   // Get peer IP Address (if known)
  virtual  bool     NodeIsPeer(Node*) = 0;   // Find out if node is a peer
  virtual  Count_t  NodePeerIndex(Node*) = 0;// Get peer index for node
  virtual  IPAddr_t DefaultPeer(Interface*) = 0; // Get IP Addr of default peer
  void     BitErrorRate(Mult_t b) { bitErrorRate = b;}
  Mult_t   BitErrorRate() { return bitErrorRate;}
  void     ResetUtilization();
  Mult_t   Utilization();
  void     AddNotify(NotifyHandler*,void*);  // Add a notify client
  Weight_t Weight() { return weight;}
  void     Weight(Weight_t w) { weight = w;}
  void     Jitter(const Random&);            // Set random jitter factor
  Time_t   NextFreeTime();                   // Get time link is free
  virtual Count_t  NeighborCount(Node*) = 0; // Number of rtg neighbors
  virtual void     Neighbors(NodeWeightVec_t&, Interface*, bool) = 0;
  virtual void     AllNeighbors(NodeWeightVec_t&) = 0;
  DCount_t         BytesSent() { return bytesSent;}
  DCount_t         PktsSent() { return pktsSent;}
protected:
  virtual void     TransmitComplete(Count_t); // Prior tx finished
  virtual void     SendNotification();        // Notify link is free
  // Helper for tx
  virtual bool     TransmitHelper(Packet*, Interface*,
                                   Node*, Interface*, bool bCast = false);
public:
  virtual MACAddr  IPToMac(IPAddr_t) = 0;     // Convert peer IP to peer MAC
  virtual bool     RxBroadcast() { return false;} // True if receive own bcast
private: 
  void ConstructorHelper(Rate_t, Time_t);
protected:
  bool     busy;             // True if busy sending a packet
  Random*  jitter;           // Small noise to add to transmission time
protected:
  Rate_t   bw;               // Link bandwidth (bits/sec)
  Time_t   delay;            // Propagation delay
  Weight_t weight;           // Link weight (for some routing protocols)
  Mult_t   bitErrorRate;     // Associated bit error rate
  DCount_t bytesSent;        // Total bytes sent on this link
  DCount_t pktsSent;         // Total packets sent on this link
  Time_t   utilizationStart; // Start of utilization measurement interval
  NList_t  notifications;    // Deque of pending notifications
  NotifyHandler* notify;     // Object to notify on packet tx complete
  Time_t   nextFreeTime;     // Earliest time link is free
protected:
  bool     debugVerbose;     // True if verbose debug requested
};

#include "l2proto.h"

#endif

