/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Link base class
// George F. Riley.  Georgia Tech, Spring 2002

// Virtual Base class for all Link types.

#ifndef __link_h__
#define __link_h__

#include <deque>

#include "common-defs.h"
#include "handler.h"
#include "l2proto802.3.h"
#include "event.h"
#include "notifier.h"
#include "rng.h"
#include "node.h"

// Link class
// This is a base class; actual links (peer-to-peer, Ethernet, 802.11)
// derive from this.
class Queue;

//Doc:ClassXRef
class Link : public Handler {
  //Doc:Class Class {\em Link} is a virtual base class, describing funtionality
  //Doc:Class needed by all \GTNS\ links.  Since there are several pure
  //Doc:Class virtual  functions, no objects of class {\tt Link} can
  //Doc:Class be created.  The \GTNS\ objects {\tt Linkp2p}, {\tt Ethernet}
  //Doc:Class and {\tt WirelessLink} are subclasses and provide the
  //Doc:Class specific functionality of those types of links.
public:
  // Constructors
  //Doc:Method
  Link();
    //Doc:Desc Creates a link with the default bandwidth and delay.

  //Doc:Method
  Link(Rate_t b, Time_t d);
    //Doc:Desc Creates a link with the default bandwidth and delay.
    //Doc:Arg1 Bandwidth for the link.
    //Doc:Arg2 Propagation delay for the link

  // Destructor
  virtual ~Link() { };
  // Make a copy of this link
  //Doc:Method
  virtual Link*  Copy() const = 0;        // All subclasses must provide a copy
    //Doc:Desc Make a copy of this {\tt Link} object.
    //Doc:Return A pointer to the  copied object.

  virtual void  SetPeer(Interface* i) = 0;
  
  // Managing the link
  //Doc:Method
  virtual Interface* GetPeer(Packet*) = 0;// Get peer interface for packet
    //Doc:Desc Get the interface pointer for the remote endpoint of
    //Doc:Desc this link for the specified packet.
    //Doc:Arg1 The packet being transmitted.
    //Doc:Return A pointer to the receiving interface for the specified packet.

  //Doc:Method
  virtual Interface* GetPeer(Count_t) = 0;// Get peer interface for peer index
    //Doc:Desc Get the interface pointer for the remote endpoint of
    //Doc:Desc this link for the specified peer index
    //Doc:Arg1 Which peer desired (0 <- i < NeighborCount())
    //Doc:Return A pointer to the receiving interface for the specified peer

  //Doc:Method
  virtual  bool Transmit(Packet*, Interface*, Node*) = 0; // Transmit a packet
    //Doc:Desc Transmit a packet.
    //Doc:Arg1 The packet to transmit.
    //Doc:Arg2 A pointer to the local (transmitting) interface
    //Doc:Arg3 A pointer to the local (transmitting) node.
    //Doc:Return True if the transmit succeeded, false if packet was dropped.

  //Doc:Method
  virtual  bool Transmit(Packet*, Interface*, Node*, Rate_t) = 0;
    //Doc:Desc Transmit a packet.
    //Doc:Arg1 The packet to transmit.
    //Doc:Arg2 A pointer to the local (transmitting) interface
    //Doc:Arg3 A pointer to the local (transmitting) node.
    //Doc:Arg4 The rate at which the packet is transmitted.
    //Doc:Return True if the transmit succeeded, false if packet was dropped.

  // Transmit including queuing delay
  virtual  bool TransmitWithDelay(Packet*, Time_t) = 0;

  //Doc:Method
  void     Handle(Event*, Time_t) = 0;
    //Doc:Desc Link objects are subclass of {\tt Handler} and must define
    //Doc:Desc a {\tt Handle} method to handle events scheduled for this
    //Doc:Desc link.
    //Doc:Arg1 The event to handle.
    //Doc:Arg2 The current simulation time.

  //Doc:Method
  virtual void     Bandwidth(Rate_t b) = 0;
    //Doc:Desc Set the link bandwidth to the specified value.
    //Doc:Arg1 Link bandwidth to set.

  //Doc:Method
  virtual   Rate_t   Bandwidth() const = 0;
    //Doc:Desc Return the bandwidth for this link.
    //Doc:Return The bandwidth for the link.

  virtual   void     Delay(Time_t d) = 0;
    //Doc:Desc Set the link propagation delay to the specified value.
    //Doc:Arg1 Link delay to set.

  virtual   Time_t   Delay() const = 0;
    //Doc:Desc Return the propagation delay for this link.
    //Doc:Return The the propagation delay for the link.

  //Doc:Method
  virtual   bool     Busy() const = 0;
    //Doc:Desc Determine if the link is  currently busy transmitting a packet.
    //Doc:Return True if link busy, false if not.

  //Doc:Method
  virtual  Count_t  PeerCount() = 0; // Return number of peers
    //Doc:Desc Returns the number of peers (nodes) that this link provides
    //Doc:Desc connectivity to. \purev
    //Doc:Return Count of peers for this link.

  //Doc:Method
  virtual  IPAddr_t PeerIP(Count_t) = 0;// Get peer IP Addr(if known)
    //Doc:Desc Get the \IPA\ of the specified peer. \purev
    //Doc:Arg1 Index of the peer desired, in the range [0..PeerCount).
    //Doc:Return \IPA\ of the requested peer.
    //Doc:Return Returns \IPAN\ if peer does not exist.

  //Doc:Method
  virtual  IPAddr_t NodePeerIP(Node*) = 0;   // Get peer IP Address (if known)
    //Doc:Desc Get the \IPA\ of the specified peer. \purev
    //Doc:Arg1 Pointer to desired peer node.
    //Doc:Return \IPA\ of the requested peer.
    //Doc:Return Returns \IPAN\ if specified node is not a peer.

  //Doc:Method
  virtual  bool     NodeIsPeer(Node*) = 0;   // Find out if node is a peer
    //Doc:Desc Determine if specified node is a peer. \purev
    //Doc:Arg1 Pointer to desired peer node.
    //Doc:Return True if node is a peer, false if not.

  //Doc:Method
  virtual  Count_t  NodePeerIndex(Node*) = 0;// Get peer index for node
    //Doc:Desc Get the peer index, in the range [0..PeerCount) for the
    //Doc:Desc specified node.
    //Doc:Arg1 Pointer to desired peer node.
    //Doc:Return Peer index for specified node.

  //Doc:Method
  virtual  IPAddr_t DefaultPeer(Interface*) = 0; // Get IP Addr of default peer
    //Doc:Desc Find the \IPA\ of the default peer.  Link classes with
    //Doc:Desc multiple peers (such as an Ethernet link) can specify
    //Doc:Desc one of the peers as the {\em Default Gateway}.  This
    //Doc:Desc simplifies routing decisions in some cases. \purev
    //Doc:Arg1 Interface pointer for the interface of this link.
    //Doc:Return \IPA\ of the default peer.
    //Doc:Return Returns \IPAN\ if no default peer exists.

  //Doc:Method
  virtual void     BitErrorRate(Mult_t b) = 0;
    //Doc:Desc Specify a bit error rate for this link.
    //Doc:Arg1 The desired bit error rate.

  //Doc:Method
  virtual   Mult_t   BitErrorRate() = 0;
    //Doc:Desc Return the bit error rate for this link.
    //Doc:Return The bit error rate assigned to this link.

  //Doc:Method
  virtual   void     ResetUtilization() = 0;
    //Doc:Desc Reset the link utilization statistics for this link to zero.

  //Doc:Method
  virtual   Mult_t   Utilization() = 0;
    //Doc:Desc Return the utilization for this link, in the range [0..1].
    //Doc:Return The link utilization for this link.

  //Doc:Method
  virtual   void     AddNotify(NotifyHandler*,void*) = 0;// Add a notify client
    //Doc:Desc Higher layer protocol entities can request a notification
    //Doc:Desc when links go non--busy.  This method adds a notification
    //Doc:Desc handler to the list.
    //Doc:Arg1 Any object subclassed from {\tt NotifyHandler}.  The
    //Doc:Arg1 {\tt Notify} method for that handler will be called
    //Doc:Arg1 when this link goes non--busy.
    //Doc:Arg2 A blind pointer to be passed to the {\tt Notify} method.

  //Doc:Method
  virtual   Weight_t Weight() = 0;
    //Doc:Desc Query the link weight (used by some routing protocols)
    //Doc:Desc for this link.
    //Doc:Return The weight for this link.

  //Doc:Method
  virtual   void   Weight(Weight_t w) = 0;
    //Doc:Desc Assign a weight (used by some routing protocols) to this link.
    //Doc:Arg1 Weight to assign to this link.

  //Doc:Method
  virtual   void   Jitter(const Random&) = 0;  // Set random jitter factor
    //Doc:Desc Specify a random number generator object used for {\em jitter}.
    //Doc:Desc Jitter is used  to randomly delay the packet transmission time
    //Doc:Desc by small pertubations, to model CPU processing time in
    //Doc:Desc routers.  No jitter value is used unless this method is used.
    //Doc:Arg1 Any random number generator object.  The object is copied.

  //Doc:Method
  virtual   Time_t NextFreeTime() = 0;         // Get time link is free
    //Doc:Desc Predict the time this link will be free.  This is useful
    //Doc:Desc only for links without jitter, and without contention.  
    //Doc:Return The predicted time this link will be free.

  //Doc:Method
  virtual Count_t  NeighborCount(Node*) = 0;   // Number of rtg neighbors
    //Doc:Desc Return the number of routing neighbors on this link.
    //Doc:Arg1 Node querying neighbors (significant for Ethernet links)
    //Doc:Return Count of routing neighbors.

  // Build the neighbor and interface list        
  //Doc:Method
  virtual void     Neighbors(NodeWeightVec_t&, Interface*, bool) = 0;
    //Doc:Desc Get a vector of all routing neighbors.
    //Doc:Arg1 Returned vector of routing neighbors.
    //Doc:Arg2 Interface pointer to associated interface.
    //Doc:Arg3 True if include all neighbors, even for single  gateway
    //Doc:Arg3 Ethernet.

  // AllNeighbors is used by broadcasts to find all peers on a link
  // even if they are not routing peers
  //Doc:Method
  virtual void     AllNeighbors(NodeWeightVec_t&) = 0;
    //Doc:Desc Find all neighbors for this link, even if they are
    //Doc:Desc not routing neighbors.  Used by broadcasts to find
    //Doc:Desc all receivers.
    //Doc:Arg1 Returned vector of neighbors.

  // Statistics
  //Doc:Method
  virtual   DCount_t         BytesSent() = 0;
    //Doc:Desc Get a count of the total number of bytes sent on this link.
    //Doc:Return Count of total bytes sent on this link.

  //Doc:Method
  virtual   DCount_t         PktsSent() = 0;
    //Doc:Desc Get a count of the total number of packets sent on this link.
    //Doc:Return Count of total packets sent on this link.

  //Doc:Method
  virtual MACAddr  IPToMac(IPAddr_t) = 0;     // Convert peer IP to peer MAC
    //Doc:Desc Convert a peer \IPA\ to peer MAC address.
    //Doc:Arg1 \IPA\ of desired peer.
    //Doc:Return MAC address for that peer.

  //Doc:Method
  virtual bool     RxBroadcast() = 0;         // True if receive own bcast
    //Doc:Desc Query if this link should receive a copy of its own
    //Doc:Desc broadcasts.
    //Doc:Return True if receive own broadcast, false if not.

public:
  //Doc:Method
  static void     DefaultRate(Rate_t r);
    //Doc:Desc Set a default transmission rate to be used for any
    //Doc:Desc new link created.
    //Doc:Arg1 Desired default transmission rate (bits/second)

  //Doc:Method
  static Rate_t   DefaultRate()          { return defaultRate;}
    //Doc:Desc Return the default transmission rate.
    //Doc:Return The default transmission rate (bits/second).

  //Doc:Method
  static void     DefaultDelay(Time_t t);
    //Doc:Desc Set a default propagation delay to be used for any
    //Doc:Desc new link created.
    //Doc:Arg1 Desired default propagation delay (second)

  //Doc:Method
  static Time_t   DefaultDelay()         { return defaultDelay;}
    //Doc:Desc Return the default propagation delay
    //Doc:Return The default propagation delay (second).

  //Doc:Method
  static void     DefaultJitter(const Random& j);
    //Doc:Desc Set a default random variable for link transmission jitter.
    //Doc:Arg1 Any random number generator object.

  //Doc:Method
  static Random*  DefaultJitter();
    //Doc:Desc Get a pointer to the default jitter random number generator.
    //Doc:Return Pointer to the default jitter random number generator.

  //Doc:Method
  static void     Default(const Link& l);
    //Doc:Desc Set a default link object to use when a link object is needed
    //Doc:Desc and none has been specified.
    //Doc:Arg1 Link object to use for the default link.

  //Doc:Method
  static const Link& Default();
    //Doc:Desc Get a reference to the default link object.
    //Doc:Return A constant reference to the default link object.

  //Doc:Method
  static void     SeedBER(Seed_t s); // Seed the bit error rate RNG
    //Doc:Desc Seed the bit error rate random number generator
    //Doc:Arg1 A random seed value for te bit error rate
    //Doc:Arg1 random number generator.

  //Doc:Method
  static void     UseSeqEvents(bool u) { useSeqEvents = u;}
    //Doc:Desc A small performance improvement can be gained by leveraging
    //Doc:Desc the fact that link receive events are always in strictly
    //Doc:Desc increasing timestamp order for a single link.  Setting
    //Doc:Desc UseSeqEvents to true enables this optimization.
    //Doc:Desc NOTE.  The GTNetS packet animation code depends on this setting
    //Doc:Desc to  find which packets are in--flight on the links.  If this
    //Doc:Desc setting is FALSE, no packets will be displayed on the
    //Doc:Desc animation display.
    //Doc:Arg1 True if using sequential events optimization, false if not.

protected:
  // Defaults
  static Rate_t   defaultRate;      // Default link speed
  static Time_t   defaultDelay;     // Default delay
  static Link*    defaultLink;      // Default link type
  static Random*  defaultJitter;    // Default jitter rng
  static Random*  berRng;           // Random numbers for bit-error-rate
  static bool     useSeqEvents;     // True if bypass central ev. list
};

#include "l2proto.h"

#endif

