/* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2005, The Boeing Company
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - LinkMatrix connectivity

#include <iostream>

#include "debug.h"
#include "linkmatrix.h"
#ifdef SIM_ETRACE_STAT
#include "sim.h"
#endif //SIM_ETRACE_STAT

using namespace std;

// Static members
Random* LinkMatrix::errRng;

//LinkMatrixEvent constructor method
LinkMatrixEvent::LinkMatrixEvent(Count_t s, Count_t d, Mult_t r)
 : src_index(s), dst_index(d), lossrate(r) 
{
}

//LinkMatrixEvent destructor method
LinkMatrixEvent::~LinkMatrixEvent()
{
}


LinkMatrix::LinkMatrix(Count_t n,     // Node count
               IPAddr_t i,    // IP addr start
               Mask_t m,      // mask
               SystemId_t id, // Responsible system
               Rate_t r, 
               Time_t dly, 
               Detail_t d) : Ethernet(n, i, m, id, r, dly, d)
{
  int j, k;
  //Matrix initialization
  for (j = 0; j < Size(); j++)
    for (k = 0; k < Size(); k++)
      matrixVec.push_back(0);
  if (!errRng)
    errRng = new Uniform();  //XXX could change this to any RV
}

void LinkMatrix::Neighbors(NodeWeightVec_t&, Interface*, bool)
{ // Add to neighbor list
  // For wireless LAN's, return an empty list.
}

bool LinkMatrix::TransmitHelper(Packet* p, Interface* self_if, Node* n,
                                Interface* i, bool bCast)
{
  Mult_t errorRate;
  // Note the object to notify when tx complete
  notify = p->notification;
  p->notification = nil;

  busy = true; // Set link busy
  Time_t txTime = ((Rate_t)p->Size() * 8) / Bandwidth(); // Time to transmit
  // Create and schedule the packet transmit complete event
  LinkEvent* evTx = new LinkEvent(LinkEvent::PACKET_TX);
  evTx->numberBytes = p->Size();
  Simulator::instance->Schedule(evTx, txTime, this);
  DEBUG(4,(cout << "Link txTime " << txTime
           << " bandwidth " << Bandwidth()
           << " delay " << Delay()
           << " pktSize " << p->Size()
           << " pktUid " << p->uid << endl));
  if (debugVerbose) cout << "Link scheduling link free at "
                         << Simulator::Now() << " for future "
                         << txTime << endl;
  if (debugVerbose) cout << "pktSize " << p->Size()
                         << " bw " << Bandwidth()
                         << endl;
  nextFreeTime = Simulator::Now() + txTime; // Note time link will be free
  // Packet arrives at destination at the sum of three times
  // 1) Time to transmit the bits (size/bw)
  // 2) Propagation delay
  // 3) Jitter (random variation specified by caller)
  Time_t rxTime = txTime + Delay() + jitter->Value();
  //Simulator::instance->Schedule(evRx, rxTime, i);
  // New way below...more efficient than old way
  // Pass this directly to the interface for inclusion in fifo event queue
  // This works since p2p and ethernet links are FIFO.  Packets are received
  // at the far end of the link in the order sent.

  Time_t* ptime = new Time_t(nextFreeTime);
  //Interface* self_if = n->GetIfByLink(this); // INEFFICIENT!
  AddNotify(self_if, ptime); // GFR Comment out later

  if (!bCast)
  { // Not broadcast, just send single packet
    // Create and schedule the packet receive event
#ifdef  TEST_BACKPLANE
    // For testing the backplane, export and import the message
    Packet* p1 = TestImportExport(p);
    p = p1;
#endif
    LinkEvent* evRx = new LinkEvent(LinkEvent::PACKET_RX, p);

    // random loss
    errorRate = MatrixNeighborLossRate(self_if,i);
    Random_t rv = errRng->Value();
    evRx->hasBitError = (rv < errorRate);

    if (useSeqEvents)
      i->ScheduleRx(evRx,rxTime);
    else
      Simulator::instance->Schedule(evRx, rxTime, i);
  }
  else
  { // Send to all neighbors on this link
    NodeWeightVec_t nwv;
    AllNeighbors(nwv);
    //Node* thisNode = i->GetNode(); // For ignoring self
    for (NodeWeightVec_t::size_type j = 0; j < nwv.size(); ++j)
    {
      Node* node = nwv[j].node;
      Interface* iface = nwv[j].iface;
      if (!iface)
      {
        cout << "HuH?  Link broadcast has nil interface" << endl;
        continue;
      }
      if (node != n || RxBroadcast())
      { // Don't sent to self unless we receive our own broadcasts
        LinkEvent* evRx = new LinkEvent(LinkEvent::PACKET_RX, p->Copy());

        // random loss
        errorRate = MatrixNeighborLossRate(self_if,iface);
        Random_t rv = errRng->Value();
        evRx->hasBitError = (rv < errorRate);

        // debug follows
        if (useSeqEvents)
          iface->ScheduleRx(evRx, rxTime);
        else
          Simulator::instance->Schedule(evRx, rxTime, iface);
      }
    }
    delete p; // Since we sent only copies of the packet, we must delete
  }
  Stats::pktsTransmitted++;
  return true;
}

Mult_t LinkMatrix::MatrixNeighborLossRate(Interface *src_if, Interface *dst_if)
{
  Mult_t lossrate;
  Count_t src_index = NodePeerIndex(src_if->GetNode());
  Count_t dst_index = NodePeerIndex(dst_if->GetNode());

  lossrate = matrixVec.at(src_index * Size() + dst_index);

  return lossrate;
}

void LinkMatrix::SetNeighborLossRate(Count_t src_index, 
                                        Count_t dst_index, 
                                        Mult_t lossrate)
{
  if (src_index < 0 || src_index >= Size() ||
      dst_index < 0 || dst_index >= Size() || 
      lossrate < 0 || lossrate > 1)
  {
    printf("Improper input parameters in SetNeighborLossRate\n");
    printf("src_index=%d, dst_index=%d, lossrate=%f\n", 
           src_index, dst_index, lossrate);
    exit(0);
  }
    
  matrixVec[src_index * Size() + dst_index] = lossrate;
#ifdef SIM_ETRACE_STAT
  TraceEvent_sim(2,"node %d to %d lossrate=%f", src_index, dst_index, lossrate);
#endif //SIM_ETRACE_STAT
}

void LinkMatrix::ScheduleMatrixChange(Time_t t, Count_t src_index, 
                                      Count_t dst_index, Mult_t lossrate) 
{
  if (src_index < 0 || src_index >= Size() ||
      dst_index < 0 || dst_index >= Size() || 
      lossrate < 0 || lossrate > 1)
  {
    printf("Improper input parameters in ScheduleMatrixChange\n");
    printf("src_index=%d, dst_index=%d, lossrate=%f\n",
           src_index, dst_index, lossrate);
    exit(0);
  }

  LinkMatrixEvent *ev = new LinkMatrixEvent(src_index, dst_index, lossrate);
  timer.Schedule(ev, t, (TimerHandler *) this);
}

void LinkMatrix::Timeout(TimerEvent *pTimerEvent)
{
  LinkMatrixEvent *ev = (LinkMatrixEvent *) pTimerEvent;
  SetNeighborLossRate(ev->src_index, ev->dst_index, ev->lossrate);
  delete ev;
} 
