/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Point to Point Link class
// George F. Riley.  Georgia Tech, Spring 2002

#include "debug.h"
#include "linkp2p.h"
#include "simulator.h"
#include "macaddr.h"

using namespace std;

Linkp2p* Linkp2p::defaultp2p; // Default p2p link pointer
Count_t  Linkp2p::count = 0;

Linkp2p::Linkp2p()
  : LinkReal(), pPeer(0)
{
  count++;
}

Linkp2p::Linkp2p(Rate_t b, Time_t d)
    : LinkReal(b, d), pPeer(0)
{ 
  count++;
}

Linkp2p::Linkp2p(const Linkp2p& c) : LinkReal(c)
{
  pPeer = nil; // Don't copy the peer pointer
  count++;
}

Linkp2p::~Linkp2p()
{
  count--;
}

Link* Linkp2p::Copy() const
{
  return new Linkp2p(*this);
}

bool Linkp2p::Transmit(Packet* p, Interface* i, Node* n)
{
  DEBUG0((cout << "Linkp2p::Transmit to peer " << GetPeer(p) << endl));
  return TransmitHelper(p, i, n, GetPeer(p));
}

bool Linkp2p::Transmit(Packet* p, Interface* i, Node* n, Rate_t)
{ // Transmit with rate is used on wireless links only
  return Transmit(p, i, n);
}

MACAddr Linkp2p::IPToMac(IPAddr_t)
{ // Convert peer IP to peer MAC
  // For point to point, this is just MAC address of peer
  if (pPeer) return pPeer->GetMACAddr();
  return MACAddr::NONE; // Else none
}

Count_t Linkp2p::NeighborCount(Node* n)
{ // Get number of routing neighbors (always 1 for p2p links)
  return 1;
}

void Linkp2p::Neighbors(NodeWeightVec_t& nwv, Interface* i, bool)
{ // Add to neighbor list
  // Point to point has only one peer
  // last param (forceAll) meaningless on p2p link.
  nwv.push_back(NodeIfWeight(pPeer->GetNode(), i, Weight()));
}

void Linkp2p::AllNeighbors(NodeWeightVec_t& nwv)
{ // Add to neighbor list
  // Point to point has only one peer.  Also note interface is
  // peer, not current as in Neighbors above.
  nwv.push_back(NodeIfWeight(pPeer->GetNode(), pPeer, Weight()));
}

Count_t  Linkp2p::PeerCount()       // Return number of peers
{
  return 1; // Point to point links have exactly 1 peer
}

IPAddr_t Linkp2p::PeerIP(Count_t npeer)  // Get peer IP Address (if known)
{
  if (npeer > 0)
    return IPADDR_NONE;      // Oops, value out of range
  return pPeer->GetIPAddr(); // Return one and only peer's ip address  
}

IPAddr_t Linkp2p::NodePeerIP(Node* n)  // Get peer IP Address (if known)
{ // Find the peer's ip address for peer n on this link
  // For point to point, there is only one peer, so it must match
  Node* peerNode = pPeer->GetNode();
  if (n != peerNode)
    return IPADDR_NONE;      // Oops, not correct peer
  return pPeer->GetIPAddr(); // Return one and only peer's ip address  
}

bool     Linkp2p::NodeIsPeer(Node* n)
{ // Return true if specified node is a peer
  Node* peerNode = pPeer->GetNode();
  return (peerNode == n);
}

Count_t  Linkp2p::NodePeerIndex(Node* n)
{
  return 0; // Point to point links always have only one peer
}

IPAddr_t Linkp2p::DefaultPeer(Interface*)
{ // Get IP Addr of default peer
  return PeerIP(0); // For p2p links, default is same as only
}

// Static methods
Linkp2p& Linkp2p::Default()
{ // Get the default link
  if (!defaultp2p) defaultp2p = new Linkp2p(); // Allocate it
  return *defaultp2p;                           // Return it
}

Linkp2p& Linkp2p::Default(const Linkp2p& l)    // Set new default
{
  if (defaultp2p) delete defaultp2p;           // Delete if already exists
  defaultp2p = (Linkp2p*)l.Copy();             // Set new one
  return *defaultp2p;
}

#ifdef UNUSED
Linkp2p* Linkp2p::Default(Linkp2p* l)          //.Set new default
{
  if (defaultp2p) delete defaultp2p;           // Delete if already exists
  defaultp2p = (Linkp2p*)l->Copy();            // Set new one
  return defaultp2p;
}
#endif
