/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator -  MAC Address Class
// George F. Riley.  Georgia Tech, Spring 2002

#include <stdio.h>

#include "macaddr.h"

using namespace std;

// Static members
MACAddr MACAddr::next(0,1); // Next available MAC
MACAddr MACAddr::NONE(0,0);
char    MACAddr::work[20];

// Constructors
MACAddr::MACAddr()
#ifndef COMPACT
  : upper(0), lower(0)
#else
  : upper(0), lower(0)
#endif
{
}

MACAddr::MACAddr(MACAddr_t u, MACAddr_t l)
#ifndef COMPACT
  : upper(u), lower(l)
#else
  : upper(0), lower(l)
#endif
{
}

MACAddr::MACAddr(const MACAddr& c) 
  : upper(c.upper)
#ifndef COMPACT
  , lower(c.lower)
#endif
{
}

// Typecast to string
#ifdef OLD_HEX
MACAddr::operator std::string() const
{
  string r("0x");
  if (Upper()) 
    {
      sprintf(work, "%lx", Upper());
      r += string(work);
    }
  sprintf(work, "%lx", Lower());
  return r + string(work);
}
#endif

// Changed to decimal by GFR Mar 8,2004
MACAddr::operator std::string() const
{
  string r;
  if (Upper())
    {
      sprintf(work, "%ld", Upper());
      r += string(work);
    }
  sprintf(work, "%ld", Lower());
  return r + string(work);
}


// Typecast to MACAddr_t
MACAddr::operator MACAddr::MACAddr_t() const
{
  return Lower(); // Just lower bits
}

// Assignment
MACAddr& MACAddr::operator=(const MACAddr& s)
{ 
#ifndef COMPACT
  upper = s.upper; 
#endif
  lower = s.lower;
  return *this;
}

// Equivalence test
bool   MACAddr::operator==(const MACAddr& r) const
{
#ifndef COMPACT
  return upper == r.upper && lower == r.lower;
#else
  return lower == r.lower;
#endif
 ;
}

// De fine some arithmetic operators
MACAddr MACAddr::operator++()
{ lower++; 
#ifndef COMPACT 
 if (lower == 0) upper++;
#endif
 return *this;
}

MACAddr MACAddr::operator++(int)
{ 
  MACAddr ma(Upper(),Lower());
  ++(*this); 
  return ma;
}

// Member access
MACAddr::MACAddr_t MACAddr::Upper() const
{
#ifndef COMPACT
  return upper;
#else
  return 0;
#endif
}

MACAddr::MACAddr_t MACAddr::Lower() const
{
  return lower;
}

void MACAddr::Upper(MACAddr_t u)
{
#ifndef COMPACT
  upper = u;
#endif
}

void MACAddr::Lower(MACAddr_t l)
{
  lower = l;
}

void MACAddr::IncUpper()
{
#ifndef COMPACT
  upper++;
#endif
}

bool MACAddr::IsBroadcast()
{
  if (lower != 0xffffffff) return false;
#ifndef COMPACT
  return (upper == 0xffff);
#else
  return true;
#endif
}

void MACAddr::SetBroadcast()
{
  lower = 0xffffffff;
#ifndef COMPACT
  upper = 0xffff;
#endif
}

// Non-Member Comparison operators

bool operator< (const MACAddr& l, const MACAddr& r)
{
  if (l.Upper() < r.Upper()) return true;
  if (l.Upper() > r.Upper()) return false;
  return (l.Lower() < r.Lower());
}

bool operator<=(const MACAddr& l, const MACAddr& r)
{
  if (l.Upper() < r.Upper()) return true;
  if (l.Upper() > r.Upper()) return false;
  return (l.Lower() <- r.Lower());
}

bool operator> (const MACAddr& l, const MACAddr& r)
{
  if (l.Upper() > r.Upper()) return true;
  if (l.Upper() < r.Upper()) return false;
  return (l.Lower() > r.Lower());
}

bool operator>=(const MACAddr& l, const MACAddr& r)
{
  if (l.Upper() > r.Upper()) return true;
  if (l.Upper() < r.Upper()) return false;
  return (l.Lower() >= r.Lower());
}

MACAddr operator-(const MACAddr& l, const MACAddr& r)
{ // Assumes never more than 32 bits difference
  if (l.Lower() >= r.Lower())
    return MACAddr(0,l.Lower() - r.Lower());
  return MACAddr(0,r.Lower() - l.Lower());
}

MACAddr operator+(const MACAddr& l, const Count_t r)
{ // Assumes never more than 32 bits r
  MACAddr m(l.Upper(), l.Lower() + r);

#ifndef COMPACT
  if (m.Lower() < l.Lower())
    { // Wrapped, bump upper
      m.IncUpper();
    }
#endif
  return m;
}


// Static functions
MACAddr MACAddr::Allocate()
{ // Allocate a new MAC address
  return next++;
}
