/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Random waypoint mobility models
// George F. Riley.  Georgia Tech, Spring 2002

// Define the random waypoint mobility model

#include <math.h>

#include "common-defs.h"
#include "simulator.h"
#include "rng.h"
#include "node.h"
#include "wireless-grid.h"
#include "mobility-random-waypoint.h"

using namespace std;

#define REPEATABLE_MOBILITY_START_TIME 1 // Seconds
//#define DEBUG_REPEATABLE_MOBILITY 

// Constructor 
// Note:  Typically, copy constructor is the relevant one to use
// (see node-real.cc)
RandomWaypoint::RandomWaypoint(
                               WirelessGrid& g,  // Associated grid
                               const Random& p,  // Pause time
                               const Random& v)  // Velocity
: grid(g), pauseRng(p.Copy()), velocityRng(v.Copy())
{
  lastTime = Simulator::Now();
  nextTime = lastTime + pauseRng->Value();
#ifdef REPEATABLE_MOBILITY
  paused = true;
  timeoutevent = nil;
#endif
}

// Copy constructor
RandomWaypoint::RandomWaypoint(const RandomWaypoint& c)
  : Mobility(c), nextTime(c.nextTime), grid(c.grid),
    pauseRng(c.pauseRng->Copy()), velocityRng(c.velocityRng->Copy())
{
  lastTime = Simulator::Now();
  nextTime = lastTime + pauseRng->Value();
#ifdef REPEATABLE_MOBILITY
  paused = true;
  timeoutevent = nil;
#endif
}

#ifdef REPEATABLE_MOBILITY
// The problem with the original code is that the act of looking at
// the location can cause it to unpause.  If radio range is different
// then Location will be looked at with different frequency
void RandomWaypoint::UpdateLocation(Node* n)
{ // Update node's location to current time
  Time_t now = Simulator::Now();
  if (!timeoutevent) {
    Time_t nexttimeout_ = pauseRng->Value();
    ScheduleNextTimeout(nexttimeout_);
  }
  if (paused)
    return;
  Location c = node_->GetLocation();         // Get current location
#ifdef DEBUG_REPEATABLE_MOBILITY
  cout << "DEBUG_REPEATABLE_MOBILITY Node " << n->Id() << " time " << now << " at " << c.X() << "," << c.Y() << " moving, last time " << lastTime << endl;
#endif
  Move(n, now - lastTime);
  lastTime = now;
  return;
}

void RandomWaypoint::ScheduleNextTimeout(Time_t t_)
{
  if (!timeoutevent) timeoutevent = new TimerEvent();
  timer.Schedule(timeoutevent, t_, this);
#ifdef DEBUG_REPEATABLE_MOBILITY
  cout << "DEBUG_REPEATABLE_MOBILITY Time " << Simulator::Now() << " Node " << node_->Id() << " timing out at " << (Simulator::Now() + t_) << endl;
#endif
}

void
RandomWaypoint::Timeout(TimerEvent *pTimerEvent)
{
  Time_t now = Simulator::Now();
  if (paused) {
    // Do all the things like getting new location etc.
          Location l = grid.GetRandomLocation(); // Get new location
          UpdateLocation(node_);
          Location c = node_->GetLocation();         // Get current location
          Meters_t v = velocityRng->Value();     // Velocity
          ComputeVelocities(l, c, v);            // X/Y velocities
          Meters_t d = Distance(l, c);           // Distance from old to new
          Time_t   t = d / v; // Time to travel the distance
          lastTime = now;
          nextTime = now + t;
#ifdef DEBUG_REPEATABLE_MOBILITY
          cout << "DEBUG_REPEATABLE_MOBILITY Node " << node_->Id() << " time " << now << " at " << c.X() << "," << c.Y() << " moving distance " << d << " to " << l.X() << "," << l.Y() << " at velocity " << v << " until time " << nextTime << endl;
#endif
    ScheduleNextTimeout(t);
    paused = false;
  } else {
    // Move the rest of the way, then pause
   // Move(node_, now - lastTime);
    UpdateLocation(node_);
    lastTime = now;
    Time_t pausevalue = pauseRng->Value();
    nextTime = now + pausevalue;
#ifdef DEBUG_REPEATABLE_MOBILITY
       Location c = node_->GetLocation();         // Get current location
       cout << "DEBUG_REPEATABLE_MOBILITY Node " << node_->Id() << " time " << now << " at " << c.X() << "," << c.Y() << " pausing until time " << nextTime << endl;
#endif
    ScheduleNextTimeout(pausevalue);
    paused = true;
  }
  return;
}

#else // ifndef REPEATABLE_MOBILITY -- original code

void RandomWaypoint::UpdateLocation(Node* n)
{ // Update node's location to current time
  Time_t now = Simulator::Now();
  while(1)
    {
      while(paused)
        { // Check if pause time passed
          if (now < nextTime) return;   // Not time to start, nothing to do
          Time_t extra = now - nextTime;// Time to move after un-pause
          Location l = grid.GetRandomLocation(); // Get new location
          Location c = n->GetLocation();         // Get current location
          Meters_t v = velocityRng->Value();     // Velocity
          ComputeVelocities(l, c, v);            // X/Y velocities
          Meters_t d = Distance(l, c);           // Distance from old to new
          Time_t   t = d / v; // Time to travel the distance
          DEBUG0((cout << "Unpausing, time " << now
               << " extra " << extra
               << " new loc " << l.X() << ", " << l.Y()
               << " vel " << v 
               << " velx " << velocityX 
               << " vely " << velocityY
               << " distance " << d
                  << " time " << t << endl));
          lastTime = now;
          if (extra < t)
            { // Move for the extra time
              Move(n, extra);
              paused = false;
              nextTime = now + t - extra;
              return;
            }
          else
            { // Just move to new location and set new nextTime, remain paused
              n->SetLocation(l);
              nextTime = extra - t + pauseRng->Value();
            }
        }
      // Not paused
      if (now < nextTime)
        { // Easy case, just move since last update
          Move(n, now - lastTime);
          lastTime = now;
          return;
        }
      Time_t t = nextTime - lastTime;  // Time to move before pause
      Move(n, t);
      lastTime = nextTime;
      paused = true;
      nextTime = lastTime + pauseRng->Value();
    }
}
#endif

Mobility* RandomWaypoint::Copy() const
{ // Make a copy of this object
  return new RandomWaypoint(*this);
}

void RandomWaypoint::Initialize()
{ // None necessary
#ifdef REPEATABLE_MOBILITY
  ScheduleNextTimeout(MOBILITY_START_TIME);
#endif
}

Angle_t RandomWaypoint::Direction() const
{
  return 0; // code later
}


