/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/


#include <stdio.h>
#include <math.h>
#include "nixvector.h"
#include "simulator.h"

using namespace std;

// Default setting
Word_t CNCList::defaultNCListSize = NC_LIST_SIZE;
Word_t CReplyList::defaultReplyListSize = REPLY_LIST_SIZE;
Word_t CNbrTable::defaultMaxNeighbor = MAX_NEIGHBOR;


// bitcopy global function
int bitcpy(Long_t* dst, int dp, Long_t* src, int sp, int len)
{
  if (!len) return 0;

  int ofs_d, ofs_s, p_d, p_s;
  ofs_d = dp / 32;  p_d = dp % 32;
  ofs_s = sp / 32;  p_s = sp % 32;

  int size = bit2dw(len)+1;
  DWn mask(size), s(size);
  for(int i=0; i<size-1; i++) mask.dw[i] = (Long_t)-1;
  mask.dw[size-2] >>= (32-len%32);

  memcpy(s.dw, src+ofs_s, bit2dw(p_s+len)*sizeof(Long_t));
  s >>= p_s;
  s &= mask;
  mask <<= p_d;
  s <<= p_d;

  for (int i=0; i<bit2dw(p_d+len); i++) {
	dst[i+ofs_d] &= ~mask.dw[i];
	dst[i+ofs_d] |= s.dw[i];
  }
  return len;
}


// DWn

DWn::DWn(int sz) : size(sz)
{
  dw = new Long_t[sz];
  memset(dw, 0, sz*sizeof(Long_t));
}

DWn& DWn::operator = (DWn& r)
{
  size = r.size;
  if (dw) delete[] dw;
  dw = new Long_t[size];
  memcpy(dw, r.dw, size*sizeof(Long_t));
  return *this;
}

DWn& DWn::operator <<= (int n)
{
  if (!n) return *this;
  if (n >= 32) {
	cout << "shift bits error n = " << n << endl;
	return *this;
  }

  Long_t carry = 0, temp;
  for (int i=0; i<size; i++)
  {
	temp = dw[i];
	dw[i] = carry | (temp << n);
	carry = temp >> (32-n);
  }
  return *this;
}

DWn& DWn::operator >>= (int n)
{
  if (!n) return *this;
  if (n >= 32) {
	cout << "shift bits error n = " << n << endl;
	return *this;
  }

  Long_t carry = 0, temp;
  for (int i=size-1; i>=0; i--)
  {
	temp = dw[i];
	dw[i] = carry | (dw[i] >> n);
	carry = temp << (32-n);
  }
  return *this;
}

DWn& DWn::operator |= (DWn& r)
{
  // Assume the same size
  for (int i=0; i<size; i++) dw[i] |= r.dw[i];
  return *this;
}

DWn& DWn::operator &= (DWn& r)
{
  // Assume the same size
  for (int i=0; i<size; i++) dw[i] &= r.dw[i];
  return *this;
}


// CNCList

CNCList::CNCList() : nItem(0)
{
  Item.reserve(defaultNCListSize);
}

int CNCList::Add(IPAddr_t src, Byte_t id, MACAddr& inf)
{
  // Ring buffer, nItem indicates next position
  if (Item.size() >= defaultNCListSize) {
	Item[nItem++] = CNCItem(id, src, 1, inf);
	nItem %= defaultNCListSize;
  }
  else
	Item.push_back(CNCItem(id, src, 1, inf));
  return nItem;
}

/*
int CNCList::Add(IPAddr_t src, Byte id)
{
	if (nItem >= defaultNCListSize) return -1;

	NCVec_t::size_type i;
	for (i=0; i<Item.size(); ++i)
	  if (!Item[i].Valid()) break;
	if (i < Item.size()) Item[i] = CNCItem(id,src,1);
	else Item.push_back(CNCItem(id,src,1));
	nItem++;
	return i;
}
*/

int CNCList::Remove(IPAddr_t addr, Byte_t nc_id)
{
  int i = Find(addr, nc_id);
  if (i >= 0) {
	Item[i] = CNCItem();
	nItem--;
  }
  return i;
}

int CNCList::Find(IPAddr_t src, Byte_t id)
{
  for (NCVec_t::size_type i=0; i<Item.size(); ++i)
	if (Item[i].Valid() && Item[i].source==src && Item[i].nc_id==id)
	  return i;
  return -1;
}


// CReplyList

CReplyList::CReplyList() : nItem(0)
{
  Item.reserve(defaultReplyListSize);
}

int CReplyList::Add(IPAddr_t s, Byte id, Byte t, Byte mid, CMetricVec& mv)
{
  // Ring buffer, nItem indicates next position
  if (Item.size() >= defaultReplyListSize) {
	Item[nItem++] = CRepItem(t,id,mid,mv,s,1);
	nItem %= defaultReplyListSize;
  }
  else
	Item.push_back(CRepItem(t,id,mid,mv,s,1));
  return nItem;
}

/*
int CReplyList::Add(IPAddr_t s, Byte id, Byte t, Byte mid, CMetricVec& mv)
{
	if (nItem >= defaultReplyListSize) return -1;

	RepVec_t::size_type i;
	for (i=0; i<Item.size(); i++) if (!Item[i].Valid()) break;
	if (i < Item.size()) Item[i] = CRepItem(t,id,mid,mv,s,1);
	else Item.push_back(CRepItem(t,id,mid,mv,s,1));
	nItem++;
	return i;
}
*/

int CReplyList::Remove(IPAddr_t src, Byte id, Byte type)
{
  CRepItem* i = Find(src, id, type);
  if (i) {
	*i = CRepItem();
	nItem--;
	return 0;
  }
  return -1;
}

CRepItem* CReplyList::Find(IPAddr_t src, Byte id, Byte t)
{
  for (RepVec_t::iterator i = Item.begin(); i != Item.end(); ++i) {
	if (!i->Valid()) continue;
	if (i->type == t && i->nc_id == id && i->source == src)
	  return &(*i);
  }
  return nil;
}

void CReplyList::Update(int i, CMetricVec& mv)
{
  Item[i].mv = mv;
}


// CNixVector

CNixVector::CNixVector(const CNixVector& r) :
  nvl(r.nvl), nvu(r.nvu)
{
  if (r.nv) {
	int dwAlloc = bit2dw(r.nvl);
	nv = new Long_t[dwAlloc];
	memcpy(nv, r.nv, dwAlloc*sizeof(Long_t));
  }
  else
	nv = 0;
}

CNixVector::CNixVector(Word_t l, Word_t u, Long_t* v) :
  nvl(l), nvu(u)
{
  if (v) {
	int dwAlloc = bit2dw(l);
	nv = new Long_t[dwAlloc];
	memcpy(nv, v, dwAlloc*sizeof(Long_t));
  }
  else
	nv = 0;
}

CNixVector::~CNixVector()
{
  if (nv) delete[] nv;
}

void CNixVector::ExpandBits(int len)
{
  if (len <= 0) {
	//cout << "CNixVector::ExpandBits() error len = " << len << endl;
	return;
  }
  int free_bits = bit2dw(nvl)*32 - nvl;
  if (free_bits >= len) return;    // enough space

  int dwAlloc = bit2dw(nvl+len);
  Long_t* old = nv;
  nv = new Long_t[dwAlloc];
  memset(nv, 0, dwAlloc*sizeof(Long_t));
  if (old) {
	memcpy(nv, old, bit2dw(nvl)*sizeof(Long_t));
	delete[] old;
  }
}

void CNixVector::Set(const MACAddr& mac)
{
  int len = NO_COLOR_BITS + NO_MAC_BITS;
  ExpandBits(len);

  int dp = nvl + NO_COLOR_BITS;

  Long_t s[2];
  s[0] = mac.Lower();
  s[1] = mac.Upper();
  bitcpy(nv, dp, s, 0, NO_MAC_BITS);
  nvl += len;
}

void CNixVector::Set(int nix)
{
  nix++;    // nix starts from 1 to account for hidden bit

  /* results in log2(8) = 2, not 3 in Linux
  Long_t b = (Long_t)(log10(nix)/log10(2));
  */
  double d = log10((double)nix)/log10(2.0);
  Long_t b = (Long_t)d;

  int len = NO_COLOR_BITS + b;
  ExpandBits(len);

  Long_t dwNix = nix;
  dwNix = (dwNix << NO_COLOR_BITS) | (0xf & (b+1));
  bitcpy(nv, nvl, &dwNix, 0, len);
  nvl += len;
};

int CNixVector::Get(MACAddr& aix)
{
  Long_t dwNix = 0;
  Byte_t color = PeekColor();

  if (color != WHITE_COLOR) {
  // Other color, contains nix
	color--;
	bitcpy(&dwNix, 0, nv, nvu+NO_COLOR_BITS, color);
	nvu += NO_COLOR_BITS + color;
	Long_t hidden = 1 << color;
	dwNix += hidden;
	aix.Lower(dwNix-1);
	aix.Upper(0);

	//cout << "{\n";
	//cout << "color " << color << " memory nix " << dwNix-1 << endl;
	//cout << "}\n";

	return 1;
  }

  // White color, contains MAC address
  Long_t mac[2];
  mac[0] = mac[1] = 0;
  bitcpy(mac, 0, nv, nvu+NO_COLOR_BITS, NO_MAC_BITS);
  nvu += NO_COLOR_BITS + NO_MAC_BITS;
  aix.Lower(mac[0]);
  aix.Upper(mac[1]);
  return 0;
}

bool CNixVector::Advance(int n)
{
  // This function assumes symmetric routing
  if (nvu) {
	//cout << "HuH? nvu is not zero nvu " << nvu << endl;
	return false;
  }
  while (n > 0) {
	if (nvl < nvu + NO_COLOR_BITS) {
	  //cout << "HuH? nvl < nvu + NO_COLOR_BITS"
	  //<< " nvl " << nvl << " nvu " << nvu << " n " << n << endl;
	  return false;
	}
	Byte_t color = PeekColor();
	if (color == WHITE_COLOR) {
	  // This should NOT happen!!!
	  //cout << "HuH? color " << color << endl;
	  return false;
	}
	nvu += NO_COLOR_BITS + color - 1;
	n--;
  }
  return true;
}

Byte CNixVector::PeekColor()
{
  Long_t mask, color;
  int ofs = nvu / 32;
  int bp =  nvu % 32;

  if (bp <= 32-NO_COLOR_BITS) {
	mask = 0xf;
	color = (nv[ofs] >> bp) & mask;
  }
  else {
	long m = 0x80000000;
	mask = (m >> (31-bp));
	color = (nv[ofs] & mask) >> bp;
	mask = (Long_t)-1;
	mask = ~(mask << (bp+NO_COLOR_BITS-32));
	color |= (nv[ofs+1] & mask) << (32-bp);
  }
  return (Byte)color;
}

void CNixVector::Dump()
{
  int len;
  int gp1, gp2;
  Long_t color=0;
  int dwAlloc = bit2dw(nvl);

  Long_t* out = new Long_t[dwAlloc];
  memset(out, 0, sizeof(Long_t)*dwAlloc);

  gp2 = sizeof(Long_t)*dwAlloc*8;
  for (gp1 = 0; gp1 < nvl; gp1 += len) {
	len = NO_COLOR_BITS;
	gp2 -= len;
	bitcpy(&color, 0, nv, gp1, len);
	gp1 += len;
	bitcpy(out, gp2, &color, 0, len);
	len = color==WHITE_COLOR ? NO_MAC_BITS : color-1;
	gp2 -= len;
	bitcpy(out, gp2, nv, gp1, len);
  }

  Word_t* wout = (Word_t*)out;
  for (int i=dwAlloc*2-1; i>=0; i--)
	printf("%04x ", wout[i]);
  cout << endl;
  delete[] out;
}

CNixVector& CNixVector::operator = (const CNixVector& r)
{
  if (r.nv) {
	int dwAlloc = bit2dw(r.nvl);
	//cout << "dwAlloc = " << dwAlloc << endl;
	if (dwAlloc != bit2dw(nvl)) {
	  if (nv) delete[] nv;
	  nv = new Long_t[dwAlloc];
	}
	memcpy(nv, r.nv, dwAlloc*sizeof(Long_t));
  }
  else {
	if (nv) delete[] nv;
	nv = 0;
  }
  nvl = r.nvl;
  nvu = r.nvu;
  return *this;
}


// CNVList

CNVList::CNVList() :
  nItem(0)
{
  Item.reserve(10);
}

CNVList::~CNVList()
{
  for(RIVec_t::iterator i=Item.begin(); i!=Item.end(); ++i)
	if (*i) delete *i;
}

int CNVList::Add(Byte_t mid, IPAddr_t dst, IPAddr_t org, const CMetricVec& mv,
	const CNixVector& nv, Byte_t attr)
{
  CRouteInfo* ri = new CRouteInfo();
  ri->mid = mid;
  ri->attr = attr;
  ri->target = dst;
  ri->origin = org;
  ri->mv = mv;
  ri->nv = nv;

  RIVec_t::size_type i;
  for (i=0; i<Item.size(); ++i) if (!Item[i]) break;
  if (i < Item.size()) Item[i] = ri;
  else Item.push_back(ri);
  nItem++;
  return i;
}

int CNVList::Remove(IPAddr_t tar)
{
  int num_deleted = 0;
  for (RIVec_t::size_type i = 0; i < Item.size(); ++i) {
	if (!Item[i]) continue;
	if (Item[i]->target == tar) {
	  delete Item[i];
	  Item[i] = 0;
	  nItem--;
	  num_deleted++;
	}
  }
  return num_deleted;
}

int CNVList::Remove(IPAddr_t tar, IPAddr_t org)
{
  for (RIVec_t::size_type i = 0; i < Item.size(); ++i) {
	if (!Item[i]) continue;
	if (Item[i]->target==tar && Item[i]->origin==org) {
	  delete Item[i];
	  Item[i] = nil;
	  nItem--;
	  return i;
	}
  }
  return -1;
}

int CNVList::Remove(Byte mid, IPAddr_t tar)
{
  for (RIVec_t::size_type i=0; i<Item.size(); ++i) {
	if( !Item[i] ) continue;
	if (Item[i]->mid==mid && Item[i]->target==tar) {
	  delete Item[i];
	  Item[i] = nil;
	  nItem--;
	  return i;
	}
  }
  return -1;
}

int CNVList::Remove(int i)
{
  // Remove by index
  if (i < 0 || !Item[i]) return -1;
  delete Item[i];
  Item[i] = nil;
  nItem--;
  return i;
}

int CNVList::Find(IPAddr_t tar)
{
  for (RIVec_t::size_type i=0; i<Item.size(); ++i)
	if (Item[i] && Item[i]->target==tar) return i;
  return -1;
}

int CNVList::Find(IPAddr_t tar, IPAddr_t org)
{
  for (RIVec_t::size_type i=0; i<Item.size(); ++i)
	if (Item[i] && Item[i]->target==tar && Item[i]->origin==org) return i;
  return -1;
}

int CNVList::Search(Byte mid, IPAddr_t target)
{
  for (RIVec_t::size_type i=0; i<Item.size(); ++i)
  {
	if (!Item[i]) continue;
	if (Item[i]->mid==mid && Item[i]->target==target) return i;
  }
  return -1;
}

void CNVList::Update(int i, Byte_t hop, CNixVector& nv)
{
  if (!Item[i]) return;
  Item[i]->mv.hop = hop;
  Item[i]->nv = nv;
}

void CNVList::Update(int i, IPAddr_t target, Byte_t hop, CNixVector& nv)
{
  if (!Item[i]) return;
  Item[i]->target = target;
  Item[i]->mv.hop = hop;
  Item[i]->nv = nv;
}


// CNbrTable

CNbrTable::CNbrTable() : nNbr(0)
{
  Nbr.reserve(10);
}

int CNbrTable::Add(const MACAddr& mac)
{
  /* This function is called if Find() returns -1
   */

  if (nNbr >= defaultMaxNeighbor) return -1;

  // Find an empty slot
  NbrVec_t::size_type i;
  for (i = 0; i < Nbr.size(); ++i)
	if (Nbr[i].Empty()) break;
  if (i == Nbr.size())  // no empty slot
	Nbr.push_back(CNeighbor());  // add one

  // Store
  Nbr[i] = CNeighbor(mac, true);
  nNbr++;
  return i;
}

int CNbrTable::Invalidate(int i)
{
  Nbr[i].val = false;
  return i;
}

int CNbrTable::Invalidate(const MACAddr& mac)
{
  int i = Find(mac);
  if (i >= 0) Nbr[i].val = false;
  return i;
}

int CNbrTable::Find(const MACAddr& mac)
{
  for (NbrVec_t::size_type i = 0; i < Nbr.size(); ++i)
	if (Nbr[i].mac == mac) return i;
  return -1;
}

Size_t CNbrTable::Size() const
{
  Size_t sz;
  sz = Nbr.size()*sizeof(CNeighbor);
  return sz;
}

// CVQue

#define T_HISTORY		((Time_t) 10.0)    // sec
#define DEFAULT_VQ_BW	((Rate_t) 125000)  // bytes/sec, 1Mb/s
#define VQ_LIMIT		5000               // bytes
#define MAX_HISTORY		40

CVQue::CVQue()
  : Byteseconds(0), AverageLen(0)
{
  //Bw = DEFAULT_VQ_BW;
  Bw = 16384/8;
  Init(0,0);
}

void CVQue::Init(Size_t s0, Time_t t0)
{
  if (!Buf.empty()) Buf.clear();
  Buf.reserve(MAX_HISTORY);
  Buf.push_back(History(s0, t0));
  Next = 0;
}

bool CVQue::Update(Size_t sz)
{
  if (!sz) return true;
  Time_t now = Simulator::Now();
  //Time_t now = Now;
  int last;
  if (Buf.size() < MAX_HISTORY)
	last = Buf.size()-1;
  else
	last = Next>0 ? Next-1 : MAX_HISTORY-1;

  if (now == Buf[last].Time) {
	Buf[last].Size += sz;
	if (Buf[last].Size > VQ_LIMIT) Buf[last].Size = VQ_LIMIT;
	return true;
  }
  if (now < Buf[last].Time) {
	cout << "Event corrupted  last " << Buf[last].Time << " now " << now
		 << endl;
	return false;
  }

  Time_t elap = now - Buf[last].Time;
  long size;
  if (elap >= T_HISTORY) {
  // Initialize, start from now - T_HISTORY
	long height = Buf[last].Size - (Size_t)(Bw*(elap-T_HISTORY));
	if (height < 0) height = 0;
	size = height - (long)(Bw*T_HISTORY);
	Byteseconds = CalcArea(height, T_HISTORY);
	AverageLen = (Long_t)(Byteseconds / T_HISTORY);
	Init(height, now-T_HISTORY);
  }
  else {
  // Compute total area during T_HISTORY
	size = Buf[last].Size - (long)(Bw*elap);
	CalcTotalArea();
  }
  if (size < 0) size = 0;
  size += sz;
  if (size > VQ_LIMIT) size = VQ_LIMIT;
  AddHistory(size, now);
  return true;
}

Long_t CVQue::CalcArea(long height, Time_t period)
{
  Long_t area;
  if (height <= 0 || period <= 0)
	area = 0;
  else {
	Time_t t0 = height / Bw;
	if (t0 <= period) {
	  // Area of triangle
	  area = (Long_t)(height * t0 / 2);
	}
	else {
	  // Area of trapezoid
	  long ds = (long)(Bw * period);
	  long h2  = height - ds;
	  if( h2 < 0 ) h2 = 0;
	  area = (Long_t)(period * (height+h2) / 2);
	}
  }
  return area;
}

void CVQue::CalcTotalArea()
{
  // Calculate area from start to now
  Time_t now = Simulator::Now();
  //Time_t now = Now;
  Time_t start = now - T_HISTORY;
  if (start < 0) start = 0;
  Byteseconds = 0;

  int i = FindHistory(start);
  if (i < 0) {
	// No history
	AverageLen = 0;
	return;
  }

  if (start < Buf[Next].Time) start = Buf[Next].Time;
  int j = (i+1) % Buf.size();
  while (Buf[i].Time < Buf[j].Time) {
	Byteseconds += CalcArea(Buf[i].Size, Buf[j].Time-Buf[i].Time);
	i = j;
	j = (i+1) % Buf.size();
  }
  Byteseconds += CalcArea(Buf[i].Size, now-Buf[i].Time);
  AverageLen = (Long_t)(Byteseconds/(now-start));
}

void CVQue::AddHistory(Size_t sz, Time_t t)
{
  if (Buf.size() < MAX_HISTORY)
	Buf.push_back(History(sz, t));
  else {
	Buf[Next++] = History(sz, t);
	Next %= MAX_HISTORY;
  }
}

int CVQue::FindHistory(Time_t t)
{
  unsigned int j=0;
  for (HVec_t::size_type i = Next; j < Buf.size(); ++j) {
	if (Buf[i].Time >= t) return i;
	i = (i+1) % Buf.size();
  }
  return -1;
}
