/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/


#ifndef __NIXVECTOR_H
#define __NIXVECTOR_H

#include <vector>
#include <list>
#include "common-defs.h"
#include "ipaddr.h"
#include "macaddr.h"
#include "ipv4.h"

#define NC_LIST_SIZE		16
#define REPLY_LIST_SIZE		32
#define NVREQ_LIST_SIZE		20
#define MAX_NEIGHBOR		32000
#define SEND_BUF_SIZE		64

#define	NO_COLOR_BITS	4
#define	WHITE_COLOR		0x0
#define NO_MAC_BITS		48

#define	NVR_OPT_SYM		0x1    // symmetric routing
#define	NVR_OPT_ERR		0x2    // containing routing fail info
#define	NVR_OPT_RNV		0x4    // build RNV while routing

#define	NV_ATTR_UNI		0x1    // NV set up partially

#define METRIC_NONE		0
#define METRIC_HOP		1
#define METRIC_LOSS		2

#define TICK_EV_NVLIST	1
#define TICK_EV_NBR		1

#define TTL_INIT		3
#define TTL_MAX			35

#define NVREQ_TO_INIT	0.5
#define NVREQ_TO_MAX	3

inline int bit2dw(int len) { return len/32 + (len%32 ? 1:0); }
int bitcpy(Long_t* dst, int dp, Long_t* src, int sp, int len);

class Packet;
class CRouteInfo;
class NVREvent;

typedef unsigned char Byte;
typedef std::vector<CRouteInfo*> RIVec_t;


// n-tuple double word class
class DWn
{
public:
	DWn() : size(0), dw(0) {}
	DWn(int sz);
	virtual ~DWn() { if(dw) delete[] dw; }
	DWn& operator =   (DWn& r);
	DWn& operator <<= (int n);
	DWn& operator >>= (int n);
	DWn& operator &=  (DWn& r);
	DWn& operator |=  (DWn& r);

public:
	int     size;
	Long_t  *dw;
};

// Metric Vector
class CMetricVec
{
public:
	CMetricVec() : hop(0), sig(0), pks(0) {}
	CMetricVec(Byte_t h, DCount_t s, DCount_t p)
		: hop(h), sig(s), pks(p) {}
	CMetricVec(const CMetricVec& r)
		: hop(r.hop), sig(r.sig), pks(r.pks) {}
	virtual ~CMetricVec() {}
	CMetricVec operator == (CMetricVec r)
		{ hop=r.hop; sig=r.sig; pks=r.pks; return *this; }

public:
	Byte_t  hop;    // hop count
	float   sig;    // signal strength
	float   pks;    // average # packets
};

// CNCItem
class CNCItem
{
public:
  CNCItem() : nc_id(0), source(0), lifetime(0) {}
  CNCItem(Byte id, IPAddr_t src, Time_t lt, MACAddr& inf)
	: nc_id(id), source(src), lifetime(lt), from_inf(inf) {}
  virtual ~CNCItem() {}
  bool Valid() { return lifetime > 0; }

public:
  Byte_t   nc_id;
  IPAddr_t source;
  Time_t   lifetime;

  // debug
  MACAddr  from_inf;
};

// CNCList
class CNCList
{
public:
  CNCList();

  int  Add(IPAddr_t source, Byte_t nc_id, MACAddr& inf=MACAddr::NONE);
  int  Remove(IPAddr_t source, Byte_t nc_id);
  int  Find(IPAddr_t source, Byte_t nc_id);
  CNCItem* At(int i) { return &Item[i]; }

public:
  typedef std::vector<CNCItem> NCVec_t;

  int     nItem;
  NCVec_t Item;

  static Word_t defaultNCListSize;
};

// CRepItem
class CRepItem
{
public:
  CRepItem() : type(0), nc_id(0), mid(METRIC_NONE), source(0), lifetime(0) {}
  CRepItem(Byte t, Byte id, Byte m, CMetricVec& mvec, IPAddr_t s, long lt)
	: type(t), nc_id(id), mid(m), mv(mvec), source(s), lifetime(lt) {}
  virtual ~CRepItem() {}
  bool Valid() { return lifetime > 0; }

public:
  Byte       type;     // packet type
  Byte       nc_id;    // NC id
  Byte       mid;      // metric id
  CMetricVec mv;       // metric vector
  IPAddr_t   source;   // src addr
  long       lifetime;
};

// CReplyList
class CReplyList
{
public:
  CReplyList();

  int  Add(IPAddr_t s, Byte id, Byte t, Byte mid, CMetricVec& mv);
  int  Remove(IPAddr_t source, Byte id, Byte type);
  void Update(int i, CMetricVec& mv);    // update metric vector
  CRepItem* Find(IPAddr_t source, Byte id, Byte type);

public:
  typedef std::vector<CRepItem> RepVec_t;
  static Word_t defaultReplyListSize;

private:
  int      nItem;
  RepVec_t Item;
};

// CNixVector
class CNixVector
{
public:
  CNixVector() : nvl(0), nvu(0), nv(0) {}
  CNixVector(const CNixVector& r);
  CNixVector(Word_t nvl, Word_t nvu, Long_t* nv);
  virtual ~CNixVector();

  void   ExpandBits(int bitlen);
  void   Set(const MACAddr& mac);
  void   Set(int nix);
  int    Get(MACAddr& aix);
  bool   Advance(int n);
  Byte   PeekColor();
  void   Clear()       { if(nv) delete[] nv; nvl=0; nvu=0; nv=0; }
  Size_t Size() const  { return nvl/8 + (nvl%8 ? 1:0); }    // byte
  //Size_t SSize() const { return sizeof(nvl)+sizeof(nvu)+Size(); }
  Size_t SSize() const { return sizeof(nvl)+Size(); }
  bool   IsValid()     { return nvl > 0; }
  bool   IsEnd()       { return nvu >= nvl; }
  void   Dump();    // dump to console
  CNixVector& operator = (const CNixVector& r);

public:
  Word_t  nvl;
  Word_t  nvu;
  Long_t* nv;
};

// CNeighbor
class CNeighbor
{
public:
  CNeighbor() : val(false) {}
  CNeighbor(const MACAddr& m, bool v) : mac(m), val(v) {}
  virtual ~CNeighbor() {}
  bool Valid() { return val; }
  bool Empty() { return mac == MACAddr::NONE; }

public:
  MACAddr mac;
  bool    val;    // validity
};

typedef std::vector<CNeighbor> NbrVec_t;

// CNbrTable
class CNbrTable
{
public:
  CNbrTable();
  virtual ~CNbrTable() {}

  int  Add(const MACAddr& mac);
  int  Invalidate(int i);
  int  Invalidate(const MACAddr& mac);
  int  Find(const MACAddr& mac);
  Size_t Size() const;    // byte

public:
  int       nNbr;    // # of neighbors
  NbrVec_t  Nbr;

  static Word_t defaultMaxNeighbor;
};

// CRouteInfo
class CRouteInfo
{
public:
  Byte_t      mid;      // metric id
  Byte_t      attr;     // NV attribute
  IPAddr_t    target;   // destination
  IPAddr_t    origin;   // opposite end point to target
  CMetricVec  mv;       // metric vector
  CNixVector  nv;       // NV
};

// CNVList
class CNVList
{
public:
  CNVList();
  virtual ~CNVList();

  int  Add(Byte_t mid, IPAddr_t dst, IPAddr_t org, const CMetricVec& mv,
		   const CNixVector& nv, Byte_t attr=0);
  int  Remove(IPAddr_t target);
  int  Remove(IPAddr_t target, IPAddr_t origin);
  int  Remove(Byte mid, IPAddr_t target);
  int  Remove(int i);
  int  Find(IPAddr_t target);
  int  Find(IPAddr_t target, IPAddr_t origin);
  int  Search(Byte mid, IPAddr_t target);
  void Update(int i, Byte_t hop, CNixVector& nv);
  void Update(int i, IPAddr_t dst, Byte_t hop, CNixVector& nv);
  bool IsBetter(int i, Byte_t hop)  { return (hop < Item[i]->mv.hop); }
  bool IsWorse(int i, Byte_t hop)   { return (hop > Item[i]->mv.hop); }
  bool IsShorter(int i, Word_t nvl) { return (nvl < Item[i]->nv.nvl); }

public:
  int      nItem;
  RIVec_t  Item;
};

// Virtual Que Class
// History
class History
{
public:
  History(Size_t sz, Time_t t) : Size(sz), Time(t) {}

public:
  Size_t  Size;
  Time_t  Time;
};

typedef std::vector<History> HVec_t;

// CVQue
class CVQue
{
public:
  CVQue();
  bool    Update(Size_t sz);
  Long_t  Load() { return AverageLen; }
  void    Bandwidth(Rate_t bw) { Bw = bw; }
  Rate_t  Bandwidth() { return Bw; }

private:
  void    Init(Size_t s0, Time_t t0);
  Long_t  CalcArea(long height, Time_t period);
  void    CalcTotalArea();
  void    AddHistory(Size_t sz, Time_t t);
  int     FindHistory(Time_t t);

public:
  Rate_t  Bw;            // transmit rate in bytes/sec
  long    Byteseconds;
  Long_t  AverageLen;    // average que length in bytes
  int     Next;
  HVec_t  Buf;
};

// NVReqItem
class NVReqItem
{
public:
  NVReqItem() : dst(0), id(0), ttl(0), time(0), timer(nil) {}
  NVReqItem(IPAddr_t d, Byte_t i, Byte_t t, Time_t ts, NVREvent* e)
	: dst(d), id(i), ttl(t), time(ts), timer(e) {}

public:
  IPAddr_t   dst;         // target
  Byte_t     id;          // NVREQ ID
  Byte_t     ttl;         // ttl in NVREQ
  Time_t     time;        // time when NVREQ was sent
  //Time_t   lifetime;    // remaining sec till timeout, not used
  NVREvent*  timer;       // timer instead of lifetime
};

// CNCQueItem
class CNCQueItem
{
public:
  CNCQueItem() : p(nil), timer(nil) {}
  CNCQueItem(Packet* pkt, NVREvent* t) : p(pkt), timer(t) {}

public:
  Packet*   p;
  NVREvent* timer;
};

// SendBufItem
class SendBufItem
{
public:
  SendBufItem() : p(nil) {}
  SendBufItem(Packet* pkt, IPV4ReqInfo& i) : p(pkt), info(i) {}

public:
  Packet*      p;
  IPV4ReqInfo  info;
  //Time_t     lifetime;
  //NVREvent*  timer;        // instead of lifetime
};

#endif
