/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Node Implementation class
// George F. Riley.  Georgia Tech, Spring 2002

// Defines some common functionality between Real nodes
// and Ghost nodes.

#include <iostream>
#include <unistd.h>

#ifdef HAVE_QT
#include <qcanvas.h>
#include <qpoint.h>
#endif

//#define DEBUG_MASK 0x02
#include "debug.h"
#include "node-real.h"
#include "protograph.h"
#include "portdemux.h"
#include "ipaddr.h"
#include "duplexlink.h"
#include "application.h"
#include "mobility.h"
#include "link-rti.h"
#include "mask.h"
#include "interface-real.h"
#include "interface-wireless.h"
#include "interface-ghost.h"
#include "interface-ethernet.h"

using namespace std;

NodeImpl::NodeImpl(Node* n)
  : pNode(n), ipAddr(IPADDR_NONE), neighborCount(0)
{
}

NodeImpl::~NodeImpl()
{
}

// IP Address Management
IPAddr_t    NodeImpl::GetIPAddr()
{  // Get the IPAddres of this node 
  return ipAddr;   // Return IP Address....may be IPADDR_NONE
}

void        NodeImpl::SetIPAddr(IPAddr_t i)
{ // Set the IPAddres of this node
  ipAddr = i;
}

void NodeImpl::IPAddrs(IPMaskVec_t& a)
{ // Create a list of IP Addresses
  bool localFound = false;  // Insure the nodes "ipAddr" is included
  for (IFVec_t::size_type j = 0; j < interfaces.size(); ++j)
    {
      Interface* i = interfaces[j];
      if (i->GetIPAddr() != IPADDR_NONE)
        {
          a.push_back(IpMask(i->GetIPAddr(), i->GetIPMask()));
          if ((ipAddr & i->GetIPMask()) == (i->GetIPAddr() & i->GetIPMask()))
            {
              localFound = true;
            }
        }
    }
  if (!localFound && (ipAddr != IPADDR_NONE))
    { // Need to add node's also
      a.push_back(IpMask(ipAddr, MASK_ALL));
    }
}

bool NodeImpl::LocalIP( IPAddr_t ip)
{ // Check if specified ip address is local
  if (ipAddr == IPADDR_NONE) return false; // This node has no ip address
  if (ip == ipAddr) return true; // Found specified local
  NodeReal::count4 += interfaces.size();
  for (IFVec_t::size_type i = 0; i < interfaces.size(); ++i)
    {
      if ((IPAddr_t)interfaces[i]->GetIPAddr() == ip)
        {
          return true; // Found local
        }
    }
  return false; // Not found, not local
}

bool NodeImpl::IPKnown()
{ // True if non-default ip is assigned
  return ipAddr != IPADDR_NONE;
}

// Interfaces

Interface* NodeImpl::AddInterface(const L2Proto& l2, bool bootstrap)
{
  return AddInterface(l2, IPADDR_NONE, MASK_ALL, MACAddr::Allocate(),
		      bootstrap);
}

Interface* NodeImpl::AddInterface(const L2Proto& l2,
                                  IPAddr_t i, Mask_t mask,
                                  MACAddr m, bool bootstrap)
{
#ifdef SIM
	  return AddInterface(l2, i, mask, m, bootstrap, NULL);
}

Interface* NodeImpl::AddInterface(const L2Proto& l2,
                                  IPAddr_t i, Mask_t mask,
                                  MACAddr m, bool bootstrap, char *name)
{
		
#endif // SIM
  Interface* pif;

  if (IsReal())
    { // Real node, use either wired or wireless interface
      if (l2.IsWireless())
        {
          DEBUG(2,(cout << "Node " << pNode->Id()
                    << " adding wireless interface" << endl));
          pif = new InterfaceWireless(l2, i, mask, m, bootstrap);
        }
      else
        {
          DEBUG(2,(cout << "Node " << pNode->Id()
                   << " adding real interface" << hex << i << endl));
          pif = new InterfaceReal(l2, i, mask, m, bootstrap);
        }
    }
  else
    { // Use a ghost interface
      DEBUG(2,(cout << "Node " << pNode->Id()
               << " adding ghost interface" << endl));
      pif = new InterfaceGhost(l2, i, mask, m, bootstrap);
    }
  
  // Add to list  of interface by Mac addresses
  Interface::Add(pif, m.Lower());
  pif->SetNode(pNode);
  interfaces.push_back(pif);
  if (!IPKnown())
    { // Default IP address to first interface
      SetIPAddr(i);
    }

#ifdef SIM
  pif->SetIfIndex(InterfaceCount());
  if (name == NULL) 
  {
    char name_str[5];
    sprintf(name_str, "eth%d", pif->GetIfIndex());
    name = name_str;
  }
  pif->SetName(name);
#endif //SIM

  return pif;
}

Interface* NodeImpl::AddInterface(const L2Proto& l2, const Interface& iface,
                                  IPAddr_t i, Mask_t mask,
                                  MACAddr m, bool bootstrap)
{
  Interface* pif;
  // if the interface parameter is of type interface ethernet
  if(iface.IsEthernet())
     {
        DEBUG(2,(cout << "Node " << pNode->Id()
                   << " adding Ethernet interface" << hex << i << endl));
       pif = new InterfaceEthernet(l2, i, mask, m, bootstrap);
    }
  else
    {
      return AddInterface(l2, i, mask, m, bootstrap);
    }

  pif->SetNode(pNode);
  interfaces.push_back(pif);
  if (!IPKnown())
    { // Default IP address to first interface
      SetIPAddr(i);
    }
  return pif;
}

Count_t NodeImpl::InterfaceCount()
{
  return interfaces.size();
}

const IFVec_t& NodeImpl::Interfaces()
{ // Get the complete list of interfaces
  return interfaces;
}

Interface* NodeImpl::GetIfByLink(Link* l)
{
  for (IFVec_t::size_type i = 0; i < interfaces.size(); ++i)
    {
      if (interfaces[i]->GetLink() == l) return interfaces[i]; // Found it
    }
  return nil;  // Does not exist
}

void NodeImpl::UseWormContainment(bool b) 
{
  usewormcontainment = b;
}



Interface* NodeImpl::GetIfByNode(Node* n)
{ // Find the interface used to get to the specified node
  for (IFVec_t::size_type i = 0; i < interfaces.size(); ++i)
    {
      Link* pLink = interfaces[i]->GetLink();
      if (!pLink) continue;
      if (pLink->NodeIsPeer(n)) return interfaces[i]; // found
    }
  return nil;
}

Interface* NodeImpl::GetIfByIP(IPAddr_t ip)
{ // Find the interface with the specified IP Address
  for (IFVec_t::size_type i = 0; i < interfaces.size(); ++i)
    {
#ifdef SIM
			IPAddr_t IP = (IPAddr_t)interfaces[i]->GetIPAddr(); 
      if (IP == ip || IP == (ip & 0x7fffffff)) //ipv6 linklocal
#else
      if ((IPAddr_t)interfaces[i]->GetIPAddr() == ip)
#endif //SIM
        {
          return interfaces[i]; // found
        }
    }
  return nil;
}

#ifdef SIM
Interface* NodeImpl::GetIfByName(char *name)
{ // Find the interface with the specified Name
  for (IFVec_t::size_type i = 0; i < interfaces.size(); ++i)
    {
      if (strcmp(name, interfaces[i]->GetName())==0)
        {
          return interfaces[i]; // found
        }
    }
  return NULL;
}

Interface* NodeImpl::GetIfByIndex(int index)
{ // Find the interface with the specified Index
  for (IFVec_t::size_type i = 0; i < interfaces.size(); ++i)
    {
      if (index == interfaces[i]->GetIfIndex())
        {
          return interfaces[i]; // found
        }
    }
  return NULL;
}
#endif //SIM

// Links.
// All of the following are just a syntactic shortcut to the DuplexLink object.
// Each creates a local DuplexLink object, which is destroyed on exit.
Interface* NodeImpl::AddDuplexLink(Node* rn)
{ // Add duplex link to/from remote
  DuplexLink dl(pNode, rn);
  return dl.localif;
}

Interface* NodeImpl::AddDuplexLink(Node* rn, const Linkp2p& l)
{ // Add duplex link to/from remote
  DuplexLink dl(pNode, l, rn, l);
  return dl.localif;
}

Interface* NodeImpl::AddDuplexLink(Interface* li, Interface* ri)
{ // Add a link between existing interfaces
  DuplexLink dl(li, ri);
  return dl.localif;
}

Interface* NodeImpl::AddDuplexLink(Interface* li, Interface* ri,
                               const Linkp2p& l)
{
  DuplexLink dl(li, l, ri, l);
  return dl.localif;
}

Interface* NodeImpl::AddDuplexLink(Node* rn,
                            IPAddr_t li, Mask_t lm,
                            IPAddr_t ri, Mask_t rm)
{
  DuplexLink dl(pNode, li, lm, rn, ri, rm);
  return dl.localif;
}

Interface* NodeImpl::AddDuplexLink(Node* rn, const Linkp2p& l,
                            IPAddr_t li, Mask_t lm,
                            IPAddr_t ri, Mask_t rm)
{
  DuplexLink dl(pNode, li, lm, l, rn, ri, rm, l);
  return dl.localif;
}

Link* NodeImpl::GetLink(Node* n)
{
  Interface* i = GetIfByNode(n); // Interface to specified node
  if (!i) return nil;            // Not found
  return i->GetLink();           // Get the associated link
}

Interface* NodeImpl::AddRemoteLink(IPAddr_t i, Mask_t m)
{
  Interface* iFace = AddInterface(L2Proto802_3(), i, m);
  iFace->SetLink(new LinkRTI(i, m, iFace));
  return iFace;
}

Interface* NodeImpl::AddRemoteLink(IPAddr_t i, Mask_t m, Rate_t r, Time_t d)
{
  Interface* iFace = AddInterface(L2Proto802_3(), i, m);
  iFace->SetLink(new LinkRTI(i, m, iFace, r, d));
  //iFace->SetLink(new LinkRTI(i, m, iFace, r, d, 0));
  return iFace;
}

// Simplex links, for modeling one-way links
Interface* NodeImpl::AddSimplexLink(Node* n)
{
  return AddSimplexLink(n, Linkp2p::Default(), IPADDR_NONE, Mask(32));
}

Interface* NodeImpl::AddSimplexLink(Node* n, const Linkp2p& l)
{
  return AddSimplexLink(n, l, IPADDR_NONE, Mask(32));
}

Interface* NodeImpl::AddSimplexLink(Node* n, const Linkp2p& l,
                                    IPAddr_t i,Mask_t m)
{
  Interface* localif = AddInterface(L2Proto802_3(), i, m);
  Interface* remoteif = n->AddInterface(L2Proto802_3(), IPADDR_NONE, Mask(32));
  Linkp2p* lp = (Linkp2p*)l.Copy();
  localif->SetLink(lp);
  lp->SetPeer(remoteif);
  return localif;
}

// Neighbor management
void NodeImpl::Neighbors(NodeWeightVec_t& n, bool noLeaf)
{ // Ask each interface to add to neighbor list
  NodeReal::count3++;
  for (IFVec_t::size_type i = 0; i != interfaces.size(); ++i)
    {
      Interface* iface = interfaces[i];
      if (noLeaf)
        { // Check for leaf neighbors with local route,
          // Skip if so
          if (iface->NeighborCount(pNode) <= 1)
            { // Only one neighbor, find it
              // ?? Why is this so hard?
              Link* l = iface->GetLink();
#ifdef BUGFIX
              if (0)
#else
              if (l)
#endif //BUGFIX
                {
                  Interface* pPeer = l->GetPeer((Packet*)nil);
                  if (pPeer)
                    {
                      Node* n = pPeer->GetNode();
                      if (n)
                        {
                          if (n->NeighborCount() <= 1)
                            { // Finally, found a leaf
                              IPAddr_t ip = n->GetIPAddr();
                              // If it's a local route from here, ignore it
                              if (iface->IsLocalIP(ip)) continue;
                            }
                        }
                    }
                }
            }
        }
      DEBUG(1,(cout << "Node " << pNode->Id() 
               << " asking if about neighbors " << endl));
      iface->Neighbors(n);
    }
}

Count_t NodeImpl::NeighborCount()
{ // Count number of routing neighbors
  // Done by summing routing neighbors for each interface
  // First see if previously calcualted
  if (neighborCount) return neighborCount;
   
  Count_t t = 0;
  for (IFVec_t::iterator i = interfaces.begin(); i != interfaces.end(); ++i)
    {
      t += (*i)->NeighborCount(pNode);
    }
  neighborCount = t;
  return t;
}

void NodeImpl::NeighborsByIf(Interface* i, IPAddrVec_t& v)
{
  NodeReal::count2++;
  NodeWeightVec_t nwv;
  // Get neighbors for this interface
  i->Neighbors(nwv, true); // Force all even for single-gateway ethernet
  for (NodeWeightVec_t::iterator j = nwv.begin(); j != nwv.end(); ++j)
    {
      v.push_back(j->node->GetIPAddr()); // Return list of neighbor IP's
    }
}


// Routing Interface
Interface* NodeImpl::LocalRoute(IPAddr_t ip)
{ // Find if destination ip is directly connected on any local interace
  DEBUG0((cout << "NodeImpl::LocalRoute, id " << pNode->Id() << endl));
  NodeReal::count1 += interfaces.size();
  for (IFVec_t::iterator i = interfaces.begin(); i != interfaces.end(); ++i)
    {
      if ((*i)->IsLocalIP(ip)) return *i; // found it
    }
  return nil; // Not local
}

Count_t    NodeImpl::GetNix(Node* n) const
{ // Get Neighbor Index for spec'd node
  Count_t nix = 0;
  // Scan the interfaces looking for specified neighbor
  for (IFVec_t::size_type i = 0; i < interfaces.size(); ++i)
    {
      if (interfaces[i]->NodeIsPeer(n))
        { // Found it
          return nix + interfaces[i]->NodePeerIndex(n);
        }
      nix += interfaces[i]->NeighborCount(pNode); // Advance neighbor count
    }
  return MAX_COUNT; // Not found
}

CustomShape_t NodeImpl::CustomShape()
{
  return nil; // Default to no custom shape if not overridden.
}

void NodeImpl::CustomShape(CustomShape_t)
{ // Ignored if not overridden
}

bool NodeImpl::CustomShapeFile(const char*)
{ // Overridden by node-real
  return true;
}

bool NodeImpl::ICMPEnabled() const
{
  return false; // Over-ridden by node-real
}

void NodeImpl::DisableICMP()
{
  // Over-ridden by node-real
}

void NodeImpl::Down()
{
  // Over-ridden by node-real
}

void NodeImpl::Up()
{
  // Over-ridden by node-real
}

bool NodeImpl::IsDown()
{
  // Over-ridden by node-real
  return false;
}


  
