/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Node Implementation class
// George F. Riley.  Georgia Tech, Fall 2002

// Defines an abstract base class for node implementations

#ifndef __nodeimpl_h__
#define __nodeimpl_h__

#include "common-defs.h"
#include "macaddr.h"
#include "node.h"

class Link;
class Node;
class Interface;
class Routing;
class Protocol;
class ProtocolGraph;
class PortDemux;
class Linkp2p;
class Queue;
class Application;

class RoutingAODV;
class AODVRoutingEntry;

//Doc:ClassXRef
class NodeImpl {
public:
  NodeImpl(Node*);
  virtual ~NodeImpl();

  NodeId_t Id() { return pNode->Id(); }

  // IP Address Management
  virtual IPAddr_t GetIPAddr();          // Get the IPAddres of this node
  virtual void     SetIPAddr(IPAddr_t i);// Set the IPAddres of this node
  virtual void     IPAddrs(IPMaskVec_t&);// Get all IP/Masks for this node
  virtual bool     LocalIP(IPAddr_t);    // True if specified ip is local
  virtual bool     IPKnown();            // True if non-default ip assigned

  // Real/Ghost test
  virtual bool     IsReal() = 0;         // True if real node


  // switching functions
  virtual void IsSwitchNode(bool) = 0;
  virtual bool IsSwitchNode() = 0;

  virtual void PacketRX(Packet* p, Interface* iface) = 0;



  // Interfaces
  virtual Interface* AddInterface(const L2Proto&, bool bootstrap=false);
  virtual Interface* AddInterface(const L2Proto&,
                                  IPAddr_t, Mask_t,
                                  MACAddr = MACAddr::Allocate(),
																	bool bootstrap = false);
#ifdef SIM
  virtual Interface* AddInterface(const L2Proto&,
                                  IPAddr_t, Mask_t,
                                  MACAddr,
                                  bool bootstrap,
																	char *name);
#endif //SIM
  virtual Interface* AddInterface(const L2Proto&,const Interface&,
                                  IPAddr_t, Mask_t,
                                  MACAddr = MACAddr::Allocate(),
                                  bool bootstrap = false);
  virtual Count_t    InterfaceCount();   // Get count of interfaces
  virtual const IFVec_t& Interfaces();   // Get complete list of interfaces
  virtual Interface* GetIfByLink(Link*); // Get iface to the specified link
  virtual Interface* GetIfByNode(Node*); // Get iface to the specified node
  virtual Interface* GetIfByIP(IPAddr_t);// Get iface with specified IP
#ifdef SIM
  //	virtual Interface* NodeImpl::GetIfByName(char *name);  
  //	virtual Interface* NodeImpl::GetIfByIndex(int index);  
	virtual Interface* GetIfByName(char *name);  
	virtual Interface* GetIfByIndex(int index);  
#endif //SIM

  // Broadcast packet on all interfaces
  virtual void       Broadcast(Packet*, Proto_t) = 0;

  // Links.  All return local interface
  // Add duplex link to/from remote
  virtual Interface* AddDuplexLink(Node*);
  virtual Interface* AddDuplexLink(Node*, const Linkp2p&);
  // Next two specify new links between existing interfaces
  virtual Interface* AddDuplexLink(Interface*,Interface*);
  virtual Interface* AddDuplexLink(Interface*,Interface*, const Linkp2p&);
  // Next two specify local and remote IPAddress and Masks
  virtual Interface* AddDuplexLink(Node*,
                                   IPAddr_t, Mask_t, IPAddr_t, Mask_t);
  virtual Interface* AddDuplexLink(Node*, const Linkp2p&,
                                   IPAddr_t, Mask_t, IPAddr_t, Mask_t);
  virtual Link*      GetLink(Node*);    // Get link to specified node
  // Next two are for distributed simulations
  virtual Interface* AddRemoteLink(IPAddr_t, Mask_t);
  virtual Interface* AddRemoteLink(IPAddr_t, Mask_t, Rate_t, Time_t);

  // Simplex links, for modeling one-way links
  virtual Interface* AddSimplexLink(Node*);
  virtual Interface* AddSimplexLink(Node*, const Linkp2p&);
  virtual Interface* AddSimplexLink(Node*, const Linkp2p&,
                                    IPAddr_t, Mask_t);

  // Queues
  virtual Queue*     GetQueue() = 0;        // Return queue if only one i/f
  virtual Queue*     GetQueue(Node*) = 0;   // Get the queue to specified node

  // Applications
  virtual Application* AddApplication(const Application& a) = 0;

  // Neighbor management (used by some routing protocols)
  virtual void        Neighbors(NodeWeightVec_t &, bool); // List of neighbors
  virtual Count_t     NeighborCount();       // Count of routing neighbors
  // Get neighbors on given iface
  virtual void        NeighborsByIf(Interface*, IPAddrVec_t&);

  // Routing
  virtual void DefaultRoute(RoutingEntry) = 0;     // Specify default route
  virtual void DefaultRoute(Node*) = 0;            // Specify default route
  virtual void AddRoute( IPAddr_t, Count_t, Interface*, IPAddr_t) = 0;
  virtual RoutingEntry LookupRoute(IPAddr_t) = 0;  // Find a routing entry
  virtual RoutingEntry LookupRouteNix(Count_t) = 0;// Lookup route by NIx
  virtual Routing::RType_t RoutingType() = 0;      // Report type rtg in use
  virtual Interface* LocalRoute(IPAddr_t);         // If dst directly connected
  virtual void       InitializeRoutes() = 0;       // Routing init (if any)
  virtual Count_t    RoutingFibSize() const = 0;   // Statistics, size of FIB
  virtual Count_t    GetNix(Node*) const;          // Neighbor Index for node
  virtual RoutingNixVector* GetNixRouting() = 0;   // Returns Nix rtg pointer
  virtual Routing*   GetRouting() = 0;             // Returns routing object

  //AODV Routing
  //Specially for AODV routing, it is not compatiable with
  //other route entry.
  //
  virtual AODVRoutingEntry *LookupRouteAODV(IPAddr_t ip) =0;
  virtual void      SetRoutingAODV(void *pRouting)= 0;
  virtual RoutingAODV *GetRoutingAODV() = 0;

  // Protocol Graph Interface
  virtual Protocol* LookupProto(Layer_t, Proto_t) = 0;// Lookup proto by layer
  virtual void      InsertProto(Layer_t, Proto_t, Protocol*) = 0;

  // Port Demux Interface
  // Register port usage
  virtual bool      Bind(Proto_t, PortId_t, Protocol*) = 0;
  virtual bool      Bind(Proto_t,
                         PortId_t, IPAddr_t,
                         PortId_t, IPAddr_t,
                         Protocol*) = 0;           // Register port usage
  virtual PortId_t  Bind(Proto_t, Protocol*) = 0;  // Choose port, register
  virtual bool      Unbind(Proto_t, PortId_t, Protocol*) = 0;// Remove binding
  virtual bool      Unbind(Proto_t,
                           PortId_t, IPAddr_t,
                           PortId_t, IPAddr_t,
                           Protocol*) = 0;         // Remove port binding
  virtual Protocol* LookupByPort(Proto_t, PortId_t) = 0;
  virtual Protocol* LookupByPort(Proto_t,
                                 PortId_t, IPAddr_t,
                                 PortId_t, IPAddr_t) = 0;  // Lookup by s/d port/ip

  // Packet tracing interface
  // Trace header if enabled
  virtual bool      TracePDU(Protocol*, PDU*, Packet*, char*) = 0;
  // Set trace level this node
  virtual void      SetTrace(Trace::TraceStatus) = 0;

  // Node location (x, y, z) interface
  virtual void      SetLocation(Meters_t, Meters_t, Meters_t) = 0;
  virtual void      SetLocation(const Location&) = 0;
  virtual void      SetLocationLongLat(const Location&) = 0;
  virtual bool      HasLocation() = 0;       // True if location assigned
  virtual Meters_t  LocationX() = 0;
  virtual Meters_t  LocationY() = 0;
  virtual Meters_t  LocationZ() = 0;
  virtual Location  GetLocation() = 0;       // Return current location
  virtual Location  UpdateLocation() = 0;    // Get updated location
  virtual Mobility* AddMobility(const Mobility&) = 0;    // Add mobility model
  virtual Mobility* GetMobility() const = 0; // Get the current mobility model
  virtual bool      IsMobile() = 0;          // True if mobility enabled
  virtual bool      IsMoving() = 0;          // True if not pausing
  virtual bool      WirelessTx() = 0;        // True if wireless transmitting
  virtual bool      WirelessRx() = 0;        // True if wireless receiving
  virtual bool      WirelessCx() = 0;        // True if wireless colliding
  virtual bool      WirelessRxMe() = 0;      // True if wireless receiving
  virtual bool      WirelessRxZz() = 0;      // True if wireless receiving

  void UseWormContainment(bool);
  bool UseWormContainment() { return usewormcontainment; }
  virtual void SetWormContainment(WormContainment*) = 0;
  virtual WormContainment* GetWormContainment() = 0;

  // QTWindow information
  virtual void           Show(bool) = 0;          // Turn on/off display
  virtual bool           Show() = 0;              // True if display enabled
  virtual QCanvasItem*   Display(QTWindow*) = 0;
  virtual QCanvasItem*   Display(const QPoint&, QTWindow*) = 0;
  virtual void           WirelessTxColor(const QColor&) = 0;
  virtual const QColor&  WirelessTxColor() = 0;
  virtual bool           PushWirelessTx(QCanvasItem*) = 0;
  virtual QCanvasItem*   PopWirelessTx() = 0;
  virtual void           PixelSize(Count_t) = 0;
  virtual Count_t        PixelSizeX() = 0;
  virtual Count_t        PixelSizeY() = 0;
  virtual void           Shape(Node::Shape_t) = 0;
  virtual Node::Shape_t  Shape() = 0;
  virtual CustomShape_t  CustomShape();
  virtual void           CustomShape(CustomShape_t);
  virtual bool           CustomShapeFile(const char*);
  virtual void           Color(const QColor&) = 0;
  virtual bool           HasColor() = 0;
  virtual QColor         Color() = 0;
  virtual NodeAnimation* GetNodeAnimation() const = 0;

  // ICMP related
  virtual bool   ICMPEnabled() const;
  virtual void   DisableICMP();

  // Failure management
  virtual void   Down();
  virtual void   Up();
  virtual bool   IsDown();

  // Wireless transmitter information
  virtual Meters_t  Distance(Node*) = 0;
  virtual void      BuildRadioInterfaceList(void) = 0;
  virtual const RadioVec_t& GetRadioInterfaceList() = 0;
  virtual void      SetRadioRange(Meters_t ) =0;
  virtual Meters_t  GetRadioRange(void ) =0;

  virtual bool      CBEnable() = 0;
  virtual void      CBEnable(bool) = 0;
  virtual void      CBListAdd(IPCallBack) = 0;
  virtual void      CBListDelete(IPCallBack) = 0;
  virtual void      PreProcessCB(Packet*) = 0; // loop over the list
  virtual const CallBackList_t& GetCBList() = 0;

  virtual void      setBattery (Joules_t) = 0;
  virtual Joules_t  getBattery (void) = 0;

  virtual void      setComputePower (double) = 0;
  virtual double    getComputePower (void) = 0;

protected:
  Node*    pNode;                   // Points to corresponding node interface object
  bool     usewormcontainment;      // Does this node use worm containment
  // Data items common between real and ghost nodes
  IPAddr_t ipAddr;                  // IP Address for this node
  IFVec_t  interfaces;              // All defined network interfaces (may be ghosts)
  Count_t  neighborCount;           // Total number of neighbors
};

#endif
