/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Node class
// George F. Riley.  Georgia Tech, Spring 2002

#ifndef __nodereal_h__
#define __nodereal_h__

#include <map>
#include <deque>
#include <string>

#include "common-defs.h"
#include "node-impl.h"

#include "handler.h"
#include "routing.h"
#include "trace.h"
#include "macaddr.h"
#include "routing-aodv.h"

#ifdef HAVE_QT
#include <qcolor.h>
#endif

#define DEFAULT_SW_TIMEOUT 10   // default timeout for the switch map entries in seconds


class Link;
class Node;
class Interface;
class Routing;
class Protocol;
class ProtocolGraph;
class PortDemux;
class Linkp2p;
class Queue;
class Application;
class QCanvasItem;
class RoutingNixVector;
class Mobility;
class Node;
class QCanvasPolygonalItem;
class QCanvasSprite;

class NodeAnimation {
public:
  NodeAnimation() : show(false),
                    pixelSizeX(Node::DefaultPixelSize()),
                    pixelSizeY(Node::DefaultPixelSize()),
                    shape(Node::DefaultShape()),
                    csCallback(nil),
#ifdef HAVE_QT
                    color(Qt::black),
#else
                    color(),
#endif
                    hasColor(false), placeImage(false), nodePolygon(nil),
                    nodeSprite(nil) {}

public:
  bool          show;         // True if animation display desired
  Count_t       pixelSizeX;   // Size of node in pixels (X)
  Count_t       pixelSizeY;   // Size of node in pixels (Y)
  Node::Shape_t shape;        // Shape of node
  CustomShape_t csCallback;   // Custom Shape Callback
  QColor        color;        // Color for this node on animation display
  bool          hasColor;     // True if color specified
  bool          placeImage;   // True if image sprite needs to be moved
  std::string   imageFileName;// Name of custom image file
  QCanvasPolygonalItem*  nodePolygon;// Polygonal item (if not image)
  QCanvasSprite*         nodeSprite; // Image item (for IMAGE shape only)
};

typedef unsigned int Port_t;

class PortPair {
public:
  PortPair() : port(0), time(0) {}
  PortPair(Port_t p, Time_t t) : port(p), time(t) {}

public:
  Port_t  port;
  Time_t  time;
};

typedef std::map<MACAddr, PortPair> MACPort_t;


// Define a map for cacheing nixvector route lookups
typedef std::map<Count_t, RoutingEntry> RENixMap_t;
// And a vector of RoutingEntries for Nix lookup (obviates above)
typedef std::vector<RoutingEntry>       RENixVec_t;

// Define a deque for managing the QCanvasItem pointers for
// wireless animation
typedef std::deque<QCanvasItem*>        CanvasItemDeq_t;


//Doc:ClassXRef
class NodeReal : public NodeImpl {
public:
  NodeReal(Node*);
  ~NodeReal();
public:
  // IP Address Management is all done by node-impl

  // Real/Ghost test
  bool     IsReal();                     // True if node is real

  // Interface management is all done by node-impl

  // Broadcast a packet
  void       Broadcast(Packet*, Proto_t);// Broadcast packet on all interfaces

  // All link managment done by node-impl.h

  // Queues
  Queue*     GetQueue();                    // Return queue if only one i/f
  Queue*     GetQueue(Node*);               // Get the queue to specified node

  // Applications
  Application* AddApplication(const Application& a);

  // Neighbor management, done by node-impl

  // Routing
  void DefaultRoute(RoutingEntry);       // Specify default route
  void DefaultRoute(Node*);              // Specify default route
  void AddRoute(IPAddr_t, Count_t, Interface*, IPAddr_t);
  RoutingEntry LookupRoute(IPAddr_t);    // Find a routing entry
  RoutingEntry LookupRouteNix(Count_t);  // Lookup route by Neighbor Index
  Routing::RType_t RoutingType();        // Report type of routing in use
  // LocalRoute used from node-impl
  void       InitializeRoutes();         // Routing initialization (if any)
  Count_t    RoutingFibSize() const;     // Statistics, size of FIB
  // GetNix used from node-impl
  RoutingNixVector* GetNixRouting();     // Returns Nix rtg pointer (if Nix)
  Routing*   GetRouting() { return pRouting;} // Returns the routing object

  //AODV Routing
  //Specially for AODV routing, it is not compatiable with
  //other route entry.
  //
  AODVRoutingEntry *LookupRouteAODV(IPAddr_t);
  void      SetRoutingAODV(void *pRouting);
  RoutingAODV * GetRoutingAODV(); //return the RoutingAODV object pointer.

  // Protocol Graph Interface
  Protocol* LookupProto(Layer_t, Proto_t);// Lookup protocol by layer
  void      InsertProto(Layer_t, Proto_t, Protocol*); // Insert a protocol

  // Port Demux Interface
  bool      Bind(Proto_t, PortId_t, Protocol*);    // Register port usage
  bool      Bind(Proto_t,
                 PortId_t, IPAddr_t,
                 PortId_t, IPAddr_t,
                 Protocol*);                       // Register port usage
  PortId_t  Bind(Proto_t, Protocol*);              // Choose port and register
  bool      Unbind(Proto_t, PortId_t, Protocol*);  // Remove port binding
  bool      Unbind(Proto_t,
                   PortId_t, IPAddr_t,
                   PortId_t, IPAddr_t,
                   Protocol*);                     // Remove port binding
  Protocol* LookupByPort(Proto_t, PortId_t);       // Lookup by port
  Protocol* LookupByPort(Proto_t,
                         PortId_t, IPAddr_t,
                         PortId_t, IPAddr_t);      // Lookup by s/d port/ip

  // Packet tracing interface
  // Trace header if enabledP
  bool      TracePDU(Protocol*, PDU*, Packet*, char*);
  void      SetTrace(Trace::TraceStatus);          // Set trace level this node

  // Node location (x, y) interface
  void      SetLocation(Meters_t, Meters_t, Meters_t);  // Set x/y/z location
  void      SetLocation(const Location&);          // Set x/y/z location
  void      SetLocationLongLat(const Location&);   // Set x/y/z location
  bool      HasLocation();                         // True if location assigned
  Meters_t  LocationX();
  Meters_t  LocationY();
  Meters_t  LocationZ();
  Location  GetLocation();                         // Return current location
  Location  UpdateLocation();                      // Get updated location
  Mobility* AddMobility(const Mobility&);          // Add a mobility model
  Mobility* GetMobility() const;                   // Get the mobility model
  bool      IsMobile();                            // True if mobility enabled
  bool      IsMoving();                            // True if not pausing
  bool      WirelessTx();                          // True if wireless transmit
  bool      WirelessRx();                          // True if wireless receive
  bool      WirelessCx();                          // True if wireless collid
  bool      WirelessRxMe();                        // True if wireless receive
  bool      WirelessRxZz();                        // True if wireless receive

  // QTWindow information
  void           Show(bool);                        // Turn on/off display
  bool           Show();                            // True if display enabled
  QCanvasItem*   Display(QTWindow*);
  QCanvasItem*   Display(const QPoint&, QTWindow*);
  void           WirelessTxColor(const QColor&);
  const          QColor& WirelessTxColor();
  bool           PushWirelessTx(QCanvasItem*);
  QCanvasItem*   PopWirelessTx();
  void           PixelSize(Count_t);                // Set/query node size
  Count_t        PixelSizeX();
  Count_t        PixelSizeY();
  void           Shape(Node::Shape_t);              // Set/query node shape
  Node::Shape_t  Shape();
  CustomShape_t  CustomShape();                     // Custom Shape Callback
  void           CustomShape(CustomShape_t);        // Specify Sustom Shape CB
  bool           CustomShapeFile(const char*);      // Custom shape image file
  void           Color(const QColor&);
  bool           HasColor();
  QColor         Color();
  NodeAnimation* GetNodeAnimation() const;

  // ICMP related
  bool   ICMPEnabled() const;
  void   DisableICMP();

  // Failure management
  void   Down();
  void   Up();
  bool   IsDown();

  Meters_t  Distance(Node*);
  void      BuildRadioInterfaceList(void);
  const RadioVec_t& GetRadioInterfaceList();
  void      SetRadioRange(Meters_t range) { radioRange = range; }
  Meters_t  GetRadioRange() { return radioRange; }

  bool      CBEnable();
  void      CBEnable(bool);
  void      CBListAdd(IPCallBack);
  void      CBListDelete(IPCallBack);
  void      PreProcessCB(Packet*);
  void      setBattery(Joules_t joules) { battery = joules;}
  Joules_t  getBattery(void) { return battery; }
  void      setComputePower(double pow) { computePower = pow; }
  double    getComputePower(void) { return computePower; }

  const CallBackList_t& GetCBList();

  // Switching functions
  void        SetTimeout(Time_t);
  int         FindSRCPort(Interface*);
  void        PacketRX(Packet*, Interface*);
  MACPort_t*  GetPortMap(){ return &portmap; }
  Time_t      GetTimeout(){ return timeout; }
  bool        IsSwitchNode() { return isswitch; }  // True if node is a switch
  void        IsSwitchNode(bool);

  // worm containment
  void SetWormContainment(WormContainment*);
  WormContainment* GetWormContainment() { return wormcontainment; } 



private:
  bool        CheckTrace(Protocol*, PDU*);         // Check if trace enabled

  // Data Members
private:
  Routing*   pRouting;    // Points to Routing Object for this node
  //RENixMap_t nixCache;    // Cache for nix lookups
  RENixVec_t nixLookup;   // Cache for nix lookups

  //specially for AODV
  RoutingAODV*  pRoutingAODV;

  Meters_t x;             // x location
  Meters_t y;             // y location
  Meters_t z;             // z location
  bool     hasLocation;   // True if location assigned
#ifndef COMPACT
  ProtocolGraph* graph;   // Private protocol graph (if not null)
#endif
  Trace::TraceStatus traceStatus; // Trace level for this node
  Mobility* mobility;     // Current mobility model
public:
  NodeAnimation*   animInfo;
  PortDemux*       demux;           // Port Demultiplexer map
  Meters_t         radioRange;      // this belongs in intefaces, not nodes
  Joules_t         battery;         // in case node is powered by battery
  double           computePower;    // CPU power cocnsumption
  bool             enableCB;        // True if callbacks enabled
  RadioVec_t       radioInterfaces; // List of nodes in range
  RadioVec_t       radioInf2;       // List of nodes in range potentially
  Time_t           radioMarginTime;
  CallBackList_t   cbList;          // List of callbacks for this node
  CanvasItemDeq_t  canvasItems;     // The circles for animating  wireless tx
  QColor           wirelessColor;   // Color for wireless tx animation
  bool             icmpEnabled;     // True if ICMP enabled here, false if not
  bool             down;            // True if node is down

private:
  // switch members
  MACPort_t        portmap;
  Time_t           timeout;
  bool             isswitch;
  WormContainment* wormcontainment;

public:
  static Count_t count;    // debug
  static Count_t count1;   // debug
  static Count_t count2;   // debug
  static Count_t count3;   // debug
  static Count_t count4;   // debug
};

#include "interface.h"

#endif
