/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Node class
// George F. Riley.  Georgia Tech, Spring 2002

#include <iostream>
#include <math.h>

//#define DEBUG_MASK 0x02
#include "debug.h"
#include "node.h"
#include "node-impl.h"
#include "node-real.h"
#include "node-ghost.h"
#include "protograph.h"
#include "portdemux.h"
#include "ipaddr.h"
#include "duplexlink.h"
#include "application.h"
#include "mobility.h"
#include "routing-nixvector.h"
#include "interface-wireless.h"
#include "simulator.h"
#ifdef SIM//
#include "application-ospf6d.h"
#endif //SIM

using namespace std;

NodeId_t  Node::nextId = 0;  // Next id to assign
NodeVec_t Node::nodes;       // Vector of all known nodes
Count_t   Node::defaultPixelSize = 10;
Node::Shape_t Node::defaultShape = Node::SQUARE;
double    Node::defaultMaxSpeed = 20;  // Default max speed of a node
#ifdef SIM
vector<Node*> Node::NodeVecByRouterId(256);
#endif

// Comparison operator for locations
bool operator==(const Location& l, const Location& r)
{
  return (l.X() == r.X() && l.Y() == r.Y() && l.Z() == r.Z());
}

Node::Node(SystemId_t sysId) : id(nextId++)
{
  if (sysId != Simulator::SystemId())
    { // Mapped to a different logical process.  Use a ghost.
      pImpl = new NodeGhost(this);
    }
  else
    { // Real node
      pImpl = new NodeReal(this);
    }
#ifdef SIM
		ospf6d_inst = NULL;
#endif //SIM
  nodes.push_back(this);                  // Add to vector of known nodes
}

#ifdef REMOVE_THIS
Node::Node(IPAddr i, SystemId_t sysId) : id(nextId++)
{
  if (sysId != Simulator::SystemId())
    { // Mapped to a different logical process.  Use a ghost.
      pImpl = new NodeGhost(this);
    }
  else
    { // Real node
      pImpl = new NodeReal(this);
    }
  pImpl->SetIPAddr(i);
#ifdef SIM
		ospf6d_inst = NULL;
#endif //SIM
  nodes.push_back(this);                  // Add to vector of known nodes
}
#endif

Node::~Node()
{
  delete pImpl;                           // Delete the corresponding impl
#ifdef SIM
	delete ospf6d_inst;
#endif //SIM
}

// IP Address Management
IPAddr_t    Node::GetIPAddr()
{  // Get the IPAddres of this node
  return pImpl->GetIPAddr();   // Return IP Address....may be IPADDR_NONE
}

void        Node::SetIPAddr(IPAddr_t i)
{ // Set the IPAddres of this node
  pImpl->SetIPAddr(i);
}

void Node::IPAddrs(IPMaskVec_t& a)
{ // Create a list of IP Addresses
  pImpl->IPAddrs(a);
}

bool Node::LocalIP( IPAddr_t ip)
{ // Check if specified ip address is local
  return pImpl->LocalIP(ip);
}

bool Node::IPKnown()
{ // True if non-default ip is assigned
  return pImpl->IPKnown();
}

bool Node::IsReal()
{ // True if real node
  return pImpl->IsReal();
}

void Node::IsSwitchNode(bool b)
{ // True if real node
  return pImpl->IsSwitchNode(b);
}

bool Node::IsSwitchNode()
{ // True if real node
  return pImpl->IsSwitchNode();
}

void Node::PacketRX(Packet* p, Interface* iface)
{ // True if real node
  return pImpl->PacketRX(p, iface);
}

// Interfaces

Interface* Node::AddInterface(const L2Proto& l2, bool bootstrap)
{
  return pImpl->AddInterface(l2, bootstrap);
}

Interface* Node::AddInterface(const L2Proto& l2,
                              IPAddr_t i, Mask_t mask,
															MACAddr m, bool bootstrap)
{
  return pImpl->AddInterface(l2, i, mask, m, bootstrap);
}

#ifdef SIM
Interface* Node::AddInterface(const L2Proto& l2,
                              IPAddr_t i, Mask_t mask,
															MACAddr m, bool bootstrap,
															char *name)
{
  return pImpl->AddInterface(l2, i, mask, m, bootstrap, name);
}
#endif //SIM

Interface* Node::AddInterface(const L2Proto& l2,const Interface& iface,
                              IPAddr_t i, Mask_t mask, MACAddr m,
			      bool bootstrap)
{
  return pImpl->AddInterface(l2, iface, i, mask, m, bootstrap);
}

Count_t Node::InterfaceCount()
{
  return pImpl->InterfaceCount();
}

const IFVec_t& Node::Interfaces()
{ // Get the complete list of interfaces
  return pImpl->Interfaces();
}

Interface* Node::GetIfByLink(Link* l)
{
  return pImpl->GetIfByLink(l);
}

Interface* Node::GetIfByNode(Node* n)
{ // Find the interface used to get to the specified node
  return pImpl->GetIfByNode(n);
}

Interface* Node::GetIfByIP(IPAddr_t ip)
{ // Find the interface used to get to the specified node
  return pImpl->GetIfByIP(ip);
}

#ifdef SIM
Interface* Node::GetIfByName(char * name)
{ // Find the interface used to get to the specified node
  return pImpl->GetIfByName(name);
}

Interface* Node::GetIfByIndex(int index)
{ // Find the interface used to get to the specified node
  return pImpl->GetIfByIndex(index);
}


void Node::SetOspf6dInst(OSPF6DInstance *inst)
{
 ospf6d_inst   =  inst;
}

void Node::SetOspf6dGlobals()
{
  if (ospf6d_inst) 
		ospf6d_inst->SetGlobals();
	else
		{
			printf("Can not set ospf6 globals, exit ...\n");
			exit(0);
		}
}
#endif //SIM

void       Node::Broadcast(Packet* p, Proto_t proto)
{ // Broadcast a packet on all interfaces
  return pImpl->Broadcast(p, proto);
}

// Links.
// All of the following are just syntactic shortcuts to the DuplexLink object.
// Each creates a local DuplexLink object, which is destroyed on exit.
Interface* Node::AddDuplexLink(Node* rn)
{ // Add duplex link to/from remote
  return pImpl->AddDuplexLink(rn);
}

Interface* Node::AddDuplexLink(Node* rn, const Linkp2p& l)
{ // Add duplex link to/from remote
  return pImpl->AddDuplexLink(rn, l);
}

Interface* Node::AddDuplexLink(Interface* li, Interface* ri)
{ // Add a link between existing interfaces
  return pImpl->AddDuplexLink(li, ri);
}

Interface* Node::AddDuplexLink(Interface* li, Interface* ri,
                               const Linkp2p& l)
{
  return pImpl->AddDuplexLink(li, ri, l);
}

Interface* Node::AddDuplexLink(Node* rn,
                            IPAddr_t li, Mask_t lm,
                            IPAddr_t ri, Mask_t rm)
{
  return pImpl->AddDuplexLink(rn, li, lm, ri, rm);
}

Interface* Node::AddDuplexLink(Node* rn, const Linkp2p& l,
                            IPAddr_t li, Mask_t lm,
                            IPAddr_t ri, Mask_t rm)
{
  return pImpl->AddDuplexLink(rn, l, li, lm, ri, rm);
}

Link* Node::GetLink(Node* n)
{
  return pImpl->GetLink(n);
}

Interface* Node::AddRemoteLink(IPAddr_t i, Mask_t m)
{
  return pImpl->AddRemoteLink(i, m);
}

Interface* Node::AddRemoteLink(IPAddr_t i, Mask_t m, Rate_t r, Time_t d)
{
  return pImpl->AddRemoteLink(i, m, r, d);
}

// Simplex links, for modeling one-way links
Interface* Node::AddSimplexLink(Node* n)
{
  return pImpl->AddSimplexLink(n);
}

Interface* Node::AddSimplexLink(Node* n, const Linkp2p& l)
{
  return pImpl->AddSimplexLink(n, l);
}

Interface* Node::AddSimplexLink(Node* n, const Linkp2p& l, IPAddr_t i,Mask_t m)
{
  return pImpl->AddSimplexLink(n, l, i, m);
}

// Queue management
Queue*     Node::GetQueue()
{ // Return queue if only one i/f
  // This is simply a syntactic shortcut to GetQueue(Node*) below
  return pImpl->GetQueue();
}

Queue*     Node::GetQueue(Node* n)
{ // Get the queue to specified node
  return pImpl->GetQueue(n);
}

// Applications
Application* Node::AddApplication(const Application& a)
{
  return pImpl->AddApplication(a);
}


void Node::UseWormContainment(bool b) 
{
  return pImpl->UseWormContainment(b);
}


bool Node::UseWormContainment()
{ 
  return pImpl->UseWormContainment();
}

void Node::SetWormContainment(WormContainment* cont)
{
  return pImpl->SetWormContainment(cont);
}

WormContainment* Node::GetWormContainment()
{
  return pImpl->GetWormContainment();
}


// Neighbor management
void Node::Neighbors(NodeWeightVec_t& n, bool noLeaf)
{ // Ask each interface to add to neighbor list
  pImpl->Neighbors(n, noLeaf);
}

void Node::AddNeighbor(Node*, Weight_t)
{ // Non-ghost nodes do not keep neighbor lists
  // Need to think about this a bit more
}

Count_t Node::NeighborCount()
{ // Count number of routing neighbors
  // Done by summing routing neighbors for each interface
  return pImpl->NeighborCount();
}


void Node::NeighborsByIf(Interface* i, IPAddrVec_t& v)
{ // Get list of neighbors for a given interface
  return pImpl->NeighborsByIf(i, v);
}


// Routing Interface
void Node::DefaultRoute( RoutingEntry r)
{ // Specify default route
  pImpl->DefaultRoute(r);
}

void Node::DefaultRoute( Node* n)
{ // Specify default route
  pImpl->DefaultRoute(n);
}

void Node::AddRoute( IPAddr_t ip, Count_t mask, Interface* i, IPAddr_t ip1)
{ // Routing entry
  pImpl->AddRoute(ip, mask, i, ip1);
}

RoutingEntry Node::LookupRoute(IPAddr_t ip)
{ // Find a routing entry
  return pImpl->LookupRoute(ip);
}

RoutingEntry Node::LookupRouteNix(Count_t nix)
{ // Lookup route by Neighbor Index
  return pImpl->LookupRouteNix(nix);
}

//Specially for AODV routing
//
AODVRoutingEntry *
Node::LookupRouteAODV(IPAddr_t ip) {
	return pImpl->LookupRouteAODV(ip);
}

void
Node::SetRoutingAODV(void *pRouting){
	return pImpl->SetRoutingAODV(pRouting);
}

RoutingAODV *
Node::GetRoutingAODV(){
	return pImpl->GetRoutingAODV();
}

Routing::RType_t Node::RoutingType()
{
  return pImpl->RoutingType();
}

Interface* Node::LocalRoute(IPAddr_t ip)
{ // Find if destination ip is directly connected on any local interace
  return pImpl->LocalRoute(ip);
}

void       Node::InitializeRoutes()
{ // Routing initialization (if any)
  pImpl->InitializeRoutes();
}

Count_t    Node::RoutingFibSize() const
{ // Get the size of the routing FIB for statistics measurement
  return pImpl->RoutingFibSize();
}

Count_t    Node::GetNix(Node* n) const
{ // Get Neighbor Index for spec'd node
  return pImpl->GetNix(n);
}

Routing* Node::GetRouting()
{
  return pImpl->GetRouting();
}

RoutingNixVector* Node::GetNixRouting()
{ // Returns a pointer to the routing object if it's NixVector routing,
  // otherwise return nil
  return pImpl->GetNixRouting();
}

void Node::DumpNV()
{
  RoutingNixVector* r = pImpl->GetNixRouting();
  if (!r) return;
  r->DBDump(this);
}

// Protocol Graph Interface
#ifdef COMPACT
// No private protocol graphs in compact mode
Protocol* Node::LookupProto(Layer_t l, Proto_t p) // Lookup protocol by layer
{
  return pImpl->LookupProto(l, p);
}

void      Node::InsertProto(Layer_t l, Proto_t p, Protocol* pp) // Insert proto
{
  pImpl->InsertProto(l, p, pp);
}
#else
Protocol* Node::LookupProto(Layer_t l, Proto_t proto)
{ // Lookup protocol by layer
  return pImpl->LookupProto(l, proto);
}

void Node::InsertProto(Layer_t l, Proto_t p, Protocol* pp)
{ // Insert proto
  pImpl->InsertProto(l, p, pp);
}
#endif

// Port Demux Interface
bool Node::Bind(Proto_t proto, PortId_t port , Protocol* p)
{
  return pImpl->Bind(proto, port, p);
}

bool Node::Bind(Proto_t proto,
                PortId_t lport, IPAddr_t laddr,
                PortId_t rport, IPAddr_t raddr,
                Protocol* p)
{ // Register port usage
  return pImpl->Bind(proto, lport, laddr, rport, raddr, p);
}

PortId_t Node::Bind(Proto_t proto, Protocol* p)
{ // Choose available port and register
  return pImpl->Bind(proto, p);
}

bool  Node::Unbind(Proto_t proto, PortId_t port, Protocol* p)
{ // Remove port binding
  return pImpl->Unbind(proto, port, p);
}

bool  Node::Unbind(Proto_t proto,
                PortId_t lport, IPAddr_t laddr,
                PortId_t rport, IPAddr_t raddr,
                Protocol* p)
{ // Remove port binding
  return pImpl->Unbind(proto, lport, laddr, rport, raddr, p);
}

Protocol* Node::LookupByPort(Proto_t proto, PortId_t port)
{
  return pImpl->LookupByPort(proto, port);
}

Protocol* Node::LookupByPort(Proto_t proto,
                             PortId_t sport, IPAddr_t sip,
                             PortId_t dport, IPAddr_t dip)
{
  return pImpl->LookupByPort(proto, sport, sip, dport, dip);
}

// Packet tracing interface
// Trace header if enabled
bool Node::TracePDU(Protocol* proto, PDU* h, Packet* p, char* s)
{
  return pImpl->TracePDU(proto, h, p, s);
}

void Node::SetTrace(Trace::TraceStatus t)        // Set trace level this node
{
  pImpl->SetTrace(t);
}

// Node Location Interface
void Node::SetLocation(Meters_t xl, Meters_t yl, Meters_t zl)
{ // Set x/y/z location
  pImpl->SetLocation(xl, yl, zl);
}

void Node::SetLocation(const Location& l)
{
  pImpl->SetLocation(l);
}

void Node::SetLocationLongLat(const Location& l)
{
  pImpl->SetLocationLongLat(l);
}

bool Node::HasLocation()
{
  return pImpl->HasLocation();
}

Meters_t Node::LocationX()
{
  return pImpl->LocationX();
}

Meters_t Node::LocationY()
{
  return pImpl->LocationY();
}

Meters_t Node::LocationZ()
{
  return pImpl->LocationZ();
}

Location Node::GetLocation()
{
  return pImpl->GetLocation();
}

Location Node::UpdateLocation()
{
  return pImpl->UpdateLocation();
}

Mobility* Node::AddMobility(const Mobility& m)
{ // Add a new mobility model for this node
  return pImpl->AddMobility(m);
}

Mobility* Node::GetMobility()
{
  return pImpl->GetMobility();
}

bool Node::IsMoving()
{
  return pImpl->IsMoving();
}

bool Node::IsMobile()
{
  return pImpl->IsMobile();
}

void Node::Show(bool s)                          // Turn on/off display
{
  pImpl->Show(s);
}

bool Node::Show()                                // True if display enabled
{
  return pImpl->Show();
}

QCanvasItem* Node::Display(QTWindow* qc)
{
  return pImpl->Display(qc);
}

QCanvasItem* Node::Display(const QPoint& qp, QTWindow* qc)
{
  return pImpl->Display(qp, qc);
}

void Node::WirelessTxColor(const QColor& c)
{
  pImpl->WirelessTxColor(c);
}

const QColor& Node::WirelessTxColor()
{
  return pImpl->WirelessTxColor();
}

bool Node::PushWirelessTx(QCanvasItem* c)
{
  return pImpl->PushWirelessTx(c);
}

QCanvasItem* Node::PopWirelessTx()
{
  return pImpl->PopWirelessTx();
}

void Node::PixelSize(Count_t nps)  // Set node size
{
  pImpl->PixelSize(nps);
}

Count_t Node::PixelSizeX()
{
  return pImpl->PixelSizeX();
}

Count_t Node::PixelSizeY()
{
  return pImpl->PixelSizeY();
}

void Node::Shape(Shape_t ns) // Set node shape
{
  pImpl->Shape(ns);
}

Node::Shape_t Node::Shape()
{
  return pImpl->Shape();
}

CustomShape_t Node::CustomShape()
{
  return pImpl->CustomShape();
}

void Node::CustomShape(CustomShape_t cs)
{
  pImpl->CustomShape(cs);
}

bool Node::CustomShapeFile(const char* fn)
{
  return pImpl->CustomShapeFile(fn);
}

void Node::Color(const QColor& c)
{
#ifdef HAVE_QT
  pImpl->Color(c);
#endif
}

bool Node::HasColor()
{
#ifdef HAVE_QT
  return pImpl->HasColor();
#else
  return false;
#endif
}

#ifndef HAVE_QT
static QColor noColor;
#endif

QColor Node::Color()
{
#ifdef HAVE_QT
  return pImpl->Color();
#else
  return noColor;
#endif
}

NodeAnimation* Node::GetNodeAnimation() const
{
  return pImpl->GetNodeAnimation();
}

bool Node::ICMPEnabled() const
{
  return pImpl->ICMPEnabled();
}

void Node::DisableICMP()
{
  pImpl->DisableICMP();
}

void Node::Down()
{
  pImpl->Down();
}

void Node::Up()
{
  pImpl->Up();
}

bool Node::CBEnable()
{
  return pImpl->CBEnable();
}

void Node::CBEnable(bool e)
{
  pImpl->CBEnable(e);
}

void Node::CBListAdd(IPCallBack cb)
{
  pImpl->CBListAdd(cb);
}

void Node::CBListDelete(IPCallBack cb)
{
  pImpl->CBListDelete(cb);
}

void Node::PreProcessCB(Packet* p)
{
  pImpl->PreProcessCB(p);
}

const CallBackList_t& Node::GetCBList()
{
  return pImpl->GetCBList();
}

bool Node::WirelessTx()
{
  return pImpl->WirelessTx();
}

bool Node::WirelessRx()
{
  return pImpl->WirelessRx();
}

bool Node::WirelessCx()
{
  return pImpl->WirelessCx();
}

bool Node::WirelessRxMe()
{
  return pImpl->WirelessRxMe();
}

bool Node::WirelessRxZz()
{
  return pImpl->WirelessRxZz();
}

Meters_t Node::Distance(Node* n)
{
  return pImpl->Distance(n);
}

void Node::BuildRadioInterfaceList(void)
{
  pImpl->BuildRadioInterfaceList();
}

const RadioVec_t& Node::GetRadioInterfaceList()
{
  return pImpl->GetRadioInterfaceList();
}

void Node::SetRadioRange(Meters_t range)
{
  pImpl->SetRadioRange(range);
}

Meters_t Node::GetRadioRange()
{
  return pImpl->GetRadioRange();
}

Joules_t Node::getBattery()
{
  return pImpl->getBattery();
}

void Node::setBattery(Joules_t joules)
{
  return pImpl->setBattery(joules);
}

void Node::setComputePower(double pow)
{
  return pImpl->setComputePower(pow);
}

double Node::getComputePower(void)
{
  return pImpl->getComputePower();
}

// Static Methods

Node* Node::GetNode(NodeId_t i)
{
  if (i >= nodes.size()) return nil; // Out of range
  return nodes[i];
}

void Node::Clear()
{
  nodes.clear();
}

void Node::DefaultPixelSize(Count_t dps)
{
  defaultPixelSize = dps;
}

Count_t Node::DefaultPixelSize()
{
  return defaultPixelSize;
}

void Node::DefaultShape(Shape_t ds)
{
  defaultShape = ds;
}

Node::Shape_t Node::DefaultShape()
{
  return defaultShape;
}

void Node::DefaultMaxSpeed(double speed)
{
  defaultMaxSpeed = speed;
}

double Node::DefaultMaxSpeed()
{
  return defaultMaxSpeed;
}


// The following is for Proxy Routing alogorithms

void Node::SetProxyRoutingConfig(IPAddr_t ip, Mask mask)
{
  proxyIP = ip;
  proxyMask = mask;
}

bool Node::CanProxyRouteIP(IPAddr_t ip)
{
  return (proxyIP & proxyMask) == (ip & proxyMask);
}

IPAddr_t Node::GetProxyIP()
{
  return proxyIP;
}

Mask Node::GetProxyMask()
{
  return proxyMask;
}

int Node::GetLongestPrefixLength(IPAddr_t ip)
{
  int left = 0, right = proxyMask.NBits(), m;
  Mask_t mask1, mask2;

  if (!HasProxyRoutingConfig()) return 0;

  // I have used a binary search method that works in O(logN) time (max(N)=32)

  do {
    m=(left+right)/2;
    if (m==32) break;

    mask1 = 0xFFFFFFFF << (32-m);
    if ((proxyIP & mask1) != (ip & mask1))
      right = m-1;
    else {
      mask2 = 0xFFFFFFFF << (32- (m+1));
      if ((proxyIP & mask2) == (ip & mask2))
        left = m+1;
      else
        break; // we have found it
    }
  } while (left<=right);
  return m;
}
