/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Node class
// George F. Riley.  Georgia Tech, Spring 2002

#ifndef __node_h__
#define __node_h__

#include <vector>
#include <deque>
#include <list>

#include "common-defs.h"
#include "l2proto802.3.h"
#include "mask.h"
#include "location.h"
//#include "worm-containment.h"


class Link;
class Node;
class Interface;
class InterfaceWireless;
class Routing;
class Protocol;
class ProtocolGraph;
class PortDemux;
class Linkp2p;
class Queue;
class Application;
class NodeImpl;
class RoutingAODV;
class AODVRoutingEntry;
class QTWindow;
class NodeAnimation;
class WormContainment;
// Definitions from Qt library.  Unused if no Qt configured.
class  QCanvasPolygonalItem;
class  QCanvasItem;
class  QPoint;
class  QCanvas;
class  QColor;
#ifdef SIM
class OSPF6DInstance;
#endif //SIM

#ifndef HAVE_QT
class QColor
{
public:
  QColor(){}
};
#endif

// For radio interface list
class RadioItem
{
public:
  RadioItem() : inf(nil), dist(0) {}
  RadioItem(InterfaceWireless* i, Meters_t d) : inf(i), dist(d) {}

public:
  InterfaceWireless* inf;
  Meters_t           dist;
};

//typedef std::vector<RadioItem> RadioVec_t;
typedef std::list<RadioItem> RadioVec_t;

//Doc:ClassXRef
class NodeIfWeight { // Used by routing protocols
public:
  NodeIfWeight(Node* n, Interface* i, Weight_t w)
    : node(n), iface(i), weight(w) { }
public:
  Node*      node;
  Interface* iface;
  Weight_t   weight;
};

//Doc:ClassXRef
class NodeIf { // Node and Interface used to get there, used by routing
public:
  NodeIf(Node* n, Interface* i)
    : node(n), iface(i) { }
public:
  Node*      node;
  Interface* iface;
};

//Doc:ClassXRef
class IpMask { // IP Address and mask
public:
  IpMask(IPAddr_t i, Mask_t m)
    : ip(i), mask(m) { }
public:
  IPAddr_t ip;
  Mask_t   mask;
};

typedef std::vector<Link*>        LinkVec_t;
typedef std::vector<Node*>        NodeVec_t;
typedef std::deque<NodeIf>        NodeDeque_t;
typedef std::vector<IPAddr_t>     IPAddrVec_t;
typedef std::vector<IpMask>       IPMaskVec_t;
typedef std::vector<Interface*>   IFVec_t;
typedef std::vector<NodeIfWeight> NodeWeightVec_t; // Neighbors,Iface w/weights
typedef std::vector<NodeIf>       NodeIfVec_t;     // Nodes and Iface to reach
typedef std::vector<Dist_t>       DistVec_t;
  // For functionality like bpf pseudodevice
typedef int (*IPCallBack)(Packet*); // define a type of callback
typedef std::list<IPCallBack> CallBackList_t;

// Define a callback for custom shapes in animation
typedef QCanvasPolygonalItem*(*CustomShape_t)(QCanvas*, Node*,
                                              const QPoint&,
                                              QCanvasPolygonalItem*);

const Dist_t   INF  = 0xffffffff;

#include "routing.h"
#include "trace.h"
#include "macaddr.h"

class RoutingNixVector;
class Mobility;
class Packet;
//Doc:ClassXRef
class Node {
  //class Node {
  //Doc:Class Objects of class {\tt Node} are used in \GTNS\ to represent all
  //Doc:Class nodes in the simulated topology.  Nodes represent end-- systems,
  //Doc:Class wireless devides, hubs, and routers.  When a node is created, it
  //Doc:Class automatically has an associated routing object (the default
  //Doc:Class is \NV\ routing) and a layer 3 protocol object (the default
  //Doc:Class is {\em IPV4}).  After nodes are created, interfaces and links
  //Doc:Class can be added to describe the connectivity of nodes, using the
  //Doc:Class public members described below.

public:
  // Define the node shapes for animation
  typedef enum { NONE, CIRCLE, SQUARE, HEXAGON, OCTAGON,
                 CUSTOM, // Displayed by the custom callback
                 IMAGE   // Displayed with a Qt image file
  } Shape_t;

public:
  //Doc:Method
  Node(SystemId_t = 0);
    //Doc:Desc Creates a node in the simulated topology. The optional parameter
    //Doc:Desc is useful only for distributed simulations, and specifies
    //Doc:Desc which  simulator is responsible for maintaining the state
    //Doc:Desc for this node.  If the id specified matches the system
    //Doc:Desc id specified on the {\tt Simulator} constructor, then
    //Doc:Desc a full--state node is created.  If not, a reduced state
    //Doc:Desc node, with only connectivity information, is created.
    //Doc:Arg1 The system responsible for this node (distributed simulations
    //Doc:Arg1 only.

#ifdef REMOVE_THIS
  //Doc:Method
  Node(IPAddr, SystemId_t sysId = 0);
    //Doc:Desc Creates a node with the specified \IPA.  Normally, the \IPA\ for
    //Doc:Desc a node is assigned to be the \IPA\ of the first associated
    //Doc:Desc interface.  This contructor allows for assignment of a
    //Doc:Desc specific \IPA\ to the node.
    //Doc:Arg1 \IPA\ assigned to this node.
    //Doc:Arg2 The system responsible for this node (distributed simulations
    //Doc:Arg2 only.
#endif
  virtual ~Node();
public:
  //Doc:Method
  NodeId_t Id() { return id;}
    //Doc:Desc Each node in a \GTNS\ simulation has an associated unique
    //Doc:Desc integer identifier, starting from zero for the first
    //Doc:Desc node created.
    //Doc:Return The node identifier for this node.

  // IP Address Management
  //Doc:Method
  IPAddr_t GetIPAddr();                  // Get the IPAddres of this node
    //Doc:Desc Returns the \IPA\ assigned to this node.  Nodes can have
    //Doc:Desc multiple \IPA{s}, since \IPA{s} are assigned to interfaces and
    //Doc:Desc nodes can have multiple interfaces.  This returns the first
    //Doc:Desc \IPA\ assigned to this node.
    //Doc:Return The first \IPA\ assigned to this node.

  //Doc:Method
  void     SetIPAddr(IPAddr_t);          // Set the IPAddres of this node
    //Doc:Desc Specifies an \IPA\ for this node.
    //Doc:Arg1 The \IPA\ to be assigned to this node.

  // Return vector of all known IPAddr's and associated masks for this node
  //Doc:Method
  void     IPAddrs(IPMaskVec_t&);
    //Doc:Desc Each interface assigned to a node has an associated \IPA\ and
    //Doc:Desc address mask.  This method returns a vector of the \IPA\ and
    //Doc:Desc masks for each interface on this node.
    //Doc:Arg1 An {\em out} parameter that is populated with the
    //Doc:Arg1 \IPA\ and mask information.

  //Doc:Method
  bool     LocalIP(IPAddr_t);            // True if specified ip is local
    //Doc:Desc Determines if the specified \IPA\ is local to this node.
    //Doc:Arg1 The \IPA\ to check for locality.

  //Doc:Method
  bool     IPKnown();                    // True if non-default ip is assigned
    //Doc:Desc Determines if this node has an \IPA\ assigned.

  //Doc:Method
  bool     IsReal();                     // True if node is not a Ghost
    //Doc:Desc   Find out if this node is "real", versus a Ghost
    //Doc:Return True if Real node, false if Ghost.

 //Doc:Method
  bool     IsSwitchNode();                     // True if node is a switch
  //Doc:Desc   Find out if this node is switch.

 //Doc:Method
  virtual  void IsSwitchNode(bool);
  //Doc:Desc   Set the node to a switch or not

 //Doc:Method
  virtual void PacketRX(Packet*, Interface*);
  //Doc:Desc   called when a packet is receive on a switch node


  // Interfaces
  //Doc:Method
  Interface* AddInterface(const L2Proto& = L2Proto802_3(),
			  bool bootstrap=false);
    //Doc:Desc Assign a new interface to this node, with no associated
    //Doc:Desc \IPA\ or mask.
    //Doc:Return A pointer to the newly assigned interface.
    //Doc:Arg1 A layer 2 protocol object to be assigned to this interface.
    //Doc:Arg1 The object passed in is {\em copied}, so the same object
    //Doc:Arg1 may be passed to multiple interfaces without problems.

  //Doc:Method
  Interface* AddInterface(const L2Proto&,
                          IPAddr_t, Mask_t, MACAddr m = MACAddr::Allocate(),
													bool bootstrap=false);
  //Doc:Desc Assign a new interface to this node, with the  specified
  //Doc:Desc \IPA\ and mask.
  //Doc:Return A pointer to the newly assigned interface.
  //Doc:Arg1 A layer 2 protocol object to be assigned to this interface.
  //Doc:Arg1 The object passed in is {\em copied}, so the same object
  //Doc:Arg1 may be passed to multiple interfaces without problems.
  //Doc:Arg2 The \IPA\ assigned to this interface.
  //Doc:Arg3 The address mask assigned to this interface.
  //Doc:Arg4 The {\em MAC} address assigned (normally the default
  //Doc:Arg4 should be used)

#ifdef SIM
  Interface* AddInterface(const L2Proto&,
                          IPAddr_t, Mask_t, MACAddr,
													bool bootstrap, char *name);
#endif //SIM

  //Doc:Method
  Interface* AddInterface(const L2Proto&,const Interface&,
			  IPAddr_t, Mask_t,
			  MACAddr = MACAddr::Allocate(),
			  bool bootstrap = false);
  //Doc:Desc Assign a new interface to this node, with the  specified
  //Doc:Desc \IPA\ and mask.
  //Doc:Return A pointer to the newly assigned interface.
  //Doc:Arg1 A layer 2 protocol object to be assigned to this interface.
  //Doc:Arg1 The object passed in is {\em copied}, so the same object
  //Doc:Arg1 may be passed to multiple interfaces without problems.
  //Doc:Arg2 Reference to what type of Interface to add
  //Doc:Arg3 The \IPA\ assigned to this interface.
  //Doc:Arg4 The address mask assigned to this interface.
  //Doc:Arg5 The {\em MAC} address assigned (normally the default
  //Doc:Arg5 should be used)



  //Interface* LocalSubnet(IPAddr_t);    // Determine if local subnet
  //Doc:Method
  Count_t    InterfaceCount();           // Get count of interfaces
    //Doc:Desc Returns the number of interfaces assigned to this node.
    //Doc:Return The number of interfaces assigned to this node.

  //Doc:Method
  const IFVec_t& Interfaces();           // Get the complete list of interfaces
    //Doc:Desc Returns a reference to a vector of all interfaces assigned
    //Doc:Desc to this node.
    //Doc:Return A vector of interface pointers for each interfae assigned
    //Doc:Return to this node.

  //Doc:Method
  Interface* GetIfByLink(Link*);         // Get interface to the specified link
    //Doc:Desc Returns the interface associated with the specified link.
    //Doc:Arg1 A pointer to the link object to search for.
    //Doc:Return The interface pointer associated with the specified link.
    //Doc:Return Returns {\em nil} if not found.

  //Doc:Method
  Interface* GetIfByNode(Node*);         // Get interface to the specified node
    //Doc:Desc Returns the interface used to communicate with the specified
    //Doc:Desc node.
    //Doc:Arg1 A pointer to the node to search for.
    //Doc:Return A pointer to the interface used to communicate with the
    //Doc:Return specified node.  Returns {\em nil} if not found.

  //Doc:Method
  Interface* GetIfByIP(IPAddr_t);        // Get interface by IP Address
    //Doc:Desc Returns the interface with the specified \IPA.
    //Doc:Arg1 The \IPA\ to find.
    //Doc:Return A pointer to the interface with the specified \IPA.
    //Doc:Return Returns {\em nil} if not found.

#ifdef SIM
  //Doc:Method
Interface* GetIfByName(char * name);  // Get Interface by Name
    //Doc:Desc Returns the interface with the specified Name.
    //Doc:Arg1 The Name to find.
    //Doc:Return A pointer to the interface with the specified Name.
    //Doc:Return Returns {\em nil} if not found.
  
  //Doc:Method
Interface* GetIfByIndex(int index);  // Get Interface by Index
    //Doc:Desc Returns the interface with the specified Index.
    //Doc:Arg1 The Index to find.
    //Doc:Return A pointer to the interface with the specified Index.
    //Doc:Return Returns {\em nil} if not found.

  //Doc:Method
void SetOspf6dInst(OSPF6DInstance *);  // Set ospf6 inst pointer
    //Doc:Desc Set ospf6 inst pointer
    //Doc:Arg1 the pointer to set
    //Doc:Return nothing

OSPF6DInstance * GetOspf6dInst() {return ospf6d_inst;}

  //Doc:Method
//void Node::SetOspf6dGlobals();  // Set ospf6 globals
void SetOspf6dGlobals();  // Set ospf6 globals
    //Doc:Desc Set ospf6 global variables
    //Doc:Arg NONE
    //Doc:Return nothing

  static void RouterIdToNodeInit(int idx, Node* n) { 
                                                NodeVecByRouterId[idx] = n; }
  static Node* RouterIdToNode(int idx) { return NodeVecByRouterId[idx]; }
#endif //SIM
	
  //Doc:Method
  void       Broadcast(Packet*, Proto_t);// Broadcast packet on all interfaces
    //Doc:Desc Broadcasts a packet on all associated interfaces.
    //Doc:Arg1 The packet to send.  Each interface gets a separate copy of the
    //Doc:Arg1 packet, so the caller is still responsible for the memory
    //Doc:Arg1 for the packet, and should delete this packet when no
    //Doc:Arg1 longer needed.
    //Doc:Arg2 The layer 3 protocol number to insert in the layer 2
    //Doc:Arg2 header for thiis packet.

  // Links.  All return local interface
  // Add duplex link to/from remote
  //Doc:Method
  Interface* AddDuplexLink(Node*);
    //Doc:Desc Adds a new duplex link (and  an associated interface) to
    //Doc:Desc this node, using the default link type.
    //Doc:Arg1 A node pointer to the peer node for this link.
    //Doc:Return A pointer to the  newly added  interface
    //Doc:Return on the local node.

  //Doc:Method
  Interface* AddDuplexLink(Node*, const Linkp2p&);
    //Doc:Desc Adds a new duplex link (and  an associated interface) to
    //Doc:Desc this node, using the specified link type.
    //Doc:Arg1 A node pointer to the peer node for this link.
    //Doc:Arg2 The type of link to add. The passed argument is copied,
    //Doc:Arg2 so an anonymous temporary object can be passed, or
    //Doc:Arg2 the same link object can be passed multiple times.
    //Doc:Return A pointer to the  newly added  interface
    //Doc:Return on the local node.

  // Next two specify new links between existing interfaces
  //Doc:Method
  Interface* AddDuplexLink(Interface*,Interface*);
    //Doc:Desc Add a new duplex link between two existing interfaces,
    //Doc:Desc using the default link type.
    //Doc:Arg1 First existing interface.
    //Doc:Arg2 Second existing interface.
    //Doc:Return A pointer to the first interface specified.

  //Doc:Method
  Interface* AddDuplexLink(Interface*,Interface*, const Linkp2p&);
    //Doc:Desc Add a new duplex link between two existing interfaces,
    //Doc:Desc using the specified link object type.
    //Doc:Arg1 First existing interface.
    //Doc:Arg2 Second existing interface.
    //Doc:Arg3 The link object to use.  This object is copied, so an
    //Doc:Arg3 anonymous temporary object can be used, or the same
    //Doc:Arg3 object can be passed multiple times.
    //Doc:Return A pointer to the first interface specified.

  // Next two specify local and remote IPAddress and Masks
  //Doc:Method
  Interface* AddDuplexLink(Node*,
                           IPAddr_t, Mask_t = Mask(32),
                           IPAddr_t = IPADDR_NONE, Mask_t = Mask(32));
    //Doc:Desc Adds a new duplex link (and  an associated interface) to
    //Doc:Desc this node, using the default link type, and specifying
    //Doc:Desc the \IPA\ and mask for both new interfaces.
    //Doc:Arg1 A node pointer to the peer node for this link.
    //Doc:Arg2 The \IPA\ for the local interface.
    //Doc:Arg3 The address mask for the local interface.
    //Doc:Arg4 The \IPA\ for the remote interface.
    //Doc:Arg5 The address mask for the remote interface.
    //Doc:Return A pointer to the  newly added  interface
    //Doc:Return on the local node.

  //Doc:Method
  Interface* AddDuplexLink(Node*, const Linkp2p&,
                           IPAddr_t, Mask_t = Mask(32),
                           IPAddr_t = IPADDR_NONE, Mask_t = Mask(32));
    //Doc:Desc Adds a new duplex link (and  an associated interface) to
    //Doc:Desc this node, using the specified link type, and specifying
    //Doc:Desc the \IPA\ and mask for both new interfaces.
    //Doc:Arg1 A node pointer to the peer node for this link.
    //Doc:Arg2 The type of link to add. The passed argument is copied,
    //Doc:Arg2 so an anonymous temporary object can be passed, or
    //Doc:Arg2 the same link object can be passed multiple times.
    //Doc:Arg3 The \IPA\ for the local interface.
    //Doc:Arg4 The address mask for the local interface.
    //Doc:Arg5 The \IPA\ for the remote interface.
    //Doc:Arg6 The address mask for the remote interface.
    //Doc:Return A pointer to the  newly added  interface
    //Doc:Return on the local node.

  // Next two add "LinkRTI" objects for distributed simulations
  //Doc:Method
  Interface* AddRemoteLink(IPAddr_t, Mask_t);
    //Doc:Desc Used for distributed simulations only.  Defines a link from
    //Doc:Desc this node to another node defined and managed on another
    //Doc:Desc simulator process.  The default link bandwidth and
    //Doc:Desc link delay are used.
    //Doc:Arg1 The \IPA\ for the local interface.
    //Doc:Arg2 The address mask for the local interface.

  //Doc:Method
  Interface* AddRemoteLink(IPAddr_t, Mask_t, Rate_t, Time_t);
    //Doc:Desc Used for distributed simulations only.  Defines a link from
    //Doc:Desc this node to another node defined and managed on another
    //Doc:Desc simulator process.
    //Doc:Arg1 The \IPA\ for the local interface.
    //Doc:Arg2 The address mask for the local interface.
    //Doc:Arg3 The bandwidth for the remote link.
    //Doc:Arg4 The propagation delay for the remote link.

  //Doc:Method
  Link*      GetLink(Node*);                // Get link to specified node
    //Doc:Desc Returns a pointer to the link used to communicate with
    //Doc:Desc the specified node.
    //Doc:Arg1 Node pointer to find a link for.
    //Doc:Return The link pointer for the link used to communicate with
    //Doc:Return the specified node.  Returns {\em nil} is none is found.

  // Simplex links, for modeling one-way links
  //Doc:Method
  Interface* AddSimplexLink(Node*);
    //Doc:Desc Add a simplex link from this node to the specified node, using
    //Doc:Desc the default link object and with no \IPA\ or mask.
    //Doc:Arg1 The node pointer for the remote node.
    //Doc:Return An interface pointer for the local interface added
    //Doc:Return for this link.

  //Doc:Method
  Interface* AddSimplexLink(Node*, const Linkp2p&);
    //Doc:Desc Add a simplex link from this node to the specified node, using
    //Doc:Desc the specified link object and with no \IPA\ or mask.
    //Doc:Arg1 The node pointer for the remote node.
    //Doc:Arg2 The link object to copy for this simplex link.
    //Doc:Return An interface pointer for the local interface added
    //Doc:Return for this link.

  //Doc:Method
  Interface* AddSimplexLink(Node*, const Linkp2p&, IPAddr_t, Mask_t);
    //Doc:Desc Add a simplex link from this node to the specified node, using
    //Doc:Desc the specified link object and with
    //Doc:Desc the specified \IPA\ and mask.
    //Doc:Arg1 The node pointer for the remote node.
    //Doc:Arg2 The link object to copy for this simplex link.
    //Doc:Arg3 The \IPA\ for the local interface.
    //Doc:Arg4 The address mask for the local interface.
    //Doc:Return An interface pointer for the local interface added
    //Doc:Return for this link.

  // Queues
  //Doc:Method
  Queue*     GetQueue();                    // Return queue if only one i/f
    //Doc:Desc Get a pointer to the queue object for this node.  Since nodes
    //Doc:Desc can have multiple interfaces and queues, this is useful
    //Doc:Desc only for nodes with a single interface.
    //Doc:Return A pointer to the single queue for this node.
    //Doc:Return Returns {\em nil}
    //Doc:Return if no interfaces, or more than one interface.

  //Doc:Method
  Queue*     GetQueue(Node*);               // Get the queue to specified node
    //Doc:Desc Get a pointer to the queue object used to buffer packets
    //Doc:Desc destined for the specified node.
    //Doc:Arg1 The target node to locate the queue for.
    //Doc:Return A pointer to the queue object uesd to buffer packets
    //Doc:Return destined for the specified node.  Returns {\em nil} if
    //Doc:Return none is found.

  // Applications
  //Doc:Method
  Application* AddApplication(const Application& a);
    //Doc:Desc Adds an application associated with this node.
    //Doc:Arg1 The application object to use.
    //Doc:Arg1 This application object is copied,
    //Doc:Arg1 so an anonymous temporary object can be specified, or the
    //Doc:Arg1 same application object can be used multiple times.
    //Doc:Return A pointer to the newly added application.

  // Neighbor management (used by some routing protocols)
  //Doc:Method
  void        Neighbors(NodeWeightVec_t &, bool = false); // Get list of neighbors
    //Doc:Desc Get a list of all neighbors for this node.
    //Doc:Desc An output parameter, returning a vector of nodes and link
    //Doc:Arg1 weights to each directly connected neighbor.
    //Doc:Arg1 True if skip leaf neighbors desired

  void        AddNeighbor(Node*, Weight_t); // Ghost nodes keep neighbor list
  //Doc:Method
  Count_t     NeighborCount();              // Get count of routing neighbors
    //Doc:Desc Get a count of directly connected neighbors.
    //Doc:Desc     //Doc:Arg1 The count of directly connected neighbors.

  // Get neighbors on given iface
  //Doc:Method
  void        NeighborsByIf(Interface*, IPAddrVec_t&);
    //Doc:Desc Get a list of \IPA{s} neighbors connected using the specified
    //Doc:Desc interface.
    //Doc:Arg1 The interface to return the neighbors for.
    //Doc:Arg2 An output parameter giving a vector of \IPA{s} connected
    //Doc:Arg2 on the specified interface.

  // Routing
  //Doc:Method
  void DefaultRoute(RoutingEntry);    // Specify default route
    //Doc:Desc Specifies the default route for packets leaving
    //Doc:Desc this node.
    //Doc:Arg1 The routing entry (interface and \IPA) to use when
    //Doc:Arg1 no other routing information can be found.

  //Doc:Method
  void DefaultRoute(Node*);    // Specify default route
    //Doc:Desc Specifies the default route for packets leaving
    //Doc:Desc this node.
    //Doc:Arg1 The next hop neighbor to use for routing when
    //Doc:Arg1 no other routing information can be found.

  //Doc:Method
  void AddRoute(IPAddr_t, Count_t, Interface*, IPAddr_t);
    //Doc:Desc Add a new routing table entry.
    //Doc:Arg1 The \IPA\ specifying the destination address this
    //Doc:Arg1 routing entry  is for.
    //Doc:Arg2 The mask for the destination \IPA, which allows
    //Doc:Arg2 route aggregation.
    //Doc:Arg3 The interface for the next hop target for the
    //Doc:Arg3 specified destination.
    //Doc:Arg4 The \IPA\ of the next hop target.

  //Doc:Method
  RoutingEntry LookupRoute(IPAddr_t);    // Find a routing entry
    //Doc:Desc Look up the routing table entry for a specified destination
    //Doc:Desc \IPA.
    //Doc:Arg1 The destination \IPA\ to look up.
    //Doc:Return The routing table entry for the specified destination.
    //Doc:Return The interface pointer in this entry will be {\em nil}
    //Doc:Return if no routing information is found for the specified
    //Doc:Return destination.

  //Doc:Method
  RoutingEntry LookupRouteNix(Count_t);  // Lookup route by Neighbor Index
    //Doc:Desc Look up the routing table entry by a neighbor index,
    //Doc:Desc used by the \NV\ routing method.
    //Doc:Arg1 The neighbor index to look up.
    //Doc:Return The routing table entry for the specified neighbor index.
    //Doc:Return The interface pointer in this entry will be {\em nil}
    //Doc:Return if no routing information is found for the specified
    //Doc:Return destination.

  //Doc:Method
  Routing::RType_t RoutingType();        // Report type of routing in use
    //Doc:Desc Return the routing type in use at this node.
    //Doc:Return The routing type in use.  The enumerated type {\tt RType_t}
    //Doc:Return is defined in {\tt routing.h}.

  //Doc:Method
  Interface* LocalRoute(IPAddr_t);       // Find if dst is directly connected
    //Doc:Desc Determine if the specified \IPA\ is on a directly connected
    //Doc:Desc neighbor.
    //Doc:Arg1 The \IPA\ to search for.
    //Doc:Return An pointer to the interface used to communicate  with the
    //Doc:Return specified \IPA.  Returns {\em nil} if the specified
    //Doc:Return  \IPA\ is not a directly connected neighbor.

  //AODV Routing
  //Specially for AODV routing, it is not compatiable with
  //other route entry.
  //
  AODVRoutingEntry *LookupRouteAODV(IPAddr_t);
  void      SetRoutingAODV(void *pRouting);
  RoutingAODV* GetRoutingAODV(); //return the AODV routing object pointer

  //Doc:Method
  void       InitializeRoutes();         // Routing initialization (if any)
    //Doc:Desc Perform any initializations needed by the routing protocol
    //Doc:Desc used by this node.

  //Doc:Method
  Count_t    RoutingFibSize() const;     // Statistics, size of FIB
    //Doc:Desc Determine the size of the forwarding information base (FIB)
    //Doc:Desc at this node.
    //Doc:Return The size, in bytes, of the FIB at this node.

  //Doc:Method
  Count_t    GetNix(Node*) const;        // Get Neighbor Index for spec'd node
    //Doc:Desc Get the neighbor index for the specified node.  Used by the
    //Doc:Desc \NV\ routing method.
    //Doc:Arg1 The node to find the neighbor index for.
    //Doc:Return The neighbor index for the specified node.
    //Doc:Return Returns MAX_COUNT if the specified node is not a neighbor.

  //Doc:Method
  RoutingNixVector* GetNixRouting();     // Returns Nix rtg pointer (if Nix)
    //Doc:Desc Get a pointer to the \NV\ routing object for this node.
    //Doc:Return A pointer to the \NV\ routing object for this node.
    //Doc:Return Returns {\em nil} if \NV\ routing is not used at this node.

  //Doc:Method
  Routing*   GetRouting();
    //Doc:Desc Get a pointer to the routing object used by this node.
    //Doc:Return Routing object pointer.

  void       DumpNV();                   // Debug, dump NixVector info

  // Protocol Graph Interface
  //Doc:Method
  Protocol* LookupProto(Layer_t, Proto_t);// Lookup protocol by layer
    //Doc:Desc Find the protocol object at the specified protocol layer,
    //Doc:Desc and protocol number.
    //Doc:Arg1 The protocol layer number to look up.
    //Doc:Arg2 The protocol number to look up.
    //Doc:Return A pointer to the protocol object at the specified layer and
    //Doc:Return protocol number.

  //Doc:Method
  void      InsertProto(Layer_t, Proto_t, Protocol*); // Insert a protocol
    //Doc:Desc Inserts a new protocol in the protocol graph.
    //Doc:Arg1 The protocol layer number to insert the protocol into.
    //Doc:Arg2 The protocol number to insert the protocol intol
    //Doc:Arg3 A protocol to insert.

  // Port Demux Interface
  //Doc:Method
  bool      Bind(Proto_t, PortId_t, Protocol*);    // Register port usage
    //Doc:Desc Bind a protocol to a port number at a particular protocol
    //Doc:Desc number.
    //Doc:Arg1 The protocol number for this port.
    //Doc:Arg2 The port to bind to.
    //Doc:Arg3 The protocol to bind to the port.
    //Doc:Return True if successful.

  //Doc:Method
  bool      Bind(Proto_t,
                 PortId_t, IPAddr_t,
                 PortId_t, IPAddr_t,
                 Protocol*);                       // Register port usage
    //Doc:Desc Binds a protocol to a local \IPA, port, remote \IPA, port,
    //Doc:Desc at a particular protocol number.
    //Doc:Arg1 The protocol number for this binding.
    //Doc:Arg2 The local port number.
    //Doc:Arg3 The local \IPA.
    //Doc:Arg4 The remote port number.
    //Doc:Arg5 The remote \IPA.
    //Doc:Arg6 The protocol to bind.
    //Doc:Return True if successful.

  //Doc:Method
  PortId_t  Bind(Proto_t, Protocol*);              // Choose port and register
    //Doc:Desc Bind a protocol to any available  port.
    //Doc:Arg1 The protocol number to use.
    //Doc:Arg2 The protocol to bind.
    //Doc:Return The port number bound to.

  //Doc:Method
  bool      Unbind(Proto_t, PortId_t, Protocol*);  // Remove port binding
    //Doc:Desc Remove a port binding.
    //Doc:Arg1 The protocol number to use.
    //Doc:Arg2 The currently bound port number.
    //Doc:Arg3 The protocol to unbind.
    //Doc:Return True if successful.

  //Doc:Method
  bool      Unbind(Proto_t,
                 PortId_t, IPAddr_t,
                 PortId_t, IPAddr_t,
                 Protocol*);                       // Remove port binding
    //Doc:Desc Remove a port binding by local and remote port and \IPA.
    //Doc:Arg1 The protocol number to use.
    //Doc:Arg2 Local port number.
    //Doc:Arg3 Local \IPA.
    //Doc:Arg4 Remote port number.
    //Doc:Arg5 Remote \IPA.
    //Doc:Arg6 Protocol to unbind.
    //Doc:Return True if successful.

  //Doc:Method
  Protocol* LookupByPort(Proto_t, PortId_t);       // Lookup by port
    //Doc:Desc Locate the protocol bound to a specified port and protocol
    //Doc:Desc number.
    //Doc:Arg1 The protocol number.
    //Doc:Arg2 The port to look up.
    //Doc:Return A pointer to the protocol bound to the specified port.
    //Doc:Return Returns {\em nil} if no protocol is found at the
    //Doc:Return specified protocol and port.

  //Doc:Method
  Protocol* LookupByPort(Proto_t,
                         PortId_t, IPAddr_t,
                         PortId_t, IPAddr_t);      // Lookup by s/d port/ip
    //Doc:Desc Locate the protocol bound to the specified local and remote
    //Doc:Desc port and \IPA, at the specified protocol number.
    //Doc:Arg1 The protocol number.
    //Doc:Arg2 Local port number.
    //Doc:Arg3 Local \IPA.
    //Doc:Arg4 Remote port number.
    //Doc:Arg5 Remote \IPA.
    //Doc:Return A pointer to the protocol bound as specified.
    //Doc:Return Returns {\em nil} if no protocol is found.

  // Packet tracing interface
  //Doc:Method
  // Trace header if enabled
  bool      TracePDU(Protocol*, PDU*, Packet* = nil, char* = nil);
    //Doc:Desc Log the contents of the specified protocol data unit, from
    //Doc:Desc the specified protocol, to the trace file.
    //Doc:Arg1 The protocol pointer for the protocol object requesting
    //Doc:Arg1 the trace.
    //Doc:Arg2 The protocol data unit pointer.
    //Doc:Arg3 Packet that contains this PDU (optional)
    //Doc:Arg4 Extra text information for trace file (optional)
    //Doc:Return Returns true  if trace file  exists.

  //Doc:Method
  void      SetTrace(Trace::TraceStatus);          // Set trace level this node
    //Doc:Desc Specifies the tracing status of this node.
    //Doc:Desc The trace status desired.  This can be either
    //Doc:Desc {\tt Trace::ENABLED},
    //Doc:Desc which indicates that all packets requesting tracing
    //Doc:Desc at this node will be traced;
    //Doc:Desc {\tt Trace::DISABLED} indicating that no packets will
    //Doc:Desc be traced, or {\tt Trace::DEFAULT}, indicating that the
    //Doc:Desc decision is deferred to the protocols at this node.
    //Doc:Arg1 The trace status desired for this node.

  // Node location (x, y, z) interface
  //Doc:Method
  void      SetLocation(Meters_t, Meters_t, Meters_t=0);  // Set x/y/z location
    //Doc:Desc Set the location of this node in the X--Y--Z coordinate.
    //Doc:Arg1 The X--coordinate (meters).
    //Doc:Arg2 The Y--coordinate (meters).
    //Doc:Arg3 The Z--coordinate (meters).

  //Doc:Method
  void      SetLocation(const Location&);          // Set x/y/z location
    //Doc:Desc Set the location of this node in the X--Y--Z coordinate.
    //Doc:Arg1 A {\tt Location} object specifying
    //Doc:Arg1 the X--Y--Z coordinate (meters).

  //Doc:Method
  void      SetLocationLongLat(const Location&);  // Set x/y location
    //Doc:Desc Set the location of this node in the X--Y plane.
    // Adjusts the lat/long by the cos(lat) factor for  map projections.
    //Doc:Arg1 A {\tt Location} object specifying
    //Doc:Arg1 the X--coordinate (longitued) and Y--coordinate (latitude).

  //Doc:Method
  bool      HasLocation();
    //Doc:Desc Test if the node has a location assigned.
    //Doc:Return True if location assigned for this node.

  //Doc:Method
  Meters_t  LocationX();
    //Doc:Desc Return the X--coordinate of this node.
    //Doc:Return The X--coordinate (meters) of this node.

  //Doc:Method
  Meters_t  LocationY();
    //Doc:Desc Return the Y--coordinate of this node.
    //Doc:Return The Y--coordinate (meters) of this node.

  //Doc:Method
  Meters_t  LocationZ();
  //Doc:Desc Return the Z--coordinate of this node.
  //Doc:Return The Z--coordinate (meters) of this node.

  //Doc:Method
  Location  GetLocation();                         // Return current location
    //Doc:Desc Return the location of this node.
    //Doc:Return A {\tt Location} object specifying the X, Y and Z coordinates
    //Doc:Return of this node (meters).

  //Doc:Method
  Location  UpdateLocation();                      // Get updated location
    //Doc:Desc The location of mobile nodes is not updated continually
    //Doc:Desc during the simulation.  Rather,  it  is updated when
    //Doc:Desc {\tt UpdateLocation()} is called.
    //Doc:Return A {\tt Location} object specifying the X, Y and Z coordinates
    //Doc:Return of this node (meters).

  //Doc:Method
  Mobility* AddMobility(const Mobility&);          // Add a mobility model
    //Doc:Desc Add a mobility model to this node.
    //Doc:Arg1 The {\tt Mobility} object to use.  This is copied to the
    //Doc:Arg1 node, so an anonymous temporary object can be used,
    //Doc:Arg1 or the same {\tt Mobility} object can be passed
    //Doc:Arg1 multiple times.
    //Doc:Return A pointer to the added mobility object.

  //Doc:Method
  Mobility* GetMobility();
    //Doc:Desc   Return existing mobility model.
    //Doc:Return Pointer to this nodes mobility model.  NIL if none.

  //Doc:Method
  bool      IsMobile();                            // True if mobility enabled
    //Doc:Desc Determine if this node has a mobility model.
    //Doc:Return True  if node has a mobility model.

  //Doc:Method
  bool      IsMoving();                            // True if not pausing
    //Doc:Desc Determine if this node is presently moving (rather than
    //Doc:Desc paused or temporarily stationary).
    //Doc:Return True if node is presently moving.

  //Doc:Method
  Meters_t  Distance(Node*);
    //Doc:Desc Calculate distance between this node and specified peer node.
    //Doc:Arg1 Peer node pointer.
    //Doc:Return Distance (meters).

  //Doc:Method
  void      BuildRadioInterfaceList(void);         // i/f list in radio range
    //Doc:Desc Build a list of nodes in transmission range.

  //Doc:Method
  const RadioVec_t& GetRadioInterfaceList();
    //Doc:Desc Get a reference to the current radio interface list
    //Doc:Return Reference to the current radio interface list

  //Doc:Method
  void      SetRadioRange(Meters_t range);
    //Doc:Desc Specify the power range of this wireless transmiter (meters).
    //Doc:Arg1 Power range (meters)

  //Doc:Method
  Meters_t  GetRadioRange(void);
    //Doc:Desc Return the power range of this transmitter.
    //Doc:Return Power range (meters).

  //Doc:Method
  bool      CBEnable();
    //Doc:Desc Determine if any callbacks enabled for this node.
    //Doc:Return True if callbacks enabled.

  //Doc:Method
  void      CBEnable(bool);
    //Doc:Desc Enable or disable callbacks on this node.
    //Doc:Arg1 True if callbacks to be enabled.

  //Doc:Method
  void      CBListAdd(IPCallBack);
    //Doc:Desc Add a callback function to this node.  Callbacks allow
    //Doc:Desc a function to examine every packet recieved by this node.
    //Doc:Arg1 Callback function to add.

  //Doc:Method
  void      CBListDelete(IPCallBack);
    //Doc:Desc Remove a callback function from this node.
    //Doc:Arg1 Callback function to remove.

  //Doc:Method
  void      PreProcessCB(Packet*);   // loop over the list
    //Doc:Desc Call all callback functions with this packet.
    //Doc:Arg1 Pointer to packet received.

  //Doc:Method
  const CallBackList_t& GetCBList();
    //Doc:Desc Get a reference to the callback list.
    //Doc:Return Reference to the callback list

  void UseWormContainment(bool b);
  bool UseWormContainment();
  void SetWormContainment(WormContainment*);
  WormContainment* GetWormContainment();


  //Doc:Method
  bool      WirelessTx();
    //Doc:Desc Check if the node is transmitting on a wireless link.  Used by
    //Doc:Desc the animation widow.
    //Doc:Return True if transmitting.

  //Doc:Method
  bool      WirelessRx();
    //Doc:Desc Check if the node is receiving on a wireless link.  Used by
    //Doc:Desc the animation widow.
    //Doc:Return True if receiving.

  //Doc:Method
  bool      WirelessCx();
    //Doc:Desc Check if the node is experiencing collision on a wireless link.
    //Doc:Desc Used by the animation widow.
    //Doc:Return True if colliding.

  //Doc:Method
  bool      WirelessRxMe();
    //Doc:Desc Check if the node is receiving on a wireless link.  Used by
    //Doc:Desc the animation widow.
    //Doc:Return True if receiving.

  //Doc:Method
  bool      WirelessRxZz();
    //Doc:Desc Check if the node is receiving on a wireless link.  Used by
    //Doc:Desc the animation widow.
    //Doc:Return True if receiving.

  // QTWindow information
  //Doc:Method
  void      Show(bool);                            // Turn on/off display
    //Doc:Desc Enable  or disable the animation window.
    //Doc:Arg1 True if animation window should be enabled, false if not.

  //Doc:Method
  bool      Show();                                // True if display enabled
    //Doc:Desc Determine if the node should be displayed in the anim window
    //Doc:Return True if the node should be displayed in the animation window

  //Doc:Method
  QCanvasItem* Display(QTWindow*);
    //Doc:Desc Each node is responsible for "drawing" itself on the animation
    //Doc:Desc display.  This allows the node to decide shape, size, etc.
    //Doc:Arg1 Qt Canvas to use.  The node is drawn at present Location().
    //Doc:Return Any subclass of QCanvasItem.

  //Doc:Method
  QCanvasItem* Display(const QPoint&, QTWindow*);
    //Doc:Desc Each node is responsible for "drawing" itself on the animation
    //Doc:Desc display.  This allows the node to decide shape, size, etc.
    //Doc:Arg1 Point in Qt canvas to display this node.
    //Doc:Arg2 Qt Canvas to use.
    //Doc:Return Any subclass of QCanvasItem.

  //Doc:Method
  void WirelessTxColor(const QColor&);
    //Doc:Desc Specify color for the wireless transmit animation.
    //Doc:Arg1 Desired color.

  //Doc:Method
  const QColor& WirelessTxColor();
    //Doc:Desc Get the specified  wireless transmit color
    //Doc:Return Color for wireless tx concentric  circles

  //Doc:Method
  bool PushWirelessTx(QCanvasItem*);
    //Doc:Desc Add a canvas itemm to the wireless transmit animation list
    //Doc:Arg1     Item (one of the concentric circles) to  add

  //Doc:Method
  QCanvasItem* PopWirelessTx();
    //Doc:Desc   Pop and return the oldest item in the wireless tx history.
    //Doc:Return Oldest item (nil if none).

  //Doc:Method
  void PixelSize(Count_t);
    //Doc:Desc Set the size of this node (pixels) on the animation display.
    //Doc:Arg1 Size of this node in pixels (both width and height)

  //Doc:Method
  Count_t PixelSizeX();
    //Doc:Desc Query the width of this node (pixels) on the animation display.
    //Doc:Return Size of this node in pixels.

  //Doc:Method
  Count_t PixelSizeY();
    //Doc:Desc Query the height of this node (pixels) on the animation display.
    //Doc:Return Size of this node in pixels.

  //Doc:Method
  void        Shape(Shape_t);
    //Doc:Desc Set the shape for this node on the animation display.
    //Doc:Arg1 Shape for this node.  See node.h for shape types.

  //Doc:Method
  Shape_t Shape();
    //Doc:Desc Query the shape for this node on the animation display.
    //Doc:Return Shape for this node.  See node.h for shape types.

  //Doc:Method
  CustomShape_t CustomShape();
    //Doc:Desc   Get the custom shape callback function.
    //Doc:Return Function pointer to custom shape callback function.

  //Doc:Method
  void CustomShape(CustomShape_t);
    //Doc:Desc Specify the custom shape callback function.
    //Doc:Arg1 Pointer to new custom shape callback function.

  //Doc:Method
  bool CustomShapeFile(const char*);
    //Doc:Desc Specifies a file containing the custom shape image.
    //Doc:Desc Must be in an image format supported by Qt.
    //Doc:Desc NOTE:  This {\bf must} be called {\em  after} the call
    //Doc:Desc to Simulator::StartAnimation, since it needs a QTWindow
    //Doc:Desc object.  If called before StartAnimation, will always
    //Doc:Desc return false.
    //Doc:Arg1 File name for image.
    //Doc:Return True if  successfully loaded

  //Doc:Method
  void Color(const QColor&);
    //Doc:Desc Specifiy a color for this node
    //Doc:Arg1 Desired color (see qcolor.h in qt).

  //Doc:Method
  bool HasColor();
  //Doc:Desc Query if this node has a color assigned
  //Doc:Return True if color has been specified.

  //Doc:Method
  QColor Color();
    //Doc:Desc Query color for this node.
    //Doc:Return Specified display color for this node.

  //Doc:Method
  NodeAnimation* GetNodeAnimation() const;
    //Doc:Desc   Get a pointer to the nodes NodeAnimation  information block.
    //Doc:Desc       Always nil  for Ghosts, possibly non-nil for real.
    //Doc:Return Associated NodeAnimation poiner.

  //Doc:Method
  bool   ICMPEnabled() const;
    //Doc:Desc Query if ICMP enabled for this node.
    //Doc:Return True if enabled. NOTE. This only means this node will use ICMP
    //Doc:Return if globally enabled (see icmp.h)

  //Doc:Method
  void   DisableICMP();
    //Doc:Desc   Specify this node does NOT use ICMP, even if globally enabled.

  // Node failure management

  //Doc:Method
  void   Down();
    //Doc:Desc Specifies the the node has crashed and should not forward or
    //Doc:Desc generate any packets.

  //Doc:Method
  void   Up();
    //Doc:Desc Specifies the node has recovered from failure.

  //Doc:Method
  bool   IsDown();
    //Doc:Desc Determine if node has failed or not.
    //Doc:Return       TRUE if node is down (failed)

  // Data Members
 private:
  NodeId_t  id;              // Node id for this node
 protected:
  NodeImpl* pImpl;           // Node Implementation, ghost or real
#ifdef SIM 
  OSPF6DInstance* ospf6d_inst;
  static std::vector<Node*> NodeVecByRouterId;  // For quick lookup of
                                                     // a node by Router ID
#endif //SIM
  // Static members
public:
  static NodeId_t  nextId;        // Next available node id
  static NodeVec_t nodes;         // Vector of all available nodes
  static Count_t   defaultPixelSize; // Size in pixels of animated nodes
  static Shape_t   defaultShape;     // Default shape for all nodes
  static double    defaultMaxSpeed;  // Default maximum speed of a node

public:
  //Doc:Method
  static void SetNextNodeId(NodeId_t i) { nextId = i;} // Set non-default Next
    //Doc:Desc Set a new value for the next node identifier value.
    //Doc:Arg1 The new next node identifier value.

  // Get const ref to Node vec
  //Doc:Method
  static const NodeVec_t& GetNodes() { return nodes;}
    //Doc:Desc Get the vector of all existing nodes.
    //Doc:Return A const reference to the vector of existing nodes.

  //Doc:Method
  static Node* GetNode(NodeId_t);  // Get a node pointer by id
    //Doc:Desc Get the node pointer for the specified node identifier.
    //Doc:Arg1 The node identifier desired.
    //Doc:Return A pointer to the specified node.  Returns {\em nil} if the
    //Doc:Return specified node does not exist.

  //Doc:Method
  static void Clear();
    //Doc:Desc Clear the node vector.
    //Doc:Desc Normally only called after the simulation
    //Doc:Desc   has completed, and we are trying to track memory leaks.

  //Doc:Method
  static void DefaultPixelSize(Count_t);
    //Doc:Desc Set the default size (in pixels) of the animated nodes
    //Doc:Arg1 Default size in pixels.

  //Doc:Method
  static Count_t DefaultPixelSize();
    //Doc:Desc Query the default node pixel size
    //Doc:Return The default node pixel size.

  //Doc:Method
  static void DefaultShape(Shape_t);
    //Doc:Desc Set the default shape for nodes on the animation display
    //Doc:Arg1 Default shape.  See node.h for shape types.

  //Doc:Method
  static Shape_t DefaultShape();
    //Doc:Desc Query the default node shape.
    //Doc:Return Default shape.  See node.h for shape types.

  static void   DefaultMaxSpeed(double speed);
  static double DefaultMaxSpeed();

  Joules_t getBattery (void);
  void     setBattery (Joules_t);

  double   getComputePower(void);
  void     setComputePower(double);

  // The following is for proxy routing
 private:
  IPAddr_t    proxyIP;
  Mask        proxyMask;

 public:
  void        SetProxyRoutingConfig(IPAddr_t ip, Mask mask);
  bool        HasProxyRoutingConfig() {return proxyIP!=IPADDR_NONE;};
  bool        CanProxyRouteIP(IPAddr_t ip);
  IPAddr_t    GetProxyIP();
  Mask        GetProxyMask();
  int         GetLongestPrefixLength(IPAddr_t ip);
};

#include "interface.h"

#endif
