/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Packet class
// George F. Riley.  Georgia Tech, Spring 2002

#ifdef HAVE_QT
#include <qcolor.h>
#endif
// Uncomment below to enable debug level 0
// #define DEBUG_MASK 0x01

//#define DEBUG_MASK 0x01
//#define DEBUG_MASK 0x08
#include "debug.h"
#include "packet.h"
#include "trace.h"
#include "simulator.h"
#include "globalstats.h"
#include "hex.h"
#include "l2proto802.3.h"
#include "routing-nixvector.h"

using namespace std;

Count_t Packet::nextUid = 1; // Unique id's for packets

Packet::Packet() : top(0),  retx(0), size(0), nixVec(nil), expire(0), notification(nil)
{
  top = 0; // Initialize the top of stack pointer
  uid = nextUid++;
  time = Simulator::Now();
  PDUs.reserve(4);                // Room for Data, TCP, IP and L2
  DBG((Stats::pktsAllocated++));
}

Packet::Packet(const Packet& p) :  retx(p.retx), uid(p.uid), size(0),
                                  time(p.time),
                                  nixVec(nil),
                                  expire(p.expire),
                                  notification(p.notification)
{ // Make a copy of the PDU header list
  if (p.nixVec) nixVec = (NixVectorOption*)p.nixVec->Copy();
  top = 0;
  PDUs.reserve(p.PDUs.size());
  //for (PDUVec_t::size_type i = 0; i < p.top; ++i)
  for (PDUVec_t::size_type i = 0; i < p.PDUs.size(); ++i)
    {
      PushPDU(p.PDUs[i]->Copy());
    }
  top = p.top;
  DBG((Stats::pktsAllocated++));
}

Packet::Packet(char* b, Size_t sz) : nixVec(nil), notification(nil)
{ // Construct from serialized buffer
  DEBUG(1,(cout << "Contructing packet, sz " << sz << endl));
  Construct(b, sz);
  DBG((Stats::pktsAllocated++));
}

Packet::~Packet()
{
  // Delete each pdu header
  for (PDUVec_t::const_iterator i = PDUs.begin(); i != PDUs.end(); ++i)
    {
      delete *i;
    }
  if (nixVec) delete nixVec;      // Delete nixvector if present
  Trace::Instance()->AppendEOL(); // End of trace line when pkt deleted
  DBG((Stats::pktsDeleted++));
}


Size_t Packet::SizeUpdate()
{
  size = 0;
  for (PDUVec_t::size_type i = 0; i < top; ++i)
    size += PDUs[i]->Size();
  return size;
}


#ifdef MOVED_TO_INLINE
Size_t Packet::Size()
{ // Return total size of all PDU's
  return size;
}

void Packet::PushPDU(PDU* pdu) // Add a new PDU header
{
  if (top == PDUs.size())
    { // Need to add new one
      PDUs.push_back(pdu);
      top++;
    }
  else
    { // Can reuse existing, after deleting
      if (PDUs[top] == pdu)
        { // Replacing with original
          top++;
        }
      else
        { // New one, delete and re-insert
          delete PDUs[top];    // Delete previous pdu
          PDUs[top++] = pdu;   // Replace with this one
        }
    }
  size += pdu->Size();
}
#endif

void Packet::PushPDUBottom(PDU* pdu) // Add a new PDU header to bottom
{ // During packet serialization, pdu's are serialized top to bottom.
  // This is to allow the L2 pdu "first", so we can recontruct the
  // pdu list "up" the stack.  However, this means we need to add
  // pdu's to the list starting with l2, which is normally the "top"
  // of the pdu stack.  This function allows building the pdu list
  // from top to bottom.
  if (PDUs.empty())
    { // Simplest case, just push and exit
      PushPDU(pdu);
    }
  else
    { // Non-empty, move existing up
      PDU* pd = PDUs[PDUs.size() - 1];
      PDUs.push_back(pd);
      for (PDUVec_t::size_type i = PDUs.size() - 1; i != 0; --i)
        {
          PDUs[i] = PDUs[i-1];
        }
      PDUs[0] = pdu; // Put this one on bottom
      size += pdu->Size();
      top++;
    }
}

#ifdef MOVED_TO_INLINE
PDU* Packet::PopPDU()      // Get last PDU header
{
  if (top == 0) return NULL; // Nope, nothing there
  
  PDU* p =  PDUs[--top];      // Get the pointer
  size -= p->Size();          // Reduce size
  return p;
}

PDU* Packet::PeekPDU()     // Get last PDU header without removal
{
  if (top == 0) return NULL;  // Nope, nothing there
  return PDUs[top - 1];       // Return without removing
}

//change Count_t to NCount_t in order to PEEK forward or backward.
//type conversion for avoiding the warnings.
//but potentially have problems here!
//
PDU* Packet::PeekPDU(NCount_t offset)
{ // Peek at last + offset without removal
  if ((NCount_t)top <= offset) return NULL;  // Nope, nothing there
  return PDUs[top - offset - 1];   // Return without removing
}
#endif

PDU*    Packet::FindPDU(Layer_t l)
{ // Find PDU by layer
  for (PDUVec_t::size_type i = 0; i < top; ++i)
    {
      PDU* pdu= PDUs[i];
      if (pdu->Layer() == l) return pdu; // Found
    }
  return nil; // Not found
}

PDU*    Packet::FindPDU(Layer_t l, Proto_t proto)
{ // Find PDU by layer and protocol
  for (PDUVec_t::size_type i = 0; i < top; ++i)
    {
      PDU* pdu= PDUs[i];
      if ((pdu->Layer() == l) && (pdu->Proto() == proto)) return pdu; //Found
    }
  return nil; // Not found
}


Packet* Packet::Copy()
{
  return new Packet(*this);
}

// Serialization

Size_t Packet::SSize()
{ // Size needed for serialization
  Size_t sz = sizeof(top) + sizeof(size) + sizeof(uid) 
    + sizeof(time) + sizeof(NixVectorOption*);
  if (nixVec)
    { // NixVector present, add size
      if (nixVec->Remaining())
        { // Only if not empty
          sz += nixVec->SSize() + sizeof(Size_t);
        }
    }
  // Add the size of each pdu
  for (PDUVec_t::size_type i = 0; i < top; ++i)
    {
      sz += PDUs[i]->SSize() + sizeof(Size_t);
    }
  return sz;
}

char*  Packet::Serialize(char* b, Size_t& sz)
{ // Serialize to a working buffer
  DEBUG(2,(cout << "Packet size during serialization " << Size() << endl));
  DEBUG(2,(cout << "Size before packet members " << sz << endl));
  b = SerializeToBuffer(b, sz, top);
  b = SerializeToBuffer(b, sz, uid);
  b = SerializeToBuffer(b, sz, size);
  b = SerializeToBuffer(b, sz, time);
  // Note we are serializing a pointer here, which normally is not
  // correct.  However, on reconstruction, this value is just used
  // to detect the presence of a NixVector.  If present, the nix
  // vector itself is constructed and a new pointer created
  NixVectorOption* nv = nixVec;
  if (nv)
    { // Serialize the nix vector
      if (nv->Remaining())
        { // Only if not all used
          b = SerializeToBuffer(b, sz, (unsigned long&)nv);
          Size_t s = nixVec->SSize();
          DEBUG(3,(cout << "Serializing NV, size " << s << endl));
          b = Serializable::PutSize(b, sz, s);
          sz -= s;
          b = nv->Serialize(b, s);
        }
      else
        { // Don't serialize an empty nixvector
          nv = nil;
        }
    }
  if (!nv) b = SerializeToBuffer(b, sz, (unsigned long&)nv);
  DEBUG(1,(cout << "Size after packet members " << sz << endl));
  // Now serialize the PDUs
  for (PDUVec_t::size_type i = top - 1; ; --i)
    {
      Size_t s = PDUs[i]->SSize() + sizeof(Size_t);
      DEBUG(2,(cout << "Serializing pdu size " << Hex2(s) << endl));
      b = Serializable::PutSize(b, sz, s);
      DEBUG(2,(cout << "Serializing pdu, s " << s << " sz " << sz << endl));
      sz -= (s-sizeof(Size_t));
      b = PDUs[i]->Serialize(b, s);
      DEBUG(2,(cout << "Size after pdu " << sz << endl));
      if (i == 0) break;
    }
  
  DEBUG(1,(cout << "Remaining size is " << Hex8(sz) 
           << " (should be zero)" << endl));
  return b;
}

char*  Packet::Construct(char* b, Size_t& sz)
{ // Construct from serialized buffer
  b = ConstructFromBuffer(b, sz, top);
  b = ConstructFromBuffer(b, sz, uid);
  b = ConstructFromBuffer(b, sz, size);
  b = ConstructFromBuffer(b, sz, time);
  unsigned long nvul;
  b = ConstructFromBuffer(b, sz, nvul);
  if (nvul)
    { // Nix vector is present, construct it
      DEBUG(3,(cout << "Constructing NixVector" << endl));
      Size_t thisSize  = 0;
      b = Serializable::GetSize(b, sz, thisSize);
      Size_t saveSize = thisSize;
      nixVec = new NixVectorOption(b, thisSize);
      b += saveSize;
      sz -= saveSize;
    }
  // For debugging, ignore the nixVector
  //if (nixVec) delete nixVec;
  //nixVec = nil;
  // End debugging
  DEBUG0((cout << "Packet::Construct, top " << top 
       << " uid " << uid
       << " size " << size
          << " time " << time << endl));
  // Constructing each pdu done slightly differently, since we can't
  // know at this point which type each pdu is.
  // Since (at least at this point in GTNetS development), L2802-3 
  // are the only packet types that can be sent simulator to simulator
  // we can start by assuming l2802-3 l2 pdu is first thing.  Each
  // pdu then constructs the next layer
  top = 0;  // we need to reset top here (why did we serialize it anyway?)
  size = 0; // also reset size, as it's calculated as pdu's are pushed
  new L2802_3Header(b, sz, this); // all we need here!
  DEBUG(1,(cout << "Packet::Construct, residual size " << sz << endl));
  return b + sz;
}

// Definitions for the colored packet
#ifdef HAVE_QT
ColoredPacket::ColoredPacket() : Packet(), color(nil)
{
}


ColoredPacket::ColoredPacket(const QColor& c)
  : Packet(), color(new QColor(c))
{
}

ColoredPacket::ColoredPacket(const ColoredPacket& p)
  : Packet(p), color(new QColor(p.Color()))
{
}

Packet* ColoredPacket::Copy()
{
  return new ColoredPacket(*this);
}

bool ColoredPacket::IsColored()
{
  return true;
}

Color_t ColoredPacket::R()
{
  return color->red();
}

Color_t ColoredPacket::G()
{
  return color->green();
}

Color_t ColoredPacket::B()
{
  return color->blue();
}

void    ColoredPacket::SetColor(const QColor& c)
{
  if (color) delete color;
  color = new QColor(c);
}


const QColor& ColoredPacket::Color() const
{
  return *color;
}

#endif

