/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Protocol Data Unit base class
// George F. Riley.  Georgia Tech, Spring 2002

// Class PDU is the base class for all packet headers

#ifndef __pdh_h__
#define __pdh_h__

#include <vector>

#include "debug.h"
#include "common-defs.h"
#include "tfstream.h"
#include "globalstats.h"
#include "serializable.h"
#include "memreuse.h"

//Doc:ClassXRef
class PDU : public Serializable, public ReuseBase {
//Doc:Class Class {\tt PDU} serves as the base class for all the 
//Doc:Class protocol data units. Protocol headers and data chunks that
//Doc:Class that form the packets are derived from this class.
	
public:
  //Doc:Method
  PDU() { DBG((Stats::pdusAllocated++));}
  //Doc:Desc This method is the default constructor for this class. It
  //Doc:Desc updates a static variable to reflect the number of PDUs
  //Doc:Desc allocated
  
  //Doc:Method
  PDU(const PDU&) { DBG((Stats::pdusAllocated++));}
  //Doc:Desc This constructs a PDU from the reference to an already existing
  //Doc:Desc PDU.
  //Doc:Arg1 The const reference to the existing PDU
  
  //Doc:Method
  virtual ~PDU() { DBG((Stats::pdusDeleted++));}
  
public:
  //Doc:Method
  virtual Size_t Size() const = 0;    // PDU's must have a size
  //Doc:Desc This method returns the size of the PDU
  //Doc:Return the size of the PDU
  
  //Doc:Method
  virtual PDU* Copy() = 0;            // Create a copy of this header
  //Doc:Desc This method creates a copy of this PDU
  //Doc:Return A pointer to the copy of this PDU
  
  //Doc:Method  
  virtual Layer_t Layer() { return 0;}// Some PDU's specify their layer
  //Doc:Desc Each PDU must also return a layer numer as to which protocol layer
  //Doc:Desc it belongs to according to the ISO model
  //Doc:Return the layer number
  
  //Doc:Method  
  virtual Count_t Version() { return 0;} // Some PDU's report version number
  //Doc:Desc If we are using multiple versions of the same PDU we may want
  //Doc:Desc to return the version number of the PDU we are using. Also useful
  //Doc:Desc when tracing
  //Doc:Return the version number
  
  //Doc:Method  
  virtual Proto_t Proto() { return 0;}   // Some PDU's report protocol number
  //Doc:Desc Most of the protocols are identified by the numeric values as
  //Doc:Desc listed by IANA. thie method returns that protocol number
  //Doc:Return the protocol number
  
  // Trace the contents of this pdu
  //Doc:Method  
  virtual void Trace(Tfstream&, Bitmap_t,
                     Packet* = nil, const char* = nil) { }
  //Doc:Desc This method traces the contents of the PDU
  //Doc:Arg1 File output stream descriptor
  //Doc:Arg2 Bitmap that specifies the level of detail of the trace
  //Doc:Arg3 Packet that contains this PDU (optional)
  //Doc:Arg4 Extra text information for trace file (optional)
  
  //Doc:Method  
  bool Detail(Bitmap_t d, Bitmap_t b) { return (1L << d) & b;}
  // Include the serializable interface here as a temporary patch
  // until all PDU subclasses implement this
  //Doc:Method  
  virtual Size_t SSize() { return 0;}      // Size needed for serialization
  //Doc:Desc The actual size of all the PDU contents in terms of bytes for
  //Doc:Desc serialization of the contents.
  //Doc:Return the size
  
  //Doc:Method  
  virtual char*  Serialize(char* b, Size_t&) { return b;}// Serialize to buffer
  //Doc:Desc This method is used to serialize the contents of the PDU
  //Doc:Desc into a serialized character buffer b . the size of this buffer is
  //Doc:Desc in the size arguement
  //Doc:Arg1 This is the pointer to the character buffer
  //Doc:Arg2 a reference to the size variable
  //Doc:Return the pointer to the updated character buffer
  
  //Doc:Method  
  virtual char*  Construct(char* b, Size_t&) { return b;}// Const. from buffer
  //Doc:Desc This method is the reverse of the Serialize . It constructs the
  //Doc:Desc PDU object from the character buffer that has been received over
  //Doc:Desc the network
  //Doc:Arg1 This is the pointer to the character buffer
  //Doc:Arg2 a reference to the size variable
  //Doc:Return the pointer to the character buffer of the remaining data
};

typedef std::vector<PDU*> PDUVec_t;     // Vector of packet headers

#endif


