/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Port Demultiplexer
// George F. Riley.  Georgia Tech, Spring 2002

// Implements a map of protocol instances by port number

#include <iostream>

//#define DEBUG_MASK 0x02
#include "debug.h"
#include "portdemux.h"

using namespace std;

// Bindings are by local port, local IP, remote port, remote IP
// Local port can be IPADDR_NONE
// Both remote port, remote IP can be NO_PORT, IPADDR_NONE

// Define some syntactic macros, since the lookup key is a pair of pairs
#define LOCAL_PORT  first.second
#define LOCAL_IP    first.first
#define REMOTE_PORT second.second
#define REMOTE_IP   second.first

PortDemux::PortDemux()
{
}

bool PortDemux::Bind(Proto_t proto, PortId_t port, Protocol* p)
{ // Local port only
  return Bind(proto, port, IPADDR_NONE, p);
}

bool PortDemux::Bind(Proto_t proto, PortId_t port, IPAddr_t ip, Protocol* p)
{ // Local port, local ip
  return Bind(proto, port, ip, NO_PORT, IPADDR_NONE, p);
}

bool PortDemux::Bind(Proto_t proto,
                     PortId_t lport, IPAddr_t laddr,
                     PortId_t rport, IPAddr_t raddr,
                     Protocol* p)
{ // Local port/ip, remote port/ip
  PortDemux_t::iterator i;
  SrcDstPair_t key;
  key.LOCAL_PORT = lport;
  key.LOCAL_IP = laddr;
  key.REMOTE_PORT = rport;
  key.REMOTE_IP = raddr;
  i = protoDemux[proto].find(key);
  if (i != protoDemux[proto].end())
    {
      DEBUG0((cout << "PortDemux::Bind, port " << lport 
              << " already in use" << endl));
      return false; // port already in use
    }
  
  protoDemux[proto].insert(PortDemux_t::value_type(key,p)); // Insert protocol 
  DEBUG0((cout << "After bind lip " << (string)IPAddr(laddr)
          << " lport " << lport
          << " rip " << (string)IPAddr(raddr)
          << " rport " << rport << endl));
  DEBUG0((DBDump(protoDemux[proto])));
  return true;
}

PortId_t PortDemux::Bind(Proto_t proto, Protocol* p)
{ // Bind to an available port (port must be > MIN_TRANSIENT)
  NextPort_t::const_iterator i = nextPort.find(proto);
  if (i == nextPort.end())
    { // First lookup, initialize to MIN_TRANSIENT
      nextPort.insert(NextPort_t::value_type(proto, MIN_TRANSIENT));
    }
  PortId_t port = nextPort[proto]++;
  Bind(proto, port, IPADDR_NONE, p);
  return port;
  // Need better..need to bound ports at MAX_PORT
}

bool PortDemux::Rebind(Proto_t proto,
                       PortId_t lport, IPAddr_t laddr,
                       PortId_t rport, IPAddr_t raddr,
                       Protocol* p)
{ // Change to more specific binding
  PortDemux_t::iterator i;
  i = Locate(proto, lport, laddr, rport, raddr);
  if (i == protoDemux[proto].end()) return false; // Not found
  protoDemux[proto].erase(i); // Remove old binding
  return Bind(proto, lport, laddr, rport, raddr, p);
}
                       
Protocol* PortDemux::Lookup(Proto_t proto, PortId_t port)
{
  return Lookup(proto, port, IPADDR_NONE, NO_PORT, IPADDR_NONE);
}

Protocol* PortDemux::Lookup(Proto_t proto,
                             PortId_t lport, IPAddr_t laddr,
                             PortId_t rport, IPAddr_t raddr)
{
  PortDemux_t::iterator i;
  i = Locate(proto, lport, laddr, rport, raddr);
  if (i == protoDemux[proto].end()) return NULL; // Not found
  return i->second;                // Return the protocol pointer
}

bool PortDemux::Unbind(Proto_t proto, PortId_t port, Protocol* p)
{
  return Unbind(proto, port, IPADDR_NONE, NO_PORT, IPADDR_NONE, p);
}


bool PortDemux::Unbind(Proto_t proto,
                        PortId_t lport, IPAddr_t laddr,
                        PortId_t rport, IPAddr_t raddr,
                        Protocol* p)
{
  PortDemux_t::iterator i;
  i = Locate(proto, lport, laddr, rport, raddr);
  if (i == protoDemux[proto].end()) return false; // Not bound
  if (i->second != p) return false;               // Not bound to this protocol
  protoDemux[proto].erase(i);                     // Remove binding
  DEBUG(1,(cout << "portdemux unbinding" << endl));
  return true;
}

PortDemux_t::iterator PortDemux::Locate(Proto_t proto,
                                        PortId_t lport, IPAddr_t laddr,
                                        PortId_t rport, IPAddr_t raddr)
{ // Locate the binding
  PortDemux_t::iterator i;
  SrcDstPair_t key;
  key.LOCAL_PORT = lport;
  key.LOCAL_IP = laddr;
  key.REMOTE_PORT = rport;
  key.REMOTE_IP = raddr;
  DEBUG0((DBDump(protoDemux[proto])));
  i = protoDemux[proto].find(key);
  if (i != protoDemux[proto].end()) return i; // Found it
  // Try without rport/raddr
  key.REMOTE_PORT = NO_PORT;
  key.REMOTE_IP = IPADDR_NONE;
  i = protoDemux[proto].find(key);
  if (i != protoDemux[proto].end()) return i; // Found it
  // Try without source ip
  key.LOCAL_IP = IPADDR_NONE;
  return protoDemux[proto].find(key);
}

 void      PortDemux::DBDump(PortDemux_t& dm) // debug
{
  PortDemux_t::iterator i;
  cout << "protoDemux size " 
       << dm.size() << endl;
  for (i = dm.begin(); i != dm.end(); ++i)
    {
      cout << "DM srcip " << (string)IPAddr(i->first.first.first)
           << " srdport " << i->first.first.second
           << " dstip " << (string)IPAddr(i->first.second.first)
           << " dstport " << i->first.second.second << endl;
    }

}


