/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - QT Graphics Window Interface
// George F. Riley.  Georgia Tech, Fall 2002

#ifndef __qtwindow_h__
#define __qtwindow_h__

#include <vector>

#include "common-defs.h"
#include "simulator.h"

#ifdef HAVE_QT
#include <qobject.h>
#include <qapplication.h>
#include <qcanvas.h>
#include <qlcdnumber.h>
#include <qslider.h>
#include <qvbox.h>

class Node;
class Interface;
class InterfaceWireless;
class Link;
//Doc:ClassXRef
class NNIfLink { // Two nodes, link and receiving interface, and canvasitem
public:
  NNIfLink(Node* n1, Node* n2, Interface* is, Interface* id,
           Link* ls,Link* ld, QCanvasPolygonalItem* item)
    : s(n1), d(n2), ifaceSource(is), ifaceDest(id),
      linkSource(ls), linkDest(ld), canvasItem(item) { }
public:
  Node* s;
  Node* d;
  Interface* ifaceSource;
  Interface* ifaceDest;
  Link*      linkSource;
  Link*      linkDest;
  QCanvasPolygonalItem* canvasItem;
};

typedef std::vector<QCanvasPolygonalItem*> CPItemVec_t;
typedef std::vector<QCanvasItem*>          CItemVec_t;
typedef std::vector<NNIfLink>              NNIfLVec_t;
#endif

#include "handler.h"
#include "event.h"

typedef int    QTCoord_t;
typedef double QTDCoord_t;

#define MAX_SLIDER 20

// Define the QT Update events
//Doc:ClassXRef
class QTEvent : public Event {
public:
  typedef enum { START, STOP, PAUSE, UPDATE,
                 WIRELESS_TX_START, WIRELESS_TX_END, 
                 RECORD } QTEvent_t;
  QTEvent(QTEvent_t ev) : Event(ev), src(nil), dst(nil), iFace(nil),
      range(0), iter(0) { };
public:
  Node*    src;     // Source node  for wireless transmit animation
  Node*    dst;     // Destination node  for wireless transmit animation
  InterfaceWireless* iFace; // Interface for wireless  animation
  Meters_t range;   // Range of wireless transmission
  Count_t  iter;    // Iteration for wireless tx
};


class Node;
class QTEvent;
#ifdef HAVE_QT
//Doc:ClassXRef
class QTWindow : public QObject, public Handler {
  Q_OBJECT
public:
  QTWindow();
  ~QTWindow();
public:
  void Handle(Event*, Time_t);
  void DisplayTopology();
  void DisplayTopologyAndReturn();
  void ProcessEvents();
  void AnimationUpdateRate(Time_t);
  void UpdateTopology(bool = true); // True if forced topo redisplay
  void AddBackgroundLines(const LocationVec_t&);
  bool HasQuit() {return quit;}
  void PlaybackMode(bool);
  Mult_t PlaybackRate(){return currentUpdateRate;}
  void PlaybackPause();
  QPoint  LocationToPixels(const Location&); // Convert a location to pixels
  QPoint  LocationToPixelsAbs(const Location&); // Convert a location to pixels
  QTCoord_t NodePixelSize() const { return nodePixelSize; }
  void    WirelessTxStart(Node*, InterfaceWireless*, Node*, Meters_t);
  void    WirelessTxEnd(Node*, InterfaceWireless*, Node*, Meters_t);
  QTCoord_t CanvasX() { return canvasX;}
  QTCoord_t CanvasY() { return canvasY;}
  QCanvas*  Canvas()  { return canvas; }
  
public slots:
  void NewSliderValue(int);
  void Save();
  void Record(bool);
  void Play();
  void Stop();
  void Pause();
  void Quit();
  void Exit();
  void TimerDone();
private:
  void    Initialize();        // Initialize constants
  QPoint  NodeLocation(Node*); // Get the pixel coords, CENTER of the node
  void    DrawNode(Node*, Count_t);     // Draw the node
  void    DrawP2P(Node*, Node*);        // Draw a point to point link  
public:
  void    RedrawP2P(Interface*, Interface*);  // Redraw after failure/recovery
private:
  void    DrawQ(Node*, Node*);          // Draw a queue object
  void    DisplayAllNodes(bool = false);// Draw all nodes
  void    DisplayAllPackets();          // Draw all in-flight packets
  void    Clear();                      // Clear the display
  void    DisplayOnePacket(Node*, Node*, Mult_t, Mult_t, const QColor&);
  void    WaitWhilePaused();            // Wait until un-paused
  void    WirelessTxStart(Node*, InterfaceWireless*, Node*, Meters_t, Count_t);
  void    WirelessTxEnd(Node*, InterfaceWireless*, Node*, Meters_t, Count_t);
  void    RecordNextFrame(QTEvent*);
  // Private members
  QApplication* app;
  QCanvas*      canvas;
  QCanvasView*  view;
  QLCDNumber*   updateRate;
  QSlider*      slider;
  QLCDNumber*   simTime;
  QTimer*       readyTimer;
  QVBox*        box;      // The container for the other windows

  QTCoord_t     nodePixelSize;
  QTDCoord_t    zoomMult;
  QTCoord_t     border;   // size of border, x and y, both pixels
  QTCoord_t     canvasX;
  QTCoord_t     canvasY;
  QTEvent*      qtEvent;
  // Calculated by Initialize()
  QTDCoord_t    metersToPixelsX;
  QTDCoord_t    metersToPixelsY;
  Meters_t      smallestX;
  Meters_t      smallestY;
  Meters_t      largestX;
  Meters_t      largestY;
  bool          topologyDisplayed;
  bool          updateScheduled;
  bool          ready;
  // Controlling the animation
  Time_t        currentUpdateRate;
  bool          paused;
  bool          quit;
  bool          oneShot;
  bool          playbackMode;
  bool          recording;
  Count_t       recordingFrame;
  CItemVec_t    nodeItems;
  CPItemVec_t   linkItems;
  CPItemVec_t   packetItems;
  NNIfLVec_t    displayedLinks;
  double        updateRates[MAX_SLIDER];
};

#else
class QTEvent;
//Doc:ClassXRef
    class QTWindow : public Handler {
public:
  QTWindow();
  ~QTWindow();
public:
  void Handle(Event*, Time_t);
  void DisplayTopology();
  void DisplayTopologyAndReturn();
  void ProcessEvents();
  void AnimationUpdateRate(Time_t);
};
#endif

#endif

