/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Queue base class
// George F. Riley.  Georgia Tech, Spring 2002

// Virtual Base class for all queue types.
// Provides some common functionality

#include "queue.h"
#include "droptail.h"
#include "simulator.h"
#include "link.h"

// Static members
Queue*  Queue::defaultQueue  = nil;
Count_t Queue::defaultLength = DEFAULT_DROPTAIL_LIMIT;
Count_t Queue::defaultLimitPkts = 0; // Default is no packet limit

DCount_t Queue::Average()
{ // Get average queue length
  Time_t now = Simulator::Now();
  if (now == startTime) return 0.0;    // Avoid divide by 0
  return totalByteSeconds / (now - startTime);
}

void     Queue::ResetAverage()
{ // Start a new averaging interval
  startTime = Simulator::Now();
  lastUpdate = startTime;
  totalByteSeconds = 0;
}

void     Queue::UpdateAverage()
{ // Queue is changing, update average
  Time_t now = Simulator::Now();
  totalByteSeconds += (now - lastUpdate) * Length();
  lastUpdate = now;
}

bool     Queue::Detailed()
{
  return detailed; // True if detailed packet enquing
}

void     Queue::Detailed(bool d)
{
  detailed = d;
}

Time_t   Queue::QueuingDelay()
{
  if (!interface) return 0;  // No idea if no interface assigned
  Rate_t linkBw = interface->GetLink()->Bandwidth();
  return ((Length() * 8.0) / linkBw); // Amount of time to drain queue
}

DCount_t Queue::TotalWorkload()
{
  return totalByteSeconds;
}

void Queue::SetInterface(Interface* i)
{ // Set the associated interface
  interface = i;
}

Count_t Queue::DropCount()
{
  return dropCount;
}

Count_t Queue::EnqueueCount()
{
  return totEnq;
}

void    Queue::CountEnq(Packet* p)
{ // Called by interface when link is not busy.
  // Packets bypass the queue, but should be counted for statistics
  // Packet pointer is needed by some queuing methods (dualqueue)
  // to determine which sub-queue should be used.
  totEnq++;
}

void    Queue::ResetStats()
{
  dropCount = 0;
  totEnq = 0;
}

void    Queue::AddForcedLoss(Time_t t, Count_t c, Time_t e)
{
  if (!forcedLosses) forcedLosses = new LossList_t();
  LossList_t::iterator i;
  for (i = forcedLosses->begin(); i != forcedLosses->end(); ++i)
    {
      if (t < i->time) break; // Insert before here
    }
  forcedLosses->insert(i, ForcedLoss(t, c, e));
}

bool    Queue::CheckForcedLoss(bool remove)
{
  if (!forcedLosses) return false; // No forced loss list
  if (forcedLosses->empty()) return false;  // list is empty
  LossList_t::iterator i;
  Time_t now = Simulator::Now();
  bool r = false;
  
  for (i = forcedLosses->begin(); i != forcedLosses->end(); )
    {
      if (now < i->time) return r; // Done checking
      if (now < i->expiration) r = true;
      if (remove) i->count--;
      if (i->count == 0 || now >= i->expiration)
        { // Done with this one, remove it
          LossList_t::iterator j = i;
          ++i;
          if (remove) forcedLosses->erase(j);
        }
      if (r) return r; // No need to keep checking
    }
  return r; // code later
}

bool   Queue::CheckSpoofedSource(Packet*)
{
  return false; // Not checked in "normal" queues
}



// Static Methods
void   Queue::Default(const Queue& d)
{
  if (defaultQueue) delete defaultQueue; // Delete any existing
  defaultQueue = d.Copy();
}

Queue* Queue::Default()
{
  if (!defaultQueue) defaultQueue = new DropTail();
  return defaultQueue;
}

void   Queue::DefaultLength(Count_t l)
{ 
  defaultLength = l;
  Default()->SetLimit(l);
}

void   Queue::DefaultLimitPkts(Count_t l)
{ 
  defaultLimitPkts = l;
  Default()->SetLimitPkts(l);
}


