/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Queue base class
// George F. Riley.  Georgia Tech, Spring 2002

// Virtual Base class for all queue types.

#ifndef __queue_h__
#define __queue_h__

#include <deque>
#include <list>

#include "packet.h"

class Interface;

//Doc:ClassXRef
class ForcedLoss { // Information for forcing packet losses
public:
  ForcedLoss(Time_t t, Count_t c) 
    : time(t), count(c), expiration(INFINITE_TIME) {}
  ForcedLoss(Time_t t, Count_t c, Time_t e) 
    : time(t), count(c), expiration(e) {}
public:
  Time_t  time;  // Time for loss
  Count_t count; // Number of losses starting at (time)
  Time_t  expiration; // Expiration time
};

typedef std::list<Packet*> PktList_t; // List of pending packets
typedef std::list<ForcedLoss> LossList_t; // List of forced losses

//Doc:ClassXRef
class NDInfo { // Information for non-detailed queues
public:
  NDInfo() : startTx(0), stopTx(0), size(0), updated(false) { }
  NDInfo(Time_t s, Time_t e, Size_t sz)
    : startTx(s), stopTx(e), size(sz), updated(false) { }
public:
  Time_t startTx; // Time transmit starts
  Time_t stopTx;  // Time transmit ends
  Size_t size;    // Size of packet in bytes
  bool   updated; // True if queue size already updated for this one
};

typedef std::deque<NDInfo> NDDeq_t; // Deque of Non-detailed info

//Doc:ClassXRef
class Queue {
//Doc:Class The class {\tt Queue} defines the base class for interface queues
//Doc:Class that enque packets before they are sent down on the link
//Doc:Class Different queuing disciplines may inherit from this class to
//Doc:Class define their own functionality  
public:
  //Doc:Method
  Queue(bool d = true) 
    : totalByteSeconds(0), startTime(0.0), 
    lastUpdate(0.0), detailed(d), forcedLosses(nil), verbose(false),
    dropCount(0), totEnq(0), ndDeq(nil), interface(nil)
    { }
  //Doc:Desc This method is the default constructor

  // Destructor
  virtual ~Queue() { }
  
  //Doc:Method
  virtual bool Enque(Packet*) = 0;    // Return true if enqued, false if not
  //Doc:Desc This method is used to enque a packet in the queue
  //Doc:Arg1 the pointer to the packet to be added to the queue
  //Doc:Return a boolean to indicate if the enque operation succeeded
  
  //Doc:Method
  virtual Packet* Deque() = 0;        // Return next element (NULL if none)
  //Doc:Desc This method deques a packet from the queue and returns it
  //Doc:Return Pointer to the packet
  
  //Doc:Method
  virtual Count_t Length() = 0;       // How many bytes are enqued?
  //Doc:Desc This method returns the size of the queue in terms of bytes
  //Doc:Return number of bytes
  
  //Doc:Method
  virtual Count_t LengthPkts() = 0;   // How many packets are enqued?
  //Doc:Desc This method returns the size of the queue in terms of the
  //Doc:Desc number of packets
  //Doc:Return number of packets enqued
  
  //Doc:Method
  virtual Queue* Copy() const = 0;    // Make a copy of this queue
  //Doc:Desc This method makes a copy of the queue and returns a pointer
  //Doc:Desc to the copy
  //Doc:Return pointer to a copy of the queue
  
  //Doc:Method
  virtual void SetLimit(Count_t) = 0; // Set new limit
  //Doc:Desc This method sets a limit to the size of the queue. this limit
  //Doc:Desc is in terms of bytes
  //Doc:Arg1 size of the queue in bytes
  
  //Doc:Method
  virtual Count_t GetLimit() const = 0;     // Get limit (bytes)
  //Doc:Desc This method returns the size of the queue in terms of the bytes
  //Doc:Return Max limit of the queue size in bytes
  
  //Doc:Method
  virtual void SetLimitPkts(Count_t) = 0;   // Set packets limit
  //Doc:Desc This method sets a limit to the size of the queue. this limit
  //Doc:Desc is in terms of number of packets that can be enqued
  //Doc:Arg1 size of the queue in packets
  
  //Doc:Method
  virtual Count_t GetLimitPkts() const = 0; // Get packets limit
  //Doc:Desc This method returns the size of the queue in terms of the packets
  //Doc:Return Max limit of the queue size in number of packets
  
  
  //Doc:Method
  DCount_t Average();                 // Get average queue length
  //Doc:Desc This method returns the average queue length
  //Doc:Return average queue length
  
  //Doc:Method
  void     ResetAverage();            // Start a new averaging interval
  //Doc:Desc This method resets the averages and starts a new interval
  //Doc:Desc to calculate averages
  
  //Doc:Method
  void     UpdateAverage();           // Queue is changing, update average
  //Doc:Desc This method updates the averages. ala moving averages
  
  //Doc:Method
  virtual bool Check(Size_t, Packet* = nil) = 0;// Test if buff space available
  //Doc:Desc This method checks if buffer space (of a given size) is available
  //Doc:Arg1 The size to be checked
  //Doc:Arg2 The pointer to a packet to be queued.
  //Doc:Return boolean to indicate true or false
  
  //Doc:Method
  virtual bool Detailed();            // True if detailed model w/ local queing
  //Doc:Desc This method returns a bool to indicate if detailed queue model
  //Doc:Desc is to be used
  //Doc:Return bool to indicate true or false
  
  //Doc:Method
  virtual void Detailed(bool);        // Set detailed on/off
  //Doc:Desc This method is used to set if detailed model is to be used
  //Doc:Arg1 bool to indicate if detailed model is needed
  
  //Doc:Method
  virtual Time_t   QueuingDelay();    // Calculate queuing delay at current lth
  //Doc:Desc This method calculates the current queuing delay
  //Doc:Return the queuing delay
  
  //Doc:Method
  virtual DCount_t TotalWorkload();   // Return the byte-seconds (workload)
  //Doc:Desc This method calculates the toal work load in terms of byte-secs
  //Doc:Return The workload
  
  //Doc:Method
  virtual void     SetInterface(Interface*);  // Let the q know which interface
  //Doc:Desc This method attaches the queue to an interface
  //Doc:Arg1 The interface to which it is to be attached
  
  //Doc:Method
  virtual Count_t  DropCount();       // Number of dropped packets
  //Doc:Desc This method returns the number of dropped packets.
  //Doc:Return  the number of dropped packets.
  
  //Doc:Method
  virtual Count_t  EnqueueCount();    // Number of enqueued packets
  //Doc:Desc This method returns the number of packets enqueued
  //Doc:Return  the number of enqueued packets.
  
  
  //Doc:Method
  virtual void     CountEnq(Packet* = nil); // Count an enque
  //Doc:Method

  virtual void     ResetStats();      // Reset the statistics to zero
  //Doc:Desc This method zeroes down all the statistics pertaining to this
  //Doc:Desc queue

  //Doc:Method
  void             AddForcedLoss(Time_t, Count_t = 1, Time_t = INFINITE_TIME);
    //Doc:Desc Adds information about a forced  loss.  Allows a simulation to
    //Doc:Desc force a packet loss at a particular time, to observe 
    //Doc:Desc protocol behavior in the presence of losses.
    //Doc:Arg1 Time of the loss.
    //Doc:Arg2 Number of consecutive packets to drop, stating at the time.
    //Doc:Arg3 Expiration time for this loss.  If all the requested losses
    //Doc:Arg3 are not accomplished by this time, then the remainder
    //Doc:Arg3 are not enforced.

  //Doc:Method
  bool             CheckForcedLoss(bool = true);
    //Doc:Desc Checks if a forced loss should be enforced.
    //Doc:Desc Called by each subclass
    //Doc:Desc prior to enquing a packet.
    //Doc:Desc The forced loss will be counted, and the entry
    //Doc:Desc removed if all forced losses have been enforced.
    //Doc:Arg1 True if the forced loss entry should be removed by this call.
    //Doc:Return True if time for a forced loss.

  //Doc:Method
  virtual bool CheckSpoofedSource(Packet*);
    //Doc:Desc Some queue types enforce egress filtering and drop spoofed
    //Doc:Desc  source addresses.  This checks if the packet should
    //Doc:Desc be dropped for spoofing.  Returns false for all queue
    //Doc:Desc types that don't check this, and true for those that
    //Doc:Desc do check and that find a spoofed source address.
    //Doc:Arg1 Packet to check.
    //Doc:Return True if packet should be dropped due to spoofing.


  void             Verbose(bool v = true) { verbose = v;}

public:
  // Managing the default queue
  //Doc:Method
  static void    Default(const Queue& d);
  //Doc:Desc This method sets a the default queue
  //Doc:Arg1 reference to a queue object
  
  //Doc:Method
  static Queue*  Default();
  //Doc:Desc This method returns the default queue object
  //Doc:Return Pointer to the default queue
  
  //Doc:Method
  static void    DefaultLength(Count_t l);
  //Doc:Desc This method sets the default queue length in bytes
  //Doc:Arg1 the count in bytes
  
  //Doc:Method
  static Count_t DefaultLength() { return defaultLength;}
  //Doc:Desc This method returns the default queue size in bytes.
  //Doc:Return the default size in bytes
  
  //Doc:Method
  static void    DefaultLimitPkts(Count_t l);
  //Doc:Desc This method sets the default queue length in terms of the
  //Doc:Desc number of packets
  //Doc:Arg1 the count in bytes
  

private:
  DCount_t    totalByteSeconds;     // Area under time/occupancy curve
  Time_t      startTime;            // Start time of this measurement
  Time_t      lastUpdate;           // Time of last update
  bool        detailed;             // True if using detailed queuing
  LossList_t* forcedLosses;         // List of pending forced losses
protected:
  bool        verbose;              // True if verbose debug
  Count_t     dropCount;            // Total dropped packets
  Count_t     totEnq;               // Total enqueued packets
  NDDeq_t*    ndDeq;                // Pointer to non-detailed deq info
public:
  Interface*  interface;            // Associated interface
private:
  // Static Members
  static Queue*  defaultQueue;      // Default queue type
  static Count_t defaultLength;     // Default queue length
  static Count_t defaultLimitPkts;  // Default limit (packets)
};

#endif



