/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Rate and Time Parsing classes
// George F. Riley.  Georgia Tech, Spring 2002

// Classes to allow rate and time specification with common abbreviations

#ifndef __ratetimeparse_h__
#define __ratetimeparse_h__

#include <string>

#include "common-defs.h"

//Doc:ClassXRef
class Rate {
  //Doc:Class Defines an object that can be used to define rates
  //Doc:Class (bits / second) in familiar units, such as 10Mb.
  //Doc:Class Objects of class {\tt Rate} have an automatic typecast
  //Doc:Class define to type {\tt Rate\_t}, so a {\tt Rate} object can be
  //Doc:Class used anywhere a {\tt Rate\_t} variable is needed.
public:
  //Doc:Method
  Rate(const std::string);  // Specify rate as a string with Mb, Kb, etc.
    //Doc:Desc Construct a {\tt Rate} object specifying a rate as a string.
    //Doc:Arg1 String specifying the rate.
    //Doc:Arg1 The rates can have the suffixes "b" (bits), "B" (bytes),
    //Doc:Arg1 "Kb" (kilobits), "KB" (kilobytes)
    //Doc:Arg1 "Mb" (megabits), "MB" (megabytes)
    //Doc:Arg1 "Gb" (gigabits), "GB" (gigabytes).

  Rate(const char*);        // Specify rate as a char* with Mb, Kb, etc.
public:
  operator Rate_t () const; // Provide automatic typecast to Rate_t
private:
  Rate_t Parse(const std::string); // Convert from string to rate_t
private:
  Rate_t rate;
};

//Doc:ClassXRef
class Time {
  //Doc:Class Defines an object that specifies time values with familiar
  //Doc:Class suffixes such as "ms".
  //Doc:Class Objects of class {\tt Time} have an automatic typecast
  //Doc:Class define to type {\tt Time\_t}, so a {\tt Time} object can be
  //Doc:Class used anywhere a {\tt Time\_t} variable is needed.
public:
  //Doc:Method
  Time(const std::string); // Specify time as a string with ms, us, etc.
    //Doc:Desc Construct a {\tt Time} object specifying a time as a string.
    //Doc:Arg1 String specifying the time.
    //Doc:Arg1 The times can have the suffixes "h" (hours), "m" (minutes),
    //Doc:Arg1 "s" (seconds), "sec" (seconds)
    //Doc:Arg1 "ms" (milliseconds), "us" (microseconds)
    //Doc:Arg1 "ns" (nanoseconds)
  Time(const char*);       // Specify time as a char* with ms, us, etc.
public:
  operator Time_t () const;       // Provide automatic typecast to Time_t
  operator std::string() const;   // Provide automatic typecast to string
private:
  Time_t Parse(const std::string);// Convert from string to Time_t
private:
  Time_t time;
};

#endif
