/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Ring topology
// Monirul I Sharif,  Georgia Tech, Fall 2003

// Define the ring topology using point-to-point links

#define DEBUG_MASK 0x01
#include "debug.h"
#include "ring.h"
#include "node.h"
#include "linkp2p.h"
#include "mask.h"
#include <math.h>
#include <stdio.h>


using namespace std;

// Constructors

Ring::Ring(Count_t count, IPAddr_t i)
{
  ConstructorHelper(count, Linkp2p::Default(), i);
}

Ring::Ring(Count_t count, const Linkp2p& link, IPAddr_t i)
{
  ConstructorHelper(count, link, i);
}

// Access functions
Node* Ring::GetNode(Count_t c)
{ // Get Specified node

  NodeId_t firstId = first;    // First ID for this level

  if (c < nodeCount) 
    return Node::GetNode(firstId + c);
  

  return nil; // Out of range
}


void Ring::BoundingBox(const Location& ll, const Location& ur)
{
  Meters_t xRadius = fabs(ur.X() - ll.X())/2;
  Meters_t yRadius = fabs(ur.Y() - ll.Y())/2;

  Meters_t xCenter = (ur.X() + ll.X())/2;
  Meters_t yCenter = (ur.Y() + ll.Y())/2;

  double angle = (2 * M_PI)/ nodeCount;

  NodeId_t thisNode = first;
  const NodeVec_t& nodes = Node::GetNodes();
  
  DEBUG0((cout<<"Putting Ring nodes in place"<<endl));

  // Assign locations for each level
  for (Count_t l = 0; l < nodeCount; ++l)
    {
      Meters_t yLoc = yCenter + yRadius * sin(angle * l);
      Meters_t xLoc = xCenter + xRadius * cos(angle * l);
      
      nodes[thisNode++]->SetLocation(xLoc,yLoc);

      DEBUG0((cout<<"Putting node "<<l<<" in ("<<xLoc<<","<<yLoc<<")"<<endl));
    }
}


// Private methods
void Ring::ConstructorHelper(Count_t count,
      		             const Linkp2p& link,
                             IPAddr_t leafIP)
{
  first = Node::nextId;
  Node *n = new Node ();
  Node *firstnode = n;
  nodeCount = count;

  IPAddr_t  nextIP = leafIP;

  for (Count_t l = 1; l <= nodeCount; ++l)
    { // Create each subsequent level

      Node *newnode;

      if (l!=nodeCount)
	newnode = new Node();
      else
	newnode = firstnode;

      if (nextIP == IPADDR_NONE)
	{ // No IP specified
	  n->AddDuplexLink(newnode, link);
	}
      else
	{
	  n->AddDuplexLink(newnode, link,
			   nextIP++, Mask(32), 
			   IPADDR_NONE, Mask(32));
	}
      n = newnode;
    }
  last = Node::nextId;
}

