/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/
// Georgia Tech Network Simulator - AODV Routing class
// George F. Riley.  Georgia Tech, Summer 2003

// AODV routing, maintained by Application-aodv.cc

#ifndef __routingaodv_h__
#define __routingaodv_h__

#include <list>
#include <algorithm>

#include "debug.h"
#include "common-defs.h"
#include "globalstats.h"

class Interface;
class PDU;
class Node;
class Packet;
class AODVNeighbor;
class AODVPrecursor;

typedef unsigned char   Byte; // define the byte. 
               //Don't understand why they don't use unsigned char?

#define INFINITY2               0xff

typedef std::list<AODVNeighbor *> AODVNeighborList_t; 
typedef std::list<AODVPrecursor *> AODVPrecursorList_t; 

// 
class AODVNeighbor {
	public:
		AODVNeighbor(Long_t a) { Address = a; }
	public:
		IPAddr_t        Address;
		double          Expire;
};

class AODVPrecursor {
	public:
		AODVPrecursor(Long_t a) {Address = a;}
	public:
		IPAddr_t Address;
};
		

//Doc:ClassXRef
class AODVRoutingEntry {
public:
  AODVRoutingEntry();
  AODVRoutingEntry(Interface* i, IPAddr_t a);
  AODVRoutingEntry(const AODVRoutingEntry& c)
    : pInterface(c.pInterface), Address(c.Address)
    { DBG((Stats::routingEntAllocated++));}
  ~AODVRoutingEntry();
public:
  void NeighborInsert(IPAddr_t id);
  AODVNeighbor * NeighborLookup(IPAddr_t id);
  
  void PrecursorInsert(IPAddr_t id);
  AODVPrecursor * PrecursorLookup(IPAddr_t id);
  void PrecursorDelete(IPAddr_t id);
  void PrecursorDelete(void);
  bool PrecursorEmpty(void); //to test if the Precursor list is empty
  
public:
  Interface* pInterface;  // Interface for route

  IPAddr_t  DstIPAddr;	  // Real destnation ip address, not next hop!
   
  IPAddr_t  Address;      // next hop actually, Destination address for route

	public:
      double	ReqTimeout;   // when I can send another request
      Byte    ReqCount;     // number of route requests;
	  Seq_t   SeqNo;
	  Long_t  Hops;
	  int     LastHopCount;

	  double  Expire;  //When entry expires
	  Byte    Flags;
#define RTF_DOWN 0
#define RTF_UP 1
#define RTF_IN_REPAIR 2
#define MAX_HISTORY 3
	  double  DiscLatency[MAX_HISTORY];
	  unsigned char HistoryIndex;
	  int     RequestLastTTL;
	  
	  AODVPrecursorList_t  PrecursorList;
	  // a lot of neighbors that are using this route.
	  AODVNeighborList_t   NeighborList;
};

typedef std::list<AODVRoutingEntry *> AODVFIB_t; // Forwarding Information base
//
//NOTE: we don't use Mask for routing any more 
//      because this is ad hoc network.
//TODO: right?

typedef  void (*ResolveCallback)(Packet *pPacket, void *arg);
//Doc:ClassXRef
class RoutingAODV {
public:
	typedef enum { AODV } RType_t; //routing type.
				
public:
  RoutingAODV() {/* instance = this;*/} //anyway, one instance.
  virtual ~RoutingAODV(); 
  virtual void Default( AODVRoutingEntry); // Specify default route
  virtual AODVRoutingEntry * Add( IPAddr_t, Count_t,
                    Interface*, IPAddr_t); // Add a routing entry
  virtual AODVRoutingEntry * Lookup(Node*, IPAddr_t); // Find a routing entry
  virtual RoutingAODV* Clone();                // Get a Clone of the object
  virtual RType_t Type();                  // Find out routing type
  virtual void    InitializeRoutes(Node*); // Initialization
  virtual bool    NeedInit();              // True if init needed
  virtual Size_t  Size() const;            // Total size (bytes) of FIB

  virtual void Delete(IPAddr_t id);        // Delete a route.
   
public:
  ResolveCallback  RouteResolve; 
 						 //a function pointer to AODVApplication
  void *who;			//store the pointer of the AODVApplication
		  
  AODVFIB_t     fib;                       // The routing database
  AODVRoutingEntry * defaultRoute;         // Default route

  //static RoutingAODV *instance;            //Removed?same with following
protected:
  static RoutingAODV *defaultRouting;
};

#endif

