/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

#include <iostream>
#include "debug.h"
#include "routing-nvr.h"
#include "simulator.h"
#include "node.h"
#include "globalstats.h"
#include "queue.h"
#include "interface.h"
#include "packet.h"
#include "l2proto802.11.h"
#include "mystat.h"    // yj

// Uncomment below to enable debug level 0
//#define DEBUG_MASK 0x01

//#define TEST_WIRELESS
//#define TEST_REP_DST_ONLY
//#define TEST_LOAD_BALANCING
#define TEST_NC2DST

using namespace std;

// Default setting
bool RoutingNVR::enableL2Notify = true;
bool RoutingNVR::enableMobilityManage = true;
bool RoutingNVR::enableRingSearch = true;
bool RoutingNVR::enableDelayedNVReq = true;
bool RoutingNVR::enableLoadBalance = false;
Word_t RoutingNVR::defaultNVReqListSize = NVREQ_LIST_SIZE;
Word_t RoutingNVR::defaultSendBufSize = SEND_BUF_SIZE;


// NVRHeader

NVRHeader::NVRHeader() : NVROption(NV_ROT)
{
}

NVRHeader::NVRHeader(Byte_t pmid, Byte_t pmv, Word_t pnvl0, IPAddr_t porg,
	CNixVector& pnv)
  : NVROption(NV_ROT),
	opt(0), mid(pmid), mv(pmv), cmv(0), nvl0(pnvl0), org(porg),
	nvRoute(pnv)
{
}

NVRHeader::NVRHeader(const NVRHeader& r)
  : NVROption(r.type),
	opt(r.opt), mid(r.mid), mv(r.mv), cmv(r.cmv), nvl0(r.nvl0), org(r.org),
	nvRoute(r.nvRoute)
{
}

Size_t NVRHeader::Size() const
{
  Count_t sz =
	5 +                         // type, opt, mid, mv, cmv
	sizeof(nvl0) + sizeof(org) +
	nvRoute.SSize();
  return sz;
}


// NCReqHeader

NCReqHeader::NCReqHeader()
{
}

NCReqHeader::NCReqHeader(const NCReqHeader& r)
  : NVROption(r.type),
	id(r.id), src(r.src), dst(r.dst),
	mid(r.mid), mv(r.mv), nv(r.nv)
{
	nbrlist = r.nbrlist;
}

NCReqHeader::~NCReqHeader()
{
}

Size_t NCReqHeader::Size() const
{
  Count_t sz =
	3 +          // type, id, mid
	sizeof(src) + sizeof(dst) + sizeof(mv) +
	nv.SSize() + nbrlist.Size();
  return sz;
}


// NCRepHeader

NCRepHeader::NCRepHeader()
{
}

NCRepHeader::NCRepHeader(const NCRepHeader& r)
  : NVROption(r.type),
	id(r.id), src(r.src), dst(r.dst),
	mid(r.mid), mv(r.mv), nv(r.nv)
{
}

NCRepHeader::~NCRepHeader()
{
}

Size_t NCRepHeader::Size() const
{
  Count_t sz =
	3 +          // type, id, mid
	sizeof(src) + sizeof(dst) + sizeof(mv) +
	nv.SSize();
  return sz;
}


// NCSymReqHeader

NCSymReqHeader::NCSymReqHeader() : NVROption(NC_REQ_SYM)
{
}

NCSymReqHeader::NCSymReqHeader(const NCSymReqHeader& r)
  : NVROption(r.type),
	id(r.id), src(r.src), dst(r.dst),
	mid(r.mid), mv(r.mv), from_inf(r.from_inf),
	nvRoute(r.nvRoute), rnv(r.rnv)
{
}

Size_t NCSymReqHeader::Size() const
{
  /*
  Count_t sz =
	2            // type, id, mid
	//+ sizeof(src)
	+ sizeof(dst)
	+ 9          // sizeof(mv) = 1 + 4 + 4
	+ 6          // from_inf MAC address
	+ nvRoute.SSize() + rnv.SSize();
  return sz;
  */
  return 28 + nvRoute.Size() + rnv.Size();
}


// NCSymRepHeader

NCSymRepHeader::NCSymRepHeader() : NVROption(NC_REP_SYM)
{
}

NCSymRepHeader::NCSymRepHeader(const NCSymRepHeader& r)
  : NVROption(r.type),
	id(r.id), src(r.src), dst(r.dst),
	mid(r.mid), mv(r.mv), cmv(r.cmv), from_inf(r.from_inf),
	nv(r.nv), nvRoute(r.nvRoute)
{
}

Size_t NCSymRepHeader::Size() const
{
  Count_t sz =
	3            // type, id, mid
	//+ sizeof(src)
	+ sizeof(dst)
	+ 9          // sizeof(mv)
	+ 9          // sizeof(cmv)
	+ 6          // from_inf MAC address
	+ nv.SSize() + nvRoute.SSize();
  return sz;
}


// NVRFailHeader

Size_t NVRFailHeader::Size() const
{
  Count_t sz =
	2 +          // type, mid
	sizeof(dst) + sizeof(org) +
	sizeof(end) +
	9 +          // sizeof(cmv)
	nvRoute.SSize();
  return sz;
}


// RoutingNVR

RoutingNVR::RoutingNVR()
  : NCId(0), evNVList(nil), evNbr(nil)
{
  nvreqQue.reserve(5);
  nvreqSlot = new Uniform(0, 20);           // time slot
  nvreqJitter = new Uniform(0, 0.000010);   // jitter, 10 microsec
}

RoutingNVR::~RoutingNVR()
{
  // Cancel all timers
  CancelTimer(evNVList, true);
  CancelTimer(evNbr, true);

  if (nvreqSlot) delete nvreqSlot;
  if (nvreqJitter) delete nvreqJitter;
}

void RoutingNVR::DataRequest(Node* n, Packet* p, void* v)
{
  IPV4ReqInfo* ipInfo = (IPV4ReqInfo*)v;

  DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": data request to "
         << (string)ipInfo->dst
         << " uid " << p->uid
         << " time " << Simulator::Now() << endl
         << "  data request size " << p->Size() << endl);

  CRouteInfo* ri = GetRouteInfo(ipInfo->dst);
  if (!ri)
  {
	// Need queueing this packet, process this pending packet later
	SendBufferEnque(p, *ipInfo);

	// Check if there is already pending NV request for the same dst
	if (FindNVReq(ipInfo->dst) >= 0) return;

	// Invoke NV creation process
	Byte_t ttl = enableRingSearch ? TTL_INIT : TTL_MAX;
	SendNVRequest(n, ipInfo->dst, ttl);
	return;
  }

  // Use NV to route
  CNixVector nvec = ri->nv;    // Must copy first

  // NVR option
  NVRHeader* nvr = CreateNVRHeader(ri, nvec);
  nvr->opt = NVR_OPT_SYM;

  // Unicast
  L3Transmit(n, p, ipInfo->dst, ipInfo->ttl, ipInfo->l4proto, nvr);

  MyStat::pktsSent++;
}

// Add L3 header and transmit
void RoutingNVR::L3Transmit(Node* n, Packet* p, IPAddr_t dst, Count_t ttl,
							Proto_t l4proto, NVROption* option)
{
  IPV4Header* iphdr = new IPV4Header();
  iphdr->src = n->GetIPAddr();
  iphdr->dst = dst;
  iphdr->ttl = ttl;
  iphdr->protocol = l4proto;
  // iphdr->uid = p->uid;
  iphdr->options.push_back(option);
  p->PushPDU(iphdr);
  iphdr->totalLength = p->Size();

  if ((IPAddr_t)iphdr->dst == IPAddrBroadcast) {
	// Broadcast
	n->Broadcast(p, IPV4::Instance()->Proto());
  }
  else {
	// Unicast
	// Process NVR option, only NV_ROT subtype for now
	if (option->type != NVROption::NV_ROT) {
	  DEBUG0(cout << "Option is not correct" << endl);
	  delete p;
	  return;
	}
	NVRHeader* nvropt = (NVRHeader*)option;
	MACAddr next = Lookup(nvropt->nvRoute, nvropt->opt & NVR_OPT_SYM);
	Unicast(n->Interfaces()[0], p, next);
  }
}

CRouteInfo* RoutingNVR::GetRouteInfo(IPAddr_t dst)
{
  int i = NVList.Find(dst);
  return i < 0 ? 0 : (CRouteInfo*)NVList.Item[i];
}

CRouteInfo* RoutingNVR::GetRouteInfo(IPAddr_t dst, IPAddr_t org)
{
  int i = NVList.Find(dst, org);
  return i < 0 ? 0 : (CRouteInfo*)NVList.Item[i];
}

void RoutingNVR::SetAdNbrList(NCReqHeader* hdr)
{
  hdr->nbrlist = NbrTable;
}

MACAddr RoutingNVR::Lookup(CNixVector& nvec, bool reduce)
{
  MACAddr aix, next_hop;
  Word_t old_nvu = nvec.nvu;

  if (nvec.Get(aix))
  {
	// Other color nix
	if (!NbrTable.Nbr[aix.Lower()].Valid()) {
	  DEBUG0(cout << "  trying to access invalid neighbor at index "
			 << aix.Lower() << endl);
	  return MACAddr::NONE;
	}
	next_hop = NbrTable.Nbr[aix.Lower()].mac;
	DEBUG0(cout << "next_hop " << (string)next_hop << endl);
  }
  else
  {
	// White color nix
	next_hop = aix;
	int nix = NbrTable.Find(aix);
	if (nix < 0) nix = NbrTable.Add(aix);
	else NbrTable.Nbr[nix].val = true;    // validate the neighbor
	if (nix < 0) {
	  DEBUG0(cout << "Lookup(): fail to NbrTable::Add()\n");
	  return MACAddr::NONE;
	}

	CNixVector tnv, nnv;
	tnv.Set(nix);
	nnv.ExpandBits(nvec.nvl+tnv.nvl-NO_COLOR_BITS-NO_MAC_BITS);

	bitcpy(nnv.nv, 0, nvec.nv, 0, old_nvu);
	nnv.nvl += old_nvu;
	bitcpy(nnv.nv, nnv.nvl, tnv.nv, 0, tnv.nvl);
	nnv.nvl += tnv.nvl;
	bitcpy(nnv.nv, nnv.nvl, nvec.nv, nvec.nvu, nvec.nvl-nvec.nvu);
	nnv.nvl += nvec.nvl-nvec.nvu;
	nnv.nvu = old_nvu + tnv.nvl;
	nvec = nnv;
  }

  if (reduce) {
	CNixVector tnv;
	tnv.ExpandBits(nvec.nvl-nvec.nvu);
	tnv.nvl = nvec.nvl-nvec.nvu;
	bitcpy(tnv.nv, 0, nvec.nv, nvec.nvu, tnv.nvl);
	nvec = tnv;
  }

  return next_hop;
}

void RoutingNVR::AddNix(NCReqHeader* hdr, const MACAddr& mac)
{
	int nix = hdr->nbrlist.Find(mac);
	if (nix < 0)
	  hdr->nv.Set(mac);
	else
	  hdr->nv.Set(nix);
}

bool RoutingNVR::DataIndication(Interface* i, Packet* p)
{
  IPV4Header* iphdr = (IPV4Header*)p->PeekPDU();
  NVROption* opt = (NVROption*)iphdr->options[0];
  Node* n = i->GetNode();

  VQue.Update(p->Size());

  DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": NVR_DataIndication"
		 << " type " << opt->type << " uid " << p->uid
		 << " time " << Simulator::Now() << endl
		 << "  load updated " << VQue.Load()
		 << " Byteseconds " << VQue.Byteseconds << endl);


  switch (opt->type)
	{
	case NVROption::NV_ROT:      return DoNVRot(i, p, opt);
	case NVROption::NV_REP:      DoNVRep(i, p, opt);      break;
	case NVROption::NV_ROT_FAIL: DoNVRotFail(i, p, opt);  break;
	case NVROption::NC_REQ:      DoNCReq(i, p, opt);      break;
	case NVROption::NC_REP:      DoNCRep(i, p, opt);      break;
	case NVROption::NC_REQ_SYM:  DoNCSymReq(i, p, opt);   break;
	case NVROption::NC_REP_SYM:  DoNCSymRep(i, p, opt);   break;
	default:
	  DEBUG0(cout << (string)IPAddr(n->GetIPAddr())
			 << ": unknown packet received" << endl);
	  delete p;
	}
  return true;
}

// Private Methods
bool RoutingNVR::DoNVRot(Interface* i, Packet* p, NVROption* opt)
{
  // Process NV Route Option
  // Get the node pointer from interface
  Node* n = i->GetNode();
  IPV4Header* iphdr = (IPV4Header*)p->PeekPDU();
  NVRHeader* nvrhdr = (NVRHeader*)opt;

  DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": NVR_DoNVRot"
		 << " time " << Simulator::Now() << endl
		 << "  src " << (string)iphdr->src << " dst " << (string)iphdr->dst
		 << " mv " << nvrhdr->mv << " uid " << p->uid << endl);

  if (n->GetIPAddr() == iphdr->src) {
	// Got packet originated by itself, just drop it

	DEBUG0(cout << (string)iphdr->src << ": astray packet dropped "
		   << p->uid << endl);

	delete p;
	Stats::pktsDropped++;
	return true;
  }

  // Process current metric vector
  nvrhdr->cmv++;

  if (n->GetIPAddr() == iphdr->dst) {
	// Destined for this node

	DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": data received"
		   << " uid " << p->uid << " time " << Simulator::Now() << endl
		   << " src " << (string)iphdr->src << " dst " << (string)iphdr->dst
		   << " nvl0 " << nvrhdr->nvl0 << " nvl " << nvrhdr->nvRoute.nvl
		   << " size " << p->Size() << endl);

	// Check if can return shortened NV to src
/*
	if (nvrhdr->nvRoute.nvl < nvrhdr->nvl0) {
	// NV shortened, return it only if this node has NV to src
	  const CRouteInfo* ri = GetRouteInfo(iphdr->src);
	  if (ri) {
	  // Has NV to src
		CNixVector nvec = ri->nv;    // Must copy first
		MACAddr next_mac = nvr->Lookup(nvec);   // Next hop mac addr
		// Create NV_REP packet
		Packet* rp = new Packet();
		NCRepHeader* rhdr = new NCRepHeader();
		rhdr->type = NV_REP;
		rhdr->id = 0;
		rhdr->src = cphdr.src;
		rhdr->dst = cphdr.dst;
		rhdr->mv.hop = cphdr.mv;
		rhdr->mid = cphdr.mid;
		rhdr->nv = cphdr.nv;
		rp->PushPDU(rhdr);
		// Insert routing header
		NVRHeader* rnvrhdr = new NVRHeader(ri->mid, ri->mv.hop, ri->nv.nvl,
		nvec);
		rp->PushPDU(rnvrhdr);
		// Send
		Unicast(i, rp, next_mac);
		DEBUG0(cout << "  shortened NV sent to " << (string)next_mac
		       << " uid " << rp->uid
	           << " src " << (string)rnvrhdr->src
			   << " dst " << (string)rnvrhdr->dst << endl);
	  }
	}
*/
	MyStat::pktsReceive(p->uid);
	return false;
  }

  // Needs forwarding
  //Word_t nvu0 = nvrhdr->nv.nvu;
  if (nvrhdr->nvRoute.IsEnd()) {
	// NV exhausted, not supposed to happen
	DEBUG0(cout << "  NV exhausted"
		   << " nvl " << nvrhdr->nvRoute.nvl
		   << " nvu " << nvrhdr->nvRoute.nvu << endl);
	delete p;
	return true;
  }

  MACAddr dst_mac = Lookup(nvrhdr->nvRoute, nvrhdr->opt & NVR_OPT_SYM);

  //cout << "Copying NVRHeader type " << nvrhdr->type << endl;
  //NVRHeader cphdr(*nvrhdr);    // copy NVRHeader for later use
  //cout << "Copying NVRHeader complete type " << cphdr.type << endl;

  Unicast(i, p, dst_mac);

  // If packet transmit is successful, cache NV from here to the dst
  // in case of asymmetric routing
/*
  if( cphdr.opt & NVR_OPT_SYM ) return;
  CNixVector nnv;
  int nnvl = cphdr.nvRoute.nvl - nvu0;
  nnv.ExpandBits(nnvl);
  bitcpy(nnv.nv, 0, cphdr.nvRoute.nv, nvu0, nnvl);
  nnv.nvl = nnvl;
  int hcnt = cphdr.mv - cphdr.cmv;
  int j = NVList.Find(cphdr.dst);
  if( j < 0 ) {
    // No cached NV, cache it
	CMetricVec mv(hcnt, 0, 0);
	NVList.Add(cphdr.mid, cphdr.dst, cphdr.src, mv, nnv);

	DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": nix vector cached"
	       << " time " << Simulator::Now() << endl
		   << "  dst " << (string)cphdr.dst << " hop " << hcnt
		   << " nvl " << nnv.nvl << endl
		   << "  nv ");
	DEBUG0(nnv.Dump());
  }
  else if( !IsWorse(j,hcnt) && IsShorter(j,nnv.nvl)) {
    // Has NV, update it only if this is not worse and shorter
	NVList.Update(j, hcnt, nnv);

	DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": nix vector updated"
           << " time " << Simulator::Now() << endl
		   << "  dst " << (string)cphdr.dst << " hop " << hcnt
		   << " nvl " << nnv.nvl << endl
		   << "  nv ");
	DEBUG0(nnv.Dump());
  }
*/
  return true;
}

void RoutingNVR::DoNVRep(Interface* i, Packet* p, NVROption* opt)
{
  // Get the node pointer from interface
  Node* n = i->GetNode();
  NCRepHeader* hdr = (NCRepHeader*)opt;

  // Update NV
  int j = NVList.Find(hdr->dst);
  if (j >= 0 && !IsWorse(j, hdr->mv.hop)) {
	// Update if has cached NV and this is not worse
	hdr->nv.nvu = 0;
	NVList.Update(j, hdr->mv.hop, hdr->nv);

	DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": nix vector updated"
		   << " time " << Simulator::Now() << endl
		   << "  dst " << (string)hdr->dst << " nvl " << hdr->nv.nvl << endl
		   << "  nv ");
	DEBUG0(hdr->nv.Dump());
  }
  delete p;
}

void RoutingNVR::DoNCReq(Interface* i, Packet* p, NVROption* opt)
{
  // Get the node pointer from interface
  Node* n = i->GetNode();
  NCReqHeader* hdr = (NCReqHeader*)opt;

  DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": NVREQ received" << endl
		 << "  src " << (string)hdr->src << " dst " << (string)hdr->dst
		 << " nc-id " << hdr->id << endl);

  CRepItem* item;
  if ((IPAddr_t)hdr->dst == n->GetIPAddr()) {
	// Destined for this node, return NV to source
	hdr->mv.hop++;
	item = NRList.Find(hdr->src, hdr->id, hdr->type);
	if (item && item->mv.hop <= hdr->mv.hop) {
	  delete p;
	  DEBUG0(cout << "  already processed, deleted\n");
	  return;
	}
	AddNix(hdr, i->GetMACAddr());
	// create NC reply packet
	Packet* rp = new Packet();

	NCRepHeader* rhdr = new NCRepHeader();
	rhdr->type = NVROption::NC_REP;
	rhdr->id = hdr->id;
	rhdr->src = hdr->src;
	rhdr->dst = hdr->dst;
	rhdr->mv = hdr->mv;
	rhdr->nv = hdr->nv;

	IPV4Header* iphdr = new IPV4Header();
//	iphdr->ttl = IPV4::DefaultTTL;
	iphdr->protocol = Proto();
	iphdr->src = n->GetIPAddr();
	iphdr->dst = rhdr->src;
	//  iphdr->uid = rp->uid;
	iphdr->options.push_back(rhdr);
	rp->PushPDU(iphdr);
	iphdr->totalLength = rp->Size();

	delete p;

	ReturnPacket(i, rp, rhdr->src);
	if( item ) item->mv = hdr->mv;
	else NRList.Add(rhdr->src, rhdr->id, rhdr->type, rhdr->mid, rhdr->mv);

	DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": NVREP sent" << endl
		   << "  src " << (string)rhdr->src << " dst " << (string)rhdr->dst
		   << " nvl " << rhdr->nv.nvl << endl);

  }
  else {
	// Forward NC request packet
	if (NCList.Find(hdr->src, hdr->id) >= 0) {
	  // Just ignore, already processed it
	  delete p;
	  DEBUG0(cout << "  already processed, deleted" << endl);
	  return;
	}
	// Check TTL in IP header before forwarding
	IPV4Header* iphdr = (IPV4Header*)p->PeekPDU();
	if (!iphdr->ttl) {
	  delete p;
	  DEBUG0(cout << "  TTL expired, deleted" << endl);
	  return;
	}

	// Forward, or do NC reply if has cached NV for dst
	NCList.Add(hdr->src, hdr->id);
	hdr->mv.hop++;
	AddNix(hdr, i->GetMACAddr());
	CRouteInfo* ri = GetRouteInfo(hdr->dst);
	if (!ri) {
	  // Forward NC request by broadcast
	  SetAdNbrList(hdr);
	  n->Broadcast(p, IPV4::Instance()->Proto());
	}
	else {
	  // NC reply
	  Packet* rp = new Packet();

	  NCRepHeader* rhdr = new NCRepHeader();
	  rhdr->type = NVROption::NC_REP;
	  rhdr->id = hdr->id;
	  rhdr->src = hdr->src;
	  rhdr->dst = hdr->dst;
	  rhdr->mv.hop = hdr->mv.hop + ri->mv.hop;
	  //rhdr->nv = hdr->nv + ri->nv;
	  rhdr->nv.ExpandBits(hdr->nv.nvl + ri->nv.nvl);
	  bitcpy(rhdr->nv.nv, 0, hdr->nv.nv, 0, hdr->nv.nvl);
	  rhdr->nv.nvl += hdr->nv.nvl;
	  bitcpy(rhdr->nv.nv, rhdr->nv.nvl, ri->nv.nv, 0, ri->nv.nvl);
	  rhdr->nv.nvl += ri->nv.nvl;

	  IPV4Header* iphdr = new IPV4Header();
//	  iphdr->ttl = IPV4::DefaultTTL;
	  iphdr->protocol = Proto();
	  iphdr->src = n->GetIPAddr();
	  iphdr->dst = rhdr->src;
	  // iphdr->uid = rp->uid;
	  iphdr->options.push_back(rhdr);
	  rp->PushPDU(iphdr);
	  iphdr->totalLength = rp->Size();

	  delete p;

	  ReturnPacket(i, rp, rhdr->src);
	  NRList.Add(rhdr->src, rhdr->id, rhdr->type, rhdr->mid, rhdr->mv);
	}
  }
}

void RoutingNVR::DoNCRep(Interface* i, Packet* p, NVROption* opt)
{
  // Get the node pointer from interface
  Node* n = i->GetNode();
  NCRepHeader* hdr = (NCRepHeader*)opt;

  DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": NVREP received\n"
		 << "  src " << (string)hdr->src << " dst " << (string)hdr->dst
		 << " hop " << hdr->mv.hop << " nvl " << hdr->nv.nvl << endl);

  CRepItem* item = NRList.Find(hdr->src, hdr->id, hdr->type);
  if (item && item->mv.hop <= hdr->mv.hop) {
	// Just ignore, already processed it
	delete p;
	DEBUG0(cout << "  already processed, deleted\n");
	return;
  }
  if( item ) item->mv = hdr->mv;
  else NRList.Add(hdr->src, hdr->id, hdr->type, hdr->mid, hdr->mv);

  if ((IPAddr_t)hdr->src == n->GetIPAddr())
  {
	// Destined for this node, cache NV
	IPAddr_t dst = hdr->dst;    // copy dst for pending packets
	int j = NVList.Search(hdr->mid, hdr->dst);
	if (j < 0) {
	  // First reply, cache NV
	  NVList.Add(hdr->mid, hdr->dst, hdr->src, hdr->mv, hdr->nv);
	  //NCList.Remove(hdr->src, hdr->id);

	  DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": nix vector cached"
			 << " time " << Simulator::Now() << endl
			 << "  dst " << (string)hdr->dst << " hop " << hdr->mv.hop
			 << " nvl " << hdr->nv.nvl << endl
			 << "  nv ");
	  DEBUG0(hdr->nv.Dump());
	}
	else if( IsBetter(j, hdr->mv.hop) ) {
	// Multiple replies, update only if this is a better route
	  NVList.Update(j, hdr->mv.hop, hdr->nv);

	  DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": nix vector updated"
			 << " time " << Simulator::Now() << endl
			 << "  dst " << (string)hdr->dst << " hop " << hdr->mv.hop
			 << " nvl " << hdr->nv.nvl << endl
			 << "  nv ");
	  DEBUG0(hdr->nv.Dump());
	}
	delete p;

	// Need to process pending packets
	ProcessPending(n, dst);
  }
  else
  {
	// Forward NC reply packet
	n->Broadcast(p, IPV4::Instance()->Proto());
  }
}

void RoutingNVR::DoNCSymReq(Interface* i, Packet* p, NVROption* opt)
{
  // Get the node pointer from interface
  Node* n = i->GetNode();
  NCSymReqHeader* hdr = (NCSymReqHeader*)opt;

  DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": NVREQ received"
		 << " time " << Simulator::Now() << endl
		 << "  src " << (string)hdr->src << " dst " << (string)hdr->dst
		 << " nc-id " << hdr->id
		 << " hop " << hdr->mv.hop << " load " << hdr->mv.pks << endl);

  CRepItem* item;
  //MACAddr mac = i->GetMACAddr();
  // Mimic wireless, yj
  MACAddr mac = n->Interfaces()[0]->GetMACAddr();

  if (n->GetIPAddr() == hdr->dst) {
	// Destined for this node
	hdr->mv.pks = (Long_t)(hdr->mv.pks/hdr->mv.hop);
	hdr->mv.hop++;
	item = NRList.Find(hdr->src, hdr->id, NVROption::NC_REP_SYM);

	if (item) {
#ifdef TEST_LOAD_BALANCING
	  if (item->mv.hop < hdr->mv.hop ||
		  item->mv.hop == hdr->mv.hop && item->mv.pks <= hdr->mv.pks) {
		delete p;
		DEBUG0(cout << "  already processed, deleted" << endl);
		return;
	  }
#else
	  if (item->mv.hop <= hdr->mv.hop) {
		// Without load balancing
		delete p;
		DEBUG0(cout << "  already processed, deleted" << endl);
		return;
	  }
#endif
	}

	// Construct RNV and store it
	// RNV
	CNixVector rnv;
	if (!ConstructNV(rnv, hdr->rnv, hdr->from_inf)) return;

	// Store RNV
	NVList.Add(hdr->mid, hdr->src, hdr->dst, hdr->mv, rnv);

	DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": nix vector cached"
		   << " time " << Simulator::Now() << endl
		   << "  dst " << (string)hdr->src << " hop " << hdr->mv.hop
		   << " nvl " << rnv.nvl << endl
		   << "  nv ");
	DEBUG0(rnv.Dump());

	// NV reply packet
	Packet* rp = new Packet();
	NCSymRepHeader* rhdr = CreateNCSymRep(hdr, mac);

	IPV4Header* iphdr = new IPV4Header();
	iphdr->ttl = IPV4::DefaultTTL;
	iphdr->protocol = Proto();
	iphdr->src = n->GetIPAddr();
	iphdr->dst = rhdr->src;
	// iphdr->uid = rp->uid;
	iphdr->options.push_back(rhdr);
	rp->PushPDU(iphdr);
	iphdr->totalLength = rp->Size();

	MACAddr next_hop = hdr->from_inf;
	delete p;
	Unicast(i, rp, next_hop);
	if (item)
	  item->mv = hdr->mv;
	else
	  NRList.Add(rhdr->src, rhdr->id, rhdr->type, rhdr->mid, rhdr->mv);

	DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": NVREP sent"
		   << " time " << Simulator::Now() << endl
		   << "  src " << (string)rhdr->src << " dst " << (string)rhdr->dst
		   << " hop " << rhdr->mv.hop
		   << " load " << (double)rhdr->mv.pks << endl
		   << "  nvl " << rhdr->nv.nvl
		   << " rnvl " << rhdr->nvRoute.nvl << endl);
  }
  else {
	// Check if this request was recently processed
	if (NCList.Find(hdr->src, hdr->id) >= 0) {
	  // Just ignore, already processed it
	  delete p;
	  DEBUG0(cout << "  already processed, deleted\n");
	  return;
	}

	// Forward, or do NV reply if has NV for dst
	NCList.Add(hdr->src, hdr->id, hdr->from_inf);
	DEBUG0(cout << "# of nc items " << NCList.Item.size() << endl);

	// Process currnet metric vector
	hdr->mv.hop++;
	hdr->mv.pks += VQue.Load();

	// Construct RNV
	CNixVector rnv;
	if (!ConstructNV(rnv, hdr->rnv, hdr->from_inf)) return;

	// Check if has NV for dst
#ifdef TEST_REP_DST_ONLY
	CRouteInfo* ri = 0;
#else
	CRouteInfo* ri = GetRouteInfo(hdr->dst);
#endif
	if (!ri) {
	  // Check TTL in IP header before forwarding
	  IPV4Header* iphdr = (IPV4Header*)p->PeekPDU();
	  if (!iphdr->ttl) {
		delete p;
		DEBUG0(cout << "  TTL expired, deleted" << endl);
		return;
	  }
	  hdr->rnv = rnv;
	  hdr->from_inf = mac;
	  if (enableDelayedNVReq) {
		// Delay sending out NCREQ to avoid collision by broadcast
		//Time_t nc_delay = nvreqSlot->IntValue() * 100e-6;
		Time_t nc_delay = nvreqSlot->IntValue() * 510e-6 + nvreqJitter->Value();
		DEBUG0(cout << "nc_delay " << nc_delay << endl);
		if (!nc_delay)
		  n->Broadcast(p, IPV4::Instance()->Proto());
		else {
		  AddNCDelay(p, nc_delay);
		  DEBUG0(cout << "# of delayed NVREQs " << nvreqQue.size() << endl);
		}
	  }
	  else {
		n->Broadcast(p, IPV4::Instance()->Proto());
	  }
	}
	else {
#ifdef TEST_NC2DST
	  // Forward NV request to dst using NV
	  if (!hdr->nvRoute.nvl) {
		// Copy stored NV to the header
		hdr->nvRoute = ri->nv;
	  }
	  // Unicast to dst using NV in the header
	  hdr->rnv = rnv;
	  hdr->from_inf = mac;
	  MACAddr next_hop = Lookup(hdr->nvRoute, true);
	  Unicast(i, p, next_hop);
#else
	  // Return NV reply
	  // Store RNV
	  CMetricVec mv;
	  mv.hop = hdr->mv.hop;
	  mv.pks = (Long_t)(hdr->mv.pks/hdr->mv.hop);    // wrong, yj
	  NVList.Add(hdr->mid, hdr->src, n->GetIPAddr(), mv, rnv, NV_ATTR_UNI);

	  Packet* rp = new Packet();
	  NCSymRepHeader* rhdr = CreateNCSymRep(hdr, mac, ri);
	  IPV4Header* iphdr = new IPV4Header();
	  iphdr->ttl = IPV4::DefaultTTL;
	  iphdr->protocol = Proto();
	  iphdr->src = n->GetIPAddr();
	  iphdr->dst = rhdr->src;
	  // iphdr->uid = rp->uid;
	  iphdr->options.push_back(rhdr);
	  rp->PushPDU(iphdr);
	  iphdr->totalLength = rp->Size();

	  MACAddr next_hop = hdr->from_inf;
	  delete p;
	  Unicast(i, rp, next_hop);

	  DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": NVREP sent"
			 << " time " << Simulator::Now() << endl
			 << "  src " << (string)rhdr->src << " dst " << (string)rhdr->dst
			 << " nvl " << rhdr->nv.nvl
			 << " rnvl " << rhdr->nvRoute.nvl << endl);
#endif
	}
  }
}

void RoutingNVR::DoNCSymRep(Interface* i, Packet* p, NVROption* opt)
{
  // Get the node pointer from interface
  Node* n = i->GetNode();
  IPV4Header* iphdr = (IPV4Header*)p->PeekPDU();
  NCSymRepHeader* hdr = (NCSymRepHeader*)opt;

  DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": NVREP received"
		 << " time " << Simulator::Now() << endl
		 << "  src " << (string)hdr->src << " dst " << (string)hdr->dst
		 << " hop " << hdr->mv.hop << " nvl " << hdr->nv.nvl << endl);

  if (n->GetIPAddr() == hdr->src) {
	// Destined for this node
	//hdr->cmv.hop++;

	// Construct NV
	CNixVector nnv;
	if (!ConstructNV(nnv, hdr->nv, hdr->from_inf)) return;

	// Cancel NVREQ timeout
	Time_t acq_time = 0;
	int j = FindNVReq(hdr->dst);
	if (j >= 0) {
	  acq_time = Simulator::Now() - nvreqList[j].time;
	  CancelTimer(nvreqList[j].timer, true);
	}

	// Store NV, need more comparison, yj
	//CRouteInfo* ri = GetRouteInfo(hdr->dst);
	CRouteInfo* ri = GetRouteInfo(hdr->dst, hdr->src);
	Byte_t attr = iphdr->src.ip==hdr->dst.ip ? 0 : NV_ATTR_UNI;
	if (!ri) {
	  // First reply, store NV
	  NVList.Add(hdr->mid, hdr->dst, hdr->src, hdr->mv, nnv, attr);
	  DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": nix vector cached"
			 << " acq_time " << acq_time
			 << " time " << Simulator::Now() << endl
			 << "  dst " << (string)hdr->dst << " hop " << hdr->mv.hop
			 << " load " << (double)hdr->mv.pks << " nvl " << nnv.nvl << endl
			 << "  nv ");
	  DEBUG0(nnv.Dump());
	  //PrintRoute(nnv);
	}
	else {
	  // Already has it, update only when it is better
#ifdef TEST_LOAD_BALANCING
	  if (hdr->mv.hop < ri->mv.hop ||
		  hdr->mv.hop == ri->mv.hop && hdr->mv.pks < ri->mv.pks)
#else
	  if (hdr->mv.hop < ri->mv.hop)
#endif
	  {
		ri->mid = hdr->mid;
		ri->attr = attr;
		ri->mv = hdr->mv;
		ri->nv = nnv;

		DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": nix vector updated"
			   << " time " << Simulator::Now() << endl
			   << "  dst " << (string)hdr->dst << " hop " << hdr->mv.hop
			   << " load " << (double)hdr->mv.pks << " nvl " << nnv.nvl << endl
			   << "  nv ");
		DEBUG0(nnv.Dump());
		//PrintRoute(nnv);
	  }
	}

	IPAddr_t dst = hdr->dst;    // copy dst for pending packets
	delete p;

	// Need to process pending packets
	ProcessPending(n, dst);
  }
  else {
	// Need to forward
	hdr->cmv.hop++;

	// Construct NV and store it, store RNV
	// Construct NV
	CNixVector nnv;
	if (!ConstructNV(nnv, hdr->nv, hdr->from_inf)) return;
	hdr->nv = nnv;

	// Store NV to dst
	CMetricVec mv;
	mv.hop = hdr->cmv.hop;
	mv.pks = hdr->mv.pks;
	Byte_t attr = iphdr->src.ip==hdr->dst.ip ? 0 : NV_ATTR_UNI;
	NVList.Add(hdr->mid, hdr->dst, hdr->src, mv, nnv, attr);

	DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": nix vector cached"
		   << " time " << Simulator::Now() << endl
		   << "  dst " << (string)hdr->dst << " hop " << mv.hop
		   << " load " << mv.pks << " nvl " << nnv.nvl << endl
		   << "  nv ");
	DEBUG0(nnv.Dump());

	// Store NV to src
	mv.hop = hdr->mv.hop - hdr->cmv.hop;
	mv.pks = hdr->mv.pks;   // wrong, yj
	NVList.Add(hdr->mid, hdr->src, iphdr->src, mv, hdr->nvRoute, attr);

	DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": nix vector cached"
		   << " time " << Simulator::Now() << endl
		   << "  dst " << (string)hdr->src << " hop " << mv.hop
		   << " load " << mv.pks << " nvl " << hdr->nvRoute.nvl << endl
		   << "  nv ");
	DEBUG0(hdr->nvRoute.Dump());

	MACAddr next_hop = Lookup(hdr->nvRoute, true);
	DEBUG0(cout << "next by nvRoute " << (string)next_hop << endl);

	int j = NCList.Find(hdr->src, hdr->id);
	MACAddr from;
	if (j < 0) {
	  DEBUG0(cout << "can't find nc item\n");
	  from = MACAddr::NONE;
	}
	else
	  from = NCList.At(j)->from_inf;
	DEBUG0(cout << "from " << (string)from << endl);

	if (!(from == next_hop)) {
	  //cout << "HuH? next_hop doesn't coincide with from"
	  //<< " next_hop " << (string)next_hop
	  //<< " from " << (string)from << endl;
	}

	hdr->from_inf = i->GetMACAddr();
	// Minic wireless, yj
	//hdr->from_inf = i->GetNode()->Interfaces()[0]->GetMACAddr();
	Unicast(i, p, next_hop);
  }
}

void RoutingNVR::DoNVRotFail(Interface* i, Packet* p, NVROption* opt)
{
  Node* n = i->GetNode();
  NVRFailHeader* hdr = (NVRFailHeader*)opt;

  DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": NVRFAIL received"
		 << " time " << Simulator::Now() << endl
		 << "  org " << (string)hdr->org << " dst " << (string)hdr->dst
		 << " end " << (string)hdr->end << endl);

  // Process current metric vector
  hdr->cmv.hop++;

  // Replace NV to dst with NV to end
  CRouteInfo* ri = GetRouteInfo(hdr->dst, hdr->org);
  if (ri) {
	if (ri->nv.Advance(hdr->cmv.hop)) {
	  CNixVector nnv;
	  nnv.ExpandBits(ri->nv.nvu);
	  bitcpy(nnv.nv, 0, ri->nv.nv, 0, ri->nv.nvu);
	  nnv.nvl = ri->nv.nvu;

	  ri->target = hdr->end;
	  ri->mv.hop = hdr->cmv.hop;
	  ri->nv = nnv;

	  DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": nix vector replaced"
			 << " time " << Simulator::Now() << endl
			 << "  dst " << (string)hdr->end << " hop " << ri->mv.hop
			 << " nvl " << ri->nv.nvl << endl
			 << "  nv ");
	  DEBUG0(ri->nv.Dump());
	}
	else {
	  // Something wrong in the NV
	  NVList.Remove(hdr->dst, hdr->org);
	  NVList.Remove(hdr->org, hdr->dst);
	}
  }

  if (n->GetIPAddr() == hdr->org) {
	// Destined for this node
	delete p;
  }
  else {
	// Change origin of stored NV to org into end
	ri = GetRouteInfo(hdr->org, hdr->dst);
	if (ri) ri->origin = hdr->end;

	// Need to forward
	if (hdr->nvRoute.IsEnd()) {
	  // NV exhausted
	  delete p;
	  return;
	}
	MACAddr next_hop = Lookup(hdr->nvRoute);
	Unicast(i, p, next_hop);
  }
}

void RoutingNVR::DoL2Fail(Interface* i, Packet* p)
{
  Node* n = i->GetNode();
  IPV4Header* iphdr = (IPV4Header*)p->PeekPDU();
  NVROption* opt = (NVROption*)iphdr->options[0];

  DEBUG0(cout << (string)IPAddr(n->GetIPAddr())<< ": L2 error detected"
		 << " uid " << p->uid << " time " << Simulator::Now() << endl);

  IPAddr org;
  Byte_t mid;
  switch (opt->type)
	{
	case NVROption::NV_ROT: {
	  NVRHeader* nvrhdr = (NVRHeader*)opt;
	  if (!(nvrhdr->opt & NVR_OPT_SYM)) {
		// No RNV, just drop the packet
		delete p;
		Stats::pktsDropped++;
		return;
	  }
	  DEBUG0(cout << "NV_ROT" << endl);
	  org = nvrhdr->org;
	  mid = nvrhdr->mid;
	  break;
	}

	case NVROption::NC_REP_SYM: {
	  NCSymRepHeader* rephdr = (NCSymRepHeader*)opt;
	  DEBUG0(cout << "NC_REP_SYM" << endl);
	  org = rephdr->dst;
	  mid = rephdr->mid;
	  break;
	}

	case NVROption::NV_ROT_FAIL: {
	  NVRFailHeader* failhdr = (NVRFailHeader*)opt;
	  DEBUG0(cout << "NV_ROT_FAIL" << endl);
	  org = failhdr->dst;
	  mid = failhdr->mid;

	  /* Don't break but return here.
		 Otherwise, NV back to src may not be found.
		 We need quite acurate movement detection method
		 to determine if a neighbor really moved out or not.
		 Theshold-based re-xmit could be one option.

		 break;
	  */
	}

	case NVROption::NC_REQ_SYM:
	  DEBUG0(cout << "NC_REQ_SYM" << endl);
	  // need something for unicast nv request

	default:
	  delete p;
	  Stats::pktsDropped++;
	  return;
	} //switch

  // Remove NV to dst and invalidate associated item in the neighbor table
  int i_nv = NVList.Find(iphdr->dst, org);
  if (i_nv < 0) {
	DEBUG0(cout << "HuH? fail to NVList.Find(dst, org)" << endl);
  }
  else {
    // Find associated neighbor from NV and invalidate it
	CNixVector* nv = &NVList.Item[i_nv]->nv;
	MACAddr aix;
	nv->Get(aix);
	NbrTable.Invalidate(aix.Lower());

	// Remove NV to dst
	NVList.Remove(i_nv);
	DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": nix vector removed"
		   << " time " << Simulator::Now() << endl
		   << "  dst " << (string)iphdr->dst
		   << " org " << (string)org << endl);
  }

  // Check if the packet is originated by NV origin
  if (n->GetIPAddr() == org || opt->type == NVROption::NV_ROT_FAIL) {
	delete p;
	Stats::pktsDropped++;
	return;
  }

  // Retrieve NV back to org
  CRouteInfo* ri = GetRouteInfo(org, iphdr->dst);
  if (!ri) {
	DEBUG0(cout << "  no NV back to " << (string)org << endl);
	delete p;
	Stats::pktsDropped++;
	return;
  }

  /* Don't do this.
	 Otherwise, NV back to org for in-flight packets cannot be found.
	 It is hopefully ok to leave it as it is because now it is end point.

	 // Change origin of NV back to org into current node
	 ri->origin = n->GetIPAddr();
  */

  // Return NVRFail message to origin
  Packet* ep = new Packet();
  NVRFailHeader* fhdr
	= new NVRFailHeader(iphdr->dst, org, mid, n->GetIPAddr(), ri->nv);

  IPV4Header* iph = new IPV4Header();
  iph->ttl = IPV4::DefaultTTL;
  iph->protocol = Proto();
  iph->src = n->GetIPAddr();
  iph->dst = org;
  // iph->uid = ep->uid;
  iph->options.push_back(fhdr);
  ep->PushPDU(iph);
  iph->totalLength = ep->Size();

  MACAddr next_hop = Lookup(fhdr->nvRoute);
  Unicast(i, ep, next_hop);
  delete p;
  Stats::pktsDropped++;
}

NVRHeader* RoutingNVR::CreateNVRHeader(CRouteInfo* ri, CNixVector& nv)
{
  NVRHeader* hdr =
	new NVRHeader(ri->mid, ri->mv.hop, ri->nv.nvl, ri->origin, nv);
  if (ri->attr & NV_ATTR_UNI) hdr->opt |= NVR_OPT_RNV;
  return hdr;
}

NCSymRepHeader* RoutingNVR::CreateNCSymRep(NCSymReqHeader* hdr, MACAddr& mac,
	CRouteInfo* ri)
{
  // Construct NV reply option from NV request option
  NCSymRepHeader* rhdr = new NCSymRepHeader();
  rhdr->id = hdr->id;
  rhdr->src = hdr->src;
  rhdr->dst = hdr->dst;
  rhdr->from_inf = mac;
  rhdr->nvRoute = hdr->rnv;
  if (!ri) {
	rhdr->mid = hdr->mid;
	rhdr->mv = hdr->mv;
  }
  else {
	rhdr->mid = METRIC_HOP;
	rhdr->mv.hop = hdr->mv.hop + ri->mv.hop;
	rhdr->mv.pks = hdr->mv.pks/hdr->mv.hop;
	//rhdr->mv.pks =
	//(Long_t)(hdr->mv.pks + ri->mv.pks*(ri->mv.hop-1))/(rhdr->mv.hop-1);
	rhdr->cmv = ri->mv;
	rhdr->nv = ri->nv;
  }
  return rhdr;
}

void RoutingNVR::ReturnPacket(Interface* i, Packet* p, IPAddr_t dst)
{
  Node* n = i->GetNode();
  CRouteInfo* ri = GetRouteInfo(dst);
  if (!ri) {
	// No cached NV, broadcast
	L3Transmit(n, p, IPAddrBroadcast, 255, Proto(), nil);
  }
  else {
	// Unicast using NV
	CNixVector nvec = ri->nv;    // must copy first
	NVRHeader* nvr = CreateNVRHeader(ri, nvec);
	L3Transmit(n, p, dst, IPV4::DefaultTTL, Proto(), nvr);
  }
}

// Unicast for wireless network
void RoutingNVR::Unicast(Interface* i, Packet* p, MACAddr& dst_mac)
{

  /* dst_mac was obtained from Lookup()
	 Invalid dst_mac means that corresponding neighbor was removed or incorrect
	 Thus need to process it in the same way as in case of L2 xmit failure
  */
  if (dst_mac == MACAddr::NONE) {
	DEBUG0(cout << "called from Unicast()" << endl);
	DoL2Fail(i, p);
	return;
  }

#ifdef TEST_WIRELESS
  // Mimic wireless, yj
  Node* n = i->GetNode();
  IFVec_t ifvec = n->Interfaces();
  for (IFVec_t::iterator it=ifvec.begin(); it!=ifvec.end(); ++it) {
	Node* peer = (*it)->GetLink()->GetPeer(0)->GetNode();
	MACAddr mac = peer->Interfaces()[0]->GetMACAddr();
	if (mac == dst_mac) {
	  DEBUG0(cout << (string)IPAddr(n->GetIPAddr()) << ": unicast to"
			 << " mac " << (string)dst_mac << " uid " << p->uid
			 << " time " << Simulator::Now() << endl);
	  (*it)->Send(p, dst_mac, IPV4::Instance()->Proto());
	}
  }
#else
  // Link layer unicast
  DEBUG0(cout << (string)IPAddr(i->GetIPAddr()) << ": unicast to"
		 << " mac " << (string)dst_mac << " uid " << p->uid
		 << " time " << Simulator::Now() << endl);

  // Register L2 xmit failure handler
  if (enableL2Notify)
    ((L2Proto802_11*)i->GetL2Proto())->xmitFailHandler = this;

  // Unicast
  i->Send(p, dst_mac, IPV4::Instance()->Proto());
#endif
}

void RoutingNVR::ScheduleTimer(Event_t ev, NVREvent*& tev, Time_t when)
{
  if (when <= 0) {
	cout << "HuH? schedule negative time " << when << endl;
	return;
  }

  Time_t now = Simulator::Now();
  if (tev) {
	// Check if this event already scheduled at same time, ignore if so
	// Already scheduled for correct time
	if (tev->Time() == now+when) return;
	CancelTimer(tev);
  }
  if (!tev) {
	tev = new NVREvent(ev);
  }

  DEBUG0(cout << "TimerTest " << this << " scheduling new timer "
		 << tev->event << " time " << now
		 << " for " << when << " secs in future" << endl);

  timer.Schedule(tev, when, this);
}

void RoutingNVR::CancelTimer(NVREvent*& tev, bool delTimer)
{
  if (!tev) return;
  timer.Cancel(tev);
  if (delTimer) {
	delete tev;
	tev = 0;
  }
}

void RoutingNVR::Timeout(TimerEvent* ev)
{
  NVREvent* tev = (NVREvent*)ev;
  switch (tev->event)
	{
	case NVREvent::NV_REQ: {
	  // NV request timed out
	  NVReqItem item = nvreqList[tev->h];
	  DEBUG0(cout << (string)IPAddr(node->GetIPAddr()) << ": NVREQ timeout"
			 << " id " << item.id << " ttl " << item.ttl
			 << " target " << (string)IPAddr(item.dst)
			 << " time " << Simulator::Now() << endl);
	  nvreqList[tev->h].timer = nil;    // free timer
	  delete tev;
	  // Expanding ring search
	  if (enableRingSearch) {
		if (item.ttl == TTL_INIT) SendNVRequest(node, item.dst, TTL_MAX);
	  }
	  break;
	}

	case NVREvent::NV_LIST:
	  break;

	case NVREvent::NEIGHBOR:
	  break;

	case NVREvent::NV_REQ_DELAY: {
	  // Send out delayed NV request
	  Packet* p = nvreqQue[tev->h].p;
	  nvreqQue[tev->h].timer = nil;    // free timer
	  delete tev;
	  node->Broadcast(p, IPV4::Instance()->Proto());
	  break;
	}

	default:
	  DEBUG0(cout << "Unknown NVREvent type " << tev->event << endl);
	  break;
  } //switch
}

bool RoutingNVR::ConstructNV(CNixVector& dnv, CNixVector& snv, MACAddr& mac)
{
  // dnv = nix_from_mac + snv
  int nix = NbrTable.Find(mac);
  if (nix < 0) nix = NbrTable.Add(mac);
  else NbrTable.Nbr[nix].val = true;    // validate the neighbor
  if (nix < 0) {
	DEBUG0(cout << "ConstructNV(): fail to NbrTable::Add()\n");
	return false;
  }
  dnv.Set(nix);

  DEBUG0(cout << "# of nbrs " << NbrTable.Nbr.size() << endl);
  DEBUG0(cout << "nix " << nix << " dnv ");
  DEBUG0(dnv.Dump());
  DEBUG0(cout << "from " << (string)mac
		 << " nbr " << (string)NbrTable.Nbr[nix].mac << endl);

  if (!(mac == NbrTable.Nbr[nix].mac))
	cout << "GOGO\n";

  dnv.ExpandBits(snv.nvl);
  bitcpy(dnv.nv, dnv.nvl, snv.nv, 0, snv.nvl);
  dnv.nvl += snv.nvl;
  return true;
}

void RoutingNVR::SendNVRequest(Node* n, IPAddr dst, Byte_t ttl)
{
  // NV request packet
  Packet* p = new Packet();

  NCSymReqHeader* nvreq = new NCSymReqHeader();
  nvreq->type = NVROption::NC_REQ_SYM;
  nvreq->id = GetNCReqID();
  nvreq->src = n->GetIPAddr();
  nvreq->dst = dst;
  nvreq->mid = METRIC_HOP;
  nvreq->from_inf = n->Interfaces()[0]->GetMACAddr();

  Time_t to = ttl == TTL_INIT ? NVREQ_TO_INIT : NVREQ_TO_MAX;
  AddNVReq(dst, nvreq->id, ttl, to);
  NCList.Add(nvreq->src, nvreq->id);

  // Broadcast
  L3Transmit(n, p, IPAddrBroadcast, ttl, Proto(), nvreq);


  DEBUG0(cout << (string)nvreq->src << ": NVREQ sent"
		 << " uid " << p->uid << " time " << Simulator::Now() << endl
		 << "  target " << (string)nvreq->dst << " id " << nvreq->id << endl
		 << "  nvreq size " << p->Size() << endl);
}

void RoutingNVR::PrintRoute(CNixVector& nv)
{
  cout << (string)IPAddr(node->GetIPAddr()) << ": route\n";
  CNixVector nvec = nv;

  RoutingNVR* nvr = this;
  while (nvec.nvu < nvec.nvl)
  {
	MACAddr next = nvr->Lookup(nvec);
	if (next == MACAddr::NONE) break;
	NodeVec_t::size_type i = next.Lower()-1;
	nvr = (RoutingNVR*)Node::nodes[i]->GetRouting();
	cout << " -> " << next.Lower();
  }
  cout << endl;
}

void RoutingNVR::AddNCDelay(Packet* p, Time_t delay)
{
  // Find available slot in nvreqQue
  NCQue_t::size_type i;
  for (i = 0; i < nvreqQue.size(); ++i)
	if (!nvreqQue[i].timer) break;
  if (i == nvreqQue.size()) // no available slot
	nvreqQue.push_back(CNCQueItem());  // add one

  // Store NVREQ
  nvreqQue[i].p = p;
  nvreqQue[i].timer = new NVREvent(NVREvent::NV_REQ_DELAY, i);

  // Schedule timer
  ScheduleTimer(NVREvent::NV_REQ_DELAY, nvreqQue[i].timer, delay);
}

bool RoutingNVR::AddNVReq(IPAddr_t dst, Byte_t id, Byte_t ttl, Time_t to)
{
  // Search for available slot in NV request list
  NVReqList_t::size_type i;
  for (i = 0; i < nvreqList.size(); ++i)
	if (!nvreqList[i].timer) break;
  if (i == nvreqList.size()) {  // no available one
	if (i < defaultNVReqListSize)
	  nvreqList.push_back(NVReqItem());
	else {
	  // NVReqList full
	  cout << "HuH? NVReqList full\n";
	  return false;
	}
  }

  // Store item
  nvreqList[i] = NVReqItem(dst, id, ttl, Simulator::Now(), nil);

  // Schedule NV request timeout
  ScheduleTimer(NVREvent::NV_REQ, nvreqList[i].timer, to);

  return true;
}

int RoutingNVR::FindNVReq(IPAddr_t dst)
{
  // Search NV request list for dst
  NVReqList_t::size_type i;
  for (i = 0; i < nvreqList.size(); ++i) {
	if (!nvreqList[i].timer) continue;
	if (nvreqList[i].dst == dst)
	  return i;
  }
  return -1;
}

void RoutingNVR::SendBufferEnque(Packet* p, IPV4ReqInfo& info)
{
  if (sendBuffer.size() >= defaultSendBufSize) {
	DEBUG0(cout << "send buffer full oldest item dropped" << endl);
	sendBuffer.pop_front();
  }

  // Store packet and ipv4info
  sendBuffer.push_back(SendBufItem(p,info));
}

void RoutingNVR::ProcessPending(Node* n, IPAddr_t dst)
{
  SendBuf_t::iterator i = sendBuffer.begin();
  while (i != sendBuffer.end())
  {
	if (dst == i->info.dst) {
	  DEBUG0(cout << "  pending packet sent uid " << i->p->uid
			 << " time " << Simulator::Now() << endl);
	  DataRequest(n, i->p, &i->info);
	  sendBuffer.erase(i++);
	}
	else
	  ++i;
  }
}

void RoutingNVR::Notify(void* v)
{
  // Handle transmit failure notification from L2
  //cout << "L2 xmit failure detected\n";
  Packet* p = (Packet*)v;
  IPV4Header* iphdr = (IPV4Header*)p->PeekPDU();
  if (IPAddrBroadcast == iphdr->dst) {
	// Don't process broadcast packet
	delete p;  // L2 drop
	return;
  }
  //cout << "L2 xmit failure detected\n";
  if (enableMobilityManage) {
	// Trigger the NVR mobility management function
	DoL2Fail(node->Interfaces()[0], p);
	return;
  }
  delete p;  // L2 drop
}
