/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/


#ifndef __ROUTING_NVR_H
#define __ROUTING_NVR_H

#include <string.h>
#include "common-defs.h"
#include "routing.h"
#include "nixvector.h"
#include "ipv4.h"
#include "timer.h"
#include "notifier.h"

#define PROTO_NVR	0x8FF

using namespace std;

class Uniform;

class NVREvent : public TimerEvent
{
public:
  enum NVREvent_t { NV_REQ=1, NV_LIST, NEIGHBOR, NV_REQ_DELAY };
public:
  NVREvent(Event_t ev, int i=0) : TimerEvent(ev), h(i) {}
  virtual ~NVREvent() {}

public:
  int h;    // index of associated item in a table
};

class NVROption : public IPV4Options
{
public:
  enum { Number=25 };
  enum NVROpt_t {
	NC_REQ, NC_REP, NC_REQ_SYM, NC_REP_SYM, NV_ROT, NV_REP, NV_ROT_FAIL
  };

public:
  NVROption() : IPV4Options(Number) {}
  NVROption(Byte_t t) : IPV4Options(Number), type(t) {}
  virtual Size_t Size() const = 0;
  virtual PDU* Copy() = 0;

public:
  Byte_t type;    // NVR option type
};

// NV Route Option
// NVRHeader
class NVRHeader : public NVROption
{
public :
  NVRHeader();
  NVRHeader(const NVRHeader& r);
  NVRHeader(Byte_t mid, Byte_t hop, Word_t nvl, IPAddr_t org, CNixVector& nv);
  virtual ~NVRHeader() {}
  Size_t  Size() const;
  PDU*    Copy() { return new NVRHeader(*this); }

public:
  Byte_t      opt;
  Byte_t      mid;     // metric id
  Byte_t      mv;      // metric vector from src to dst, hop count for now
  Byte_t      cmv;     // current metric vector from src to here
  Word_t      nvl0;    // NVL at src
  IPAddr      org;     // origin, opposite end point to target
  CNixVector  nvRoute; // route in the NV form
};

class NCReqHeader : public NVROption
{
public :
  NCReqHeader();
  NCReqHeader(const NCReqHeader& r);
  virtual ~NCReqHeader();

  Size_t  Size() const;
  PDU*    Copy() { return new NCReqHeader(*this); }

public:
  Byte_t     id;
  IPAddr     src;
  IPAddr     dst;
  Byte_t     mid;
  CMetricVec mv;
  Word_t     nvl;
  CNixVector nv;
  CNbrTable  nbrlist;
};

class NCRepHeader : public NVROption
{
public :
  NCRepHeader();
  NCRepHeader(const NCRepHeader& r);
  virtual ~NCRepHeader();

  Size_t  Size() const;
  PDU*    Copy() { return new NCRepHeader(*this); }

public:
  Byte_t      id;
  IPAddr      src;
  IPAddr      dst;
  Byte_t      mid;
  CMetricVec  mv;
  CNixVector  nv;
};

// NV Request Option, Symmetric
// NCSymReqHeader
class NCSymReqHeader : public NVROption
{
public :
  NCSymReqHeader();
  NCSymReqHeader(const NCSymReqHeader& r);
  virtual ~NCSymReqHeader() {}

  Size_t  Size() const;
  PDU*    Copy() { return new NCSymReqHeader(*this); }

public:
  Byte_t      id;
  IPAddr      src;        // not part of option, specified in ip header
  IPAddr      dst;
  Byte_t      mid;
  CMetricVec  mv;
  MACAddr     from_inf;
  CNixVector  nvRoute;    // used for forwarding if valid
  CNixVector  rnv;        // NV from here to src
};

// NV Reply Option, Symmetric
// NCSymRepHeader
class NCSymRepHeader : public NVROption
{
public :
  NCSymRepHeader();
  NCSymRepHeader(const NCSymRepHeader& r);
  virtual ~NCSymRepHeader() {}

  Size_t  Size() const;
  PDU*    Copy() { return new NCSymRepHeader(*this); }

public:
  Byte_t      id;
  IPAddr      src;        // not part of option, specified in ip header
  IPAddr      dst;
  Byte_t      mid;
  CMetricVec  mv;
  CMetricVec  cmv;
  MACAddr     from_inf;
  CNixVector  nv;         // requested NV
  CNixVector  nvRoute;    // used for forwarding
};

// NVRFailHeader
// NV Route Failure Option
class NVRFailHeader : public NVROption
{
public:
  NVRFailHeader() : NVROption(NV_ROT_FAIL) {}
  NVRFailHeader(IPAddr d, IPAddr o, Byte_t m, IPAddr e, const CNixVector& nv)
	: NVROption(NV_ROT_FAIL), dst(d), org(o), mid(m), end(e), nvRoute(nv) {}
  virtual ~NVRFailHeader() {}

  Size_t Size() const;
  PDU* Copy() { return new NVRFailHeader(*this); }

public:
  IPAddr      dst;        // target
  IPAddr      org;        // origin
  Byte_t      mid;
  IPAddr      end;        // error reporting node
  CMetricVec  cmv;        // current metric vector
  CNixVector  nvRoute;    // NV to origin
};

// RoutingNVR Class

class RoutingNVR : public Routing, public TimerHandler, public NotifyHandler
{
public:
  RoutingNVR();
  virtual ~RoutingNVR();

  void  Default(RoutingEntry r) {}         // Specify default route
  RoutingEntry GetDefault() { return RoutingEntry();}// Default route
  void  Add(IPAddr_t, Count_t, Interface*, IPAddr_t) {}  // Add a routing entry
  RoutingEntry Lookup(Node*, IPAddr_t) { return RoutingEntry(); }
  RoutingEntry LookupFromPDU(PDU*) { return RoutingEntry(); }
  Routing* Clone() { return new RoutingNVR(); }
  RType_t  Type() { return NVR; }
  void  InitializeRoutes(Node*) {}         // Initialization
  bool  NeedInit() { return false; }       // True if init needed
  Size_t Size() const { return 0; }

  CRouteInfo* GetRouteInfo(IPAddr_t dst);
  CRouteInfo* GetRouteInfo(IPAddr_t dst, IPAddr_t org);
  Byte  GetNCReqID() { return ++NCId; }
  void  SetAdNbrList(NCReqHeader* hdr);
  MACAddr Lookup(CNixVector& nv, bool reduce=false);
  void  AddNix(NCReqHeader* hdr, const MACAddr& mac);
  bool  IsBetter(int i, Byte_t hop)  { return NVList.IsBetter(i,hop); }
  bool  IsWorse(int i, Byte_t hop)   { return NVList.IsWorse(i,hop); }
  bool  IsShorter(int i, Word_t nvl) { return NVList.IsShorter(i,nvl); }

  void  DataRequest(Node* n, Packet* p, void* v);
  bool  DataIndication(Interface* i, Packet* p);
  void  L3Transmit(Node* n, Packet* p, IPAddr_t dst, Count_t ttl,
				   Proto_t l4proto, NVROption* option);
  virtual Proto_t Proto()   { return PROTO_NVR; }
  void  Notify(void*);    // L2 xmit failure notification handler

private:
  bool  DoNVRot(Interface*, Packet*, NVROption*);
  void  DoNVRep(Interface*, Packet*, NVROption*);
  void  DoNCReq(Interface*, Packet*, NVROption*);
  void  DoNCRep(Interface*, Packet*, NVROption*);
  void  DoNCSymReq(Interface*, Packet*, NVROption*);
  void  DoNCSymRep(Interface*, Packet*, NVROption*);
  void  DoNVRotFail(Interface*, Packet*, NVROption*);
  void  DoL2Fail(Interface*, Packet*);
  NVRHeader* CreateNVRHeader(CRouteInfo*, CNixVector&);
  NCSymRepHeader* CreateNCSymRep(NCSymReqHeader* hdr, MACAddr& mac,
	CRouteInfo* ri=0);
  void  ReturnPacket(Interface*, Packet*, IPAddr_t);
  void  Unicast(Interface* i, Packet* p, MACAddr& dst_mac);
  bool  ConstructNV(CNixVector& dnv, CNixVector& snv, MACAddr& mac);
  void  SendNVRequest(Node* n, IPAddr dst, Byte_t ttl);
  void  PrintRoute(CNixVector& nv);
  void  AddNCDelay(Packet* p, Time_t delay);

  // Timer
  void  ScheduleTimer(Event_t ev, NVREvent*& tev, Time_t when);
  void  CancelTimer(NVREvent*& tev, bool delTimer=false);
  void  Timeout(TimerEvent* ev);    // timer event handler

  // NV request list
  bool  AddNVReq(IPAddr_t dst, Byte_t id, Byte_t ttl, Time_t to);
  int   FindNVReq(IPAddr_t dst);

  // Send buffer
  void  SendBufferEnque(Packet* p, IPV4ReqInfo& info);
  void  ProcessPending(Node* n, IPAddr_t dst);

public:
  static bool    enableL2Notify;
  static bool    enableMobilityManage;
  static bool    enableRingSearch;
  static bool    enableDelayedNVReq;
  static bool    enableLoadBalance;
  static Word_t  defaultNVReqListSize;
  static Word_t  defaultSendBufSize;

private:
  typedef vector<CNCQueItem>  NCQue_t;
  typedef vector<NVReqItem>   NVReqList_t;
  typedef list<SendBufItem>   SendBuf_t;

  Uniform*    nvreqSlot;     // NV request random time slot
  Uniform*    nvreqJitter;   // NV request jitter in a time slot
  Byte_t      NCId;          // NC sequence number
  Timer       timer;
  NVREvent*   evNVList;      // NV timeout
  NVREvent*   evNbr;         // Neighbor timeout

  CNCList     NCList;        // NC list
  CReplyList  NRList;        // Reply list
  CNVList     NVList;        // NV list
  CNbrTable   NbrTable;      // Neighbor table
  CVQue       VQue;          // Virtual queue for load estimatimation
  NCQue_t     nvreqQue;      // Que for delayed NV request
  NVReqList_t nvreqList;     // NV request list
  SendBuf_t   sendBuffer;    // Buffer for pending packets
};

#endif
