/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Routing base class
// George F. Riley.  Georgia Tech, Spring 2002

// Virtual Base class for all routing interfaces

#ifndef __routing_h__
#define __routing_h__

#include <vector>
#include <map>
#include <algorithm>

#include "debug.h"
#include "common-defs.h"
#include "globalstats.h"

class Interface;
class PDU;
class Node;
class Packet;

//Doc:ClassXRef
class RoutingEntry {
//Doc:Class This class defines each entry in the routing table. Currently
//Doc:Class this class supports only IPv4 addresses. It essentially
//Doc:Class defines the destination address and the interface to use to
//Doc:Class reach it.
public:
  RoutingEntry() : interface(0), nexthop(IPADDR_NONE)
    { DBG((Stats::routingEntAllocated++));}
  RoutingEntry(Interface* i, IPAddr_t a) 
    : interface(i), nexthop(a)
    { DBG((Stats::routingEntAllocated++));}
  RoutingEntry(const RoutingEntry& c)
    : interface(c.interface), nexthop(c.nexthop)
    { DBG((Stats::routingEntAllocated++));}
  ~RoutingEntry() { DBG((Stats::routingEntDeleted++));}
public:
  Interface* interface;  // Interface for route
  IPAddr_t   nexthop;    // Next hop IP Address
};

typedef std::map<IPAddr_t, RoutingEntry> FIB_t; // Forwarding Information base
typedef std::vector<FIB_t>       FIBVec_t;      // One FIB per mask length 
typedef std::map<Count_t, FIB_t> FIBMap_t;      // One FIB per mask length

// Virtual base class for all routing
//Doc:ClassXRef
class Routing {
//Doc:Class The class {\tt Routing} is a virtual base class for all the
//Doc:Class routing protocols that may be needed for a simulation. Each of
//Doc:Class these routing protocols have to derive from this class and define
//Doc:Class the routing protocol procedures
public:
  // Routing types
#ifdef SIM
  typedef enum { STATIC, DYNAMIC, MANUAL, NIXVECTOR, NVR, DSR, OSPF6 } RType_t;
#else
  typedef enum { STATIC, DYNAMIC, MANUAL, NIXVECTOR, NVR, DSR } RType_t;
#endif //SIM
public:
  //Doc:Method
  Routing();
  //Doc:Desc This method is the default constructor for class {\tt Routing}.

  //Doc:Method
  virtual ~Routing();
  //Doc:Desc This is the default destructor. Each routing protocol must
  //Doc:Desc define its own destructor.
  
public:
  //Doc:Method
  virtual void Default( RoutingEntry r) = 0;     // Specify default route
  //Doc:Desc This method adds the default route to a particular node.
  //Doc:Arg1 The default routing entry or the default gateway
  
  //Doc:Method
  virtual RoutingEntry GetDefault() = 0;         // Get default route
  //Doc:Desc This method returns the default route for this node
  //Doc:return The default routing entry or the default gateway
  
  //Doc:Method
  virtual void Add( IPAddr_t, Count_t,
                    Interface*, IPAddr_t) = 0;   // Add a routing entry
  //Doc:Desc This method is used to add a routing entry into the routing
  //Doc:Desc table.
  //Doc:Arg1 The destination IP Address
  //Doc:Arg2 The subnet Mask of the destination IP Address
  //Doc:Arg3 The interface to use
  //Doc:Arg4 The next hop IP Address
  
  //Doc:Method
  virtual RoutingEntry Lookup(Node*, IPAddr_t) = 0; // Find a routing entry
  //Doc:Desc This method implements the lookup mechanism at a  node
  //Doc:Desc for a given destination IP address
  //Doc:Arg1 Node which is looking up the routing table
  //Doc:Arg2 Destination IP address
  //Doc:Return The corresponding routing entry or nil
  
  //Doc:Method
  virtual RoutingEntry LookupFromPDU(PDU*) = 0;  // Find from PDU (NixVector)
  //Doc:Desc This method uses the PDU to look up the next route. Such a
  //Doc:Desc mechanism is used in routing protocols like nix vector routing
  //Doc:Arg1 The pointer to the PDU
  //Doc:Return The corresponding routing entry or nil
  
  //Doc:Method
  virtual Routing* Clone() = 0;                  // Get a clone of the object
  //Doc:Desc This method is used to make copies of the routing object.
  //Doc:Desc Such a mechanism is useful for example, in topology creation
  //Doc:Return A pointer to the cloned object
  
  //Doc:Method
  virtual RType_t Type() = 0;                    // Find out routing type
  //Doc:Desc This method enables to query the routing object of the
  //Doc:Desc routing protocol it implements
  //Doc:Return the routing type
  
  //Doc:Method
  virtual void    InitializeRoutes(Node*) = 0;   // Initialization
  //Doc:Desc This method initializes the routes and the routing table at a
  //Doc:Desc given node
  //Doc:Arg1 The pointer to the Node at which the routes are to be initialized
  
  //Doc:Method
  virtual bool    NeedInit() = 0;                // True if init needed
  //Doc:Desc This method determines if initialization is needed and returns
  //Doc:Desc a bool
  //Doc:Return true if initialization is needed
  
  // Statistics
  //Doc:Method
  virtual Size_t  Size() const = 0;              // Total size (bytes) of FIB
  //Doc:Desc This method gives the size of the routing table. or the
  //Doc:Desc Forwarding Information Base
  //Doc:Desc Size of the FIB

  virtual void    DataRequest(Node* n, Packet* p, void* v) {}
  virtual bool    DataIndication(Interface* i, Packet* p) { return true; }

public:
  // Get the default routing pointer
  //Doc:Method
  static Routing* Default();
  //Doc:Desc This method returns the default routing protocol object
  //Doc:Desc used by default
  //Doc:Return A pointer to the default routing object

  //Doc:Method
  static void SetRouting( Routing*);      // Set a new active protocol
  //Doc:Desc This method is used to set a new routing protocol for a simulation
  //Doc:Arg1 A pointer to the new routing protocol object
  
public:
  Node*  node;                                   // Node to which it's attached

private:
  static Routing* defaultRouting;                // Points to active protocol
};

#endif


