/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Packet/PDU Serializaton
// George F. Riley.  Georgia Tech, Spring 2002

// Serialization interface

#include <iostream>
#include <string.h>

#include "debug.h"
#include "serializable.h"
#include "hex.h"
#include "packet.h"

using namespace std;

// Helpers

char* Serializable::SerializeToBuffer(char* b, Size_t& r, size_t& l)
{ // Serialize an unsigned long 
  CheckSize(r, sizeof(l));
  memcpy(b, &l, sizeof(l));
  return b + sizeof(l);
}

#if defined (Linux) || defined (SunOS)
char* Serializable::SerializeToBuffer(char* b, Size_t& r, unsigned long& l)
{ // Serialize an unsigned long
  CheckSize(r, sizeof(l));
  memcpy(b, &l, sizeof(l));
  return b + sizeof(l);
}
#endif

char* Serializable::SerializeToBuffer(char* b, Size_t& r, unsigned short& s)
{ // Serialize an unsigned short
  CheckSize(r, sizeof(s));
  memcpy(b, &s, sizeof(s));
  return b + sizeof(s);
}

char* Serializable::SerializeToBuffer(char* b, Size_t& r, float& f)
{ // Serialize a four byte float
  CheckSize(r, sizeof(f));
  memcpy(b, &f, sizeof(f));
  return b + sizeof(f);
}

char* Serializable::SerializeToBuffer(char* b, Size_t& r, double& d)
{ // Serialize an eight byte float
  CheckSize(r, sizeof(d));
  memcpy(b, &d, sizeof(d));
  return b + sizeof(d);
}

char* Serializable::SerializeToBuffer(char* b, Size_t& r, MACAddr& m)
{ // Serialize a 48 bit MAC Address
  MACAddr::MACAddr_t u;
  MACAddr::MACAddr_t l;
  u = m.Upper();
  l = m.Lower();
  Word_t us = u; // 16 bit upper
#ifndef COMPACT
  CheckSize(r, sizeof(us));
  memcpy(b, &us, sizeof(us));
  b += sizeof(us);
#endif
  CheckSize(r, sizeof(l));
  memcpy(b, &l, sizeof(l));
  return b + sizeof(l);
}

char* Serializable::SerializeToBuffer(char* b, Size_t& r, char& d)
{ // Serialize an eight byte float
  CheckSize(r, sizeof(d));
  memcpy(b, &d, sizeof(d));
  return b + sizeof(d);
}

char* Serializable::ConstructFromBuffer(char* b, Size_t& r, size_t& s)
{
  CheckSize(r, sizeof(s));
  memcpy(&s, b, sizeof(s));
  return b + sizeof(s);
}

#if defined (Linux) || defined (SunOS)
char*  Serializable::ConstructFromBuffer(char* b, Size_t& r, unsigned long& l)
{
  CheckSize(r, sizeof(l));
  memcpy(&l, b, sizeof(l));
  return b + sizeof(l);
}
#endif

char*  Serializable::ConstructFromBuffer(char* b, Size_t& r, unsigned short& s)
{
  CheckSize(r, sizeof(s));
  memcpy(&s, b, sizeof(s));
  return b + sizeof(s);
}

char*  Serializable::ConstructFromBuffer(char* b, Size_t& r, float& f)
{
  CheckSize(r, sizeof(f));
  memcpy(&f, b, sizeof(f));
  return b + sizeof(f);
}

char*  Serializable::ConstructFromBuffer(char* b, Size_t& r, double& d)
{
  CheckSize(r, sizeof(d));
  memcpy(&d, b, sizeof(d));
  return b + sizeof(d);
}

char*  Serializable::ConstructFromBuffer(char* b, Size_t& r, MACAddr& m)
{
#ifndef COMPACT
  Word_t u;
  CheckSize(r, sizeof(u));
  memcpy(&u, b, sizeof(u));
  b += sizeof(u);
  m.Upper(u);
#endif
  MACAddr::MACAddr_t l;
  CheckSize(r, sizeof(l));
  memcpy(&l, b, sizeof(l));
  m.Lower(l);
  return b + sizeof(l);
}

char*  Serializable::ConstructFromBuffer(char* b, Size_t& r, char& l)
{
  CheckSize(r, sizeof(l));
  memcpy(&l, b, sizeof(l));
  return b + sizeof(l);
}

char* Serializable::GetSize(char* b, Size_t& r, Size_t& s)
{ // Get the size of the next size field
  CheckSize(r, sizeof(s));
  memcpy(&s, b, sizeof(s));
  return b + sizeof(s);
}

char* Serializable::PutSize(char* b, Size_t& r, Size_t& s)
{
  DEBUG0((cout << "Putting size " << Hex8n(s) << endl));
  Size_t ss = s;
  CheckSize(r, sizeof(s));
  memcpy(b, &ss, sizeof(ss));
  return b + sizeof(s);
}


// Private methods
void Serializable::CheckSize(Size_t& r, Size_t t)
{ // r = remaining, t = thissize
  if (t > r)
    {
      cout << "Serialization error; remaining " << r
           << " thissize " << t << endl;
      exit(1);
    }
  r -= t; // Reduce remaining for next time
}

// Static members

char* Serializable::Dump(char* b, Size_t s)
{ // Dump the serialized buffer for debugging
  cout << "Dumping " << s << " bytes" << endl;
  for (Size_t t = 0; t < s; ++t)
    {
      cout << Hex2n((unsigned char)b[t]);
      if (((t+1)%16 == 0)) cout << endl;
      else                 cout << " ";
    }
  if (s % 16) cout << endl;
  return b;
}

void Serializable::Validate(Packet* p)
{
  Size_t sz = p->SSize() + sizeof(Size_t); // Include size for length hdr
  char*  b = new char[sz];                 // Allocate serialization buffer
  char*  sb = b;                           // Save for reconstruction
  Size_t ssz = sz;                         // Save for reconstruction
  b = Serializable::PutSize(b, sz, sz);    // Store the total length
  // Copy to serialization buffer
  p->Serialize(b, sz);
  if (sz)
    cout << "Serializeable::Validate, expected non, got " << sz << endl;

  // Now reconstruct
  Size_t nsz = 0;
  Size_t sssz = ssz;
  char* nb = Serializable::GetSize(sb, sssz, nsz);
  Packet* np = new Packet(nb, sssz);

  // And serialize again
  Size_t sz1 = np->SSize() + sizeof(Size_t);  // Include size for length hdr
  char*  b1 = new char[sz];                  // Allocate serialization buffer
  char*  sb1 = b1;                           // Save for reconstruction
  Size_t ssz1 = sz1;                         // Save for reconstruction
  if (ssz1 != ssz)
    { // Oops, not same size
      cout << "Serializeable::Validate, expected identical sizes, got "
           << ssz1 << " " << ssz << endl;
    }
  b = Serializable::PutSize(b1, sz1, sz1);   // Store the total length
  // Copy to serialization buffer
  np->Serialize(b1, sz1);
  if (sz1)
    cout << "Serializeable::Validate, expected zero, got " << sz1 << endl;

  // At this point sb and sb1 buffers should be identical
  if (!memcmp(sb, sb1, ssz1))
    { // Oops, no match
      cout << "Serializable::Validate buffer mismatch " << endl;
      Serializable::Dump(sb,ssz);   
      Serializable::Dump(sb1,ssz1); 
    }
  delete [] sb;
  delete [] sb1;
}
