/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Star topology
// George F. Riley.  Georgia Tech, Fall 2002

// Define a star topology using point-to-point links

#ifndef __star_h__
#define __star_h__

#include <math.h>

#include "common-defs.h"

class Node;
class Linkp2p;
class Queue;
class Location;

//Doc:ClassXRef
class Star {
  //Doc:Class Define a star topology using point--to--point links
public:
  //Doc:Method
  Star(Count_t, IPAddr_t = IPADDR_NONE, SystemId_t = 0);// Number of nodes
    //Doc:Desc Create a star topology with the specified number of leaf nodes.
    //Doc:Desc Uses the default point--to--point link object.
    //Doc:Arg1 Number of leaf nodes desired.
    //Doc:Arg2 Starting \IPA\ of leaf nodes.
    //Doc:Arg3 Responsible system ID for distributed simulation

  //Doc:Method
  Star(Count_t, const Linkp2p&, IPAddr_t = IPADDR_NONE, SystemId_t = 0);
    //Doc:Desc Create a star topology with the specified number of leaf nodes,
    //Doc:Desc and the specified point--to--point link object.
    //Doc:Arg1 Number of leaf nodes desired.
    //Doc:Arg2 Link object to use for leaf to core links.
    //Doc:Arg3 Starting \IPA\ of leaf nodes.
    //Doc:Arg4 Responsible system ID for distributed simulation

  //Doc:Method
  Star(Count_t, Node*, IPAddr_t = IPADDR_NONE, SystemId_t = 0);
    //Doc:Desc Create a star topology with the specified number of leaf nodes.
    //Doc:Desc and an existing node for the core.
    //Doc:Desc Uses the default point--to--point link object.
    //Doc:Arg1 Number of leaf nodes desired.
    //Doc:Arg2 Node pointer for core node.
    //Doc:Arg3 Starting \IPA\ of leaf nodes.
    //Doc:Arg4 Responsible system ID for distributed simulation

  //Doc:Method
  Star(Count_t, Node*, const Linkp2p&, IPAddr_t = IPADDR_NONE, SystemId_t = 0);
    //Doc:Desc Create a star topology with the specified number of leaf nodes.
    //Doc:Desc and an existing node for the core.
    //Doc:Desc Uses the specified point--to--point link object for leaf links.
    //Doc:Arg1 Number of leaf nodes desired.
    //Doc:Arg2 Node pointer for core node.
    //Doc:Arg3 Link object to use for leaf to core links.
    //Doc:Arg4 Starting \IPA\ of leaf nodes.
    //Doc:Arg5 Responsible system ID for distributed simulation

  //Doc:Method
  Node* GetHub();                     // Get the hub node
    //Doc:Desc Get a pointer to the hub node in the star.
    //Doc:Return Node pointer to hub node.

  //Doc:Method
  Node* GetLeaf(Count_t);             // Get specified leaf
    //Doc:Desc Get a pointer to a leaf node.
    //Doc:Arg1 Leaf node index to get.
    //Doc:Return Pointer to specified leaf node.

  //Doc:Method
  Linkp2p* HubLink(Count_t);          // Get link from hub to leaf router
    //Doc:Desc Get a pointer to a hub to leaf link.
    //Doc:Arg1 Leaf node index.
    //Doc:Return Pointer to point--to--point link from hub to specified leaf

  //Doc:Method
  Linkp2p* LeafLink(Count_t);         // Get link from leaf to hub
    //Doc:Desc Get a pointer to a leaf to hub link.
    //Doc:Arg1 Leaf node index.
    //Doc:Return Pointer to point--to--point link from specified leaf to hub.

  //Doc:Method
  Queue*   HubQueue(Count_t);         // Get queue from hub to leaf
    //Doc:Desc Get a pointer to a hub to leaf queue.
    //Doc:Arg1 Leaf node index.
    //Doc:Return Pointer to queue from hub to specified leaf

  //Doc:Method
  Queue*   LeafQueue(Count_t);        // Get queue from leaf to hub
    //Doc:Desc Get a pointer to a leaf to hub queue.
    //Doc:Arg1 Leaf node index.
    //Doc:Return Pointer to queue from specified leaf to hub

  //Doc:Method
  Count_t  LeafCount() { return leafCount;}
    //Doc:Desc Get a count  of the number of leaf nodes.
    //Doc:Return Leaf node count.

  //Doc:Method
  void BoundingBox(const Location&, const Location&,
      Angle_t = 0, Angle_t = 2 * M_PI);
    //Doc:Desc Define a bounding box for this star object. 
    //Doc:Desc The bounding box
    //Doc:Desc is used to assign node locations, which are used during
    //Doc:Desc the simulation animation.
    //Doc:Desc If the hub node has an existing location, then this bounding box
    //Doc:Desc is relative to that node.  Otherwise, the hub is placed in the
    //Doc:Desc center of the box, with all leaf nodes extending radially from 
    //Doc:Desc the hub with length of half the bounding box size
    //Doc:Desc (minus a small fudge factor).
    //Doc:Arg1 Lower left corner of bounding box.
    //Doc:Arg2 Upper right corner of bounding box.
    //Doc:Arg3 The angle that the radial links are centered around.
    //Doc:Arg4 The total size of the arc for the radial links

public:
  NodeId_t   first;      // First leaf node id
  NodeId_t   last;       // Just log first/last+1 entries
  Node*      hub;        // Points to hub node
  Count_t    leafCount;  // Number of leaf nodes
private:
  void ConstructorHelper(Count_t, Node*, 
                         IPAddr_t, const Linkp2p&, SystemId_t);
};

#endif


