/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - TCP Base Class
// George F. Riley.  Georgia Tech, Spring 2002

// Base class for all TCP variants

#ifndef __tcp_h__
#define __tcp_h__

#include <iostream>
#include <map>

#include "pdu.h"
#include "l4protocol.h"
#include "datapdu.h"
#include "rtt-estimator.h"
#include "timer.h"
#include "seq.h"
#include "notifier.h"
#include "backplane.h"

class TCP;
//Doc:ClassXRef
class TCPHeader : public L4PDU {
public :
  TCPHeader();
  TCPHeader(char*, Size_t&, Packet*);   // Construct form serialized buffer
  Size_t  Size() const;
  PDU*    Copy();
  Proto_t Proto() { return 6;}         // TCP is protocol 6
  void    Trace(Tfstream&, Bitmap_t,  // Trace the contents of this pdu
                 Packet* = nil, const char* = nil);
  // Serialization
  Size_t SSize();                   // Size needed for serialization
  char*  Serialize(char*, Size_t&); // Serialize to a buffer
  char*  Construct(char*, Size_t&); // Construct from buffer
public:
  PortId_t sourcePort;
  PortId_t destPort;
  Seq      sequenceNumber;
  Seq      ackNumber;
  Byte_t   headerLength;
  Byte_t   flags;
  Word_t   window;
  Word_t   checksum;
  Word_t   urgentPointer;
  FId_t    fid;           // Flow id's are not part of TCP, but useful
  Count_t  dataLength;    // Also not part of TCP, but useful for tracing
  Count_t  cWnd;          // Congestion window is NOT in tcp header,
                          // but included here for simulation analysis
  TCP*     peer;          // Debug
};

// Data structures needed by all TCP Variants

// UnAckData_t is Used to keep track of unacked data
// Senders use this to track data sent but not acknowledged
// Recievers use this to buffer out-of-sequence receives


typedef std::pair<Packet*, IPAddr_t> PacketIPPair_t;
typedef std::map<Seq, PacketIPPair_t> UnAckData_t;

// Timer events
//Doc:ClassXRef
class TCPEvent : public TimerEvent {
public:
  // The types of TCP timer events
  typedef enum { CONNECTION_TIMEOUT, RETRANSMIT_PACKET,
                 DELAYED_ACK, TIMED_WAIT, LAST_ACK } TCPTimeout_t;
public:
  TCPEvent(Event_t ev) : TimerEvent(ev) { };
  virtual ~TCPEvent();
};

// Class TimeSeq is used for statistics logging of seq/time data
//Doc:ClassXRef
class TimeSeq {
public:
  TimeSeq(Time_t t, Seq_t s) : time(t), seq(s) { };
  TimeSeq(const TimeSeq& c) : time(c.time), seq(c.seq) { }
public:
  Time_t time;
  Seq_t  seq;
};
typedef std::vector<TimeSeq>    TimeSeq_t;    // Time/sequence logs
typedef std::vector<TimeSeq_t*> TimeSeqVec_t; // Array by seq logging type

// Debug history log
typedef enum { NoEvent, Ind, App, Tout, Pkt } DBEvents_t;
struct DBHist {
  DBHist()
    : event(NoEvent), time(0), state(0), action(0),
      seq(0), ack(0), dLth(0) { }
  DBHist(DBEvents_t e, Time_t t, int s, int a, Seq sq, Seq ak, Count_t d)
    : event(e), time(t), state(s), action(a), seq(sq), ack(ak), dLth(d) { }
  DBEvents_t event;
  Time_t     time;
  int        state;
  int        action;
  Seq        seq;   // sequence  number
  Seq        ack;   // ack number
  Count_t    dLth;  // data length
};

typedef std::vector<DBHist> DBHVec_t;

//Doc:ClassXRef
class TCP : public L4Protocol, public TimerHandler, public NotifyHandler {
  //Doc:Class Class {\tt TCP} is the base class for all {\tt TCP} models.
  //Doc:Class Inherits from {\tt L4Protocol}.

public:
  // Logging detail types
  typedef enum { LOG_SEQ_TX, LOG_ACK_TX,
                 LOG_SEQ_RX, LOG_ACK_RX,
                 LOG_CWIN,   LOG_LAST } TimeSeqSelector_t;

  // TCP Header tracing selectors
  typedef enum { SOURCE_PORT, DEST_PORT, 
                 SEQ_NUMBER, ACK_NUMBER,
                 HEADER_LENGTH, FLAGS,
                 WINDOW, CHECKSUM,
                 URGENT, FID,
                 DATA_LENGTH, CWND } TCPTrace_t; // Tracing options

  // Define the flag bits
  typedef enum { FIN = 0x01, SYN = 0x02, RST = 0x04,
                 PSH = 0x08, ACK = 0x10, URG = 0x20 } Flags;
  typedef enum { MAX_FLAGS = 0x40 } TCPMaxFlags_t;  // Flags are 6 bits
  // Typedefs for state machine
  typedef enum { CLOSED,       // 0
                 LISTEN,       // 1
                 SYN_SENT,     // 2
                 SYN_RCVD,     // 3
                 ESTABLISHED,  // 4
                 CLOSE_WAIT,   // 5
                 LAST_ACK,     // 6
                 FIN_WAIT_1,   // 7
                 FIN_WAIT_2,   // 8
                 CLOSING,      // 9
                 TIMED_WAIT,   // 10
                 LAST_STATE } States_t;
  typedef enum { APP_LISTEN,   // 0
                 APP_CONNECT,  // 1
                 APP_SEND,     // 2
                 SEQ_RECV,     // 3
                 APP_CLOSE,    // 4
                 TIMEOUT,      // 5
                 ACK_RX,       // 6
                 SYN_RX,       // 7
                 SYN_ACK_RX,   // 8
                 FIN_RX,       // 9
                 FIN_ACK_RX,   // 10
                 RST_RX,       // 11
                 BAD_FLAGS,    // 12
                 LAST_EVENT } Events_t;
  typedef enum { NO_ACT,       // 0
                 ACK_TX,       // 1
                 ACK_TX_1,     // ACK response to syn
                 RST_TX,       // 2
                 SYN_TX,       // 3
                 SYN_ACK_TX,   // 4
                 FIN_TX,       // 5
                 FIN_ACK_TX,   // 6
                 NEW_ACK,      // 7
                 NEW_SEQ_RX,   // 8
                 RETX,         // 9
                 TX_DATA,      // 10
                 PEER_CLOSE,   // 11
                 APP_CLOSED,   // 12
                 CANCEL_TM,    // 13
                 APP_NOTIFY,   // 14 - Notify app that connection failed
                 LAST_ACTION } Actions_t;
  typedef enum { ProtoNumber =  6 } TCPProto_t; // TCP l4 Protocol number
  // Constructors
  //Doc:Method
  TCP();
    //Doc:Desc   Default constructor,  no arguments.

  //Doc:Method
  TCP(Node* n);
    //Doc:Desc Construct a {\tt TCP} endpoint attached to a particular node.
    //Doc:Arg1 Node ponter to attach this {\tt TCP} protocol to.

  //Doc:Method
  TCP(const TCP&);           // Copy Constructor       
    //Doc:Desc Copy constructor.
    //Doc:Arg1 {\tt TCP} protocol to copy.

  void CommonConstructor();  // Common constructor code
  virtual ~TCP();

  // Upcalls from layer 3
#ifdef SIM
  void DataIndication(Node*, Packet*, IPAddr_t, Interface*, IPAddr_t = 0);
#else
  void DataIndication(Node*, Packet*, IPAddr_t, Interface*);
#endif //SIM

  // Notifier Methods
  void Notify(void*);

  // L4Protocol Methods
  NCount_t    Send(Size_t);         // Send empty data
  NCount_t    Send(char*, Size_t);  // Send with data
  NCount_t    Send(Data&);          // Send with data
  NCount_t    SendTo(Size_t, IPAddr_t, PortId_t);       // Send to target
  NCount_t    SendTo(char*, Size_t, IPAddr_t, PortId_t);// Send w/ data
  NCount_t    SendTo(Data&, IPAddr_t, PortId_t);        // Send w/ data
  Proto_t     Proto() const { return 6;}                // Return l4 protocol
  bool        Connect(IPAddr_t, PortId_t);              // Connect to remote
  bool        Close();                                  // Close connection
  L4Protocol* Copy() const = 0;                         // Create a copy
  Proto_t     ProtocolNumber();                         // Return protocol num
  bool        IsTCP();                                  // True if TCP

  // TimerHandler Methods
  void Timeout(TimerEvent*); // Called when timers expire

  // TCP Specific functins
  //Doc:Method
  bool     Listen();         // Set to accept connections
    //Doc:Desc This this endpont to the listening state to accept incoming
    //Doc:Desc connections.
    //Doc:Return True if successful.

  void     Reset();          // Reset to known, closed state
  void     ChildComplete();  // Decrement number of children

  // Responses to connection req from peer
  // Sets state to SYN_Received, send syn/ack     
  bool     Respond(PortId_t, IPAddr_t);
  // Rejects connection, sends reset
  void     Reject(PortId_t, IPAddr_t); 

  // Parameterizing the TCP endpoint
  //Doc:Method
  void SetRTTEstimator(const RTTEstimator&);
    //Doc:Desc Set a non--default round trip time estimator.
    //Doc:Arg1 Reference to new RTT estimator for this {\tt TCP}.

  //Doc:Method
  RTTEstimator* GetRTTEstimator();
    //Doc:Desc Get a pointer to the current RTT estimator.
    //Doc:Return Poiinter to current RTT estimator.

  //Doc:Method
  void SetSSThresh(DCount_t sst) { ssThresh = sst;}
    //Doc:Desc Set the slow start threshold for this connection.
    //Doc:Arg1 Slow start threshold (bytes).

  //Doc:Method
  void SetAdvertisedWindow(Count_t w) { advWin = w;}
    //Doc:Desc Set the receiver advertised window for this connection.
    //Doc:Arg1 Advertised window (bytes) for this connection.

  //Doc:Method
  void DeleteOnComplete(bool doc = true) { deleteOnComplete = doc;}
    //Doc:Desc Specify that this connection should automatically
    //Doc:Desc delete itself when it closes.
    //Doc:Arg1 True if delete on complete desired.

  //Doc:Method
  void DeleteOnTWait(bool dotw = true) { deleteOnTWait = dotw;}
    //Doc:Desc Specify that this connection should automatically
    //Doc:Desc delete itself on the completion of the {\em Timed Wait}
    //Doc:Desc state.
    //Doc:Arg1 True if delete on timed wait desired.

  //Doc:Method
  void CloseOnEmpty(bool coe = true) { closeOnEmpty = coe;}
    //Doc:Desc Specify that this connection should initiate close 
    //Doc:Desc action when all pending data has been sent.
    //Doc:Arg1 True if close on empty desired.

  //Doc:Method
  void SetSegSize(Count_t s) { segSize = s;}
    //Doc:Desc Set the segment size for this connection.
    //Doc:Arg1 Segment size (bytes).

  //Doc:Method
  void SetTxBuffer(Count_t s) { txBuffer = s;}
    //Doc:Desc Set the size of the transmit buffer.  Note: Presently
    //Doc:Desc not enforced.
    //Doc:Arg1 Size of transmit buffer (bytes).

  void SetRxBuffer(Count_t s) { rxBuffer = s;}
    //Doc:Desc Set the size of the receive buffer.  Note: Presently
    //Doc:Desc not enforced.
    //Doc:Arg1 Size of receive buffer (bytes).

  //Doc:Method
  void SetTwTimeout(Time_t t) { twTimeout = t;}
    //Doc:Desc Set the length of the {\em Timed Wait} state.
    //Doc:Arg1 Length of timed wait state (seconds).

  //Doc:Method
  void SetConnTimeout(Time_t t) { cnTimeout = t;}
    //Doc:Desc Set the length of the connection timeout.
    //Doc:Arg1 Length of connection state (seconds).

  //Doc:Method
  void SetChildLimit(Count_t l) { childLimit = l;}
    //Doc:Desc Set a limit on the number of simulataneous child
    //Doc:Desc (spawned by a listening endpoing).
    //Doc:Arg1 Desired child limit.

  // Statistics management
  //Doc:Method
  Rate_t  GoodPut();           // Return goodput so far
    //Doc:Desc Return the current goodput (bytes / second) for this
    //Doc:Desc connection.
    //Doc:Return Goodput in bytes per second.

  //Doc:Method
  Count_t RetransmitCount();   // Return number of retransmits
    //Doc:Desc Query the number of packet retransmissions for this
    //Doc:Desc connection.
    //Doc:Return Retransmitted packet count.

  //Doc:Method
  void    EnableTimeSeq (TimeSeqSelector_t); // Enable stats collection
    //Doc:Desc The {\tt TCP} model in \GTNS\ will keep detailed sequence
    //Doc:Desc versus time history information for various counters.  This
    //Doc:Desc method enables this data collection.  This information can
    //Doc:Desc later be streamed to a data file with {\tt LogTimeSeq}.
    //Doc:Desc More than one data collection can be enabled at a time with
    //Doc:Desc multiple  calls to {\tt EnableTimeSeq}.
    //Doc:Arg1 Desired level of  history connection.  Can be one of:
    //Doc:Arg1 LOG_SEQ_TX (sequence number transmitted),
    //Doc:Arg1 LOG_ACK_TX (ack number transmitted),
    //Doc:Arg1 LOG_SEQ_RX (sequence number received),
    //Doc:Arg1 LOG_ACK_RX (ack number received),
    //Doc:Arg1 LOG_CWIN (congestion window size).

  //Doc:Method
  void    DisableTimeSeq(TimeSeqSelector_t); // Disable stats collection
    //Doc:Desc Stop history data collection for the specified type of sequence
    //Doc:Desc number.
    //Doc:Arg1 Desired history level to stop, same levels as in
    //Doc:Arg1 {\tt EnableTimeSeq}.

  //Doc:Method
  void    ResetTimeSeq(TimeSeqSelector_t);   // Remove existing stats
    //Doc:Desc Reset the collected information for the specified time/sequence
    //Doc:Desc collection.
    //Doc:Arg1 Desired history level to reset,  same levels as in
    //Doc:Arg1 {\tt EnableTimeSeq}.

  //Doc:Method
  void    LogTimeSeq(TimeSeqSelector_t, std::ostream&,
                     Seq_t div = 0, Seq_t mod = 0, 
                     char sep = ' ');
    //Doc:Desc Log previously collected time/sequence data to a file.
    //Doc:Arg1 Desired Time/Sequence type.
    //Doc:Arg2 Output stream to log the data.
    //Doc:Arg3 If non--zero, each sample is divided by this value.
    //Doc:Arg4 If non--zero, each sample is modded by this value.
    //Doc:Arg5 Separator character between time and sequence number

  // Various debug
  //Doc:Method
  DCount_t CWin() { return cWin;}
    //Doc:Desc Query the current value of the congestion window.
    //Doc:Return Congestion window value (bytes, double).

  void    DBDump();                  // Debug dump information
  //Doc:Method
  TCP*    Peer() { return peer;}     // Debug..get peer pointer
    //Doc:Desc Get a pointer to remote peer endpoint.  This information
    //Doc:Desc is not always available, since distributed simulations may
    //Doc:Desc not have access to the peer object.
    //Doc:Return Pointer to peer {\tt TCP} object.  {\tt nil} if unknown.

  //Doc:Method
  int     State() const { return state;}
    //Doc:Desc Query the current {\em state} of this connection.
    //Doc:Return Connection state.  See {\tt tcp.h} for a list of the
    //Doc:Return state values.

  //Doc:Method
  Count_t TimeoutCount() const { return timeoutCount;}
    //Doc:Desc Query a count of timeouts for this connection.
    //Doc:Return Number of timeouts.

  //Doc:Method
  Time_t  TimeoutDelay() const { return lastTimeoutDelay;}
    //Doc:Desc Query the length of the last scheduled timeout.
    //Doc:Return Length of last scheduled timeout (seconds).

#ifdef USE_DEBUG_HISTORY
  void    DebugHistory(const DBHist&); // Add an entry to the debug hist
#else
  inline void    DebugHistory(const DBHist&) { }; // Add an entry to the debug hist
#endif

  void    PrintDBH();                  // Print the dbhist
protected:
  // Many of these should be overridden by the TCP variation implementations
  // Window Management
  virtual Count_t  UnAckDataCount(); // Return count of number of unacked bytes
  virtual DCount_t WindowD();        // Return window size (floating point)
  virtual Count_t  Window();         // Return window size (integer)
  virtual Count_t  AvailableWindow();// Return unfilled portion of window
  virtual void     NewCWnd(Count_t); // Adjust congestion window due to new ack

  // Manage State Machine
  // Process event        
  Actions_t ProcessEvent(Events_t);
  // Process action       
  bool      ProcessAction(Actions_t, Packet* p = NULL,
                          TCPHeader* tcp = NULL, IPAddr_t ip = IPADDR_NONE);

  // Packet Sending Functions
  void SendEmptyPacket(Byte_t, Seq_t, Seq_t);
  void SendEmptyPacketTo(Byte_t, Seq_t, Seq_t,
                         IPAddr_t dstip, PortId_t dstport);
  void SendPacket(Byte_t, Seq_t, Seq_t, Packet*);
  void SendPacketTo(Byte_t, Seq_t, Seq_t, Packet*,
                    IPAddr_t dstip, PortId_t dstport);
  // Send as much pending data as windows allow, include ack flag if requested
  // Return true if sent anything
  bool SendPendingData(bool withAck = false);

  // Ack Management
  virtual void NewAck(Seq);    // Process new acknowledgement received
  virtual void DupAck(const TCPHeader&, Count_t) = 0; // Duplicate ack received
  virtual void SendAck(Seq);   // Send the specified ack

  // New Recieved Data
  virtual void NewRx(Packet*, TCPHeader*, IPAddr_t);

  // Timer management
  void ScheduleTimer(Event_t, TCPEvent*&, Time_t);
  void CancelTimer(TCPEvent*&, bool delTimer = false);
  void CancelAllTimers();

  // Retransmissions
  virtual void ReTxTimeout() = 0; // Retransmit timeout (supplied by subclass)
  virtual void Retransmit();      // Retransmit the oldest pending packet

  // Statistics gathering
  void NoteTimeSeq(TimeSeqSelector_t, Seq_t );

public:
  // Setting defaults
  //Doc:Method
  static void DefaultSegSize (Count_t s)  { defaultSegSize = s;}
    //Doc:Desc Set the default segment size.  The default
    //Doc:Desc value is 512 bytes unless a new value is assigned.
    //Doc:Arg1 Default segment size (bytes).

  //Doc:Method
  static void DefaultAdvWin  (Count_t s)  { defaultAdvWin = s;}
    //Doc:Desc Set the default advertised window size.
    //Doc:Arg1 Default advertised window size (bytes).The default
    //Doc:Desc value is 65k bytes unless a new value is assigned.

  //Doc:Method
  static void DefaultSSThresh(DCount_t s) { defaultSSThresh = s;}
    //Doc:Desc Set the default  slow--start threshold size.
    //Doc:Arg1 Default slow--size size (bytes).The default
    //Doc:Desc value is 65k bytes unless a new value is assigned.

  //Doc:Method
  static void DefaultTxBuffer(Count_t s)  { defaultTxBuffer = s;}
    //Doc:Desc Set the default transmit buffer size.
    //Doc:Arg1 Default transmit buffer size (bytes).The default
    //Doc:Desc value is 4Gb unless a new value is assigned.

  //Doc:Method
  static void DefaultRxBuffer(Count_t s)  { defaultRxBuffer = s;}
    //Doc:Desc Set the default receive buffer size.
    //Doc:Arg1 Default receive buffer size (bytes).The default
    //Doc:Desc value is 4Gb unless a new value is assigned.

  //Doc:Method
  static void DefaultTwTimeout(Time_t t)  { defaultTwTimeout = t;}
    //Doc:Desc Set the default value of the the timed wait timeout.
    //Doc:Arg1 Default time wait value (seconds).The default
    //Doc:Desc value is 5 seconds unless a new value is assigned.

  //Doc:Method
  static void DefaultConnTimeout(Time_t t)  { defaultConnTimeout = t;}
    //Doc:Desc Set the default value of the the connection timeout.
    //Doc:Arg1 Default time to wait for syn-ack (seconds).The default
    //Doc:Desc value is 6 seconds unless a new value is assigned.

  //Doc:Method
  static void LogFlagsText(bool b)        { logFlagsText = b;}
    //Doc:Desc Specifies that the trace file entries for this connection
    //Doc:Desc should show  the flags field in text, such as "SYN|ACK".
    //Doc:Desc If not specified, flags values are shown in 8 bit hex.
    //Doc:Arg1 True if  text flags desired.

  //Doc:Method
  static void UseTimerBuckets(bool u = true) { useTimerBuckets = u;}
    //Doc:Desc Specifies that all timeout timers should be rounded to
    //Doc:Desc integral {\em bucket} values.  If specified, timers are rounded
    //Doc:Desc to units of 10ms, which is slightly more efficient in the
    //Doc:Desc simulator, and is a bit more realistic.  If not specified,
    //Doc:Desc timers are nano-second accurate.
    //Doc:Arg1 True if timer buckets desired.

  //Doc:Method
  static void DefaultConnCount(Count_t c) { defaultConnCount = c;}
    //Doc:Desc Specify the default number of connection retries.
    //Doc:Arg1 Default connection retry. The default
    //Doc:Desc value is 3 unless a new value is assigned.

  //Doc:Method
  static void Default(const TCP&); // Set default TCP variation
    //Doc:Desc Specify the default {\tt TCP} variation to use for all
    //Doc:Desc new {\tt TCP} endpoints.  Default value is {\tt TCPTahoe} 
    //Doc:Desc unless a new value is specified.
    //Doc:Arg1 Reference to any {\tt TCP} variant to set as default.

  //Doc:Method
  static TCP& Default();           // Get default TCP variation
    //Doc:Desc Return a reference to the default {\tt TCP} variation.
    //Doc:Return Reference to default {\tt TCP} variant.

  // Backplane related methods
  // Register data items with the dynamic simulation backplane
  static void  RegisterBackplane(DSHandle_t);
  static int   TCPExportQuery(char*, int, void*);
  static int   TCPDefaultCallback(char*, void*);
  static int   TCPExportCallback(char*, char*, int, void*);
  static int   TCPImportCallback(char*, int, char*, void*);
protected: // Members
  // History data
  UnAckData_t   bufferedData;    // Received, but out-of-sequence
  Data*         pendingData;     // Data sent by application, but not sent
  Seq           firstPendingSeq; // First sequence number in pendingData

  // Sequence Information, sender side
public: // for debugging
  Seq           nextTxSeq;       // Next sequence to send
  Seq           highTxMark;      // TX High water mark, for counting retx
  Seq           highestRxAck;    // largest ack received
  Seq           lastRxAck;       // For dupack testing
protected:
  Count_t       dupAckCount;     // Number of dup acks in a row
  Count_t	fastReTxFlag;	 // True if Fast ReTransmit occured
  Count_t	tflag;           // Added by Ishwar A.  need documentation
  bool          needAck;         // True if need ACK bit transmit

  // Sequence Information, receiver side
  Seq           nextRxSeq;       // Next expected sequence

  // Window Management
  Count_t       segSize;         // SegmentSize
  Count_t       rxWin;           // Window as received from peer
  Count_t       advWin;          // Window to advertise to peer
  DCount_t      cWin;            // Congestion window
  DCount_t      ssThresh;        // Slow Start Threshold
  // Round Trip Time estimation
  RTTEstimator* rtt;             // Points to any subclass of RTTEstimator

  // Timers
  Timer         timer;           // Timer for various TCP timer events
  TCPEvent*     connTimeout;     // Pending connection timeout event
  TCPEvent*     retxTimeout;     // Retransmit timeout
  TCPEvent*     delackTimeout;   // Delayed ack timeout
  TCPEvent*     timedWaitTimeout;// Timeout for timed-wait state
  TCPEvent*     lastAckTimeout;  // Timeout for last-ack state
  Time_t        twTimeout;       // Timeout period for timed-wait state
  Time_t        cnTimeout;       // Timeout period for connection retry

  // Counters
  Count_t       retryCount;      // Limit the re-tx retries
  Count_t       connCount;       // Connection retry count

  // State
  //States_t      state;           // Current state
  int           state;           // Current state
  bool          closeOnEmpty;    // True if send FIN pkt on empty data
  bool          deleteOnComplete;// True if should delete object on close
  bool          deleteOnTWait;   // True if delete after timed wait
  bool          pendingClose;    // True if close pending
  bool          closeNotified;   // True if close has been notified
  bool          closeReqNotified;// True if close request has been notified

  // Buffer limits
  Count_t       txBuffer;        // Size of tx buffer
  Count_t       rxBuffer;        // Size of rx buffer
  
  // Statistics
  Count_t       totalAck;        // Total bytes acked
  Time_t        openTime;        // Time connection was opened
  Time_t        lastRxTime;      // Time of last pkt receipt
  Time_t        lastAckTx;       // Time last ack sent
  Time_t        synTime;         // Time SYN sent
  Count_t       retransmitCount; // Count of retransmits
  Count_t       timeoutCount;    // Count of timeouts
  Count_t       pktsSent;        // Count of packets sent
  Count_t       pktsReceived;    // Count of packets received
  TimeSeqVec_t* timeSeqStats;    // Pointer to Array of time/seq stats
  TCP*          parent;          // Parent (server) tcp (if exists)
  TCP*          peer;            // Debug pointer to peer
  Count_t       nChildren;       // Number of child TCP's (if server)
  Count_t       childLimit;      // Maximum number of allowed children
  Count_t       nFin;            // Number fin packets rx, just testing
  Time_t        lastTimeoutDelay;// Timeout delay of last scheduled retx timer
  Time_t        lastMeasuredRTT; // Time of last rtt measurement
  DBHVec_t      dbVec;           // Debug history vector
  DBHVec_t::size_type dbInd;     // next available entry in dbVec
public:
  bool          isServer;        // debug..remove later
  bool          nRequested;      // notification is requested already
public:
  // Static members
  static Count_t       defaultSegSize;  // Segment size
  static Count_t       defaultAdvWin;   // Advertised window
  static DCount_t      defaultSSThresh; // SlowStart Threshold
  static Count_t       defaultTxBuffer; // Maximum number of pending tx bytes
  static Count_t       defaultRxBuffer; // Maximum number of pending rx bytes
  static Time_t        defaultTwTimeout;// Default timed-wait timeout
  static Time_t        defaultConnTimeout; // Default connection timeout
  static bool          logFlagsText;    // True if text mode flag logging
  static FId_t         nextFlowId;      // Next unique tcp flow id
  static bool          useTimerBuckets; // True if using timer buckets
  static Count_t       defaultConnCount;// Number of connection retries
  static TCPHeader*    bpHeader;        // Imported header from backplane
private:
  static TCP*          defaultTCP;      // Default TCP variant
 };

#endif


