/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Trace fstream object class
// George F. Riley.  Georgia Tech, Spring 2002

// Define trace stream object

#include <stdio.h>

#include <iostream>

#include "tfstream.h"
#include "simulator.h"
#include "node.h"

using namespace std;

Tfstream::Tfstream() : pendingEOL(false), timePrecision(5),
                       node(0), ipDotted(true), digitsThreshold(10.0),
                       format("%7.5f")
{
}

void Tfstream::TimePrecision(Count_t d)
{
  char work[20];

  timePrecision = d; // Number of significant digits
  if (timePrecision > 10) timePrecision = 10; // Bound at 10 for sprintf
  sprintf(work, "%c%ld%c%ld%c", '%', timePrecision+2, '.', timePrecision, 'f');
  format = string(work);
}

// Output operators

Tfstream& Tfstream::operator<<(char c)
{
  CheckLogTime();
  (ostream&)*this << c;
  pendingEOL = true;
  return *this;
}

Tfstream& Tfstream::operator<<(const char *s)
{
  CheckLogTime();
  (ostream&)*this << s;
  pendingEOL = true;
  return *this;
}

Tfstream& Tfstream::operator<<(const void *p)
{
  CheckLogTime();
  (ostream&)*this << p;
  pendingEOL = true;
  return *this;
}

Tfstream& Tfstream::operator<<(int n)
{
  CheckLogTime();
  (ostream&)*this << n;
  pendingEOL = true;
  return *this;
}

Tfstream& Tfstream::operator<<(unsigned int n)
{
  CheckLogTime();
  (ostream&)*this << n;
  pendingEOL = true;
  return *this;
}

Tfstream& Tfstream::operator<<(long n)
{
  CheckLogTime();
  (ostream&)*this << n;
  pendingEOL = true;
  return *this;
}

Tfstream& Tfstream::operator<<(unsigned long n)
{
  CheckLogTime();
  (ostream&)*this << n;
  pendingEOL = true;
  return *this;
}

Tfstream& Tfstream::operator<<(double n)
{
  CheckLogTime();
  (ostream&)*this << n;
  pendingEOL = true;
  return *this;
}

Tfstream& Tfstream::operator<<(IPAddr& i)
{
  CheckLogTime();
  if (ipDotted) (ostream&)*this << ((string)i).c_str();
  else          (ostream&)*this << (unsigned long)i;
  pendingEOL = true;
  return *this;
}

bool Tfstream::AppendEOL()
{
  if (pendingEOL)
    {
      (ostream&)*this << '\n';
      pendingEOL = false;
      return true; // appended
    }
  return false; // not needed
}

void Tfstream::NewNode(Node* n) // Log the node for protocol logging
{
  if (n == node) return;        // Same as existing
  AppendEOL();
  node = n;
}

void Tfstream::SetNode(Node* n) // Set node pointer for subsequent traces
{
  node = n;
}

// Private methods
void Tfstream::LogSimTime()
{
  char work[256];
  Time_t now = Simulator::Now();
  if (now >= digitsThreshold)
    { // Need to make new format
      Count_t newTot = 0;
      if      (now >= 1000000) newTot = 8 + timePrecision;
      else if (now >= 100000 ) newTot = 7 + timePrecision;
      else if (now >= 10000  ) newTot = 6 + timePrecision;
      else if (now >= 1000   ) newTot = 5 + timePrecision;
      else if (now >= 100    ) newTot = 4 + timePrecision;
      else if (now >= 10     ) newTot = 3 + timePrecision;
      else                     newTot = 2 + timePrecision;
      sprintf(work, "%c%ld%c%ld%c", '%', newTot, '.', timePrecision, 'f');
      format = string(work);
      switch (newTot - timePrecision) {
        case 8 :
          digitsThreshold = 10000000;
          break;
        case 7 :
          digitsThreshold = 1000000;
          break;
        case 6 :
          digitsThreshold = 100000;
          break;
        case 5 :
          digitsThreshold = 10000;
          break;
        case 4 :
          digitsThreshold = 1000;
          break;
        case 3 :
          digitsThreshold = 100;
          break;
        case 2 :
          digitsThreshold = 10;
          break;
      }
    }
  sprintf(work, format.c_str(), Simulator::Now());
  (ostream&)*this << work;
  (ostream&)*this << " N";
  (ostream&)*this << node->Id();
  pendingEOL = true;
}

