/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Global trace control
// George F. Riley.  Georgia Tech, Spring 2002

#include <iostream>

#include "trace.h"
#include "protograph.h"
#include "node.h"

using namespace std;

Trace*            Trace::instance;        // Single global trace object
bool              Trace::enabled = false; // Global enabled flag
Trace::Callback_t Trace::callback = nil;  // Tracing callback

Trace::Trace() 
  : trace(ProtocolGraph::N_LAYERS, DISABLED), node(0)
{
}

Trace::~Trace()
{
  Close(); // Close the file if open
}

bool Trace::Open(const char* fn)
{
  os.open(fn, ios::out);
  if (os) enabled = true;
  return os;  // True if opened successfully
}

void Trace::Close()
{
  if (os.is_open())
    {
      os.AppendEOL();
      os.close();
    }
  enabled = false;
}

void Trace::On()
{ // Enable all
  for (TSVec_t::size_type i = 0; i < trace.size(); ++i)
    {
      trace[i] = ENABLED;
    }
}

void Trace::On(Layer_t l)
{
  if (l > ProtocolGraph::N_LAYERS) return;  // Out of range
  trace[l-1] = ENABLED;
}

void Trace::Off()
{ // Disable all
  for (TSVec_t::size_type i = 0; i < trace.size(); ++i)
    {
      trace[i] = DISABLED;
    }
}

void Trace::Off(Layer_t l)
{
  if (l > ProtocolGraph::N_LAYERS) return;  // Out of range
  trace[l-1] = DISABLED;
}

bool Trace::IsOn(Layer_t l)
{ // Enquire trace status by layer
  if (l > ProtocolGraph::N_LAYERS) return false;  // Out of range
  if (!os.is_open()) return false;  // Not enabled if trace file not open
  return trace[l-1] == ENABLED;
}

bool Trace::IsEnabled()
{
  return(os.is_open());
}

void Trace::AppendEOL()
{
  if(os)os.AppendEOL(); // Add the end of line if needed
}

void Trace::NewNode(Node* n)
{ // A packet has arrived at a new node, see if log needed
  os.NewNode(n);
}

void Trace::SetNode(Node* n)
{ // Set node pointer for subsequent traces
  os.SetNode(n);
}

// Static methods
Trace* Trace::Instance()
{
  // If no instance, create one
  if (!instance) instance = new Trace();
  return instance;
}

