/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - UDP Base Class
// George F. Riley.  Georgia Tech, Spring 2002

// Base class for all UDP variants

#include <iostream>

// Uncomment below to enable debug level 0
// #define DEBUG_MASK 0x01
#define DEBUG_MASK 0x02
#include "debug.h"
#include "udp.h"
#include "application.h"
#include "datapdu.h"
#include "portdemux.h"
#include "hex.h"

using namespace std;

// UDPHeader methods

UDPHeader::UDPHeader()
  : sourcePort(0), destPort(0),
    messageLength(0), checksum(0), fid(0)
{
}

UDPHeader::UDPHeader(PortId_t s, PortId_t d, Word_t m, Word_t c)
  : sourcePort(s), destPort(d),
    messageLength(m), checksum(c), fid(0)
{
}

UDPHeader::UDPHeader(PortId_t s, PortId_t d, Word_t m, Word_t c, FId_t f)
  : sourcePort(s), destPort(d),
    messageLength(m), checksum(c), fid(f)
{
}

UDPHeader::UDPHeader(char* b, Size_t& sz, Packet* p)
{
  cout << "Constructing udp, initial size " << sz << endl;
  Size_t s = 0;
  b = Serializable::GetSize(b, sz, s);
  cout << "l4pdu size is " << s << " (" << Hex8(s) << ") " << endl;
  sz -= s;
  b = Construct(b, s);
  cout << "Constructing udp, final  size " << sz << endl;
  p->PushPDUBottom(this);  // Add to packet
  if (sz)
    { // Remaining size, must be data
      new Data(b, sz, p);
    }
}

Size_t UDPHeader::Size() const
{
  return 8;
}


PDU*   UDPHeader::Copy()
{
  return new UDPHeader(*this);
}

void   UDPHeader::Trace(Tfstream& tos, Bitmap_t b, Packet* p, const char* s)
{ // Trace the contents of this pdu
  tos << " ";
  if (s)
    tos << s;
  tos << "L4-UDP";
  if (Detail(UDP::SOURCE_PORT, b))     tos << " " << sourcePort;
  if (Detail(UDP::DEST_PORT, b))       tos << " " << destPort;
  if (Detail(UDP::MSG_LENGTH, b))      tos << " " << messageLength;
  if (Detail(UDP::CHECKSUM, b))        tos << " " << Hex4(checksum);
  if (Detail(UDP::FID, b))             tos << " " << fid;
}

// Serialization
Size_t UDPHeader::SSize()
{ // Size needed for serialization
  return
    sizeof(sourcePort) + 
    sizeof(destPort) + 
    sizeof(messageLength) + 
    sizeof(checksum) + 
    sizeof(fid);
}

char*  UDPHeader::Serialize(char* b, Size_t& sz)
{ // Serialize to a buffer
  b = SerializeToBuffer(b, sz, sourcePort);
  b = SerializeToBuffer(b, sz, destPort);
  b = SerializeToBuffer(b, sz, messageLength);
  b = SerializeToBuffer(b, sz, checksum);
  b = SerializeToBuffer(b, sz, fid);
  return b;
}

char*  UDPHeader::Construct(char* b, Size_t& sz)
{ // Construct from buffer
  b = ConstructFromBuffer(b, sz, sourcePort);
  b = ConstructFromBuffer(b, sz, destPort);
  b = ConstructFromBuffer(b, sz, messageLength);
  b = ConstructFromBuffer(b, sz, checksum);
  b = ConstructFromBuffer(b, sz, fid);
  return b;
}


// UDPPending Methods - Manage the pending data queue
UDPPending::UDPPending() 
  : data(nil), ip(IPADDR_NONE), port(NO_PORT), sent(0)
{
}

UDPPending::UDPPending(Data* d, IPAddr_t i, PortId_t p) 
    : data(d), ip(i), port(p), sent(0)
{
}

UDPPending::UDPPending(const UDPPending& c) // Copy constructor
    : data(c.data), ip(c.ip), port(c.port), sent(c.sent)
{
  DEBUG(0,(cout << "Hello from UDPPending cctor" << endl));
}

UDPPending::~UDPPending()
{
  DEBUG(0,(cout << "Hello from UDPPending dtor" << endl));
  // Data is specifically not deleted...users of UDPPending must do this
}



// UDP methods

// Static members
Size_t UDP::defaultPacketSize = 512;

UDP::UDP() : L4Protocol(), pktSize(defaultPacketSize), pendingData(nil)
{
  DetailOff(CHECKSUM); // By default, don't trace checksum
}

UDP::UDP(Node* n) : L4Protocol(n), pktSize(defaultPacketSize), pendingData(nil)
{
  DetailOff(CHECKSUM); // By default, don't trace checksum
}

UDP::UDP(const UDP& c) 
    : L4Protocol(c), pktSize(c.PacketSize()), pendingData(nil)
{
  DEBUG0((cout << "UDP CopyConst " << this
          << " localApp " << localApplication << endl));
}

UDP::~UDP()
{
  if (localPort != NO_PORT)
    { // Remove binding
      Unbind(ProtoNumber, localPort);
    }
}

// Notifyer Methods
void UDP::Notify(void*)
{ // We have been notified by the interface that queue space is available
  DEBUG(0,(cout << "Hello from UDP::Notify" << endl));
  SendPending();
}

#ifdef SIM
void UDP::DataIndication(Node* n, Packet* p, IPAddr_t ip, Interface*, IPAddr_t dst)
#else
void UDP::DataIndication(Node* n, Packet* p, IPAddr_t ip, Interface*)
#endif //SIM
{ // From lower layer
  PDU* h = p->PopPDU();
  if (localNode)
    { // Should always exist, but just be sure
      Trace::Instance()->NewNode(localNode);
      localNode->TracePDU(this, h, p, "+");
    }

  if (localApplication)
    { // Application exists, pass the packet up the stack
      localApplication->Receive(p, this);
    }
  else
    { // End of the line, delete the packet
      delete p;
    }
}

// L4Protocol methods
NCount_t  UDP::Send(Size_t s)            // Send empty data
{
  if (peerIP == IPADDR_NONE)
    { // Oops, no connect peer, can't do this
      return -1; 
    }
  DEBUG0((cout << "UDP Sending " << s << " bytes" << endl));
  pendingData.push_back(UDPPending(new Data(s), peerIP, peerPort));
  SendPending();
  return s;
}

NCount_t  UDP::Send(char* d, Size_t s)   // Send with data
{
  if (peerIP == IPADDR_NONE)
    { // Oops, no connect peer, can't do this
      return -1; 
    }
  DEBUG0((cout << "UDP Sending " << s << " bytes" << endl));
  pendingData.push_back(UDPPending(new Data(s,d), peerIP, peerPort));
  SendPending();
  return s;
}

NCount_t UDP::Send(Data& d)
{
  if (peerIP == IPADDR_NONE)
    { // Oops, no connect peer, can't do this
      return -1; 
    }
  DEBUG0((cout << "UDP Sending " << d.Size() << " bytes" << endl));
  Size_t s = d.Size();
  pendingData.push_back(UDPPending((Data*)d.Copy(), peerIP, peerPort));
  SendPending();
  return s;
}

NCount_t  UDP::SendTo(Size_t s, IPAddr_t ip, PortId_t port)
{
  pendingData.push_back(UDPPending(new Data(s), ip, port));
  SendPending();
  return s;
}

NCount_t  UDP::SendTo(char* d, Size_t s, IPAddr_t ip, PortId_t port)
{
  pendingData.push_back(UDPPending(new Data(s,d), ip, port));
  SendPending();
  return s;
}

NCount_t  UDP::SendTo(Data& d, IPAddr_t ip, PortId_t port)
{
  Size_t s = d.Size();
  pendingData.push_back(UDPPending((Data*)d.Copy(), ip, port));
  SendPending();
  return s;
}

bool        UDP::Connect(IPAddr_t ip, PortId_t p)
{ // Connect to peer.  For UDP just update remote ip/port
  DEBUG0((cout << "UDP Connecting to peer " << (string)IPAddr(ip)
          << " port " << p << endl));
  peerIP = ip;
  peerPort =  p;
  if (localApplication) localApplication->ConnectionComplete(this);
  return true;  // Success
}

bool        UDP::Close()
{
  return true; // UDP Close, no action needed
}

L4Protocol* UDP::Copy() const
{
  return new UDP(*this);
}

Proto_t     UDP::ProtocolNumber()
{ // Return UDP protocol number
  return ProtoNumber;
}

bool        UDP::IsTCP()
{
  return false;
}

void        UDP::PacketSize(Size_t s)
{ // Set pkt size for this instance
  pktSize = s;
}

// Private methods

NCount_t UDP::Send(Packet* p, IPAddr_t ip, PortId_t port)
{
  UDPHeader* udp = new UDPHeader(localPort, port, p->Size(), 0);
  p->PushPDU(udp);   // Push the UDP header on the packet header list
#ifdef BUGFIX
  Word_t messageLength = udp->messageLength;
#endif //BUGFIX
  // Trace this PDU
  if (localNode)
    { // Should always exist, but just be sure
      Trace::Instance()->NewNode(localNode);
      localNode->TracePDU(this, udp, p, "-");
    }
  // Pass to layer 3 protocol
  IPV4ReqInfo r(ip, IPADDR_NONE, ttl, ProtoNumber); // Info for l3
  DEBUG0((cout << "UDP::Send to ip " << (string)IPAddr(ip)
          << " port " << port << endl));
  if (extraTxDelay != 0)
    { // Need to schedule an event for later
      Simulator::instance->Schedule(
          new L4DelayedTxEvent(l3Proto, localNode,
                               p, r,
                               udp->messageLength,
                               true),
          extraTxDelay, this);
    }
  else
    {
      l3Proto->DataRequest(localNode, p, &r);
      if (localApplication)
        { // Application exists, notify data sent
#ifdef BUFIX
          localApplication->Sent(messageLength, this);
#else
          localApplication->Sent(udp->messageLength, this);
#endif //BUGFIX
        }
    }
#ifdef BUGFIX
  return messageLength;
#else
  return udp->messageLength;
#endif //BUGFIX
}

void UDP::SendPending()
{ // Send pending data (as much as possible)
  while(pendingData.size())
    { // While there is pending data
      DEBUG(0,(cout << "SendPending, pdsize " << pendingData.size() << endl));
      Data*  d = nil;
      Size_t s = 0;
      UDPPending& pd = pendingData.front(); // Get earliest pending 
      Size_t t = pd.data->Size();           // Total bytes in data pdu
      DEBUG(0,(cout << "SendPending, frontsize " << t 
               << "  frontsent " << pd.sent << endl));
      if (t <= PacketSize() && pd.sent == 0)
        { // Special case for efficiency, just use the pointer
          d = pd.data;
          pd.data = nil; // Do this so it's not deleted later
          s = d->Size();
        }
      else
        { // Calculate remaining size, bounded by packet size
          s = min(PacketSize(), pd.data->SizeFromOffset(pd.sent));
          d = new Data(s);
          DEBUG(0,(cout << "SendPending, minsize " << s 
                   << " pktsize " << PacketSize()
                   << " sfo " << pd.data->SizeFromOffset(pd.sent)
                   << endl));
        }
      if (BufferAvailable(s))
        { // Buffer space available on output link, process this
          Trace::Instance()->AppendEOL(); // Each on separate line
          if (!d) d = pd.data->CopyFromOffset(s, pd.sent);
          DEBUG0((cout << "Advancing sent by " << s << endl));
          pd.sent += s;
          Packet* p = NewPacket();      // Create a new packet
          DEBUG0((cout << "UDP NewPacket, notif " << p->notification << endl));
          
          p->PushPDU(d);                 // Add the data
          Send(p, pd.ip, pd.port);       // Send the packet
          // delete deq entry if finished...code later
          if (pd.sent == t)
            { // Send the entire pending data item, time to delete
              if (pd.data) delete pd.data; // Delete data if still there
              pendingData.pop_front();     // Remove from deque
            }
        }
      else
        { // Ask for notification when buffer space becomes available
          DEBUG(0,(cout << "SendPending, NBS" << endl));
          if(!pd.data) pd.data = d;       // Put back the data pointer
          RequestNotification(this, nil); // Notify when space available
          break;
        }
    }
}


// Static methods
void UDP::DefaultPacketSize(Size_t s)
{
  defaultPacketSize = s;
}
