/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - UDP Base Class
// George F. Riley.  Georgia Tech, Spring 2002

// Base class for all UDP variants

#ifndef __udp_h__
#define __udp_h__

#include <deque>

#include "pdu.h"
#include "l4protocol.h"
#include "datapdu.h"
#include "notifier.h"

//Doc:ClassXRef
class UDPHeader : public L4PDU {
public :
  UDPHeader();
  UDPHeader(PortId_t, PortId_t, Word_t, Word_t);
  UDPHeader(PortId_t, PortId_t, Word_t, Word_t, FId_t);
  UDPHeader(char* b, Size_t& sz, Packet* p);
  Size_t  Size() const;
  PDU*    Copy();
  Proto_t Proto() { return 17;}        // UDP is protocol 17
  // Serialization
  Size_t  SSize();                     // Size needed for serialization
  char*   Serialize(char*, Size_t&);   // Serialize to a buffer
  char*   Construct(char*, Size_t&);   // Construct from buffer
  void    Trace(Tfstream&, Bitmap_t, // Trace the contents of this pdu
                Packet* = nil, const char* = nil);
public:
  PortId_t sourcePort;
  PortId_t destPort;
  Word_t   messageLength;
  Word_t   checksum;
  FId_t    fid;           // Flow id's are not part of UDP, but useful
};

// Typedefs for the pending data queue
//Doc:ClassXRef
class UDPPending {
public:
  UDPPending();
  UDPPending(Data*, IPAddr_t, PortId_t) ;
  UDPPending(const UDPPending&);   // Copy constructor
  ~UDPPending();                   // Destructor
public:
  Data*    data;
  IPAddr_t ip;
  PortId_t port;
  Size_t   sent;   // Sent so far
};
typedef std::deque<UDPPending> Pending_t;

//Doc:ClassXRef
class UDP : public L4Protocol, public NotifyHandler {
  //Doc:Class Class {\tt UDP} defines a model of the User Datagram Protocol.
  //Doc:Class {\tt UDP} inherits from {\tt L4Protocol} and implements most of
  //Doc:Class the methods defined there.
public:
  typedef enum { ProtoNumber =  17 } UDPProto_t; // UDP Layer4 Protocol number
  typedef enum { SOURCE_PORT, DEST_PORT, MSG_LENGTH,
                 CHECKSUM, FID } UDPTrace_t; // Tracing options
  // Constructors
  //Doc:Method
  UDP();
    //Doc:Desc Default constructor, no arguments.

  //Doc:Method
  UDP(Node* n);
    //Doc:Desc Constructor, specify a node to attached to.
    //Doc:Arg1 Node this {\tt UPD} protocol is attached to.

  //Doc:Method
  UDP(const UDP&);
    //Doc:Desc Copy constructor.
    //Doc:Arg1 {\tt UDP} object to copy.

  // Destructor
  virtual ~UDP(); 
public:
  // Notifier Methods
  void Notify(void*);

  // L4Protocol methods
  // Upcall From lower layer
#ifdef SIM
  void DataIndication(Node*, Packet*, IPAddr_t, Interface*, IPAddr_t = 0); 
#else
  void DataIndication(Node*, Packet*, IPAddr_t, Interface*); 
#endif //SIM
  NCount_t Send(Size_t);                         // Send empty data
  NCount_t Send(char*, Size_t);                  // Send with data
  NCount_t Send(Data&);                          // Send with data
  NCount_t SendTo(Size_t, IPAddr_t, PortId_t);   // Send to specified target
  NCount_t SendTo(char*, Size_t, IPAddr_t, PortId_t); // Send w/data, to target
  NCount_t SendTo(Data&, IPAddr_t, PortId_t);    // Send w/ data
  Proto_t  Proto() const { return 17;}           // Return the layer 4 protoco
  bool     Connect(IPAddr_t, PortId_t);          // Connect to remote host
  bool     Close();                              // Close connection
  L4Protocol* Copy() const;                      // Create a copy of protocol
  Proto_t  ProtocolNumber();                     // Return UDP protocol number
  bool     IsTCP();                              // True if TCP
  //Doc:Method
  void     PacketSize(Size_t);                   // Set packet size
    //Doc:Desc Set the packet size for this {\tt UDP} endpoint.
    //Doc:Arg1 Packet size (bytes).

  //Doc:Method
  Size_t   PacketSize() const { return pktSize;} // Get packet size
    //Doc:Return Query packet size for this {\tt UDP} endpoint.

private:
  NCount_t Send(Packet*,IPAddr_t,PortId_t);      // Common send code
  void     SendPending();                        // Send pending data
private:
  Size_t    pktSize;                             // Size of packets
protected:
  Pending_t pendingData;                         // Data to be sent
public:
  // Static methods/members
  //Doc:Method
  static void   DefaultPacketSize(Size_t);       // Set the default pkt size
    //Doc:Desc Set default packet size for all {\tt UDP} protocol objects.
    //Doc:Arg1 Default packet size (bytes).  Default size is 512 bytes unless
    //Doc:Arg1 otherwise specified.

  static Size_t defaultPacketSize;               // Global default
};



#endif


