/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - VirusThrottle class
// Mohamed Abd Elhafez.  Georgia Tech, Spring 2004

// Implements the Virus throttle algorithm for worm containment

#ifndef __virus_throttle_h__
#define __virus_throttle_h__

#include <queue>

#include "common-defs.h"
#include "pdu.h"
#include "ipaddr.h"
#include "ipv4.h"
#include "node.h"
#include "tcp.h"
#include "worm-containment.h"

#ifdef HAVE_QT
#include "qnamespace.h"
#include <qcolor.h>
#endif

#define ALLOW true
#define DISALLOW false
#define TCP_PROTOCOL 6
#define UDP_PROTOCOL 17

typedef std::pair<Packet*, int> PacketPair_t;
typedef std::queue<PacketPair_t> DelayQueue_t;
typedef std::vector<IPAddr_t> WorkingSet_t;

// Define the event subclass for throttle  events
//Doc:ClassXRef
class ThrottleEvent : public Event {
 public:
  typedef enum { PROCESS_QUEUE } ThrottleEvent_t;
 public:
  ThrottleEvent(): dst(0)    { }
  ThrottleEvent(Event_t ev)
    : Event(ev), dst(0), iface(0)
    { }
  ThrottleEvent(Event_t ev, IPAddr_t ip, Interface* interface)
    : Event(ev), dst(ip), iface(interface)
    { }
  virtual ~ThrottleEvent() { }
 public:
  IPAddr_t dst;
  Interface* iface;
};

//Doc:ClassXRef
class VirusThrottle : public WormContainment, public Handler{
//Doc:Class The class {\tt VirusThrottle} implements the virus throttle algorithm
//Doc:Class It derives 	from the class {\tt WormContainment}. 
	
public:
  //Doc:Method
  VirusThrottle();
  //Doc:Desc Default constructor

  // Handler
  virtual void Handle(Event*, Time_t);

  //Doc:Method
  void ProcessOutPacket(Packet*, IPAddr_t, int, Interface*);
  //Doc:Desc Process the given packet for TCP SYN packets or UDP packets
  
  // void ProcessInPacket(Packet*, Interface*) {}

  bool InWorkingSet(IPAddr_t);
 
  void Delay(IPV4Header*, Packet* , IPAddr_t , int, Interface*);
    


  static void SetWorkingSetSize(int size)
    {
      workingsetsize = size;
    }
  
  static void SetDelayQueueLimit(int size)
    {
      delayqueuelimit = size;
    }

 static Count_t  TotalBlocked() {return blocked;}
 static Count_t  FalseDetected() {return falsedetection;}
 static void SetRouterThrottle(bool b){ routerthrottle = b;}

 private:
  static Trace::TraceStatus traceStatus; // For tracing
  DelayQueue_t delayqueue;
  WorkingSet_t workingset;
  bool conn;
 public:
  static int delayqueuelimit;
  static int workingsetsize;
  static int blocked;
  static int falsedetection;
  static bool routerthrottle;

};

#endif

