/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Grid Network Object
// George F. Riley.  Georgia Tech, Spring 2002

// Define a two dimensional grid of wireless nodes

#include "wireless-grid-rectangular.h"
#ifdef BUGFIX
#include "wlan.h"
#endif //BUGFIX

using namespace std;

// Constructors
WirelessGridRectangular::WirelessGridRectangular(
                           const Location& l, // Lower left corner
                           const Random& s, // Rng for number of nodes
                           const Random& p, // Rng for x/y location
                           IPAddr_t      i, 
                           bool create)
  : WirelessGrid(i), 
    lowerLeft(l),
    sizeRng(s.Copy()),
    xLocationRng(p.Copy()), yLocationRng(p.Copy())
{
#ifdef MANET
  l2 = ieee802_11;
#endif
  if(create)CreateNodes(); // Create the nodes
}

WirelessGridRectangular::WirelessGridRectangular(
                           const Location& l, // Lower left corner
                           const Random& s, // Rng for number of nodes
                           const Random& x, // Rng for x location
                           const Random& y, // Rng for y location
                           IPAddr_t      i, 
                           bool create)
  : WirelessGrid(i),
    lowerLeft(l),
    sizeRng(s.Copy()),
    xLocationRng(x.Copy()), yLocationRng(y.Copy())
{
#ifdef MANET
  l2 = ieee802_11;
#endif
  size = sizeRng->IntValue();
  if(create)CreateNodes(); // Create the nodes
}

#ifdef MANET
WirelessGridRectangular::WirelessGridRectangular(
                           L2ProtocolType mac,
                           const Location& l, // Lower left corner
                           const Random& s, // Rng for number of nodes
                           const Random& p, // Rng for x/y location
                           IPAddr_t      i, 
                           bool create)
  : WirelessGrid(i), 
    lowerLeft(l),
    l2 (mac),
    sizeRng(s.Copy()),
    xLocationRng(p.Copy()), yLocationRng(p.Copy())
{
  if(create)CreateNodes(); // Create the nodes
}
WirelessGridRectangular::WirelessGridRectangular(
                           L2ProtocolType mac,
                           const Location& l, // Lower left corner
                           const Random& s, // Rng for number of nodes
                           const Random& x, // Rng for x location
                           const Random& y, // Rng for y location
                           IPAddr_t      i, 
                           bool create)
  : WirelessGrid(i),
    lowerLeft(l),
    l2 (mac),
    sizeRng(s.Copy()),
    xLocationRng(x.Copy()), yLocationRng(y.Copy())
{
  size = sizeRng->IntValue();
  if(create)CreateNodes(); // Create the nodes
}
#endif // MANET

WirelessGridRectangular::~WirelessGridRectangular()
{
}

Location WirelessGridRectangular::GetRandomLocation()
{
  return Location(xLocationRng->Value() + lowerLeft.X(),
                  yLocationRng->Value() + lowerLeft.Y());
}

// Private Procedures
void WirelessGridRectangular::CreateNodes()
{
  while(1)
    {
      size = sizeRng->IntValue();
      if (size) break; // Insure non-zero size
    }
  first = Node::nextId;
#ifdef MANET
  WirelessLink* wlan;
  if (l2 == ieee802_11) 
    wlan = new  WirelessLink(size, startingIP, Mask(24), L2Proto802_11());
  else if (l2 == usap)
    wlan = new  WirelessLink(size, startingIP, Mask(24), L2ProtoUsap());
  else if (l2 == simple) 
    wlan = new  SimpleWirelessLink(size, startingIP, Mask(24), L2ProtoSimple());
  else
  {
    cout << "Wrong mac protocol type" <<endl;
    exit(0);
  }
#else
  WirelessLink* wlan = new  WirelessLink(size, startingIP, Mask(24));
#endif //MANET

  for (Count_t i = 0; i < size; ++i)
    {
#ifdef BUGFIX
      Node*  n = wlan->GetNode(i);
      Location l = GetRandomLocation();
      n->SetLocation(l.X(),l.Y());
      cout << "Node location is " << n->LocationX()
              << " " << n->LocationY() << endl;
#else
      Node* n = new Node();
      n->SetLocation(xLocationRng->Value(), yLocationRng->Value());
#endif //BUGFIX
      DEBUG0((cout << "Node location is " << n->LocationX()
              << " " << n->LocationY() << endl));
      
#ifdef USAP_ADDON
      if (l2 == usap)
      { 
        IFVec_t vInterface =  n->Interfaces();
        for (IFVec_t::iterator it=vInterface.begin(); it!=vInterface.end(); ++it)
        {
          // Need to init here after node add intf. in WirelessLink constructor
          // This is because Init() needs iface->GetNode() to initialize the 
          // frame descriptor
          if ((*it)->GetL2Proto()->IsWireless())
          { // for wireless only, not lo
          	L2ProtoUsap *usapptr = (L2ProtoUsap *) (*it)->GetL2Proto();
          	if (usapptr) usapptr->Init();
          }
        }
      }
#endif // USAP_ADDON
      if (startingIP != IPADDR_NONE)
        {
          n->SetIPAddr(startingIP + i);
        }
    }
}
