/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/


#ifndef __wlan_h__
#define __wlan_h__

#include "common-defs.h"
#include "link-real.h"
#include "ratetimeparse.h"
#include "l4protocol.h"
#include "l2proto802.11.h"
#ifdef MANET
#include "l2protosimple.h"
#include "l2protousap.h"
#endif // MANET

typedef enum {RESERVED, IDLE} LinkState_t;

class Propagation;

//Doc:ClassXRef
class WirelessLink : public LinkReal {
  //Doc:Class   Class {\em WirelessLink} is a subclass derived from class
  //Doc:Class   {\tt Link} and it describes a typical wireless link.
  //Doc:Class   It currently makes some assumptions which are valid only
  //Doc:Class   in the case of a wireless LAN but may be extended to
  //Doc:Class   any wireless physical link.
public:
  typedef enum { NONE, PARTIAL, FULL } Detail_t;
  // Constructors
  // This one specifies how many nodes
  //Doc:Method
  WirelessLink(Count_t,             // Node count
	       IPAddr_t, Mask_t,    // IP addr start, mask
#ifdef USAP_ADDON
               const L2Proto& l2 = L2Proto802_11(),// L2 proto to bind to each
#endif // USAP_ADDON
	       L4Protocol* l4 = nil,// Layer 4 protocol to bind to each node 
	       Rate_t r = Rate("11Mb"), 
	       Time_t dly = Time("1us"),
	       Detail_t d = PARTIAL,
	       bool bootstrap = false);
  //Doc:Desc This is the default constructor for this link. It constructs
  //Doc:Desc a default WirelessLink of a given number of nodes.
  //Doc:Arg1 Number of nodes in the LAN
  //Doc:Arg2 The IP Address of the first node in a given mask
  //Doc:Arg3 The subnet mask
  //Doc:Arg4 The layer 4 protocol to bind to each node. Currently just nil
  //Doc:Arg5 The Rate of the link.
  //Doc:Arg6 The delay in the link.
  //Doc:Arg7 The level of detail we want to model. Currently there is no
  //Doc:Arg7 difference in the model based on this argument.
  
  ~WirelessLink();      // Destructor
  // Link methods

  //Doc:Method
  virtual    Link*  Copy() const;          // All subclasses must have Copy
  //Doc:Desc This method is used to make rigid copies of the object
  //Doc:Return A pointer to the newly created copy

  //Doc:Method
  virtual    bool   Transmit(Packet*, Interface*, Node*) { return false; }
  //Doc:Desc This method is used to transmit the Packet from a give node
  //Doc:Arg1 The packet to be transmitted
  //Doc:Arg2 The interface from which the packet is transmitted
  //Doc:Arg3 The node from which the packet is transmitted

  //Doc:Method
  virtual  bool Transmit(Packet*, Interface*, Node*, Rate_t);
    //Doc:Desc Transmit a packet.
    //Doc:Arg1 The packet to transmit.
    //Doc:Arg2 A pointer to the local (transmitting) interface
    //Doc:Arg3 A pointer to the local (transmitting) node.
    //Doc:Arg4 The rate at which the packet is transmitted.

  //Doc:Method
  Time_t     SpeedOfLight(Node*, Node*); //Speed of light delay, node1 to node2
  //Doc:Desc This method returns the speed of light delay between the two
  //Doc:Desc nodes.
  //Doc:Arg1 The first node
  //DOc:Arg2 The second node

  //Doc:Method
  Count_t    PeerCount();                      // Return number of peers
  //Doc:Desc Returns the number of peers on this link
  //Doc:Return number of peers

  //Doc:Method
  Count_t    NodeCount();
    //Doc:Desc Returns the total number of nodes on this link
    //Doc:Return Node Count;

  //Doc:Method
  IPAddr_t   PeerIP (long unsigned int);
  //Doc:Desc Since each node on this link is also associated with a count
  //Doc:Desc this returns the IP address of the n'th  peer
  //Doc:Arg1 the count index n
  //Doc:Return the corresopnding IPAddress

  //Doc:Method
  IPAddr_t   NodePeerIP(Node*);
  //Doc:Desc Return the IP address of the node if it is on this link and
  //Doc:Desc is found
  //Doc:Arg1 The pointer to the node object
  //Doc:Return The IP Address
  
  //Doc:Method
  IPAddr_t   DefaultPeer(Interface*) { return firstIP; }
  //Doc:Desc Find the default peer's IP Address, this will find the default
  //Doc:Desc gateway's IP Address in the same subnet
  //Doc:Desc This method is valid only if we have a staic wireless LAN
  //Doc:Arg1 The interface whose default peerIP is to be known
  //Doc:Return The default peer's IP

  //Doc:Method
  void       AllNeighbors(NodeWeightVec_t & nw);
  //Doc:Desc AllNeighbors is used by broadcasts even if they are
  //Doc:Desc not routing peers. Use this to find all the peers
  //Doc:Arg1 The NodeWeightVec element, with respect to which,
  //Doc:Arg1 we need all the peers in this link

  //Doc:Method
  void       Neighbors (NodeWeightVec_t &, Interface *, bool);
  //Doc:Desc AllNeighbors is used by broadcasts even if they are
  //Doc:Desc not routing peers. Use this to find all the peers
  //Doc:Arg1 The NodeWeightVec element, with respect to which,
  //Doc:Arg1 we need all the peers in this link

  //Doc:Method
  bool       NodeIsPeer(Node*); // Find out if node is a peer
  //Doc:Desc Find out if node is a peer
  //Doc:Arg1 The Node
  //Doc:Return boolean result of the test

  
  //Doc:Method
  Count_t    NodePeerIndex(Node*);             // Get peer index for node
  //Doc:Desc Returns the index of the node in this link
  //Doc:Arg1 The pointer to the node
  //Doc:Return The index of the peer

  
  //Doc:Method  
  Count_t    NeighborCount(Node*);             // Number of routing neighbors
  //Doc:Desc Returns the number of neighbors on this link which are routers
  //Doc:Desc technically, count of routing peers
  //Doc:Arg1 Node querying neighbors (significant for Ethernet links)
  //Doc:Return The number of routing peers.

  
  //Doc:Method  
  Interface* GetIfByMac(MACAddr );
  //Doc:Desc Get a pointer to the interface by its MAC address
  //Doc:Arg1 the MAC address
  //Doc:Return pointer to the interface

  
  //Doc:Method  
  MACAddr    IPToMac(IPAddr_t);                // Convert peer IP to peer MAC
  //Doc:Desc Convert peer IP to peer MAC
  //Doc:Arg1 The IP Address whose MAC is to be determined
  //Doc:Return The corresponding MAC address
  
  
  //Doc:Method  
  void       ReserveLink(void) { state = RESERVED; }
  //Doc:Desc This method reserves the link for a given transmission interval
  
  //Doc:Method  
  void       FreeLink(void) { state = IDLE;}
  //Doc:Desc This method frees the link into IDLE state after a transmission

  //Doc:Method  
  Interface* GetPeer(Packet*);
  //Doc:Desc This method returns the destination interface by looking
  //Doc:Desc at the destination MAC address of the packet.
  //Doc:Arg1 The packet to be transmitted.
  //Doc:Return The destination interface.

  //Doc:Method
  virtual Interface* GetPeer(Count_t);    // Get peer interface for peer index
    //Doc:Desc Get the interface pointer for the remote endpoint of
    //Doc:Desc this link for the specified peer index
    //Doc:Arg1 Which peer desired (0 <- i < NeighborCount())
    //Doc:Return A pointer to the receiving interface for the specified peer

  virtual void  SetPeer(Interface*) { };

  //Doc:Method  
  bool       IsIdle();
  //Doc:Desc This method returns a bool to indicate whether a link is idle/busy
  //Doc:Return true if the link is idle, false otherwise

  //Doc:Method  
  IPAddr_t   GetIP(Count_t);
  //Doc:Desc Get IP of the node by the index of the node
  //Doc:Arg1 int index of the node
  //Doc:Return IP address if the index is within range, else IPADDR_NONE

  
  //Doc:Method  
  Node*      GetNode(Count_t);
  //Doc:Desc  Get the node by its index
  //Doc:Arg1  The index to the list of nodes
  //Doc:Return Pointer to the node
  
  
  //Doc:Method  
  Interface* GetIf(Count_t);
  //Doc:Desc Get the interface of the node connected to this link by its index
  //Doc:Arg1 the index of the node
  //Doc:Return the interface of the node that attaches to this link

#ifdef BUGFIX
  void SetAlpha(double);
#endif // BUGFIX

public:
  NodeId_t   first;    // Since there is a global node list, we don't need one 
  NodeId_t   last;     // Just log first/last+1 entries
  Rate_t     rate;     // Bandwidth
  Time_t     delay;    // Propogation delay (1us if detail NONE or PARTIAL)
  IPAddr_t   firstIP;  // First IP Address
  MACAddr    firstMac; // First MAC Address
  IFVec_t    IF_list;
  Detail_t   detail;
  Propagation*  propagation;  // radio propagation model
  double     csThreshold;     // carrier sense threshold, mW
  double     rxThreshold;     // rx threshold, mW
  double     snrThreshold;    // SNR or SINR threshold, dB
#ifdef SIM
  double     alpha;
  Uniform*   rngProp;
#endif // SIM

private:
  void ConstructorHelper(
                         Count_t,          // Node count
                         IPAddr_t, Mask_t, // IP addr start, mask
#ifdef USAP_ADDON
                         const L2Proto &l2, // Layer 2 protocol to bind to each
#endif // USAP_ADDON
                         L4Protocol* l4,   // Layer 4 protocol to bind to each 
                         Rate_t r, Time_t dly,
			 Detail_t d, bool bootstrap=false );
  LinkState_t state;
};

#ifdef MANET
class SimpleWirelessLink : public WirelessLink {
public:
  SimpleWirelessLink(Count_t n,             // Node count
	       IPAddr_t i, Mask_t m,    // IP addr start, mask
#ifdef USAP_ADDON
               const L2Proto& l2 = L2Proto802_11(),// L2 proto to bind to each
#endif // USAP_ADDON
	       L4Protocol* l4 = nil,// Layer 4 protocol to bind to each node 
	       Rate_t r = Rate("11Mb"), 
	       Time_t dly = Time("1us"),
	       Detail_t d = PARTIAL,
	       bool bootstrap = false) : WirelessLink(n, i, m,
#ifdef USAP_ADDON
               l2,
#endif // USAP_ADDON
               l4, r, dly, d, bootstrap) {}
  virtual  bool Transmit(Packet*, Interface*, Node*, Rate_t);
  bool IsAddrMatch(IPAddr_t, IPAddr_t);
 
};
#endif //MANET

#endif //__wlan_h_
