/*

* GTNetS provides a portable C++ class library for network simulations
*
* Copyright (C) 2003 George F. Riley
*
* No guarantees or warranties or anything are provided or implied in any way
* whatsoever.  Use this program at your own risk.  Permission to use this
* program for any purpose is given, as long as the copyright is kept intact.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

* Permission is hereby granted for any non-commercial use of this code

*/

// Georgia Tech Network Simulator - Worm Target Vector classes
// Monirul I Sharif.  Georgia Tech, Fall 2003

// These classes are to be used with descendents of the class Worm

#ifndef __worm_targetvector_h__
#define __worm_targetvector_h__

#include "common-defs.h"
#include "rng.h"
#include "application.h"


// Base class of target vectors
// All classes are to be descendents of this one.

class WormTargetVector : public Object {
public:
  WormTargetVector(Count_t range);
  WormTargetVector(Mask m);
  
  // The following methods are to be overriden for changing the 
  // functionality
  virtual WormTargetVector* Copy() const;
  virtual IPAddr_t Generate();
  virtual void Initialize(IPAddr_t);
 private:
  bool initialized;
 protected:
  Count_t scanrange;
  IPAddr_t localIP;
};


// Uniform Target Vector class
// The worm generates IP addresses uniformly distributed in the
// scanrange
class WTVUniform : public WormTargetVector {
 public:
  WTVUniform(Count_t range);
  WTVUniform(Mask m);

  virtual WormTargetVector* Copy() const;
  virtual IPAddr_t Generate();
  virtual void Initialize(IPAddr_t);

 private:
  Random *rngT;
};


// Local Preference scanning class
// The worm generates IP address that are either in the 
// scanrange, or local range based on a defined probability

class WTVLocalPref : public WormTargetVector {
 public:
  WTVLocalPref(Mask, double, Mask);
  WTVLocalPref(Count_t, double, Mask);
  // selects local preference 
  //  when prob<=p

  virtual WormTargetVector* Copy() const;
  virtual IPAddr_t Generate();
  virtual void Initialize(IPAddr_t);

 private:
  Count_t localscanrange;
  double prob;

  Random *rngP;
  Random *rngT1;
  Random *rngT2;
};


// Sequential Scanning 
// Selects a random start address which is uniformly distributed
// in the scan range and then generates sequential addresses after it

class WTVSequential : public WormTargetVector {
 public:
  WTVSequential(Mask);
  WTVSequential(Count_t);

  virtual WormTargetVector* Copy() const;
  virtual IPAddr_t Generate();
  virtual void Initialize(IPAddr_t);

 private:
  IPAddr_t currentIP;
};


// Sequential Scanning with local preference
// Selects a random start address which is either uniformly distributed
// in the scan range or in the local range of IP with a fixed probability 
// and then generates sequential addresses after it

class WTVSequentialWithLocalPref : public WormTargetVector {
 public:
  WTVSequentialWithLocalPref(Mask, double, Mask);
  WTVSequentialWithLocalPref(Count_t, double, Mask);

  virtual WormTargetVector* Copy() const;
  virtual IPAddr_t Generate();
  virtual void Initialize(IPAddr_t);

 private:
  IPAddr_t currentIP;
  double prob;
  Count_t localscanrange;
};


#endif
