// Generates figures shown in Chapter 7 of the documentation
//
// Copyright 2004, The Boeing Company

#include <iostream>
#include <stdlib.h>

#include "simulator.h"
#include "wireless-grid-rectangular.h"
#include "ipaddr.h"
#include "rng.h"
#include "node.h"
#include "mobility-random-waypoint.h"
#include "application-cbr.h"
#include "validation.h"  
#include "application-ospf6d.h"
#include "l2proto802.11.h"
#include "etrace.h"
#include "estat.h"
#include "args.h"
#include "ratetimeparse.h"

// Basic configuration
int default_seed = 1;                  // Random seed
double default_start_time = 1800;      // seconds.  All statistics discarded
                                       //   for first "start_time" seconds
double default_stop_time = 3600;       // seconds
int default_pktsize = 40;              // CBR packet size (bytes)
int default_pktrate = 10;              // CBR packets per second overall 
std::string default_tag = "";          // string tag for appending output files
                                       //   in the simulator
// Scenario configuration
int default_num_nodes = 20;            // Number of nodes in the topology
DCount_t default_grid_length = 500.0;  // meters (square grid)
DCount_t default_pause_time = 40;      // random waypoint pause time (seconds)
DCount_t default_velocity = 10;        // random waypoint velocity (meters/sec)
int default_radio_range = 250;         // meters (usually not changed)
// OSPF configuration
int default_wireless_interface = 1;    // interface type:  1=ptmp, 2=manet
int default_rxmt_int = 10;             // retransmit inteval (seconds)
int hello_interval = 10;               // hello interval (seconds)
int dead_interval = 40;                // dead interval (seconds)
int cost = 1;                          // ospfv3 interface cost
int interface_jitter = 100;            // jitter of pkt sends (msec)
int interface_flood_delay = 100;       // LSA coalescing timer (msec)

void usage(int argc, char** argv) {
  cout << "Usage:  " << argv[0] << " [seed=value]"<< endl;
  cout << "          [start_time=value] [stop_time=value]"<< endl;
  cout << "          [pktsize=value] [pktrate=value]"<< endl;
  cout << "          [tag=string]"<< endl;
  cout << "          [num_nodes=value] [grid_length=value]"<< endl;
  cout << "          [pause_time=value] [velocity=value]"<< endl;
  cout << "          [wireless_interface=value] [rxmt_int=value] "<< endl;
  cout << "          [diff_hellos] [tracing] [etrace] [ospfv3log]"<< endl;
  cout << " Options:"<<endl;
  cout << "     seed             : random seed, default: " 
                << default_seed << endl;
  cout << "     start_time       : statistics discarded before this time, "
                << "default: " << default_start_time << endl;
  cout << "     stop_time        : simulation stop time, default: " 
                << default_stop_time << endl;
  cout << "     pktsize          : packet size in bytes, default: " 
                << default_pktsize << endl;
  cout << "     pktrate          : number of user data pkts sent/sec (across "
                << endl;
  cout << "                        whole simulation), default: " 
                << default_pktrate << endl;
  cout << "     tag              : unique string to append to trace/statfile "
                << endl;
  cout << "                        names, default:null" << endl; 
  cout << "     num_nodes        : number of nodes, default: " 
                << default_num_nodes << endl;
  cout << "     grid_length      : length in meters of square grid, default: "
                << default_grid_length << endl;
  cout << "     pause_time       : seconds to pause after movement, default: "
                << default_pause_time << endl;
  cout << "     wireless_interface: type of interface ptmp=1, manet=2, "
                << "default: " << default_wireless_interface <<endl;
  cout << "     rxmt_int         : ospf6 retransmit interval, default: " 
                << default_rxmt_int << endl;
  cout << "     diff_hellos      : enable diff hellos on wireless interface, "
                " default:off" <<  endl;
  cout << "     tracing          : enable per-packet tracing, default: off" 
                << endl;
  cout << "     etrace           : enable extended tracing, default: off" 
                << endl;
  cout << "     ospfv3log        : enable ospfv3 logging, default: off" 
                << endl;
  cout << endl;
}

void set_global_seeds(int seed)
{
  // turn one seed into 6 repeatable seeds in the simulator using srand() 
  unsigned long seeds[6];
  srand(seed);

  seeds[0] = rand(); 
  seeds[1] = rand(); 
  seeds[2] = rand(); 
  seeds[3] = rand(); 
  seeds[4] = rand(); 
  seeds[5] = rand(); 

  Random::GlobalSeed(seeds[0],seeds[1],seeds[2],seeds[3],seeds[4],seeds[5]);
}

int main(int argc, char** argv)
{
  Validation::Init(argc, argv);
  Simulator s;
  Interface *inter;
  IFVec_t vInterface;
  std::string tracefile, statfile;
  char hostname[20], logfilename[20];
  std::string tag; 
  IPAddr_t first_rtrid;

  Count_t seed, nNodes, radio_range;
  DCount_t grid_length, velocity, pause_time;
  Count_t w_int_type, rxmt_int, pktrate, pktsize;
  DCount_t start_time, stop_time; 
  Arg("seed", seed, default_seed);
  Arg("num_nodes", nNodes, default_num_nodes);
  Arg("radio_range", radio_range, default_radio_range);
  Arg("grid_length", grid_length, default_grid_length);
  Arg("pause_time", pause_time, default_pause_time);
  Arg("velocity", velocity, default_velocity);
  Arg("wireless_interface", w_int_type, default_wireless_interface);
  Arg("rxmt_int", rxmt_int, default_rxmt_int);
  Arg("pktrate", pktrate, default_pktrate);
  Arg("pktsize", pktsize, default_pktsize);
  Arg("tag", tag, default_tag);
  Arg("start_time", start_time, default_start_time);
  Arg("stop_time", stop_time, default_stop_time);
  
  Arg::ProcessArgs(argc, argv);

  set_global_seeds(seed);

  if (Arg::Specified("help") || 
      Arg::Specified("-help") || 
      Arg::Specified("-h") || 
      Arg::Specified("h") || 
      Arg::Specified("?"))
  {
    usage(argc, argv);
    return 0;
  }
  cout << "Running " << nNodes << " nodes with random seed " << seed <<  endl;
  cout << "The ospf6 retransmit interval is " << rxmt_int << " secs" << endl;
  cout << "The wireless interface type is ";
  if (w_int_type == 1)
    cout << "point-to-multipoint" << endl;
  else if (w_int_type == 2)
    cout << "manet" << endl;
  else 
  {
    cout << "bad interface type: " << w_int_type << endl;
    usage(argc, argv);
    return 0;
  } 
  cout << endl;
  
  // Prepare output file names
  tracefile.append(argv[0]);
  if (tag.size()) {
    tracefile.append("-");
    tracefile.append(tag.c_str());
  }
  tracefile.append(".tr");
  statfile.append(argv[0]);
  if (tag.size()) {
    statfile.append("-");
    statfile.append(tag.c_str());
  }
  statfile.append(".stat");

  // ipv4 addresses must be less than 128.0.0.0 for ospf6 because we used 
  // the first bit in the ipv4 address to indicate a linklocal address 
  IPAddr_t firstIP = IPAddr("10.0.0.1");

  Location l(0,0); // lower left corner of rectangular grid

  // Randomize the distribution of nodes uniformly on square grid
  WirelessGridRectangular g(l, 
                            Constant(nNodes), //number of nodes
                            Uniform(0.0, grid_length), 
                            firstIP); 

  // Mobility:  Pause time is second argument, velocity is third argument
  g.AddMobility(RandomWaypoint(g, Constant(pause_time), Uniform(0,velocity)));

  // Per-packet tracing configuration
  if (Arg::Specified("tracing")) {
    ETrace* egs = ETrace::Instance();
    egs->Open(tracefile.c_str());
    if (Arg::Specified("etrace"))
      egs->SetExtended(true);
    egs->LayerOff(1); //don't etrace this layer
    egs->LayerOff(2); //don't etrace this layer
    egs->SetTraceEventLevel(2);
  }

  // Statistics collection
  EStat* estat = EStat::Instance();
  estat->CollectEStats(statfile.c_str(), start_time);
  estat->LayerOff(1);
  estat->LayerOff(2);
  
  // OSPF6 initialization
  OSPF6DApplication ospf6d;
  ospf6d.CollectStats(statfile.c_str(), start_time);
  ospf6d.LogNeighbors(0.5);  // Log physical layer neighbors every 500 ms
  
  OSPF6DInstance *ospf6_inst;
  Uniform u(0,start_time);
  first_rtrid = IPAddr("0.0.0.1"); 
          
  // CBR traffic initialization
  IPAddr remoteIP; 
  char buf[32];
  Count_t rate = (pktrate * pktsize * 8)/(nNodes*(nNodes-1)); 
  // start traffic generators at a random time
  Uniform cbrstart(start_time-(double)(pktsize*8)/rate, start_time); 
  
  for (Count_t i = 0; i < g.Size(); ++i)
  {
    Node* n = g.GetNode(i);
    n->SetRadioRange(radio_range);
     
    // OSPF6 Router Configuration
    ospf6_inst = ospf6d.AddNode(n);

    sprintf(hostname, "ospf6_%d", n->Id()+1);
    sprintf(logfilename, "log_ospf6_%d.log", n->Id()+1);

    ospf6_inst->SetHostName(hostname);
    if (Arg::Specified("ospfv3log"))
      ospf6_inst->SetLogFileName(logfilename);
    ospf6_inst->SetRouterId(n->Id()+first_rtrid);

    vInterface = n->Interfaces();
    for (IFVec_t::iterator it=vInterface.begin(); it!=vInterface.end(); ++it)
    {
      ospf6_inst->SetInterfaceHelloInterval(*it, hello_interval);
      ospf6_inst->SetInterfaceDeadInterval(*it, dead_interval);
      ospf6_inst->SetInterfaceRetransmitInterval(*it, rxmt_int);
      ospf6_inst->SetInterfaceArea(*it, 0);
      ospf6_inst->SetInterfaceCost(*it, cost);
      // too much jitter can cause stale ospf6 state to be sent
      ospf6_inst->SetInterfaceJitter(*it, interface_jitter); //msec
      ospf6_inst->SetInterfaceFloodDelay(*it, interface_flood_delay); //msec

      if ((*it)->GetL2Proto()->IsWireless())
      {
        if(w_int_type == 1)
          ospf6_inst->SetInterfaceType(*it, "point-to-multipoint");
        else if (w_int_type == 2)
        {
          ospf6_inst->SetInterfaceType(*it, "manet-reliable");
          if (Arg::Specified("diff_hellos"))
            ospf6_inst->SetInterfaceDiffHellos(*it, true);
        }
      }
      else if (strcmp((*it)->GetName(), "lo") == 0)
      {
        ospf6_inst->SetInterfaceType(*it, "loopback");
        if (rate > 0 && pktsize > 0)
        {
          // CBR traffic from this router to all other router's loopback addr
          for (Count_t j = 0; j < g.Size(); ++j)
          {
            if (i == j)
              continue;
            sprintf(buf, "0.0.%d.0", j+1); //loopback addr of routers
            remoteIP = IPAddr(buf); 
            CBRApplication* cbr = (CBRApplication*) n->AddApplication(
                          CBRApplication(remoteIP,1000,NO_PORT,rate,pktsize));
            cbr->Start(cbrstart.Value()); 
            cbr->Stop(stop_time - 5);
          }
        } 
      }
      else
        ospf6_inst->SetInterfaceType(*it, "broadcast");
    }
    ospf6_inst->Start(u.Value());  // random start time of each ospf6 router
  }
 if (!Validation::noAnimation)
    {
      s.AnimateWirelessTx(true);
      s.StartAnimation(0, true);
      //s.PauseAnimation(100);
      s.AnimationUpdateInterval(Time("1ms")); // 1ms initial update rate
    }
  s.StopAt(stop_time);
  s.Progress(40);
  s.Run();
}
