// Copyright 2004, The Boeing Company
/*
*
*            0     1
*            | \ \ |
*            2- 4 -3
*            \  |  /
*               5
*               
*           8 - 6 - 7
* At time 1900 the link is broken between nodes 5 and 6.
* After the neighbor relationship between 5 & 6 expires, a new router LSA is
* originated by node 5.  This LSA is flooded through out the network in
* an LSU.  The MPR or relay node that floods the LSU from node 5 is node
* 4.  The LSU does not reach node 0 or 1 when it is flooded from node 4.
* Therefore, the LSA is sent out by node 2 and 3 when their pushback timers
* expire.  
* When nodes 0 and 1 receive the LSA then they acknowledges it.
* This LSA has now reached all nodes, but nodes 0 and 1 do not know that
* node 4 has received this LSA, 

so it pushes back the LSA and then
* sends it upon pushback expiration.  Node 3 does not respond to the
* LSA because it previously implicitly acked the LSA, so node 0 unicast
* retransmits the LSA to node 3.  Node 3 finally acknowledges the
* reception to node 0.
*
*/


#include <iostream>
#include <stdlib.h>

#include "application-cbr.h"
#include "application-ospf6d.h"
#include "args.h"
#include "estat.h"
#include "etrace.h"
#include "ipaddr.h"
#include "linkmatrix.h"
#include "mobility-random-waypoint.h"
#include "node.h"
#include "rng.h"
#include "simulator.h"
#include "validation.h"  
#include "wireless-grid-rectangular.h"
#include "wlan.h"

void set_global_seeds(int seed)
{
  // turn one seed into 6 repeatable seeds in the simulator using srand() 
  unsigned long seeds[6];
  srand(seed);

  seeds[0] = rand(); 
  seeds[1] = rand(); 
  seeds[2] = rand(); 
  seeds[3] = rand(); 
  seeds[4] = rand(); 
  seeds[5] = rand(); 

  Random::GlobalSeed(seeds[0],seeds[1],seeds[2],seeds[3],seeds[4],seeds[5]);
}

int main(int argc, char** argv)
{
  struct ospf6_config_args ca;
  Validation::Init(argc, argv);
  Simulator s;
  Interface *inter;
  IFVec_t vInterface;
  std::string tracefile, statfile;
  char hostname[20], logfilename[20];
  IPAddr_t first_rtrid;

  if (!set_ospf6_config_arguments(argc, argv, &ca))
    return 0;
  
  printf("\nNOTE:  Overriding command line input for validation!\n\n");
  ca.trace = true;
  ca.etrace = true;
  ca.nNodes = 9;
  ca.w_int_type = 2;
  ca.w_flooding = 2;
  double trace_on_time = 1900;
  double trace_off_time = 2000;

  ca.HelloInterval = 2;
  ca.DeadInterval = 6;
  ca.RxmtInterval = 5;
  ca.AckInterval = 1800;
  ca.PushbackInterval = 2000;
  ca.BackupWaitInterval = 2000;

  set_global_seeds(ca.seed);

  // Prepare output file names
  tracefile.append(argv[0]);
  if (ca.tag.size()) {
    tracefile.append("-");
    tracefile.append(ca.tag.c_str());
  }
  tracefile.append(".tr");
  statfile.append(argv[0]);
  if (ca.tag.size()) {
    statfile.append("-");
    statfile.append(ca.tag.c_str());
  }
  statfile.append(".stat");

  // ipv4 addresses must be less than 128.0.0.0 for ospf6 because we used 
  // the first bit in the ipv4 address to indicate a linklocal address 
  IPAddr_t firstIP = IPAddr("10.0.0.1");

  //Matrix 
  LinkMatrix LM(ca.nNodes, firstIP, Mask(24)); 
  LM.ScheduleMatrixChange(0, 0, 1, 1);  //(Time,src_node,dst_node,error rate)
  LM.ScheduleMatrixChange(0, 0, 3, 1);
  LM.ScheduleMatrixChange(0, 0, 5, 1);
  LM.ScheduleMatrixChange(0, 0, 6, 1);
  LM.ScheduleMatrixChange(0, 0, 7, 1);
  LM.ScheduleMatrixChange(0, 0, 8, 1);

  LM.ScheduleMatrixChange(0, 1, 0, 1);
  LM.ScheduleMatrixChange(0, 1, 2, 1);
  LM.ScheduleMatrixChange(0, 1, 5, 1);
  LM.ScheduleMatrixChange(0, 1, 6, 1);
  LM.ScheduleMatrixChange(0, 1, 7, 1);
  LM.ScheduleMatrixChange(0, 1, 8, 1);

  LM.ScheduleMatrixChange(0, 2, 1, 1);
  LM.ScheduleMatrixChange(0, 2, 3, 1);
  LM.ScheduleMatrixChange(0, 2, 6, 1);
  LM.ScheduleMatrixChange(0, 2, 7, 1);
  LM.ScheduleMatrixChange(0, 2, 8, 1);

  LM.ScheduleMatrixChange(0, 3, 0, 1);
  LM.ScheduleMatrixChange(0, 3, 2, 1);
  LM.ScheduleMatrixChange(0, 3, 6, 1);
  LM.ScheduleMatrixChange(0, 3, 7, 1);
  LM.ScheduleMatrixChange(0, 3, 8, 1);

  LM.ScheduleMatrixChange(0, 4, 6, 1);
  LM.ScheduleMatrixChange(0, 4, 7, 1);
  LM.ScheduleMatrixChange(0, 4, 8, 1);

  LM.ScheduleMatrixChange(0, 5, 0, 1);
  LM.ScheduleMatrixChange(0, 5, 1, 1);
  LM.ScheduleMatrixChange(0, 5, 6, 1);
  LM.ScheduleMatrixChange(0, 5, 7, 1);
  LM.ScheduleMatrixChange(0, 5, 8, 1);

  LM.ScheduleMatrixChange(0, 6, 0, 1);
  LM.ScheduleMatrixChange(0, 6, 1, 1);
  LM.ScheduleMatrixChange(0, 6, 2, 1);
  LM.ScheduleMatrixChange(0, 6, 3, 1);
  LM.ScheduleMatrixChange(0, 6, 4, 1);
  LM.ScheduleMatrixChange(0, 6, 5, 1);

  LM.ScheduleMatrixChange(0, 7, 0, 1);
  LM.ScheduleMatrixChange(0, 7, 1, 1);
  LM.ScheduleMatrixChange(0, 7, 2, 1);
  LM.ScheduleMatrixChange(0, 7, 3, 1);
  LM.ScheduleMatrixChange(0, 7, 4, 1);
  LM.ScheduleMatrixChange(0, 7, 5, 1);

  LM.ScheduleMatrixChange(0, 8, 0, 1);
  LM.ScheduleMatrixChange(0, 8, 1, 1);
  LM.ScheduleMatrixChange(0, 8, 2, 1);
  LM.ScheduleMatrixChange(0, 8, 3, 1);
  LM.ScheduleMatrixChange(0, 8, 4, 1);
  LM.ScheduleMatrixChange(0, 8, 5, 1);

  // force flood of LSA from node 4 at time
  LM.ScheduleMatrixChange(1900, 5, 6, 0);
  LM.ScheduleMatrixChange(1900, 6, 5, 0);

  //force loss of LSA from node 3 to node 0
  //LM.ScheduleMatrixChange(1938, 4, 0, 1);
  //LM.ScheduleMatrixChange(1938, 4, 1, 1);
  //LM.ScheduleMatrixChange(1939, 4, 0, 0);
  //LM.ScheduleMatrixChange(1939, 4, 1, 0);

// */
  // Per-packet tracing configuration
  if (ca.trace) {
    ETrace* egs = ETrace::Instance();
    egs->Open(tracefile.c_str());
    if (ca.etrace)
      egs->SetExtended(true);
    egs->LayerOff(1); //don't etrace this layer
    egs->LayerOff(2); //don't etrace this layer
    egs->SetTraceEventLevel(2);
    egs->SetTraceOnTime(trace_on_time);
    egs->SetTraceOffTime(trace_off_time);
  }

  // Statistics collection
  EStat* estat = EStat::Instance();
  estat->CollectEStats(statfile.c_str(), ca.start_time);
  estat->LayerOff(1);
  estat->LayerOff(2);
  
  // OSPF6 initialization
  OSPF6DApplication ospf6d;
  ospf6d.CollectStats(statfile.c_str(), ca.start_time);
  ospf6d.LogNeighbors(0.5);  // Log physical layer neighbors every 500 ms
  
  OSPF6DInstance *ospf6_inst;
  Uniform u(0,ca.start_time);
  first_rtrid = IPAddr("0.0.0.1"); 
          
  // CBR traffic initialization
  IPAddr remoteIP; 
  char buf[32];
  Count_t rate = (ca.pktrate * ca.pktsize * 8)/(ca.nNodes*(ca.nNodes-1)); 
  // start traffic generators at a random time
  Uniform cbrstart(ca.start_time-(double)(ca.pktsize*8)/rate, ca.start_time); 
  
  for (Count_t i = 0; i < LM.Size(); ++i)
  {
    Node* n = LM.GetNode(i);
     
    // OSPF6 Router Configuration
    ospf6_inst = ospf6d.AddNode(n);

    sprintf(hostname, "ospf6_%d", n->Id()+1);
    sprintf(logfilename, "log_ospf6_%d.log", n->Id()+1);

    ospf6_inst->SetHostName(hostname);
    if (ca.ospfv3log)
      ospf6_inst->SetLogFileName(logfilename);
    ospf6_inst->SetRouterId(n->Id()+first_rtrid);
    ospf6_inst->SetMinLSInterval(ca.MinLSInterval);
    ospf6_inst->SetMinLSArrival(ca.MinLSArrival);

    vInterface = n->Interfaces();
    for (IFVec_t::iterator it=vInterface.begin(); it!=vInterface.end(); ++it)
    {
      ospf6_inst->SetInterfaceHelloInterval(*it, ca.HelloInterval);
      ospf6_inst->SetInterfaceDeadInterval(*it, ca.DeadInterval);
      ospf6_inst->SetInterfaceRetransmitInterval(*it, ca.RxmtInterval);
      ospf6_inst->SetInterfaceAckInterval(*it, ca.AckInterval);
      ospf6_inst->SetInterfaceArea(*it, 0);
      ospf6_inst->SetInterfaceCost(*it, ca.cost);
      // too much jitter can cause stale ospf6 state to be sent
      ospf6_inst->SetInterfaceJitter(*it, ca.interface_jitter); //msec
      ospf6_inst->SetInterfaceFloodDelay(*it, ca.interface_flood_delay); //msec

      if (strcmp((*it)->GetName(), "lo") == 0)
      {
        ospf6_inst->SetInterfaceType(*it, "loopback");
        if (rate > 0 && ca.pktsize > 0)
        {
          // CBR traffic from this router to all other router's loopback addr
          for (Count_t j = 0; j < LM.Size(); ++j)
          {
            if (i == j)
              continue;
            sprintf(buf, "0.0.%d.0", j+1); //loopback addr of routers
            remoteIP = IPAddr(buf);
            CBRApplication* cbr = (CBRApplication*) n->AddApplication(
                        CBRApplication(remoteIP,1000,NO_PORT,rate,ca.pktsize));
            cbr->Start(cbrstart.Value());
            cbr->Stop(ca.stop_time - 5);
          }
        }
      }
      else
      {
        if(ca.w_int_type == 1)
          ospf6_inst->SetInterfaceType(*it, "point-to-multipoint");
        else if (ca.w_int_type == 2)
        {
          ospf6_inst->SetInterfaceType(*it, "manet-reliable");
          ospf6_inst->SetInterfaceDiffHellos(*it, ca.diff_hellos);
          //set flooding type
          if (ca.w_flooding == 0)
            ospf6_inst->SetInterfaceFlooding(*it, "broadcast");
          else if (ca.w_flooding == 1)
          {
            ospf6_inst->SetInterfaceFlooding(*it, "mpr");
            ospf6_inst->SetInterfacePushBackInterval(*it, ca.PushbackInterval);
          }
          else if (ca.w_flooding == 2)
          {
            ospf6_inst->SetInterfaceFlooding(*it, "mdr");
            ospf6_inst->SetInterfaceBackupWaitInterval(*it,
                                                       ca.BackupWaitInterval);
            ospf6_inst->SetInterfaceTwoHopRefresh(*it, ca.TwoHopRefresh);
            ospf6_inst->SetInterfaceHelloRepeatCount(*it, ca.HelloRepeatCount);
            if (ca.AdjConnectivity == 0)
              ospf6_inst->SetInterfaceAdjConnectivity(*it, "fully");
            else if (ca.AdjConnectivity == 1)
              ospf6_inst->SetInterfaceAdjConnectivity(*it, "uniconnected");
            else
              ospf6_inst->SetInterfaceAdjConnectivity(*it, "biconnected");
            if (ca.LSAFullness == 0)
              ospf6_inst->SetInterfaceLSAFullness(*it, "minlsa");
            else if (ca.LSAFullness == 1)
              ospf6_inst->SetInterfaceLSAFullness(*it, "minhoplsa");
            else if (ca.LSAFullness == 2)
              ospf6_inst->SetInterfaceLSAFullness(*it, "mdrfulllsa");
            else
              ospf6_inst->SetInterfaceLSAFullness(*it, "fulllsa");
            ospf6_inst->SetInterfaceNonPersistentMDR(*it, ca.NonPersistentMDR);
          }
        }
        if (i==0)
        {
          WirelessLink *l = (WirelessLink *) (*it)->GetLink();
          l->SetAlpha(ca.alpha);
        }
      }
    }
    ospf6_inst->Start(u.Value());  // random start time of each ospf6 router
  }
  s.StopAt(ca.stop_time);
  s.Progress(40);
  s.Run();
}
